;;;; Cu
;;;; LastChange: Thu Apr 12 23:41:23 2007


;;; sck-ast-format

(defconst sck=ast-formatters
  '(("%[-0-9.]*s" sck-ast-fmt-s)
    ("%p" sck-ast-fmt-p)))

(defun sck-ast-fmt-s (fmt arg)
  (cond
   ((sck-ast-ident-p arg)
    (format fmt (sck-ast-ident-name arg)))
   (t
    (format fmt arg))))

(defun sck-ast-fmt-p (fmt pos)
  (assert (eq (car-safe pos) '$))
  (format "%s:%d" (elt sck*ast-message-file-names (cadr pos)) (caddr pos)))

(defun sck-ast-format (fmt &rest args)
  (fmt fmt sck=ast-formatters args))


;;; bZ[W

(defvar sck*ast-messageq nil)
(defvar sck*ast-message-file-names nil)
(defvar sck*ast-error-count 0)
(defvar $ (list '$ 0 0)) ;  special variable

(defun sck-ast-message-init (filenames)
  (setq sck*ast-messageq (make-que)
	sck*ast-message-file-names filenames
	sck*ast-error-count 0))

(defun sck-ast-message (fmt &rest args)
  (let ((msg (concat (if $ (sck-ast-format "%p: " $) "")
		     (apply 'sck-ast-format fmt args))))
    (append-que sck*ast-messageq msg)
    msg))

(defun sck-ast-error (fmt &rest args)
  (incf sck*ast-error-count)
  (apply 'sck-ast-message fmt args))

(defun sck-ast-warning (fmt &rest args)
  (apply 'sck-ast-message (concat "warning: " fmt) args))

(defun sck-ast-message-strings ()
  (que-to-list sck*ast-messageq))

(defun sck-ast-sort-message-strings (strs)
  ;; (sck-ast-sort-message-strings '("a.c:100: a" "a.c:20: z" "a.c:20: b"))
  ;; => ("a.c:20: z" "a.c:20: b" "a.c:100: a")
  (sort strs #'(lambda (x y)
		 (string< (sck-ast-sort-message-strings/key x)
			  (sck-ast-sort-message-strings/key y)))))

(defun sck-ast-sort-message-strings/key (str)
  ;; (sck-ast-sort-message-strings/key "a.c:100: error")
  ;; => "a.c:0000000100"
  (unless (string-match "\\(^[^ ]+\\):\\([0-9]+\\):" str)
    (assert nil))
  (format "%s:%010d"
	  (match-string 1 str)
	  (string-to-int (match-string 2 str))))


(defun sck-ast-error-in-messages-p (messages)
  ;; warning ȊO̓G[
  ;; (sck-ast-error-in-messages-p '("a.c:3: warning: foo")) => nil
  ;; (sck-ast-error-in-messages-p '("a.c:3: warning: foo" "a.c:3: bar")) => t
  (block return
    (dolist (msg messages)
      (unless (string-match "^[^ ]+: warning: " msg)
	(return-from return t)))
    nil))


;;; O̐

(defvar sck*ast-genname-counters nil)

(defun sck-ast-genname-init ()
  (setq sck*ast-genname-counters nil))

(defun sck-ast-genname (prefix)
  ;; PREFIX:<String>|<Symbol> ƂɈӓIȖO PREFIX<Fixnum> 𐶐B
  ;; (sck-ast-genname "L") => "L0"
  ;; (sck-ast-genname "L") => "L1"
  ;; (sck-ast-genname "struct ") => "struct 0"
  (when (symbolp prefix)
    (setq prefix (symbol-name prefix)))
  (let ((a (assoc prefix sck*ast-genname-counters)))
    (unless a
      (setq a (list prefix 0))
      (push a sck*ast-genname-counters))
    (prog1
	(format "%s%d" prefix (cadr a))
      (incf (cadr a)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; q

(defun sck-ast-integral-zero-expr-p (expr)
  (and (eq (sck-ast-expr-op expr) 'const)
       (sck-ast-integral-type-p (sck-ast-expr-type expr))
       (bn-= 0 (sck-ast-expr-value expr))))

(defun sck-ast-iconst-expr-p (expr)
  (and (eq (sck-ast-expr-op expr) 'const)
       (sck-ast-integral-scalar-type-p (sck-ast-expr-type expr))))

(defun sck-ast-fconst-expr-p (expr)
  (and (eq (sck-ast-expr-op expr) 'const)
       (sck-ast-floating-type-p (sck-ast-expr-type expr))))

(defun sck-ast-nconst-expr-p (expr)
  (or (sck-ast-iconst-expr-p expr)
      (sck-ast-fconst-expr-p expr)))

(defun sck-ast-nconst-zero-expr-p (expr)
  (and (sck-ast-nconst-expr-p expr)
       (bn-= 0 (sck-ast-expr-value expr))))

(defun sck-ast-sconst-expr-p (expr)
  (and (eq (sck-ast-expr-op expr) 'const)
       ;; ̂Ƃ const & array łAAA
       (sck-ast-array-type-p (sck-ast-expr-type expr))))

(defun sck-ast-lvalue-expr-p (expr)
  (or (sck-ast-sconst-expr-p expr)
      (memq (sck-ast-expr-op expr) '(vref mem pmember))
      (and (eq (sck-ast-expr-op expr) 'member)
	   (sck-ast-lvalue-expr-p  (car (sck-ast-expr-args expr))))))


(defun sck-ast-asmconst-expr-p (expr)
  ;; AZuł̒萔ǂ肷B
  ;; 萔AO[oϐ̃AhXA add/sub
  (setq expr (sck-ast-fold-const expr)) ; 
  (let ((op (sck-ast-expr-op expr))
	(type (sck-ast-expr-type expr))
	(args (sck-ast-expr-args expr)))
    (case op
     (const
      (sck-ast-nconst-expr-p expr))
     ((add sub)
      (every 'sck-ast-asmconst-expr-p args))
     (conv
      ;;  ltype gȂ悤
      (and (eq (sck-ast-ltypeof type)
	       (sck-ast-ltypeof (sck-ast-expr-type (car args))))
	   (sck-ast-asmconst-expr-p (car args))))
     ((addr decay)
      (or (sck-ast-static-variable-expr-p (car args))
	  (and (eq (sck-ast-expr-op (car args)) 'mem)
	       (sck-ast-asmconst-expr-p
		(car (sck-ast-expr-args (car args)))))))
     (t
      nil))))

(defun sck-ast-static-variable-expr-p (expr)
  (let ((op (sck-ast-expr-op expr))
	(type (sck-ast-expr-type expr))
	(args (sck-ast-expr-args expr)))
    (case op
      (const
       (sck-ast-sconst-expr-p expr))
      (vref
       (let ((ident (sck-ast-expr-value expr)))
	 (or (eq (sck-ast-ident-sc ident) 'static)
	     ;;  sck*cps-global-env
	     ;; (eq (sck-ast-ident-env ident) sck*cps-global-env)
	     (null (sck-ast-env-parent (sck-ast-ident-env ident)))
	     )))
      (member
       (sck-ast-static-variable-expr-p (car args)))
      (t
       nil))))


(defun sck-ast-register-expr-p (expr)
  (and (eq (sck-ast-expr-op expr) 'vref)
       (eq (sck-ast-ident-sc (sck-ast-expr-value expr)) 'register)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ̌`̎B
;;; ^G[ꍇ͎wEA error ^̎ƂĂƂɂB

(defun sck-ast-make-error-expr ()
  ;; G[Nꏊӂ߂ error ^̎B
  (sck-ast-make-vref-expr
   (sck-ast-make-ident `(ident ,$ "@error")
		       :kind 'var
		       :type 'error)))

(defun* sck-ast-make-iconst-expr (type value &key overflowmsg)
  ;; sck-ast-integral-scalar-type-p Ȓ萔̍쐬B
  ;; I[o[t[̃[jǑ
  (assert (sck-ast-integral-scalar-type-p type))
  (when (bn-floating-p value) ; 
    (setq value (bn-truncate value)))
  (assert (bn-integer-p value))
  ;;
  (sck-ast-make-expr 'const
		     :type type
		     :value (sck-ast-make-iconst-expr-value
			     type
			     value
			     :overflowmsg overflowmsg)))

(defun* sck-ast-make-iconst-expr-value (type value &key overflowmsg)
  ;; Return: <Integer>
  (assert (sck-ast-integral-scalar-type-p type))
  (let* ((imptype (cond
		   ((sck-ast-pointer-type-p type)
		    sck=ast-pointer-implement-type)
		   ((sck-ast-enum-type-p type)
		    sck=ast-enum-implement-type)
		   (t type)))
	 (retval (bn-try-to-elnum
		 (funcall (if (sck-ast-signed-type-p imptype)
			      'bn-signed
			    'bn-unsigned)
			  value
			  (* sck=ast-byte-in-bits
			     (sck-ast-sizeof imptype))))))
    (when (and overflowmsg (bn-/= value retval))
      (sck-ast-warning overflowmsg))
    ;;
    retval))
  

(defun sck-ast-make-fconst-expr (type value)
  (assert (sck-ast-floating-type-p type))
  (when (bn-integer-p value)
    (setq value (bn-add 0.0 value)))
  (assert (bn-floating-p value))
  (sck-ast-make-expr 'const :type type :value (bn-try-to-elnum value)))

(defun* sck-ast-make-nconst-expr (type value &key overflowmsg)
  (assert (sck-ast-scalar-type-p type))
  (if (sck-ast-integral-scalar-type-p type)
      (sck-ast-make-iconst-expr type value :overflowmsg overflowmsg)
    (sck-ast-make-fconst-expr type value)))

(defun sck-ast-make-sizeof-expr (type)
  (let ((size (sck-ast-sizeof type)))
    (unless size
      (sck-ast-error "unknown size")
      (setq size 1))
    (sck-ast-make-iconst-expr sck=ast-sizeof-type size)))
    
(defun sck-ast-make-mem-expr (expr)
  (sck-ast-complete-expr
   (sck-ast-make-expr 'mem :args (list expr))))

(defun sck-ast-make-vref-expr (ident)
  (assert (eq (sck-ast-ident-kind ident) 'var))
  (sck-ast-make-expr
   'vref
   :type (sck-ast-ident-type ident)
   :value ident))

(defun sck-ast-make-decay-expr (expr)
  ;; IyhƂĎQƂꂽƂɂ C Ɠ̌^ύXB
  ;; array ܂ function Ȃ decay łȂ΂̂܂܁B
  ;; enum  sck=ast-enum-implement-type  conv
  (let* ((type0 (sck-ast-skip-typedef (sck-ast-expr-type expr)))
	 (type (sck-ast-skip-typedef-and-qual type0)))
    (case (car-safe type)
     (array
      (sck-ast-make-expr 'decay
			 :type `(pointer ,(caddr type0))
			 :args (list expr)))
     (function
      (sck-ast-make-expr 'decay
			 :type `(pointer ,type0)
			 :args (list expr)))
     (enum
      (sck-ast-make-expr 'conv
			 :type sck=ast-enum-implement-type
			 :args (list expr)))
     (t
      expr))))

(defun sck-ast-make-conv-expr (type expr)
  ;; Ƃ肠 decay ĂяoŖIɂ悢H
  (setq expr (sck-ast-make-decay-expr expr))
  ;; SɌ^Ȃ
  (if (sck-ast-type-equal-strict type (sck-ast-expr-type expr))
      ;; m[h͍Ȃ
      expr
    (sck-ast-complete-expr
     (sck-ast-make-expr 'conv
			:type type
			:args (list expr)))))


(defun sck-ast-make-conv-bool-expr (expr)
  ;; boolean ɕϊB
  (if (memq (sck-ast-expr-op expr)
	    '(lnot land lor tsteq tstne tstlt tstle tstge tstgt)) ; 
      expr
    (sck-ast-complete-expr
     ;;  int,
     ;; ̂Ył͂邪AG[bZ[WB
     (sck-ast-make-expr 'tstne
			:args (list expr
				    (sck-ast-make-iconst-expr 'int 0))))))

(defun sck-ast-make-set-expr (x y)
  (sck-ast-complete-expr
   (sck-ast-make-expr 'set
		      :args (list x y))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Constant folding

(defun sck-ast-fold-const (expr)
  ;; \Ȃ萔ɕϊB
  ;; ͔j󂵂ȂB
  ;;  struct foo {int x,y;}; int i = (int)&(*(struct foo *)0).y; 
  (let (($ (sck-ast-expr-pos expr)))
    (sck-ast-fold-const-rec expr)))

(defun sck-ast-fold-const-rec (expr)
  (let ((op (sck-ast-expr-op expr))
	(type (sck-ast-expr-type expr))
	(args (sck-ast-expr-args expr)))
    (block return
      ;;  fold ς݁B
      (when (memq op '(const vref))
	(return-from return expr))
      ;; Zp^Aenum ^ȊOȂ炠߂B
      ;; |C^萔ׂ͋
      ;; (int)(int*)123 + 4 Ƃ
      (unless (or (sck-ast-arith-type-p type)
		  (sck-ast-enum-type-p type))
	(return-from return expr))
      ;;
      ;; ċAIɎqvZB
      ;;
      (setq args (mapcar 'sck-ast-fold-const-rec args))
      ;;
      ;; 萔łȂ̂΂߂B
      ;;  non-strict ȉŽ
      (unless (every 'sck-ast-nconst-expr-p args)
	(return-from return expr))
      ;;
      ;; ȉASĂ̈͒萔B
      ;;
      (let ((vals (mapcar 'sck-ast-expr-value args)))
	(cond
	 ;;
	 ;;  x / 0,  x % 0  fold ȂB
	 ;;
	 ((and (memq op '(div mod))
	       (bn-zerop (cadr vals)))
	  expr)
	 ;;
	 (t
	  (let ((val (ecase op
		       (neg (apply 'bn-neg vals))
		       (add (apply 'bn-add vals))
		       (sub (apply 'bn-sub vals))
		       (mul (apply 'bn-mul vals))
		       (div (apply 'bn-div vals))
		       (mod (apply 'bn-mod vals))
		       (lsh (apply 'bn-ash vals))
		       (rsh (bn-ash (car vals) (bn-neg (cadr vals))))
		       (bnot (apply 'bn-lognot vals))
		       (bor (apply 'bn-logior vals))
		       (bxor (apply 'bn-logxor vals))
		       (band (apply 'bn-logand vals))
		       (lnot (if (bn-zerop (car vals)) 1 0))
		       (land (if (some 'bn-zerop vals) 0 1))
		       (lor (if (every 'bn-zerop vals) 0 1))
		       (comma (cadr vals))
		       (cond (if (bn-zerop (car vals))
				 (caddr vals)
			       (cadr vals)))
		       (tsteq (if (apply 'bn-= vals) 1 0))
		       (tstne (if (apply 'bn-/= vals) 1 0))
		       (tstlt (if (apply 'bn-< vals) 1 0))
		       (tstgt (if (apply 'bn-> vals) 1 0))
		       (tstle (if (apply 'bn-<= vals) 1 0))
		       (tstge (if (apply 'bn->= vals) 1 0))
		       (conv (car vals))
		       ((decay set addr mem member) nil))))
	    ;;
	    (if val
		(sck-ast-make-nconst-expr
		 type val :overflowmsg "integer overflow")
	      expr))))))))





;;; debug p

(defun sck-ast-printf (fmt &rest args)
  (with-current-buffer (get-buffer-create "*sck-ast-printf*")
    (apply 'insertf fmt args)))

;;;
(provide 'sck-ast-lib)
