;;;; AST
;;;; LastChange: Thu Apr 12 23:40:55 2007


(require 'sck-ast-lib)
(require 'sck-ast-types)
(require 'sck-ast-check)
(require 'sck-ast-reduce)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ident

(defstruct sck-ast-ident
  ;; name space  "struct foo" "union foo" "label foo"
  ;; kind  type ̏ꍇAstruct foo, union foo ͂ꂼ
  ;; "struct foo", "union foo" ƂOƂĈB
  name ; <String>
  kind ; var|type|const
  type ; <Type>
  
  ;; <IdentValue> ::= nil | <Integer> | <Expr> | <Inits> | <Lambda>
  ;; <Inits>      ::= (inits <<Expr>|(zeros <Integer>)>..)
  ;; <Lambda>     ::= (lambda <Type> (<Declaration>..) <BlockStmt>)
  value

  ;; <SC> ::= nil | static | extern | auto | register
  sc


  ;; o^Ă env
  env ; <sck-ast-env> | nil

  ;; & ꂽtO nil|&
  addrp

  ;; \[Xʒu
  pos ; ($ <FileNo> <LineNo>)
  )

(defun* sck-ast-make-ident (past-ident &key kind type value sc)
  ;; PAST-IDENT: (ident <Pos> <String>)
  ;; Return: <sck-ast-ident>
  (assert (eq (car-safe past-ident) 'ident))
  (assert (memq kind '(var type const)))
  (assert (memq sc '(nil extern static auto register)))
  (make-sck-ast-ident :name (caddr past-ident)
		      :kind kind
		      :type type
		      :value value
		      :sc sc
		      :pos (cadr past-ident)))

(defun sck-ast-ident-to-sexp (ident)
  (let ((sexp (list (sck-ast-ident-kind ident)
		    (sck-ast-ident-name ident)
		    (sck-ast-type-to-sexp (sck-ast-ident-type ident))
		    (sck-ast-ident-value-to-sexp (sck-ast-ident-value ident))
		    (sck-ast-ident-sc ident)
		    ;;  sexp ɂ͏oȂ
		    ;; (sck-ast-ident-addrp ident) 
		    )))
    ;; ̘A nil 폜B
    (nreverse (drop-while 'null (nreverse sexp)))))


(defun sck-ast-ident-value-to-sexp (val)
  (cond
   ;; member/union offset
   ((bn-integer-p val) val)
   ;; lambda
   ((eq (car-safe val) 'lambda)
    `(lambda ,(sck-ast-type-to-sexp (cadr val))
       ,(sck-ast-env-to-sexp (caddr val))
       ,(sck-ast-stmt-to-sexp (cadddr val))))
   ;; initializer : <expr> | (inits <<expr>|(zeros <Nbyte>)>..)
   ((sck-ast-expr-p val)
    (sck-ast-expr-to-sexp val))
   ((eq (car-safe val) 'inits)
    (cons 'inits
	  (mapcar #'(lambda (e)
		      (if (sck-ast-expr-p e)
			  (sck-ast-expr-to-sexp e)
			e))
		  (cdr val))))
   ((null val)
    nil)
   (t
    (assert nil))))


(defun sck-ast-type-to-sexp (type)
  (if (symbolp type)
      type
    (ecase (car type)
      (typename (sck-ast-ident-name (cadr type)))
      (function `(function ,(mapcar 'sck-ast-type-to-sexp
				    (cadr type))
			   ,(sck-ast-type-to-sexp (caddr type))))
      (array `(array ,(cadr type) ,(sck-ast-type-to-sexp (caddr type))))
      ((const volatile pointer)
       (mapcar 'sck-ast-type-to-sexp type))
      ((struct union)
       (if (cadr type)
	   `(,(car type)
	     ,(and (cadr type) (sck-ast-env-to-sexp (cadr type)))
	     ,(caddr type)
	     ,(cadddr type))
	 (list (car type))))
      ((enum) type)
      ;;
      (t
       (assert nil)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; env

(defstruct sck-ast-env
  parent ; <sck-ast-env> | nil
  (htab (make-hash-table :test 'equal))
  (identq (make-que)))

(defun sck-ast-make-env (parent)
  (assert (or (null parent) (sck-ast-env-p parent)))
  (make-sck-ast-env :parent parent))

(defun sck-ast-env-idents (env)
  (que-to-list (sck-ast-env-identq env)))

(defun sck-ast-env-to-sexp (env)
  (mapcar #'sck-ast-ident-to-sexp (sck-ast-env-idents env)))

(defun sck-ast-module-to-sexp (module)
  ;; ^ MODULE  S-ɕϊB
  `(module ,(cadr module)
	   ,@(sck-ast-env-to-sexp (caddr module))))

;(defun sck-ast-empty-env-p (env)
;  (null (sck-ast-env-idents env)))


(defun sck-ast-enter-ident (env ident &optional force)
  ;; FORCE  Non-nil ȂADŋIɓo^B
  ;; ̓O[o env ł̂ݎgB
  ;; ident : <sck-ast-ident>
  (assert (sck-ast-ident-p ident))
  (assert (null (sck-ast-ident-env ident)))
  (let ((name (sck-ast-ident-name ident))
	(htab (sck-ast-env-htab env))
	(identq (sck-ast-env-identq env)))
    (unless force
      (assert (null (gethash name htab))))
    (setf (gethash name htab) ident
	  (sck-ast-ident-env ident) env)
    (append-que identq ident)))

(defun sck-ast-search-ident (env ident &optional onlyhere)
  ;; ident : (ident <Pos> <String>) | <sck-ast-ident> | <String>
  ;; env y ionlyhere = nil ȂjeT[`
  (assert (or (eq (car-safe ident) 'ident)
	      (sck-ast-ident-p ident)
	      (stringp ident)))
  (let ((name (cond
	       ((sck-ast-ident-p ident) (sck-ast-ident-name ident))
	       ((stringp ident) ident)
	       (t (caddr ident)))))
    (if onlyhere
	(gethash name (sck-ast-env-htab env))
      (block nil
	(while env
	  (let ((ident (gethash name (sck-ast-env-htab env))))
	    (when ident (return ident)))
	  (setq env (sck-ast-env-parent env)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; expr

(defstruct sck-ast-expr
  op
  type
  args
  value
  pos)

(defun* sck-ast-make-expr (op &key type args value pos)
  (make-sck-ast-expr :op op
		     :type type
		     :args args
		     :value value
		     :pos (or pos $)))

(defun sck-ast-expr-to-sexp (expr)
  (list* (sck-ast-expr-op expr)
	 (sck-ast-type-to-sexp
	  (sck-ast-expr-type expr))
	 (if (sck-ast-expr-value expr)
	     (list (sck-ast-expr-value-to-sexp
		    (sck-ast-expr-value expr)))
	   (mapcar 'sck-ast-expr-to-sexp
		   (sck-ast-expr-args expr)))))

(defun sck-ast-expr-value-to-sexp (val)
  (cond
   ((sck-ast-ident-p val)
    (sck-ast-ident-name val))
   (t
    val)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; stmt

(defun sck-ast-stmt-to-sexp (stmt)
  (cond
   ((sck-ast-expr-p stmt) (sck-ast-expr-to-sexp stmt))
   ((atom stmt) stmt)
   (t
    (case (car stmt)
      (block  `(block ,(sck-ast-env-to-sexp (cadr stmt))
		 ,@(mapcar 'sck-ast-stmt-to-sexp (cddr stmt))))
      (t (mapcar 'sck-ast-stmt-to-sexp stmt))))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; sexp -> iast

(defun sck-ast-sexp-to-module (sexp)
  "S- SEXP  AST  <Module> Ƃĉ͂A\ɕϊB
S-͐ <Module> łƉ肵AG[`FbN⃊Jo͈؍sȂB
ɃR}hƂċNꂽꍇ̓|Cgȉ S-W[ƌȂA
sck-ast-module-to-sexp ŋtϊʂ\B"
  (interactive (list (save-excursion (read (current-buffer)))))
  (let* ((name (cadr sexp))
	 (iast `(module ,name ,(sck-ast-sexp-to-env nil (cddr sexp)))))
    (when (interactive-p)
      (with-display-buffer "*sck-ast-sexp-to-module*"
	(let ((sast (sck-ast-module-to-sexp iast)))
	  (insert (if (equal sexp sast)
		      ";;; Correct conversion !\n\n"
		    ";;; Incorrect conversion...\n\n"))
	  (insertf "%s" (sck-ast-pp-ast sast 2)))))
    ;;
    iast))

(defun sck-ast-sexp-to-env (env sexps)
  (let ((newenv (sck-ast-make-env env)))
    (dolist (sexp sexps)
      (let ((ident (make-sck-ast-ident
		    :kind (car sexp)
		    :name (cadr sexp)
		    :sc (elt sexp 4))))
	(sck-ast-enter-ident newenv ident
			     t ; ˂ force... ܂薳nY
			     )
	;; ċAI`̂ŁAȉ͈Uo^Ăݒ肵B
	(setf (sck-ast-ident-type ident)
	      (sck-ast-sexp-to-type newenv (caddr sexp))
	      (sck-ast-ident-value ident)
	      (sck-ast-sexp-to-ident-value newenv (cadddr sexp)))))
    newenv))

(defun sck-ast-sexp-to-type (env sexp)
  (cond
   ;; <TypeName>
   ((stringp sexp)
    ;; (sck-ast-ident-type (sck-ast-search-ident env sexp))
    (let ((ident (sck-ast-search-ident env sexp)))
      (assert ident)
      `(typename ,ident)))
   ;; <AtomicType> | void
   ((atom sexp)
    sexp)
   ;; <PointerType>
   ((eq (car sexp) 'pointer)
    `(pointer ,(sck-ast-sexp-to-type env (cadr sexp))))
   ;; <ArrayType>
   ((eq (car sexp) 'array)
    `(array ,(cadr sexp) ,(sck-ast-sexp-to-type env (caddr sexp))))
   ;; <FunctionType>
   ((eq (car sexp) 'function)
    `(function
      ,(mapcar #'(lambda (x) (sck-ast-sexp-to-type env x))
	       (cadr sexp))
      ,(sck-ast-sexp-to-type env (caddr sexp))))
   ;; <StructType> | <UnionType>
   ((memq (car sexp) '(struct union))
    `(,(car sexp)
      ,(sck-ast-sexp-to-env env (cadr sexp))
      ,@(cddr sexp)))
   ;; <EnumType>
   ((eq (car sexp) 'enum)
    sexp)
   ;; <QualifiedType> ::= (const <Type>) | (volatile <Type>)
   ((memq (car sexp) '(const volatile))
    `(,(car sexp)
      ,(sck-ast-sexp-to-env env (cadr sexp))))
   ;;
   (t
    (assert nil))))

(defun sck-ast-sexp-to-ident-value (env sexp)
  ;; <IdentValue> ::= nil | <Integer> | <Expr> | <Inits> | <Lambda>
  (cond
   ;; 
   ((null sexp) nil)
   ;; <Integer>
   ((bn-integer-p sexp) sexp)
   ;; <Inits> ::= (inits <<Expr>|(zeros <Integer>)>..)
   ((eq (car sexp) 'inits)
    `(inits ,@(mapcar #'(lambda (x)
			  (cond
			   ((eq (car-safe x) 'zeros)
			    x)
			   (t
			    (sck-ast-sexp-to-ident-value env x))))
		      (cdr sexp))))
   ;; <Lambda> ::= (lambda <Type> (<Declaration>..) <BlockStmt>)
   ((eq (car sexp) 'lambda)
    (let ((paramenv (sck-ast-sexp-to-env env (caddr sexp))))
      `(lambda
	 ,(sck-ast-sexp-to-type env (cadr sexp))
	 ,paramenv
	 ,(sck-ast-sexp-to-stmt paramenv (cadddr sexp)))))
   ;; <Expr>
   (t
    (sck-ast-sexp-to-expr env sexp))))

(defun sck-ast-sexp-to-expr (env sexp)
  (if (null sexp)
      nil
    (case (car sexp)
      ;; (const <Type> <<Number>|(<Number>..)>)
      (const
       (make-sck-ast-expr :op 'const
			  :type (sck-ast-sexp-to-type env (cadr sexp))
			  :value (caddr sexp)))
      ;; (vref <Type> <IdentName>) 
      (vref
       (let ((ident (sck-ast-search-ident env (caddr sexp))))
	 (assert ident)
	 (make-sck-ast-expr :op 'vref
			    :type (sck-ast-sexp-to-type env (cadr sexp))
			    :value ident)))
      ;; (member <Type> <Expr:suexpr> (vref <Type> <Name>))
      (member
       (let* ((type (sck-ast-sexp-to-type env (cadr sexp)))
	      (suexpr (sck-ast-sexp-to-expr env (caddr sexp)))
	      (sutype (sck-ast-struct-or-union-type-p
		       (sck-ast-expr-type suexpr))))
	 (assert sutype)
	 (make-sck-ast-expr :op 'member
			    :type type
			    :args (list suexpr
					(sck-ast-sexp-to-expr
					 (cadr sutype)
					 (cadddr sexp))))))
      ;; (<Op> <Type> <Expr>..)
      (t
       (make-sck-ast-expr :op (car sexp)
			  :type (sck-ast-sexp-to-type env (cadr sexp))
			  :args (mapcar #'(lambda (e)
					    (sck-ast-sexp-to-expr env e))
					(cddr sexp)))))))

(defun sck-ast-sexp-to-stmt (env sexp)
  (case (car sexp)
    ;; (block (<Declaration>..) <Stmt>..)
    ((block)
     (let ((newenv (sck-ast-sexp-to-env env (cadr sexp))))
       `(block
	    ,newenv
	  ,@(mapcar #'(lambda (s) (sck-ast-sexp-to-stmt newenv s))
		    (cddr sexp)))))
    ;; (if <Expr> <Stmt> <Stmt>)
    ((if)
     `(if ,(sck-ast-sexp-to-expr env (cadr sexp))
	  ,(sck-ast-sexp-to-stmt env (caddr sexp))
	,(sck-ast-sexp-to-stmt env (cadddr sexp))))
    ;; (switch <StmtName> <Expr> <Stmt>)
    ((switch)
     `(switch ,(cadr sexp)
	      ,(sck-ast-sexp-to-expr env (caddr sexp))
	      ,(sck-ast-sexp-to-stmt env (cadddr sexp))))
    ;; (case <StmtName> <Fixnum> <Stmt>)
    ((case)
     `(case ,(cadr sexp)
	,(caddr sexp)
	,(sck-ast-sexp-to-stmt env (cadddr sexp))))
    ;; (default <StmtName> <Stmt>)
    ((default)
     `(default ,(cadr sexp)
	,(sck-ast-sexp-to-stmt env (caddr sexp))))
    ;; (while <StmtName> <Expr> <Stmt>)
    ((while)
     `(while ,(cadr sexp)
	,(sck-ast-sexp-to-expr env (caddr sexp))
	,(sck-ast-sexp-to-stmt env (cadddr sexp))))
    ;; (do <StmtName> <Expr> <Stmt>)
    ((do)
     `(do ,(cadr sexp)
	  ,(sck-ast-sexp-to-expr env (caddr sexp))
	,(sck-ast-sexp-to-stmt env (cadddr sexp))))
    ;; (for <StmtName> (<ExprOpt> <ExprOpt> <ExprOpt>) <Stmt>)
    ((for)
     `(for ,(cadr sexp)
	   (,(sck-ast-sexp-to-expr env (car (caddr sexp)))
	    ,(sck-ast-sexp-to-expr env (cadr (caddr sexp)))
	    ,(sck-ast-sexp-to-expr env (caddr (caddr sexp))))
	   ,(sck-ast-sexp-to-stmt env (cadddr sexp))))
    ;; (label <Label> [<Stmt>])
    ((label)
     `(label ,(cadr sexp)
	     ,(sck-ast-sexp-to-stmt env (caddr sexp))))
    ;; (goto  <Label>)
    ((goto)
     sexp)
    ;; (break <StmtName>)
    ((break)
     sexp)
    ;; (continue <StmtName>)
    ((continue)
     sexp)
    ;; (return <ExprOpt>)
    ((return)
     `(return ,(sck-ast-sexp-to-expr env (cadr sexp))))
    ;; nil
    ((nil)
     nil)
    ;; <Expr>
    (t
     (sck-ast-sexp-to-expr env sexp))))

				    
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Pretty printer
;;;  LIR ̃veB[v^Ɠ\ɂ

(defun sck-ast-pp-ast (ast &optional col)
  (with-temp-buffer
    (sck-ast-pp-ast/rec ast (or col 0))
    (buffer-string)))

(defun sck-ast-pp-ast/rec (ast col)
  (sck-ast-pp-indent-to col)
  (cond
   ((atom ast) (insertf "%S" ast))
   ;; ((...) ...)
   ((consp (car ast))
    (sck-ast-pp-ast/list ast col '(1)))
   ;; reduce Őꂽx͑OɋsB
   ;; (label <String>)
   ((and (eq (car ast) 'label)
	 (= 2 (length ast)))
    (delete-region (point) (progn (beginning-of-line) (point)))
    (newline)
    (sck-ast-pp-indent-to col)
    (insertf "%S" ast))
   ;; ̂͐܂ȂB
   ((<= (length (format "%S" ast)) 45)
    (insertf "%S" ast))
   ;; struct/union
   ((memq (car ast) '(struct union))
    (sck-ast-pp-ast/list ast col '(nil t 1 t)))
   ;; (inits ...)
   ((eq (car ast) 'inits)
    (sck-ast-pp-ast/list ast col '(nil 7)))
   ;; (var ...) (type ...)
   ((memq (car ast) '(var type))
    (sck-ast-pp-ast/list ast col '(nil t t 1)))
   ;; (pointer/const/volatile ...)
   ((and (memq (car ast) '(pointer const volatile))
	 ;;  const Ƌʂ邽
	 (= (length ast) 2))
    (sck-ast-pp-ast/list ast col '(nil 1)))
   ;; (while/for/do ...)
   ((memq (car ast) '( while for do switch))
    (sck-ast-pp-ast/list ast col '(nil t t 1)))
   ;; (if ... (goto ...) (goto ...))
   ((and (eq (car ast) 'if)
	 (eq (car (elt ast 2)) 'goto)
	 (eq (car (elt ast 3)) 'goto))
    (sck-ast-pp-ast/list ast col '(nil t 4 t)))
   ;; (lambda ...)
   ((eq (car ast) 'lambda)
    (sck-ast-pp-ast/list ast col '(nil t 3 1)))
   ;; ftHg
   ((<= 2 (length ast))
    (sck-ast-pp-ast/list ast col '(nil t 1)))
   (t
    (insertf "%S" ast))))

(defun sck-ast-pp-ast/list (list col styles)
  ;; STYLES: (<Style> <Style>..)
  ;; <Style> ::= nil | <Fixnum> | t
  (assert (consp styles))
  ;; ֐`̑Oɂ͉sB
  (when (and (eq (car list) 'var)
	     (eq (car-safe (caddr list)) 'function))
    (delete-region (point) (progn (beginning-of-line) (point)))
    (newline)
    (sck-ast-pp-indent-to col))
  ;;
  (insert "(")
  (let ((style (pop styles)))
    (dolist (x list)
      (cond
       ;; nil Ȃ獡̏ꏊɒuB
       ((null style)
	(sck-ast-pp-ast/rec x (current-column)))
       ;; natnum Ȃ (+ col style) ɒuB
       ((natnump style)
	(sck-ast-pp-ast/rec x (+ col style)))
       ;; t ȂsEɒuB
       ((eq style t)
	(sck-ast-pp-ast/rec x (1+ (current-column))))
       (t
	(assert nil)))
      (when styles
	(setq style (pop styles)))))
  (insert ")"))

(defun sck-ast-pp-indent-to (col)
  (when (< col (current-column))
    (newline))
  (indent-to col))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; cps ̏o͂veB[vgAƂĕԂB

(defun sck-ast-pp-output (output)
  "cps ̏o͂veB[vgAƂĕԂB"
  (with-temp-buffer
    (sck-letalist output
	(ast messages)
      (insert "((ast\n")
      (insertf "%s" (sck-ast-pp-ast ast 2))
      (insert ")\n\n")
      (insert "(messages (\n")
      (dolist (msg messages)
	(insertf "%S\n" msg))
      (insert ")))\n"))
    (buffer-string)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; font-lock for AST

(defun sck-ast-font-lock ()
  (setq font-lock-defaults
	'((lisp-font-lock-keywords)
	  nil nil (("+-*/.<>=!?$%_&~^:" . "w")) beginning-of-defun
	  (font-lock-mark-block-function . mark-defun)))
  (font-lock-mode t))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ELIM v^̐ݒB

(when (featurep 'elim)
  (elim-add-printer
   #'sck-ast-ident-p
   #'(lambda (x) (sck-ast-pp-ast (sck-ast-ident-to-sexp x))))
  (elim-add-printer
   #'sck-ast-env-p
   #'(lambda (x) (sck-ast-pp-ast (sck-ast-env-to-sexp x))))
  (elim-add-printer
   #'sck-ast-expr-p
   #'(lambda (x) (sck-ast-pp-ast (sck-ast-expr-to-sexp x)))))


;;;
(provide 'sck-ast)
