;;;; fp cc
;;;; LastChange: Wed Apr 11 05:29:45 2007

(defun sck-cc-demo-file (file)
  (interactive "fC-Source: ")
  (with-temp-buffer
    (insert-file-contents file)
    (setq buffer-file-name file)
    (set-buffer-modified-p nil)
    (sck-cc-demo-buffer)))

(defun sck-cc-demo-buffer ()
  (interactive)
  (sck-cc-demo-region (point-min) (point-max)))

(defun sck-cc-demo-region (beg end)
  (interactive "r")
  (sck-cc-demo-str (buffer-substring-no-properties beg end)
		   (file-name-nondirectory
		    (file-name-sans-extension
		     (or (buffer-file-name)
			 (buffer-name))))))


(defun sck-cc-demo-header (fmt &rest args)
  (let ((str (apply 'format fmt args)))
    (insertf "%s\n;;; %s:\n" (make-string  (- (frame-width) 1) ?\;) str)
    (message "%s" str)))


(defun sck-cc-demo-str (cstr &optional name)
  " CSTR  C ̃\[XƂĒԌʂ\͂A
AZuo͂܂ŕϊsB"
  (let* ((cppstr (let ((tmpf (make-temp-file "SCK")))
		   (write-region cstr 0 tmpf nil 'nomsg)
		   (unwind-protect
		       (sck-cc-cpp tmpf)
		     (delete-file tmpf))))
	 (tokens (sck-clex `((name ,(or name "stdin")) (source ,cppstr))))
	 (plain-ast (sck-cpps tokens))
	 (ast (sck-cps plain-ast))
	 (core-ast (sck-ast-reduce ast))
	 (lirgenout (sck-lirgen core-ast)))
    (sck-letalist lirgenout
	(lir messages)
      (with-display-buffer "*sck-shcc*"
	(if (sck-ast-error-in-messages-p messages)
	    ;; V^bNXG[
	    (dolist (msg messages)
	      (insertf "%s\n" msg))
	  ;;
	  (sck-cc-demo-header "C-Source (pre-processed)")
	  (insertf "%s\n\n" cppstr)
	  ;;
	  (sck-cc-demo-header "C-Tokens")
	  (insertf "%s\n\n" (sck-clex-pp-output tokens))
	  ;;
	  (sck-cc-demo-header "C-PlainAST")
	  (insertf "%s\n\n" (sck-cpps-pp-output plain-ast))
	  ;;
	  (sck-cc-demo-header "AST")
	  (insertf "%s\n\n" (sck-ast-pp-output ast))
	  ;;
	  (sck-cc-demo-header "Core-AST")
	  (insertf "%s\n\n" (sck-ast-pp-output core-ast))
	  ;;
	  (sck-cc-demo-header "LIR")
	  (insertf "%s\n\n" (sck-lir-pp lir))
	  ;;
	  ;; R[h
	  ;;
	  (let* ((lir1 (sck-lir-foldconst-module (copy-tree lir)))
		 (lir2 (sck-tmd-restruct-module (copy-tree lir1)))
		 (defs (sck-lir-module-defs lir2))
		 (symtab (sck-lir-module-symtab lir2))
		 (body (sck-lir-module-body lir2)))
	    ;;
	    (sck-cc-demo-header "LIR after simple optimizations")
	    (insertf "%s\n\n" (sck-lir-pp lir1))
	    ;;
	    (sck-cc-demo-header "LIR after restruct")
	    (insertf "%s\n\n" (sck-lir-pp lir2))
	    ;;
	    (dorest (b body)
	      (let ((func (car b)))
		;; func ֐ <Lfunc> ȂR[hs func uB
		;; łȂ func  <Ldata> Ȃ̂ŁAȂB
		(when (eq (car-safe func) 'function)
		  ;; func : <Lfunc> ̃R[h
		  (block 'regassign
		    (while t
		      ;;
		      ;; ߑI
		      ;;
		      (setq func (sck-tmd-instsel func))
		      ;;
		      (sck-cc-demo-header "Lfunc `%s' after instsel"
					  (sck-lir-function-name func))
		      (insertf "%s\n\n" (sck-lir-pp func))
		      ;;
		      ;; WX^AP[Vyъ蓖
		      ;;
		      (sck-letalist (sck-regassign `((module-defs ,defs)
						     (lfunc ,func)))
			  (lfunc spilled-registers)
			(setq func lfunc)
			;;
			(sck-cc-demo-header "Lfunc `%s' after regassign"
					    (sck-lir-function-name func))
			(if spilled-registers
			    (insertf ";;; spilled-registers = %S\n"
				     spilled-registers)
			  (insertf ";;; Register assignment success!\n"))
			(insertf "%s\n\n" (sck-lir-pp func))
			;;
			;; spilled-registers ΏI
			;;
			(unless spilled-registers
			  (return-from 'regassign)))))
		  ;;
		  (setf (car b) func))))
	    ;;
	    ;; AZuo
	    ;;
	    (sck-cc-demo-header "Assembler output")
	    (insert (sck-tmd-asmout
		     `(module ,(sck-lir-module-name lir2)
			      ,defs
			      ,symtab
			      ,@body)))))))))


;;
(provide 'sck-cc-demo)
