;;;; C $B%3%s%Q%$%i%I%i%$%P(B
;;;; LastChange: Wed Apr 11 02:10:11 2007
;;;; $Id: sck-cc.el,v 1.7 2007/03/29 09:04:11 heavy Exp $

(require 'sck-lirgen)
(require 'sck-regassign)
(require 'sck-cc-demo)


(defvar sck*cc-buffer "*sck-cc*")
  
(defun sck-cc-file (file)
  (interactive "f:C-File ")
  (with-temp-buffer
    (insert-file-contents file)
    (sck-cc-buffer)))

(defun sck-cc-buffer ()
  (interactive)
  (sck-cc-region (point-min) (point-max)))

(defun sck-cc-region (beg end)
  (interactive "r")
  (sck-cc-str (buffer-substring-no-properties beg end)))

(defun sck-cc-str (str)
  (with-display-buffer sck*cc-buffer
    (insert (sck-cc-str-to-asm-str str))))
;;
(defun sck-cc-str-to-asm-str (str &optional file)
  ;; $B%W%j%W%m%;%C%5$rDL$7$?(BC$B%3!<%I$r%3%s%Q%$%k$7$?7k2L$rJV$9(B
  (message "Parsing...")
  (sck-letalist 
      (sck-lirgen
       (sck-ast-reduce
	(sck-cps
	 (sck-cpps
	  (sck-clex
	   (list (list 'name (if file file "stdin.c"))
		 (list 'source str)))))))
      ((lir module)
       messages)
    (concat 
     (prog1
	 (with-temp-buffer
	   (dolist (mess messages)
	     (message "%s" mess)
	     (insert (format "# %s\n" mess)))
	   (buffer-string))
       (message "Code generating..."))
     ;; $B%3!<%I@8@.(B
     ;; module1 : resutruct $B:Q$_%b%8%e!<%k(B
     ;; module2 : $B%3!<%I@8@.:Q$_%b%8%e!<%k(B
     (let* ((module1 (sck-tmd-restruct-module module))
	    (module-defs (sck-lir-module-defs module1))
	    (module-symtab (sck-lir-module-symtab module1))
	    (module2 `(module ,(sck-lir-module-name module1)
			      ,module-defs
			      ,module-symtab)))
       ;; module1 : (module <Name> <Defs> <GlobalSymtab> {<Lfunc> | <Ldata>})
       ;; $B$N(B <Lfunc> $B$H(B <Ldata> $B$N=hM}(B
       (dolist (func (sck-lir-module-body module1))
	 ;; func $B$,4X?t(B <Lfunc> $B$J$i%3!<%I@8@.$r9T$$!"(Bfunc $B$rCV$-49$($k!#(B
	 ;; $B$=$&$G$J$1$l$P(B func $B$O(B <Ldata> $B$J$N$G!"2?$b$;$:(B asmout $B$KEO$9!#(B
	 (when (eq (car-safe func) 'function)
	   ;; func : <Lfunc> $B$N%3!<%I@8@.(B
	   (block 'regassign
	     (while t
	       ;; $BL?NaA*Br(B
	       (setq func (sck-tmd-instsel func))
	       ;; $B%l%8%9%?%"%m%1!<%7%g%s5Z$S3d$jEv$F(B
	       (sck-letalist (sck-regassign `((module-defs ,module-defs)
					      (lfunc ,func)))
		   (lfunc spilled-registers)
		 (setq func lfunc)
		 ;; spilled-registers $B$,L5$1$l$P=*N;(B
		 (unless spilled-registers
		   (return-from 'regassign))))))
	 ;; lfunc : <Lfunc>|<Ldata> $B$r:G=*E*$J(B module2 $B$KDI2C$9$k!#(B
	 (setq module2 (append module2 (list func))))
       ;; $B%"%;%s%V%i=PNO(B
       (sck-tmd-asmout module2)))))

;;;
;;; 
;;;
(defun sck-cc-cpp (file)
  ;; cpp $B$N7k2L$rJ8;zNs$GJV$9(B
  (let ((errfile (expand-file-name "_sck_cc_cpperr" temporary-file-directory)))
    (prog1
	(with-temp-buffer
	  (call-process "cpp" file (list t errfile) nil)
	  ;; @@@ mcpp $B$r;H$&$H(B warning $B$,=P$F$7$^$&(B
	  ;;(call-process "mcpp" file (list t errfile) nil)
	  (buffer-string))
      (when (file-exists-p errfile)
	(unless (zerop (elt (file-attributes errfile) 7))
	  (message "%s" 
		   (with-temp-buffer
		     (insert-file-contents errfile)
		     (buffer-string))))
	;; $B2?$i$+$NM}M3$G%U%!%$%k$r>C$;$J$$>l9g$G$b%(%i!<$H$7$J$$(B
	(condition-case err
	    (delete-file errfile)
	  (error))))))

(defun sck-cc-asm-link (files outfile)
  '(message "files: %S outfile:%S" files outfile)
  (when outfile 
    (push outfile files)
    (push "-o" files))
  (with-temp-buffer
    (apply 'call-process
	   (append '("gcc" nil t nil)
		   (cons
		    (expand-file-name "gccmain.o" 
				      (concat sck=root-directory "/cc"))
		    files)))
    (let ((msg (buffer-string)))
      (when (> (length msg) 0)
	(message "%s" msg)))))

(defun sck-cc (file &optional options)
  ;; C$B%3%s%Q%$%i%I%i%$%P(B
  ;; options ::= (("-E") ...)
  (interactive "fC-File: ")
  (message "Pre-processing...")
  (let ((cpp (sck-cc-cpp file)))
    (if (assoc "-E" options)
	(princ cpp)
      (let ((asm (sck-cc-str-to-asm-str cpp file))
	    (asmfile (format "%s.s" (substring file 0 -2))))
	(with-temp-buffer
	  (insert asm)
	  (write-region (point-min) (point-max) asmfile nil 'nomsg))
	(message "\"%s\" Compile done." file)))))

(defun sck-cc-with-link (file &optional options)
  (interactive "fC-File: ")
  (sck-cc file options)
  (sck-cc-asm-link (list (format "%s.s" (substring file 0 -2))) nil))

;;;
;;; batch $B$G8F$P$l$k(B
;;;
(defun sck-cc-batch ()
  ;; % EMACS -batch -l ../sck.el -l sck-cc.el -f sck-cc-batch $argv
  (unless command-line-args-left
    (message "Usage: sck-cc [options...] srcfile")
    (kill-emacs 1))
  (let ((args command-line-args-left)
	srcfiles
	outfile
	options)
    (while args
      (let ((arg (car args)))
	(cond
	 ((string= (substring arg -2 nil) ".c")
	  (push arg srcfiles))
	 ;; outfile
	 ((string= arg "-o")
	  (setq args (cdr args))
	  (setq outfile (car args))
	  (push (list arg (car args)) options))
	 ;; options...
	 (t
	  (push (list arg) options))))
      (setq args (cdr args)))
    (dolist (file srcfiles)
      (sck-cc file options))
    (message "Assembling & Linking...")
    (when srcfiles
      (sck-cc-asm-link
       (mapcar #'(lambda (x) (format "%s.s" (substring x 0 -2))) srcfiles)
       outfile)))
  (kill-emacs 0))

(provide 'sck-cc)
