;;;; C ̃LVJAiCU
;;;; LastChange: Thu Apr 12 23:39:27 2007

;;; Export:
;;; sck-clex : <ClexInput> -> <ClexOutput>
;;; sck-clex-token-to-string : <Token> -> <String>
;;;
;;; <ClexInput> ::= ((nane <String>) (source <String>))
;;; 
;;; <ClexOutput> ::= ((name <String>)
;;;                   (tokens (<Token>..))
;;;                   (file-names (<String>..))
;;;                   (messages (<String>..)))
;;;
;;; <Token> ::= <Keyword> | <Punctuation> | <Ident> | <Literal> | <PosInfo>
;;; <Keyword> ::= <Symbol>
;;; <Punctuation> ::= <Symbol>
;;; <Ident> ::= (ident <String>)
;;; 
;;; <Literal> ::= <Iconst> | <Fconst> | <Cconst> | <Sconst>
;;; <Iconst> ::= (iconst <Integer> <ISuffixes>)
;;; <Fconst> ::= (fconst (<Ipart> <Fpart> <Epart>) <FSuffixes>)
;;; <Cconst> ::= (cconst <CharCode> <CSuffixes>)
;;; <Sconst> ::= (sconst (<CharCode>..) <SSuffixes>)
;;; <{I,F,E}part> ::= <Integer>
;;; <CharCode> ::= <Fixnum>
;;; <ISuffixes> ::= [L | LL | U | UL | ULL]
;;; <FSuffixes> ::= [F | L | LL]
;;; <CSuffixes> ::= [L]
;;; <SSuffixes> ::= [L]
;;; 
;;; <PosInfo> ::= ($ <FileNo> <LineNo>)
;;; <FileNo> ::= <Fixnum> ; index of file-names
;;; <LineNo> ::= <Fixnum>


(require 'sck-clex-readlit)


(defconst sck=clex-input-sample
  "# 1 \"demo1.c\" 
/* puncs1 */ ! % & ( ) * + , - . / : ; < = > ? [ ] ^ { | } ~ 
/* puncs2 */ -> ++ -- << >> <= >= == != && || *= /= %= += -= &= ^= |= 
/* puncs3 */ <<= >>= ...
/* keywords */ auto break case
/* identifiers */ foo1 _bar2 _b_a_z_3

# 100 \"demo2.c\"
/* iconst */ 0 012 12345 1234567890123 0xabcdefabcd 123l 1LL 1U 1uL 1ULl
/* fconst */ 3.14 09.1F .1L 1.2e3 3.4e-5 6.e-10 6e1
/* cconst */ 'a' '\\'' '\\\"' '\\\\' '\\141' '\\x62'
/* sconst */ \"\\'\\\"\\?\\\\\\a\\b\\f\\n\\r\\t\\v\" \"'\" \"\\141\\x62\"

# 100 \"errors.c\"
/* iconst */ 09 0UU 0LULL 1.2U 3.1LL
/* fconst */ 10e 10e- 0xabe.3 
/* cconst */ '\\Z' 'abc' 'd
/* sconst */ \"\\x\" \"unterminated
/* illegal characters */ \\ @ $ ` #
/* unterminated comment
")


(defun sck-clex-demo ()
  (interactive)
  (sck-clex-str sck=clex-input-sample "demo" t))

(defun sck-clex-file (file &optional name)
  (interactive "f(pre processed) C source file: ")
  (with-temp-buffer
    (insert-file-contents file)
    (setq buffer-file-name (or name file))
    (set-buffer-modified-p nil)
    (sck-clex-buffer (interactive-p))))

(defun sck-clex-buffer (&optional display)
  (interactive)
  (sck-clex-region (point-min) (point-max)
		   (or display (interactive-p))))
  
(defun sck-clex-region (beg end &optional display)
  (interactive "r")
  (sck-clex-str (buffer-substring-no-properties beg end)
		(file-name-nondirectory
		 (file-name-sans-extension
		  (or (buffer-file-name)
		      (buffer-name))))
		(or display (interactive-p))))

(defun sck-clex-str (str &optional name display)
  (let* ((output (sck-clex `((name ,(or name "stdin"))
			     (source ,str)))))
    (when display
      (with-display-buffer "*sck-clex*"
	(emacs-lisp-mode)
	(insert (sck-clex-pp-output output))))
    ;;
    output))

(defun sck-clex-pp-output (output)
  "clex ̏o OUTPUT ̃veB[v^"
  (sck-letalist output
      (name tokens file-names messages)
    (with-temp-buffer
      (insertf "((name %S)\n\n(tokens %s)\n\n"
	       name (sck-clex-tokens-pp tokens file-names))
      ;;
      (insertf "(file-names  %S)\n\n" file-names)
      (insert "(messages (\n")
      (dolist (msg messages) (insertf "%S\n" msg))
      (insert ")))\n")
      ;;
      (buffer-string))))

(defun sck-clex-tokens-pp (tokens file-names)
  "g[NXg̃veB[v^"
  (with-temp-buffer
    (insert "(\n")
    (dolist (tk tokens)
      (insertf "%-30S ; " tk)
      (if (eq (car-safe tk) '$)
	  (insertf "# %d %S\n" (caddr tk) (elt file-names (cadr tk)))
	(insertf "%s\n" (sck-clex-token-to-string tk))))
    (insert ")")
    (buffer-string)))
    




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; CLEX C

(defun sck-clex (input)
  "sck-clex : <ClexInput> -> <ClexOutput>

<ClexInput> ::= ((name <String>) (source <String>))
<ClexOutput> :: ((name <String>)
                 (tokens (<Token>..))
                 (file-names (<String>..))
                 (messages (<String>..)))"
  (sck-letalist input
      (name source)
    (let ((handle (sck-clex-open name source))
	  (tokensq (make-que))
	  (file-idtab (make-idtab))
	  (pos-prev nil)
	  (pos nil))
      (loop (sck-clex-get handle)
	    (setq pos (list '$
			    (idtab-obj-to-id file-idtab
					     (sck-clex-handle-file handle))
			    (sck-clex-handle-line handle)))
	    (unless (equal pos-prev pos)
	      (append-que tokensq pos)
	      (setq pos-prev pos))
	    (unless (eq (sck-clex-handle-token handle) 'error)
	      ;;  A-List C^tF[Xł error g[N̓XLbvB
	      ;; Input ŃXLbvȂ悤Ȏw肪ĂȂA
	      ;; ʕKvȂB
	      (append-que tokensq
			    (case (sck-clex-handle-token handle)
			      ((iconst fconst cconst sconst)
			       (list* (sck-clex-handle-token handle)
				      (sck-clex-handle-value handle)
				      (sck-clex-handle-suffs handle)))
			      (ident
			       (list 'ident (sck-clex-handle-value handle)))
			      (t
			       (sck-clex-handle-token handle)))))
	    (when (eq (sck-clex-handle-token handle) 'eof)
	      (return nil)))
      (sck-clex-close handle)
      ;;
      `((name ,name)
	(tokens ,(que-to-list tokensq))
	(file-names ,(idtab-to-list file-idtab))
	(messages ,(que-to-list (sck-clex-handle-msgs handle)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; handle ̃C^tF[X
;;; : ̃C^tF[Xł̓G[g[N error ƂĕԂB

(defstruct sck-clex-handle
  ;; \[Xobt@
  srcbuf
  ;; {Iȏ
  token           ; g[NV{
  image           ; ؂g[N̂܂܂̕
  value           ; 萔̒lBbn V^bNX̕
  suffs           ; 萔̌ɕt L Ƃ U ƂiV{̃Xgj
  ;; ʒu
  (file "undefined.c")  ; t@C
  (line 1)        ; sԍifixnumj
  ;; ͒̃G[bZ[W̃L[
  (msgs (make-que)))

(defun sck-clex-open (name str)
  ;; ͑Ώۂ̕ str  sck-clex-handle ^
  ;; nhIuWFNgB
  (make-sck-clex-handle
   :srcbuf (with-current-buffer
	       (generate-new-buffer "*sck-clex-handle-srcbuf*")
	     (insert str)
	     (goto-char (point-min))
	     (current-buffer))
   :file (concat name ".c")))

(defun sck-clex-close (handle)
  ;; nhjB
  (kill-buffer (sck-clex-handle-srcbuf handle)))

(defun sck-clex-error (handle fmt &rest args)
  ;; ͒̃bZ[Wnh msgs ֏o͂B
  (append-que (sck-clex-handle-msgs handle)
		(format "%s:%d: %s"
			(sck-clex-handle-file handle)
			(sck-clex-handle-line handle)
			(apply 'format fmt args))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; LVJAiCUC

(defun sck-clex-get (handle)
  ;; LVJAiCUnh handle g[Nǂ݁A
  ;; handle ֋L^Ahandle ɃZbgg[NiV{jԂB
  ;;
  (let ((case-fold-search nil))
    (setf (sck-clex-handle-token handle) nil
	  (sck-clex-handle-image handle) ""
	  (sck-clex-handle-value handle) nil
	  (sck-clex-handle-suffs handle) nil)
    ;;
    (with-current-buffer (sck-clex-handle-srcbuf handle)
      ;; 󔒂ǂݔ΂A
      (sck-clex-get-skip handle)
      (let ((beg (point)))
	;; |Cgȉ̃g[N͂B
	(sck-clex-get-token handle)
	;; g[NC[WZbgB
	(setf (sck-clex-handle-image handle)
	      (buffer-substring beg (point))))
      ;; g[NV{ԂB
      (sck-clex-handle-token handle))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 󔒓ǂݔ΂ # ̏

(defun sck-clex-get-skip (handle)
  ;; RgA󔒕A#... iKvȏjXLbvB
  (block break
    (while t
      (skip-chars-forward " \t")
      (case (following-char)
	(?\n
	 (incf (sck-clex-handle-line handle))
	 (forward-char 1))
	(?#
	 (let ((pos (point)))
	   (beginning-of-line)
	   (skip-chars-forward " \t")
	   (cond
	    ;; sƂ̊Ԃ󔒂݂̂Ȃ #-directive
	    ((= (point) pos)
	     (sck-clex-get-sharp handle))
	    ;; łȂΕs #
	    (t
	     (goto-char pos)
	     (return-from break)))))
	(?/
	 (if (looking-at "/\\*")
	     (sck-clex-get-skip-comment handle)
	   (return-from break)))
	(t
	 (return-from break))))))

(defun sck-clex-get-skip-comment (handle)
  ;; Rgǂݔ΂B
  (let ((beg (point)))
    (if (search-forward "*/" nil :move-to-limit)
	(incf (sck-clex-handle-line handle)
	      (sck-clex-count-newlines beg (point)))
      (sck-clex-error handle "unterminated comment"))))

(defun sck-clex-count-newlines (beg end)
  ;; [W̉s𐳂JEgB
  ;; (with-temp-buffer (insert "a\nb") (sck-clex-count-newlines 1 (point-max)))
  ;; => 1
  (save-excursion
    (save-restriction
      (narrow-to-region beg end)
      (goto-char (point-min))
      (- (buffer-size) (forward-line (buffer-size))
	 ;; forward-line ɂs𐔂@ count-lines ɂB
	 ;; ̒`ł forward-line ̎dlɂ萳m
	 ;; [W̉sJEgoȂ̂ňȉ́u␳vKvB
	 (if (bolp) 0 1)))))

(defun sck-clex-get-sharp (handle)
  ;; # 10 "file" [[3] 4]
  ;;             ??????? 
  ;; ̏B
  (cond
   ((looking-at "# *\\([0-9]+\\) *\\(\"\\([^\"]+\\)\"\\)?")
    (setf (sck-clex-handle-line handle) (string-to-number (match-string 1)))
    (when (match-beginning 2)
      (setf (sck-clex-handle-file handle) (match-string 3))))
   ((looking-at "#pragma")
    (sck-clex-error handle "#pragma is not supported"))
   (t
    (sck-clex-error handle "unknown #... directive")))
  (forward-line 1))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; |Cgȉ̃g[N͂B

(defconst sck=clex-keyword-alist
  '(("auto" auto) ("break" break) ("case" case) ("char" char) ("const" const)
    ("continue" continue) ("default" default) ("do" do) ("double" double)
    ("else" else) ("enum" enum) ("extern" extern) ("far" far) ("float" float)
    ("for" for) ("goto" goto) ("if" if) ("int" int) ("long" long)
    ("register" register) ("return" return) ("short" short) ("signed" signed)
    ("sizeof" sizeof) ("static" static) ("struct" struct) ("switch" switch)
    ("typedef" typedef) ("union" union) ("unsigned" unsigned) ("void" void)
    ("volatile" volatile) ("while" while)))

(defconst sck=clex-punc-alist
  '(("!" exc) ("%" per) ("&" amp) ("(" lpa) (")" rpa) ("*" ast) ("+" pls)
    ("," com) ("-" hyp) ("." dot) ("/" sla) (":" col) (";" sem) ("<" les)
    ("=" equ) (">" gre) ("?" que) ("[" lbk) ("]" rbk) ("^" hat) ("{" lbr)
    ("|" pip) ("}" rbr) ("~" til)
    ("->" hypgre) ("++" plspls) ("--" hyphyp) ("<<" lesles) (">>" gregre)
    ("<=" lesequ) (">=" greequ) ("==" equequ) ("!=" excequ) ("&&" ampamp)
    ("||" pippip) ("*=" astequ) ("/=" slaequ) ("%=" perequ) ("+=" plsequ)
    ("-=" hypequ) ("&=" ampequ) ("^=" hatequ) ("|=" pipequ)
    ("<<=" leslesequ) (">>=" gregreequ) ("..." dotdotdot)))

(defconst sck=clex-keyword-hashtab
  ;; (gethash "int" sck=clex-keyword-hashtab) => int
  (let ((htab (make-hash-table :test 'equal)))
    (dolist (a sck=clex-keyword-alist)
      (setf (gethash (car a) htab) (cadr a)))
    htab))
  
(defconst sck=clex-punc-hashtab
  ;; (gethash "+=" sck=clex-punc-hashtab) => plsequ
  (let ((htab (make-hash-table :test 'equal)))
    (dolist (a sck=clex-punc-alist)
      ;; (put (cadr a) 'sck-clex-punc-string (car a))
      (setf (gethash (car a) htab) (cadr a)))
    htab))

(defconst sck=clex-punc-rex
  (regexp-opt (mapcar 'car sck=clex-punc-alist)))

(defconst sck=clex-keyword-and-punc-name-hashtab
  ;; (gethash 'int sck=clex-keyword-and-punc-name-hashtab)
  ;; (gethash 'plsequ sck=clex-keyword-and-punc-name-hashtab)
  (let ((htab (make-hash-table)))
    (dolist (a (append sck=clex-keyword-alist sck=clex-punc-alist))
      (setf (gethash (cadr a) htab) (car a)))
    htab))


(defun sck-clex-get-token (handle)
  ;; |Cgȉ̃g[Nǂ݁Ahandle ֋L^B
  ;; sȕ error Ƃg[NƂB
  (cond
   ;; ȉ݂̂̃N[ÝȀłȂ΂ȂȂB
   ;;
   ;; (iconst)A_(fconst)A萔(cconst)A萔(sconst)
   ((looking-at "\\.?[0-9]\\|L?['\"]")
    (destructuring-bind ((token value &rest suffs) msgs) (sck-clex-readlit)
      (setf (sck-clex-handle-token handle) token
	    (sck-clex-handle-value handle) value
	    (sck-clex-handle-suffs handle) suffs)
      (dolist (msg msgs) (sck-clex-error handle "%s" msg))))
   ;; ؂Li.foo  .123 ̋ɒӁj
   ((looking-at sck=clex-punc-rex)
    (setf (sck-clex-handle-token handle)
	  (gethash (match-string 0) sck=clex-punc-hashtab))
    (goto-char (match-end 0)))
   ;; ʎqL[[hiL"..." Ƃ̋ɒӁj
   ((looking-at "[a-zA-Z_][a-zA-Z0-9_]*")
    (let ((name (match-string 0)))
      (setf (sck-clex-handle-token handle)
	    (or (gethash name sck=clex-keyword-hashtab) ; keyword
		;; ident
		(progn
		  (setf (sck-clex-handle-value handle) name)
		  'ident)))
      (goto-char (match-end 0))))
   ;; obt@̏I
   ((eobp)
    (setf (sck-clex-handle-token handle) 'eof))
   ;; sȕ
   (t
    (sck-clex-error handle "illegal character: '%c'" (following-char))
    (setf (sck-clex-handle-token handle) 'error)
    (forward-char 1))))

    
(defun sck-clex-token-to-string (token)
  ;; t
  ;; <Token> -> <String>
  ;; token  ident, [ifcs]const ƂV{ł悢ibZ[WpjB
  ;; Example:
  ;;  (sck-clex-token-string 'int)
  ;;  (sck-clex-token-string 'plsequ)
  (cond
   ((eq (car-safe token) '$)
    (format "/* %S */" token))
   ((eq (car-safe token) 'ident)
    (cadr token))
   ((memq (car-safe token) '(iconst fconst cconst sconst))
    (sck-clex-readlit-unparse token))
   ((eq token 'eof)
    "/* eof */")
   (t
    ;; keyword or punc
    (assert (symbolp token))
    (let ((name (gethash token sck=clex-keyword-and-punc-name-hashtab)))
      (assert name)
      name))))

;;;
(provide 'sck-clex)
