;;;
;;; $Id: c-syntax-bnf.el,v 1.1 2007/02/17 06:42:14 heavy Exp $
;;;
;;; make sck-c-syntax-output.el
;;;
(require 'sck-yayacc-lib)
(require 'sck-yayacc-yacc)

(defconst sck=c-syntax
  '((object-language "elisp")
    (associativity (
		    (left ("||"))
		    (left ("&&"))
		    (left ("|"))
		    (left ("^"))
		    (left ("&"))
		    (left ("==" "!="))
		    (left ("<" ">" "<=" ">="))
		    (left ("<<" ">>"))
		    (left ("+" "-"))
		    (left ("*" "/" "%"))
		    ))
    (attribute-grammar 
     (
      ((start	translation_unit " EOF")
       (cons 'module 
	     (cons (sck-cparse-get-pos (car $0))
		   (cons "stdin.c" $0))))
      
      ((translation_unit	external_decl)
       (list $0))
      ((translation_unit	translation_unit external_decl)
       (append $0 (list $1)))
      ((translation_unit	" ERROR")
       (sck-cparse-error nil $0 "translation_unit: error."))
      ((translation_unit	translation_unit " ERROR")
       (sck-cparse-error $0 $1 "translation_unit: translation_unit error."))
      
      ((external_decl		function_definition)
       $0)
      ((external_decl 		decl)
       $0)

      ;; (function <Pos> <Declaration> <OldParamDecls> <Block>)
      ((function_definition	decl_specs declarator decl_list compound_stat)
       (sck-cparse-make-function $0 $1 $2 $3))
      ((function_definition	           declarator decl_list compound_stat)
       (sck-cparse-make-function nil $0 $1 $2))
      ((function_definition	decl_specs declarator		compound_stat)
       (sck-cparse-make-function $0 $1 nil $2))
      ((function_definition	           declarator           compound_stat)
       (sck-cparse-make-function nil $0 nil $1))

      ;; (declaration <Pos> <Specifiers> (<Declarator> <<Initializer>|nil>)..)
      ((decl			decl_specs init_declarator_list ";")
       (let ((decl (sck-cparse-make-decl $0 $1)))
	 (sck-cparse-set-typedef-decl decl)
	 decl))
      ((decl 			decl_specs			";")
       (sck-cparse-make-decl $0 nil))

      ;;
      ((decl_list		decl)
       (list $0))
      ((decl_list		decl_list decl)
       (append $0 (list $1)))
      ((decl_list		decl_list " ERROR")
       (sck-cparse-error $0 $1 "decl_list: decl_list error."))

      ;;
      ((decl_specs		storage_class_spec decl_specs)
       (append (list $0) $1))
      ((decl_specs		storage_class_spec)
       (list $0))
      ((decl_specs		type_spec decl_specs)
       (append (list $0) $1))
      ((decl_specs		type_spec)
       (list $0))
      ((decl_specs		type_qualifier decl_specs)
       (append (list $0) $1))
      ((decl_specs		type_qualifier)
       (list $0))

      ;;
      ((storage_class_spec	"auto")
       (sck-cparse-make-keyword $0))
      ((storage_class_spec	"register")
       (sck-cparse-make-keyword $0))
      ((storage_class_spec	"static")
       (sck-cparse-make-keyword $0))
      ((storage_class_spec	"extern")
       (sck-cparse-make-keyword $0))
      ((storage_class_spec	"typedef")
       (sck-cparse-make-keyword $0))

      ;;
      ((type_spec		"void")
       (sck-cparse-make-keyword $0))
      ((type_spec		"char")
       (sck-cparse-make-keyword $0))
      ((type_spec		"short")
       (sck-cparse-make-keyword $0))
      ((type_spec		"int")
       (sck-cparse-make-keyword $0))
      ((type_spec		"long")
       (sck-cparse-make-keyword $0))
      ((type_spec		"float")
       (sck-cparse-make-keyword $0))
      ((type_spec		"double")
       (sck-cparse-make-keyword $0))
      ((type_spec		"signed")
       (sck-cparse-make-keyword $0))
      ((type_spec		"unsigned")
       (sck-cparse-make-keyword $0))
      ((type_spec		struct_or_union_spec)
       $0)
      ((type_spec		enum_spec)
       $0)
      ((type_spec		"typedef_name")
       (sck-cparse-make-typedef-name $0))

      ;;
      ((type_qualifier		"const")
       (sck-cparse-make-keyword $0))
      ((type_qualifier		"volatile")
       (sck-cparse-make-keyword $0))
      
      ;; (<struct|union> <Pos> <<Ident>|nil> <Declaration>..)
      ((struct_or_union_spec	struct_or_union "id" "{" struct_decl_list "}")
       (sck-cparse-make-struct-or-union $0 $1 $3))
      ((struct_or_union_spec	struct_or_union	     "{" struct_decl_list "}")
       (sck-cparse-make-struct-or-union $0 nil $2))
      ((struct_or_union_spec	struct_or_union "id")
       (sck-cparse-make-struct-or-union $0 $1 nil))

      ;; (<struct|union> <Pos>)
      ((struct_or_union		"struct")
       (list 'struct (sck-cparse-get-pos $0)))
      ((struct_or_union		"union")
       (list 'union (sck-cparse-get-pos $0)))

      ;; 
      ((struct_decl_list	struct_decl)
       (list $0))
      ((struct_decl_list	struct_decl_list struct_decl)
       (append $0 (list $1)))
      ((struct_decl_list	struct_decl_list " ERROR")
       (sck-cparse-error $0 $1 "struct_decl_list: struct_decl_list error."))

      ;;
      ((init_declarator_list	init_declarator)
       (list $0))
      ((init_declarator_list	init_declarator_list "," init_declarator)
       (append $0 (list $2)))
      ((init_declarator_list	init_declarator_list " ERROR")
       (sck-cparse-error
	$0 $1
	"init_declarator_list: init_declarator_list error."))
      ((init_declarator_list	init_declarator_list " ERROR" init_declarator)
       (sck-cparse-error
	(append $0 (list $2)) $1
	"init_declarator_list: init_declarator_list error init_declarator."))
      ((init_declarator_list	init_declarator_list "," " ERROR")
       (sck-cparse-error
	$0 $2 
	"init_declarator_list: init_declarator_list \",\" error."))
      
      ;; (<Declarator <<Initializer>|nil>)
      ((init_declarator		declarator)
       (list $0 nil))
      ((init_declarator		declarator "=" initializer)
       (list $0 $2))

      ;;
      ((struct_decl		spec_qualifier_list struct_declarator_list ";")
       (sck-cparse-make-decl-declarator-list $0 $1))

      ;;
      ((spec_qualifier_list	type_spec spec_qualifier_list)
       (append (list $0) $1))
      ((spec_qualifier_list	type_spec)
       (list $0))
      ((spec_qualifier_list	type_qualifier spec_qualifier_list)
       (append (list $0) $1))
      ((spec_qualifier_list	type_qualifier)
       (list $0))

      ;;
      ((struct_declarator_list	struct_declarator)
       (list $0))
      ((struct_declarator_list	struct_declarator_list "," struct_declarator)
       (append $0 (list $2)))
      ((struct_declarator_list struct_declarator_list " ERROR")
       (sck-cparse-error
	$0 $1
	"struct_declarator_list: struct_declarator_list error."))
      ((struct_declarator_list
	struct_declarator_list " ERROR" struct_declarator)
       (sck-cparse-error
	(append $0 (list $2)) $1
	"struct_declarator_list: struct_declarator_list error struct_declarator."))
      ((struct_declarator_list	struct_declarator_list "," " ERROR")
       (sck-cparse-error
	$0 $2
	"struct_declarator_list: struct_declarator_list \",\" error."))

      ;;
      ((struct_declarator	declarator)
       $0)
      ((struct_declarator	declarator ":" const_exp)
       $0)				; error
      ((struct_declarator		   ":" const_exp)
       $0)				; error

      ;; (enum <Pos> <<Ident>|nil> (<Ident> <<Expr>|nil>)..)
      ((enum_spec		"enum" "id" "{" enumerator_list "}")
       (sck-cparse-make-enum $0 $1 $3))
      ((enum_spec		"enum"	    "{" enumerator_list "}")
       (sck-cparse-make-enum $0 nil $2))
      ((enum_spec		"enum" "id")
       (sck-cparse-make-enum $0 $1 nil))

      ;; ((<Ident> <<Expr>|nil>)..)
      ((enumerator_list		enumerator)
       (list $0))
      ((enumerator_list		enumerator_list "," enumerator)
       (append $0 (list $2)))
      ((enumerator_list		enumerator_list " ERROR")
       (sck-cparse-error $0 $1
			 "enumerator_list: enumerator_list error."))

      ((enumerator_list		enumerator_list " ERROR" enumerator)
       (sck-cparse-error (append $0 (list $2)) $1
			 "enumerator_list: enumerator_list error enumerator."))
      ((enumerator_list		enumerator_list "," " ERROR")
       (sck-cparse-error $0 $2
			 "enumerator_list: enumerator_list \",\" error."))

      ;; (<Ident> <<Expr>|nil>)
      ((enumerator		"id")
       (list $0 nil))
      ((enumerator		"id" "=" const_exp)
       (list $0 $2))

      ;; <Declarator> ::= (empty <Pos>)
      ;; 	       | <Ident>
      ;; 	       | (mem <Pos> <Declarator>)
      ;; 	       | (aref <Pos> <Declarator> <<Expr>|nil>)
      ;; 	       | (call <Pos> <Declarator> <<Declaration>|varargs>..)
      ;; 	       | (const <Pos> <Declarator>)
      ;; 	       | (volatile <Pos> <Declarator>)
      ((declarator		pointer direct_declarator)
       (sck-cparse-make-decl-pointer $0 $1))
      ((declarator			direct_declarator)
       $0)

      ;;
      ((direct_declarator	"id")
       $0)
      ((direct_declarator	"(" declarator ")")
       $1)
      ((direct_declarator	direct_declarator "[" const_exp "]")
       (list 'aref (sck-cparse-get-pos $0) $0 $2))
      ((direct_declarator	direct_declarator "[" 		"]")
       (list 'aref (sck-cparse-get-pos $0) $0 nil))
      ((direct_declarator	direct_declarator "(" param_type_list ")")
       (append (list 'call (sck-cparse-get-pos $0) $0) $2))
      ((direct_declarator	direct_declarator "(" id_list	")")
       (append (list 'call (sck-cparse-get-pos $0) $0) $2))
      ((direct_declarator	direct_declarator "("		")")
       (list 'call (sck-cparse-get-pos $0) $0))

      ;; pointer $B$O%j%9%H$K$7$FJV$7$F>e$GE83+$5$;$k(B
      ;; ((pointer <Pos>)|(type_qualifier) ...)
      ((pointer			"*" type_qualifier_list)
       (cons (list 'pointer (sck-cparse-get-pos $0)) $1))
      ((pointer			"*")
       (list (list 'pointer (sck-cparse-get-pos $0))))
      ((pointer			"*" type_qualifier_list pointer)
       (append (cons (list 'pointer (sck-cparse-get-pos $0)) $1) $2))
      ((pointer			"*"			pointer)
       (append (list (list 'pointer (sck-cparse-get-pos $0))) $1))

      ;; <Specifiers>
      ((type_qualifier_list	type_qualifier)
       (list $0))
      ((type_qualifier_list	type_qualifier_list type_qualifier)
       (append $0 (list $1)))
      ((type_qualifier_list	type_qualifier_list " ERROR")
       (sck-cparse-error $0 $1
			 "type_qualifier_list: type_qualifier_list error."))

      ;; (<Declaration>..)
      ((param_type_list		param_list)
       $0)
      ((param_type_list		param_list "," "...")
       (append $0 '(varargs)))

      ;; (<Declaration>..)
      ((param_list		param_decl)
       (list $0))
      ((param_list		param_list "," param_decl)
       (append $0 (list $2)))
      ((param_list		param_list " ERROR")
       (sck-cparse-error $0 $1
			 "param_list: param_list error."))
      ((param_list		param_list " ERROR" param_decl)
       (sck-cparse-error (append $0 (list $2)) $1
			 "param_list: param_list error param_decl."))
      ((param_list		param_list "," " ERROR")
       (sck-cparse-error $0 $1
			 "param_list: param_list \",\" error."))

      ;; <Declaration>
      ((param_decl		decl_specs declarator)
       (sck-cparse-make-decl-declarator $0 $1))
      ((param_decl		decl_specs abstract_declarator)
       (sck-cparse-make-decl-declarator $0 $1))
      ((param_decl		decl_specs)
       (sck-cparse-make-decl-declarator
	$0 (sck-cparse-make-empty (sck-cparse-get-pos $0))))

      ;;
      ((id_list			"id")
       (list $0))
      ((id_list			id_list "," "id")
       (append $0 (list $2)))
      ((id_list			id_list " ERROR")
       (sck-cparse-error $0 $1
			 "id_list: id_list error."))
      ((id_list			id_list " ERROR" "id")
       (sck-cparse-error (append $0 (list $2)) $1
			 "id_list: id_list error id."))
      ((id_list			id_list "," " ERROR")
       (sck-cparse-error $0 $2
			 "id_list: id_list \",\" error."))

      ;;
      ((initializer		assignment_exp)
       $0)
      ((initializer		"{" initializer_list "}")
       (append (list 'list (sck-cparse-get-pos $0)) $1))
      ((initializer		"{" initializer_list "," "}")
       (append (list 'list (sck-cparse-get-pos $0)) $1))

      ;;
      ((initializer_list	initializer)
       (list $0))
      ((initializer_list	initializer_list "," initializer)
       (append $0 (list $2)))
      ((initializer_list	initializer_list " ERROR")
       (sck-cparse-error
	$0 $1
	"initializer_list: initializer_list error."))
      ((initializer_list	initializer_list " ERROR" initializer)
       (sck-cparse-error
	(append $0 (list $2)) $1
	"initializer_list: initializer_list error initializer."))
      ((initializer_list	initializer_list "," " ERROR")
       (sck-cparse-error
	$0 $2
	"initializer_list: initializer_list \",\" error."))

      ;; <Declaration>
      ((type_name		spec_qualifier_list abstract_declarator)
       (sck-cparse-make-decl-declarator $0 $1))
      ((type_name		spec_qualifier_list)
       (sck-cparse-make-decl-declarator
	$0 (sck-cparse-make-empty (sck-cparse-get-pos $0))))

      ;; <Declarator>
      ((abstract_declarator	pointer)
       (sck-cparse-make-decl-pointer
	$0 (sck-cparse-make-empty (sck-cparse-get-pos $0))))
      ((abstract_declarator	pointer direct_abstract_declarator)
       (sck-cparse-make-decl-pointer $0 $1))
      ((abstract_declarator		direct_abstract_declarator)
       $0)

      ;; <Declarator>
      ((direct_abstract_declarator
	"(" abstract_declarator ")")
       $1)
      ((direct_abstract_declarator
	direct_abstract_declarator	"[" const_exp "]")
       (list 'aref (sck-cparse-get-pos $0) $0 $2))
      ((direct_abstract_declarator 
			       		"[" const_exp "]")
       (list 'aref (sck-cparse-get-pos $0)
	     (sck-cparse-make-empty (sck-cparse-get-pos $0)) $1))
      ((direct_abstract_declarator
	direct_abstract_declarator	"[" "]")
       (list 'aref (sck-cparse-get-pos $0) $0 nil))
      ((direct_abstract_declarator
					"[" "]")
       (list 'aref (sck-cparse-get-pos $0)
	     (sck-cparse-make-empty (sck-cparse-get-pos $0)) nil))
      ((direct_abstract_declarator 
	direct_abstract_declarator	"(" param_type_list ")")
       (append (list 'call (sck-cparse-get-pos $0) $0) $2))
      ((direct_abstract_declarator
					"(" param_type_list ")")
       (append (list 'call (sck-cparse-get-pos $0)
		     (sck-cparse-make-empty (sck-cparse-get-pos $0)) $2)))
      ((direct_abstract_declarator
	direct_abstract_declarator	"("		")")
       (list 'call (sck-cparse-get-pos $0) $0))
      ((direct_abstract_declarator 
					"("		")")
       (list 'call (sck-cparse-get-pos $0) 
	     (sck-cparse-make-empty (sck-cparse-get-pos $0))))

      ;; statements
      ;; <Stmt> ::=  <Expr>
      ;; 	 | <Declaration>
      ;; 	 | <Block>
      ;; 	 | (label <Pos> <Ident> <Stmt>)
      ;; 	 | (case <Pos> <Expr> <Stmt>)
      ;; 	 | (default <Pos> <Stmt>)
      ;; 	 | (if <Pos> <Expr> <Stmt> <<Stmt>|nil>)
      ;; 	 | (switch <Pos> <Expr> <Stmt>)
      ;; 	 | (while <Pos> <Expr> <Stmt>)
      ;; 	 | (do <Pos> <Stmt> <Expr>)
      ;; 	 | (for <Pos> <<Expr>|nil> <<Expr>|nil> <<Expr>|nil> <Stmt>)
      ;; 	 | (goto <Pos> <Ident>)
      ;; 	 | (continue <Pos>)
      ;; 	 | (break <Pos>)
      ;; 	 | (return <Pos> <<Expr>|nil>)
      ;;         | (empty <Pos>)
      ((stat			labeled_stat)
       $0)
      ((stat			exp_stat)
       $0)
      ((stat			compound_stat)
       $0)
      ((stat			selection_stat)
       $0)
      ((stat			iteration_stat)
       $0)
      ((stat			jump_stat)
       $0)

      ;;
      ((labeled_stat		"id" ":" stat)
       (list 'label (sck-cparse-get-pos $1) $0 $2))
      ((labeled_stat		"case" const_exp ":" stat)
       (list 'case (sck-cparse-get-pos $0) $1 $3))
      ((labeled_stat		"default" ":" stat)
       (list 'default (sck-cparse-get-pos $0) $2))
      
      ;; 
      ((exp_stat		exp ";")
       $0)
      ((exp_stat		    ";")
       (sck-cparse-make-empty (sck-cparse-get-pos $0)))

      ;; <Block> ::= (block <Pos> <Stmt>..)
      ((compound_stat
	typename_stack_in "{" decl_list stat_list "}" typename_stack_out)
       (sck-cparse-make-block $1 $2 $3))
      ((compound_stat
	typename_stack_in "{"           stat_list "}" typename_stack_out)
       (sck-cparse-make-block $1 nil $2))
      ((compound_stat
	typename_stack_in "{" decl_list           "}" typename_stack_out)
       (sck-cparse-make-block $1 $2 nil))
      ((compound_stat
	typename_stack_in "{"                     "}" typename_stack_out)
       (sck-cparse-make-block $1 nil nil))

      ;; typename $B$rEPO?$9$k%9%?%C%/$r4IM}$9$k$?$a(B
      ((typename_stack_in)
       (sck-cparse-push-type-table))
      ((typename_stack_out)
       (sck-cparse-pop-type-table))

      ;;
      ((stat_list		stat)
       (list $0))
      ((stat_list		stat_list stat)
       (append $0 (list $1)))
      ((stat_list		stat_list " ERROR")
       (sck-cparse-error $0 $1 "stat_list: stat_list error."))

      ;;
      ((selection_stat		"if" "(" exp ")" stat)
       (list 'if (sck-cparse-get-pos $0) $2 $4 nil))
      ((selection_stat		"if" "(" exp ")" stat "else" stat)
       (list 'if (sck-cparse-get-pos $0) $2 $4 $6))
      ((selection_stat		"switch" "(" exp ")" stat)
       (list 'switch (sck-cparse-get-pos $0) $2 $4))

      ;;
      ((iteration_stat		"while" "(" exp ")" stat)
       (list 'while (sck-cparse-get-pos $0) $2 $4))
      ((iteration_stat		"do" stat "while" "(" exp ")" ";")
       (list 'do (sck-cparse-get-pos $0) $1 $4))
      ((iteration_stat		"for" "(" exp ";" exp ";" exp ")" stat)
       (sck-cparse-make-for $0 $2 $4 $6 $8))
      ((iteration_stat		"for" "(" exp ";" exp ";"     ")" stat)
       (sck-cparse-make-for $0 $2 $4 nil $7))
      ((iteration_stat		"for" "(" exp ";"     ";" exp ")" stat)
       (sck-cparse-make-for $0 $2 nil $5 $7))
      ((iteration_stat		"for" "(" exp ";"     ";"     ")" stat)
       (sck-cparse-make-for $0 $2 nil nil $6))
      ((iteration_stat		"for" "("     ";" exp ";" exp ")" stat)
       (sck-cparse-make-for $0 nil $3 $5 $7))
      ((iteration_stat		"for" "("     ";" exp ";"     ")" stat)
       (sck-cparse-make-for $0 nil $3 nil $6))
      ((iteration_stat		"for" "("     ";"     ";" exp ")" stat)
       (sck-cparse-make-for $0 nil nil $4 $6))
      ((iteration_stat		"for" "("     ";"     ";"     ")" stat)
       (sck-cparse-make-for $0 nil nil nil $5))

      ((jump_stat		"goto" "id" ";")
       (list 'goto (sck-cparse-get-pos $0) $1))
      ((jump_stat		"continue" ";")
       (list 'continue (sck-cparse-get-pos $0)))
      ((jump_stat		"break" ";")
       (list 'break (sck-cparse-get-pos $0)))
      ((jump_stat		"return" exp ";")
       (list 'return (sck-cparse-get-pos $0) $1))
      ((jump_stat		"return"     ";")
       (list 'return (sck-cparse-get-pos $0) nil))

      ;; Expression
      ;; <Expr> ::=  <Iconst> | <Fconst>	        /* $B@0?t!"IbF0>.?t(B */
      ;; 	   | <Cconst> | <Sconst>	        /* $BJ8;z!"J8;zNs(B */
      ;; 	   | <Ident>			        /* $B<1JL;R(B */
      ;; 	   | (<Uop> <Pos> <Expr>)	        /* $BC19`1i;;(B */
      ;; 	   | (<Bop> <Pos> <Expr> <Expr>)        /* $BFs9`1i;;(B */
      ;; 	   | (cond <Pos> <Expr> <Expr> <Expr>)  /* $B;09`1i;;(B */
      ;; 	   | (cast <Pos> <Declaration> <Expr>)  /* $B%-%c%9%H(B */
      ;; 	   | (sizeof <Pos> <Declaration>)       /* sizeof $B7?(B */
      ;; 	   | (sizeof <Pos> <Expr>)              /* sizeof $B<0(B */
      ;; 	   | (aref <Pos> <Expr> <Expr>)	        /* v[i] */
      ;; 	   | (call <Pos> <Expr> <Expr>..)       /* f(...) */
      ;; 	   | (member <Pos> <Expr> <Ident>)      /* s.i */
      ;; 	   | (pmember <Pos> <Expr> <Ident>)     /* p->i */
      ((exp			assignment_exp)
       $0)
      ((exp			exp "," assignment_exp)
       (list 'comma (sck-cparse-get-pos $1) $0 $1))

      ((assignment_exp		conditional_exp)
       $0)
      ((assignment_exp		unary_exp assignment_operator assignment_exp)
       (append $1 (list $0 $2)))

      ;; (<Bop> <Pos>)
      ((assignment_operator	"=")
       (list 'set (sck-cparse-get-pos $0)))
      ((assignment_operator	"*=")
       (list 'setmul (sck-cparse-get-pos $0)))
      ((assignment_operator	"/=")
       (list 'setdiv (sck-cparse-get-pos $0)))
      ((assignment_operator	"%=")
       (list 'setmod (sck-cparse-get-pos $0)))
      ((assignment_operator	"+=")
       (list 'setadd (sck-cparse-get-pos $0)))
      ((assignment_operator	"-=")
       (list 'setsub (sck-cparse-get-pos $0)))
      ((assignment_operator	"<<=")
       (list 'setlsh (sck-cparse-get-pos $0)))
      ((assignment_operator	">>=")
       (list 'setrsh (sck-cparse-get-pos $0)))
      ((assignment_operator	"&=")
       (list 'setband (sck-cparse-get-pos $0)))
      ((assignment_operator	"^=")
       (list 'setbxor (sck-cparse-get-pos $0)))
      ((assignment_operator	"|=")
       (list 'setbor (sck-cparse-get-pos $0)))

      ;; (cond <Pos> <Expr> <Expr> <Expr>)
      ((conditional_exp		binary)
       $0)
      ((conditional_exp		binary "?" exp ":" conditional_exp)
       (list 'cond (sck-cparse-get-pos $1) $0 $2 $4))

      ;; 
      ((const_exp		conditional_exp)
       $0)

      ;; (<Bop> <Pos> <Expr> <Expr>)
      ((binary			cast_exp)
       $0)
      ((binary			binary "||" binary)
       (list 'lor (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "&&" binary)
       (list 'land (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "|" binary)
       (list 'bor (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "^" binary)
       (list 'bxor (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "&" binary)
       (list 'band (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "==" binary)
       (list 'tsteq (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "!=" binary)
       (list 'tstne (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "<" binary)
       (list 'tstlt (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary ">" binary)
       (list 'tstgt (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "<=" binary)
       (list 'tstle (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary ">=" binary)
       (list 'tstge (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "<<" binary)
       (list 'lsh (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary ">>" binary)
       (list 'rsh (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "+" binary)
       (list 'add (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "-" binary)
       (list 'sub (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "*" binary)
       (list 'mul (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "/" binary)
       (list 'div (sck-cparse-get-pos $1) $0 $2))
      ((binary			binary "%" binary)
       (list 'mod (sck-cparse-get-pos $1) $0 $2))

      ;; (cast <Pos> <Declaration> <Expr>)
      ((cast_exp		unary_exp)
       $0)
      ((cast_exp		"(" type_name ")" cast_exp)
       (list 'cast (sck-cparse-get-pos $0) $1 $3))
       
      ;; (<Uop> <Pos> <Expr>)
      ((unary_exp		postfix_exp)
       $0)
      ((unary_exp		"++" unary_exp)
       (list 'preinc (sck-cparse-get-pos $0) $1))
      ((unary_exp		"--" unary_exp)
       (list 'predec (sck-cparse-get-pos $0) $1))
      ((unary_exp		unary_operator cast_exp)
       (append $0 (list $1)))
      ((unary_exp		"sizeof" unary_exp)
       (list 'sizeof (sck-cparse-get-pos $0) $1))
      ((unary_exp		"sizeof" "(" type_name ")")
       (list 'sizeof (sck-cparse-get-pos $0) $2))

      ;; (<Uop> <Pos>)
      ((unary_operator		"&")
       (list 'addr (sck-cparse-get-pos $0)))
      ((unary_operator		"*")
       (list 'mem (sck-cparse-get-pos $0)))
      ((unary_operator		"+")
       (list 'uadd (sck-cparse-get-pos $0)))
      ((unary_operator		"-")
       (list 'usub (sck-cparse-get-pos $0)))
      ((unary_operator		"~")
       (list 'unot (sck-cparse-get-pos $0)))
      ((unary_operator		"!")
       (list 'unot (sck-cparse-get-pos $0)))

      ;; 
      ((postfix_exp		primary_exp)
       $0)
      ((postfix_exp		postfix_exp "[" exp "]")
       (list 'aref (sck-cparse-get-pos $1) $0 $2))
      ((postfix_exp		postfix_exp "(" argument_exp_list ")")
       (append (list 'call (sck-cparse-get-pos $1) $0) $2))
      ((postfix_exp		postfix_exp "("			  ")")
       (list 'call (sck-cparse-get-pos $1) $0))
      ((postfix_exp		postfix_exp "." "id")
       (list 'member (sck-cparse-get-pos $1) $0 $2))
      ((postfix_exp		postfix_exp "->" "id")
       (list 'pmember (sck-cparse-get-pos $1) $0 $2))
      ((postfix_exp		postfix_exp "++")
       (list 'posinc (sck-cparse-get-pos $1) $0))
      ((postfix_exp		postfix_exp "--")
       (list 'posdec (sck-cparse-get-pos $1) $0))

      ;;
      ((primary_exp		"id")
       $0)
      ((primary_exp		const)
       $0)
      ((primary_exp		string_const)
       $0)
      ((primary_exp		"(" exp ")")
       $1)

      ;; $BDI2C(B
      ((string_const		"string")
       $0)
      ((string_const		string_const "string")
       (list 'sconst (sck-cparse-get-pos $0)
	     (append (butlast (caddr $0)) (caddr $1))))

      ;;
      ((argument_exp_list	assignment_exp)
       (list $0))
      ((argument_exp_list	argument_exp_list "," assignment_exp)
       (append $0 (list $2)))
      ((argument_exp_list	argument_exp_list " ERROR")
       (sck-cparse-error
	$0 $1
	"argument_exp_list: argument_exp_list error."))
      ((argument_exp_list	argument_exp_list " ERROR" assignment_exp)
       (sck-cparse-error
	(append $0 (list $2)) $1
	"argument_exp_list: argument_exp_list error assignment_exp."))
      ((argument_exp_list	argument_exp_list "," " ERROR")
       (sck-cparse-error
	$0 $2
	"argument_exp_list: argument_exp_list \",\" error."))

      ;;
      ((const			"int_const")
       $0)
      ((const			"char_const")
       $0)
      ((const			"float_const")
       $0)
      ))
    (output-filename "sck-c-syntax-output.el")))

;;;
;;;
;;;
(defun sck-c-syntax-yayacc ()
  (interactive)
  (sck-yayacc-yacc sck=c-syntax))

;;;
;;; test
;;;
(defun sck-c-syntax-test ()
  (interactive)
  (sck-letalist sck=c-syntax
      ((object-language language)
       (associativity associativity)
       (attribute-grammar grammar)
       (output-filename outfile))
    (sck-yayacc-lr (list (list 'productions (mapcar 'car grammar))))
    (sck-yayacc-dump)
    ;;(sck-yayacc-graphviz-display)
))

(defun sck-c-syntax-to-doty-file (file)
  (interactive "FFile: ")
  (sck-yayacc-alist-to-doty-file sck=c-syntax file))

(defun sck-c-syntax-action ()
  (interactive)
  (sck-letalist sck=c-syntax
      ((object-language language)
       (associativity associativity)
       (attribute-grammar grammar)
       (output-filename outfile))
    (with-display-buffer "*sck-c-syntax-action*"
      (insert (format "(sck*yayacc-cc-actions\n '%s)\n"
		      (pp-to-string (mapcar 'cadr grammar)))))))

  