;;;
;;; $Id: sck-cparse-yayacc.el,v 1.3 2007/04/09 09:50:23 heavy Exp $
;;;
;;; sck-cparse-yayacc : <CppsInput> -> <CppsOutput>
;;;
;;; Type:
;;; <CppsInput> ::= ((name <String>)
;;;                  (source <String>))
;;; <CppsOutput> ::= ((plain-ast <PlainAst>)
;;;                   (file-names (<String>..))
;;;                   (messages (<String>..)))
;;; <PlainAST> ::= ...

(require 'sck-clex)
(require 'sck-cpps-lib)

;;;
;;; yayacc ōꂽt@C
;;;
(load "sck-c-syntax-output.el")

;;;
;;; demos
;;;
(defun sck-cparse-yayacc-file (file)
  (interactive "fFile :")
  (sck-cparse-yayacc-string (with-temp-buffer
			      (insert-file-contents file)
			      (buffer-string))
			    file))

(defun sck-cparse-yayacc-buffer ()
  (interactive)
  (sck-cparse-yayacc-region (point-min) (point-max)))

(defun sck-cparse-yayacc-region (beg end)
  (interactive "r")
  (sck-cparse-yayacc-string (buffer-substring-no-properties beg end)))

(defun sck-cparse-yayacc-string (string &optional name)
  (let ((input (list (list 'name (if name name "stdin.c"))
		     (list 'source string))))
    (sck-letalist (sck-cparse-yayacc input)
	(plain-ast file-names messages)
      (with-display-buffer "*sck-cparse-yayacc-demo*"
	(emacs-lisp-mode)
	(sck-cpps-ast-font-lock)
	(insert ";;; sck-cparse-yayacc Input\n\n")
	(insertf "%s\n" (let ((pp-escape-newlines nil))
			  (pp-to-string input)))
	(insert ";;; sck-cparse-yayacc Output\n\n")
	(insert "((plain-ast\n")
	(insertf "%s" (sck-cpps-pp-ast plain-ast 2))
	(insert ")\n\n")
	(insertf "(file-names %S)\n\n" file-names)
	(insert "(messages (\n")
	(dolist (msg messages)
	  (insertf "%S\n" msg))
	(insert ")))\n")))))

;;;
;;;
;;;
(defun sck-cparse-yayacc (input)
  (sck-letalist input
      (name source)
    (sck-letalist (sck-cparse-yayacc-init-clex name source)
	((tokens sck*cparse-lex-tokens)
	 (file-names sck*cparse-lex-file-names)
	 (messages lex-msgs))
      (let ((sck*cparse-type-tables (list nil))
	    (sck*cparse-tokens sck*cparse-lex-tokens)
	    sck*cparse-token-pos
	    (sck*cparse-msgq (make-que)))
	(list (list 'plain-ast (sck-yayacc-executor))
	      (list 'file-names sck*cparse-lex-file-names)
	      (list 'messages 
		    (sort (append lex-msgs (que-to-list sck*cparse-msgq))
			  'string<)))))))

(defun sck-cparse-message (fmt &rest args)
  (let ((msg (apply 'message fmt args)))
    (append-que sck*cparse-msgq msg)))

;;
;; lex
;;
(defun sck-yayacc-lex-get-token ()

  '(message "sck-yayacc-lex-get-token tables: %S"
	   sck*cparse-type-tables)

  (when (and
	 (listp (car sck*cparse-tokens))
	 (eq '$ (caar sck*cparse-tokens)))
    (setq sck*cparse-token-pos (car sck*cparse-tokens))
    (setq sck*cparse-tokens (cdr sck*cparse-tokens)))

  '(message "sck-yayacc-lex-get-token %s" 
	   (list (car sck*cparse-tokens) sck*cparse-token-pos))

  (prog1
      (list (car sck*cparse-tokens) sck*cparse-token-pos)
    (unless (eq (car sck*cparse-tokens) 'eof)
      (setq sck*cparse-tokens (cdr sck*cparse-tokens)))))

(defun sck-yayacc-lex-token-id (token)
  '(message "sck-yayacc-lex-token-id %S" token)
  (let ((ret (sck-yayacc-lex-token-id/test token)))
    '(message "ret: %S" ret)
    ret))

(defun sck-yayacc-lex-token-id/test (token)
  ;; token-id
  (if (listp (car token))
      (ecase (caar token)
       (ident
	(if (sck-cparse-type-tables-member-p (cadar token))
	    "typedef_name"
	  "id"))
       (iconst "int_const")
       (fconst "float_const")
       (cconst "char_const")
       (sconst "string"))
    (if (eq (car token) 'eof)
	" EOF"
      (sck-clex-token-to-string (car token)))))

(defun sck-yayacc-lex-token-attr (token)
  ;; (token-id <Pos> ...)
  (if (listp (car token))
      (cons (caar token)
	    (cons (cadr token)
		  (cdar token)))
    token))

(defun sck-yayacc-lex-token-pos (token)
  ;; ("filename" lineno)
  (let ((pos (cadr token)))
    (list 
     (elt sck*cparse-lex-file-names (cadr pos))
     (caddr pos))))

;;
(defun sck-cparse-yayacc-init-clex (name string)
  (sck-clex (list 
	     (list 'name name)
	     (list 'source string))))

;;
;; Action Ŏg֐
;;
(defun sck-cparse-type-tables-member-p (id)
  ;; typedef Ă邩H

  '(message "sck-cparse-type-tables-member-p %S" id)

  (block 'ret
    (dolist (table sck*cparse-type-tables)
      (when (member id table)
	(return-from 'ret t)))
    nil))

(defun sck-cparse-set-type-table (id)

  '(message "sck-cparse-set-type-table %S" id)

  (setf (car sck*cparse-type-tables)
	(cons id (car sck*cparse-type-tables))))

(defun sck-cparse-push-type-table ()
  (push nil sck*cparse-type-tables))

(defun sck-cparse-pop-type-table ()
  (pop sck*cparse-type-tables))


(defun sck-cparse-set-typedef-decl (decl)

  '(message "sck-cparse-set-typedef-decl tables %S decl %S"
	   sck*cparse-type-tables decl)

  (assert (eq (car decl) 'declaration))
  (when (sck-cparse-set-typedef-decl/typedef-p decl)
    (let ((ids (sck-cparse-set-typedef-decl/get-idents decl)))
      (dolist (id ids)
	(when id
	  (sck-cparse-set-type-table id))))))

(defun sck-cparse-set-typedef-decl/typedef-p (decl)
  (block 'ret
    (dolist (x (caddr decl))
      (when (eq (caddr x) 'typedef)
	(return-from 'ret t)))
    nil))

(defun sck-cparse-set-typedef-decl/get-idents (decl)
  (let (ret)
    (dolist (d (cdddr decl))
      (push (sck-cparse-set-typedef-decl/get-idents/get-id (car d)) ret))

    '(message "sck-cparse-set-typedef-decl/get-idents ret %S" ret)

    ret))

(defun sck-cparse-set-typedef-decl/get-idents/get-id (declaration)

  '(message "sck-cparse-set-typedef-decl/get-idents/get-id %S"
	   declaration)

  (cond
   ((null declaration))
   ((eq (car declaration) 'ident)
    (caddr declaration))
   ((eq (car declaration) 'empty)
    nil)
   (t
    (sck-cparse-set-typedef-decl/get-idents/get-id (caddr declaration)))))

;;
(defun sck-cparse-get-pos (l)
  (if (symbolp (car l))
      (cadr l)
    (sck-cparse-get-pos (car l))))

(defun sck-cparse-make-function (decl-specs declarator decl-list compound-stat)
  (list 'function 
	(sck-cparse-get-pos declarator)
	(sck-cparse-make-decl-declarator decl-specs declarator)
	decl-list
	compound-stat))

(defun sck-cparse-make-decl (decl-specs init-declarator-list)
  (append (list 'declaration (sck-cparse-get-pos decl-specs) decl-specs)
	  init-declarator-list))

(defun sck-cparse-make-decl-declarator (decl-specs declarator)
  (list 'declaration (sck-cparse-get-pos declarator)
	decl-specs (list declarator nil)))

(defun sck-cparse-make-decl-declarator-list (decl-specs declarator-list)
  (sck-cparse-make-decl decl-specs 
			(mapcar #'(lambda (x) (list x nil)) declarator-list)))

(defun sck-cparse-make-decl-pointer (pointer core)
  (dolist (x (reverse pointer))
    (if (eq 'pointer (car x))
	(setq core (list 'mem (sck-cparse-get-pos x) core))
      (assert (eq 'keyword (car x)))
      (setq core (list (caddr x) (sck-cparse-get-pos x) core))))
  core)

(defun sck-cparse-make-keyword (key)
  (list 'keyword (sck-cparse-get-pos key) (car key)))

(defun sck-cparse-make-struct-or-union (struct-or-union id struct-decl-list)
  (append struct-or-union (list id) struct-decl-list))

(defun sck-cparse-make-enum (enum id enumerator-list)
  (append (list 'enum (sck-cparse-get-pos enum) id) enumerator-list))

(defun sck-cparse-make-empty (pos)
  (list 'empty pos))

(defun sck-cparse-make-for (f e1 e2 e3 stat)
  (list 'for (sck-cparse-get-pos f) (list e1 e2 e3) stat))

(defun sck-cparse-make-block (b decls stats)
  (append (list 'block (sck-cparse-get-pos b)) decls stats))

(defun sck-cparse-make-typedef-name (n)
  '(message "sck-cparse-make-typedef-name %S" n)
  n)

;;
(defun sck-cparse-error (ret err fmt &rest args)
  (sck-cparse-message "%s:%d: Syntax error %S. %s"
		      (caar err) (cadar err) (cadr err)
		      (apply 'format fmt args))
  ret)

(provide 'sck-cparse-yayacc)
