;;;; C Plain Parser
;;;; LastChange: Fri Apr 13 12:46:00 2007

;;; Export:
;;; sck-cpps : <CppsInput> -> <CppsOutput>
;;;
;;; Type:
;;; <CppsInput> ::= ((name <String>)
;;;                  (source <String>))
;;; CI
;;; <CppsInput> ::= <ClexOutput>
;;;
;;; <ClexOutput> ::= ((name <String>)
;;;                   (tokens (<Token>..))
;;;                   (file-names (<String>..))
;;;                   (messages (<String>..)))
;;;
;;; <CppsOutput> ::= ((plain-ast <PlainAst>)
;;;                   (file-names (<String>..))
;;;                   (messages (<String>..)))
;;; <PlainAST> ::= ...


(require 'sck-clex)

(require 'sck-cpps-getlex)
(require 'sck-cpps-lib)
(require 'sck-cpps-sys)


(defconst sck=cpps-input-example
  "# 1 \"demo.c\"
char *str = \"str\" \"ing\";

extern int add(int x, int y);

static int from = 1, to = 10;
int total;

void main()
{
  total = sum(from,to);
}

int sum(int from, int to)
{
  int i;
  int sum = 0;

  for (i = from; i < to; i++)
    sum = add(sum,i);

  return sum;
}

# 100 \"misc.c\"
int foo(a,b,c) 
  int a,b; char *c; 
{
  return a;
}

void bar(int a,...);


# 100 \"error.c\"
int a b;
int x = 1ULUU;
char *s = \"\\k\";
f(){ if 1 else return}
/* 
")


(defun sck-cpps-demo ()
  "C Plain Parser ̃fsB"
  (interactive)
  (sck-cpps-str "demo" sck=cpps-input-example))

(defun sck-cpps-file (file)
  "t@Cp[XA*sck-cpps* obt@֏o͂B"
  (interactive "f(pre processed) C source file: ")
  (with-temp-buffer
    (insert-file-contents file)
    (setq buffer-file-name file)
    (set-buffer-modified-p nil)
    (sck-cpps-buffer)))

(defun sck-cpps-buffer ()
  "Jgobt@p[XA*sck-cpps* obt@֏o͂B"
  (interactive)
  (sck-cpps-region (point-min) (point-max)))

(defun sck-cpps-region (beg end)
  "[Wp[XA*sck-cpps* obt@֏o͂B"
  (interactive "r")
  (sck-cpps-str (file-name-nondirectory
		 (file-name-sans-extension
		  (or (buffer-file-name)
		      (buffer-name))))
		(buffer-substring-no-properties beg end)))

(defun sck-cpps-str (name str)
  " STR p[XA*sck-cpps* obt@֏o͂B"
  (with-display-buffer "*sck-cpps*"
    (emacs-lisp-mode)
    (sck-cpps-ast-font-lock)
    (let ((output (sck-cpps
		   (sck-clex `((name ,name)
			       (source ,str))))))
      (insert (sck-cpps-pp-output output))
      output)))

(defun sck-cpps-pp-output (output)
  "sck-cpps ̏o OUTPUT veB[vgB"
  (with-temp-buffer
    (sck-letalist output
	(plain-ast file-names messages)
      ;; (insert ";;; CppsInput:\n\n")
      ;; (insertf "%s\n"(let ((pp-escape-newlines nil))(pp-to-string input)))
      ;; (insert ";;; Output:\n\n")
      (insert "((plain-ast\n")
      (insertf "%s" (sck-cpps-pp-ast plain-ast 2))
      (insert ")\n\n")
      (insertf "(file-names %S)\n\n" file-names)
      (insert "(messages (\n")
      (dolist (msg messages)
	(insertf "%S\n" msg))
      (insert ")))\n"))
    (buffer-string)))


(defun sck-cpps-demo-getlex ()
  (interactive)
  (with-display-buffer "*sck-cpps*"
    (sck-cpps-getlex-open (sck-clex `((name "stdin")
				      (source ,sck=cpps-input-example))))
    (sck-cpps-getlex nil)
    (loop (insertf "%S %S %S\n"
		   sck*cpps-token
		   sck*cpps-token-value
		   sck*cpps-token-pos)
	  (sck-cpps-getlex nil)
	  (when (eq sck*cpps-token 'eof) (return)))
    (sck-cpps-getlex-close)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Plain p[Tgbvx

(defun sck-cpps (clex-output)
  "clex ̏o CLEX-OUTPUT  plain parse B"
  (let ((tns nil)
	(fsys '(eof))
	(ast nil))
    (sck-cpps-message-init)
    (sck-cpps-getlex-open clex-output)
    (sck-cpps-getlex tns)
    ;;
    (setq ast (sck-cpps-decl tns fsys sck*cpps-clex-name))
    ;;
    (prog1
	`((plain-ast ,ast)
	  (file-names ,sck*cpps-clex-file-names)
	  (messages ,(sort (append sck*cpps-clex-messages
				   (sck-cpps-message-strings))
			   'string<)))
      ;; 
      (sck-cpps-getlex-close))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 錾̉

(defun sck-cpps-decl (tns fsys name)
  `(module ,sck*cpps-token-pos
	   ,name
	   ,@(sck-cpps-decl-fundecl-list tns fsys)))

(defun sck-cpps-decl-fundecl-list (tns fsys)
  ;; Return (<<Function> | <Declaration>>..)
  (let* ((ast (make-que))
	 (begsys (cons 'sem (append sck=cpps-specbegsys
				    sck=cpps-declbegsys)))
	 (fsys1 (union begsys fsys)))
    (while (memq sck*cpps-token begsys)
      (let ((fundecl (sck-cpps-decl-fundecl tns fsys1)))
	(append-que ast fundecl)
	(when (eq (car fundecl) 'declaration)
	  (setq tns (sck-cpps-enter-typename fundecl tns)))))
    (sck-cpps-getlex-skip tns fsys)
    (que-to-list ast)))

(defun sck-cpps-decl-fundecl (tns fsys)
  ;; Return <Function> | <Declaration>
  (let* (($ sck*cpps-token-pos)
	 (fsys1 (union sck=cpps-declbegsys (union '(com sem lbr) fsys)))
	 (spec (sck-cpps-decl-specifiers tns fsys1))
	 (decl (sck-cpps-decl-fundecl/declarator tns fsys1))
	 (ast nil))
    ;; (message "sck-cpps-decl-fundecl 1 %S" sck*cpps-token)
    (cond
     ;; int f() { ...
     ;;         ^
     ((eq sck*cpps-token 'lbr)
      ;; (message "sck-cpps-decl-fundecl 2 %S" sck*cpps-token)
      (setq ast `(function ,$
			   (declaration ,$ ,spec ,decl)
			   nil ; <OrdArgDecls>
			   ,(sck-cpps-stmt tns fsys))))
     ;; int f(x) int ... oldstyle 
     ;;          ^
     ((memq sck*cpps-token sck=cpps-specbegsys)
      (let* ((oldargs (sck-cpps-decl-fundecl-list tns (cons 'lbr fsys)))
	     (blockstmt `(block ,sck*cpps-token-pos)))
	(if (eq sck*cpps-token 'lbr)
	    (setq blockstmt (sck-cpps-stmt tns fsys))
	  (sck-cpps-error "syntax error"))
	(setq ast `(function ,$
			     (declaration ,$ ,spec ,decl)
			     ,oldargs
			     ,blockstmt))))
     ;; int i , ... 
     ;;       ^
     (t
      (let ((decls (list decl))
	    (declbegsys (union '(com sem) sck=cpps-declbegsys)))
	(when (eq sck*cpps-token 'com)
	  (sck-cpps-getlex tns)
	  (setq decls (append decls
			      (sck-cpps-delimited-list
			       tns fsys1 declbegsys
			       'sck-cpps-decl-fundecl/declarator
			       'com 'sep 'sem nil))))
	(sck-cpps-getlex-expect tns 'sem)
	(setq ast `(declaration ,$ ,spec ,@decls)))))
    ;;
    (sck-cpps-getlex-skip tns fsys)
    ;;
    ast))

(defun sck-cpps-decl-fundecl/declarator (tns fsys)
  ;; int  i = ...
  ;;      ^^^^^^^
  ;; Return (<Declarator> <<Initializer>|nil>)
  (let ((decl (sck-cpps-decl-declarator tns (cons 'equ fsys)))
	(init nil))
    (when (eq sck*cpps-token 'equ)
      (sck-cpps-getlex tns)
      (setq init (sck-cpps-decl-initializer tns fsys)))
    ;;
    (list decl init)))

(defun sck-cpps-decl-specifiers (tns fsys)
  ;; Return <Specifiers>
  (if (memq sck*cpps-token sck=cpps-specbegsys)
      (let ((specs (make-que)))
	(while (memq sck*cpps-token sck=cpps-specbegsys)
	  (append-que specs
			(case sck*cpps-token
			  ((struct union)
			   (sck-cpps-decl-struct-or-union tns fsys))
			  ((enum)
			   (sck-cpps-decl-enum tns fsys))
			  ((typename)
			   (prog1
			       `(ident ,sck*cpps-token-pos
				       ,@(cdr sck*cpps-token-value))
			     (sck-cpps-getlex tns)))
			  (t
			   (prog1
			       `(keyword ,sck*cpps-token-pos ,sck*cpps-token)
			     (sck-cpps-getlex tns))))))
	(que-to-list specs))
    nil))

(defun sck-cpps-decl-simple-declaration (tns fsys)
  ;; vg^Cv̈Asizeof y cast ɂ <Declaration> ̉́B
  (let (($ sck*cpps-token-pos))
    `(declaration ,$
		  ,(sck-cpps-decl-specifiers tns fsys)
		  (,(sck-cpps-decl-declarator tns fsys) nil))))


(defun sck-cpps-decl-declarator (tns fsys)
  (let* (($ sck*cpps-token-pos)
	 (fsys1 (union '(lpa lbk) fsys))
	 (ast `(empty ,$)))
    (case sck*cpps-token
      (ident
       (setq ast (sck-cpps-expr-factor tns fsys1)))
      (ast
       (sck-cpps-getlex tns)
       (setq ast `(mem ,$ ,(sck-cpps-decl-declarator tns fsys1))))
      ((const volatile)
       (let ((q sck*cpps-token))
	 (sck-cpps-getlex tns)
	 (setq ast `(,q ,$ ,(sck-cpps-decl-declarator tns fsys1)))))
      (lpa
       (sck-cpps-getlex tns)
       (setq ast (sck-cpps-decl-declarator tns (cons 'rpa fsys1)))
       (sck-cpps-getlex-expect tns 'rpa)))
    (while (memq sck*cpps-token '(lpa lbk))
      (ecase sck*cpps-token
	(lbk
	 (sck-cpps-getlex tns)
	 (setq ast `(aref ,$
			  ,ast
			  ,(if (eq sck*cpps-token 'rbk)
			       nil
			     (sck-cpps-expr tns (cons 'rbk fsys1)))))
	 (sck-cpps-getlex-expect tns 'rbk))
	(lpa
	 (let ((begsys (cons 'dotdotdot
			     (append sck=cpps-specbegsys
				     sck=cpps-declbegsys))))
	   (sck-cpps-getlex tns)
	   (setq ast `(call ,$
			    ,ast
			    ,@(sck-cpps-delimited-list
			       tns fsys begsys
			      #'(lambda (tns fsys)
				  (cond
				   ((eq sck*cpps-token 'dotdotdot)
				    (sck-cpps-getlex tns)
				    'varargs)
				   (t
				    (sck-cpps-decl-simple-declaration
				     tns fsys))))
			      'com 'sep 'rpa t)))))))
    ;;
    (sck-cpps-getlex-skip tns fsys)
    ;;
    ast))


(defun sck-cpps-decl-struct-or-union (tns fsys)
  ;; rbgtB[hT|[g
  (assert (memq sck*cpps-token '(struct union)))
  (let (($ sck*cpps-token-pos)
	(kind sck*cpps-token)
	(ident nil)
	(decls nil))
    (sck-cpps-getlex tns)
    (when (eq sck*cpps-token 'ident)
      (setq ident (sck-cpps-expr-factor tns fsys)))
    (when (eq sck*cpps-token 'lbr)
      (sck-cpps-getlex tns)
      ;; 邷邩H
      (setq decls (sck-cpps-decl-fundecl-list tns (cons 'rbr fsys)))
      (sck-cpps-getlex-expect tns 'rbr))
    ;;
    `(,kind ,$ ,ident ,@decls)))


(defun sck-cpps-decl-enum (tns fsys)
  (assert (eq sck*cpps-token 'enum))
  (let (($ sck*cpps-token-pos)
	(ast nil)
	(ident nil)
	(enums nil))
    (sck-cpps-getlex tns)
    (when (eq sck*cpps-token 'ident)
      (setq ident (sck-cpps-expr-factor tns fsys)))
    (when (eq sck*cpps-token 'lbr)
      (sck-cpps-getlex tns)
      (setq enums (sck-cpps-delimited-list
		     tns fsys
		     ;; 邷H
		     sck=cpps-exprbegsys 
		     #'(lambda (tns fsys)
			 (let ((ident `(ident ,sck*cpps-token-pos "@error"))
			       (value nil))
			   (if (eq sck*cpps-token 'ident)
			       (setq ident (sck-cpps-expr-factor
					    tns
					    (cons 'equ fsys)))
			     (sck-cpps-error "missing identifier"))
			   (when (eq sck*cpps-token 'equ)
			     (sck-cpps-getlex tns)
			     (setq value (sck-cpps-expr-cond tns fsys)))
			   `(,ident ,value)))
		     'com 'both 'rbr t)))
    ;;
    `(enum ,$ ,ident ,@enums)))
			     
(defun sck-cpps-decl-initializer (tns fsys)
  ;; @int v[] = {1,2,,3}; ̃JoЂǂB
  (let (($ sck*cpps-token-pos)
	(ast nil))
    (cond
     ((eq sck*cpps-token 'lbr)
      (sck-cpps-getlex tns)
      (setq ast `(list ,$
		       ,@(sck-cpps-delimited-list
			  tns fsys (cons 'lbr sck=cpps-exprbegsys)
			  'sck-cpps-decl-initializer
			  'com 'both 'rbr t))))
     (t
      (setq ast (sck-cpps-expr-assign tns fsys))))
    ;;
    ast))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ̉

(defun sck-cpps-stmt (tns fsys)
  (let (($ sck*cpps-token-pos)
	(ast nil))
    ;; (message "sck-cpps-stmt 1 %S" sck*cpps-token)
    (sck-cpps-getlex-skip tns sck=cpps-stmtbegsys) ;  skip2 ?
    (case sck*cpps-token
      ((if)
       (let ((con nil)
	     (th nil)
	     (el nil))
	 (sck-cpps-getlex tns)
	 (sck-cpps-getlex-expect tns 'lpa)
	 (setq con (sck-cpps-expr tns (union '(rpa sem else) fsys))) ; 
	 (sck-cpps-getlex-expect tns 'rpa)
	 (setq th (sck-cpps-stmt tns (cons 'else fsys))) ; 
	 (when (eq sck*cpps-token 'else)
	   (sck-cpps-getlex tns)
	   (setq el (sck-cpps-stmt tns fsys)))
	 (setq ast `(if ,$ ,con ,th ,el))))
      ((switch)
       (let ((expr nil)
	     (stmt nil))
	 (sck-cpps-getlex tns)
	 (sck-cpps-getlex-expect tns 'lpa)
	 (setq expr (sck-cpps-expr tns (cons 'rpa fsys)))
	 (sck-cpps-getlex-expect tns 'rpa)
	 (setq stmt (sck-cpps-stmt tns fsys))
	 (setq ast `(switch ,$ ,expr ,stmt))))
      ((case)
       (let ((expr nil)
	     (stmt nil))
	 (sck-cpps-getlex tns)
	 (setq expr (sck-cpps-expr tns (cons 'col fsys)))
	 (sck-cpps-getlex-expect tns 'col)
	 (setq stmt (sck-cpps-stmt tns fsys))
	 (setq ast `(case ,$ ,expr ,stmt))))
      ((default)
       (sck-cpps-getlex tns)
       (sck-cpps-getlex-expect tns 'col)
       (setq ast `(default ,$ ,(sck-cpps-stmt tns fsys))))
      ((while)
       (let ((expr nil)
	     (stmt nil))
	 (sck-cpps-getlex tns)
	 (sck-cpps-getlex-expect tns 'lpa)
	 (setq expr (sck-cpps-expr tns (cons 'rpa fsys)))
	 (sck-cpps-getlex-expect tns 'rpa)
	 (setq stmt (sck-cpps-stmt tns fsys))
	 (setq ast `(while ,$ ,expr ,stmt))))
      ((for)
       (let ((fsys1 (union '(sem rpa) fsys))
	     (expr1 nil)
	     (expr2 nil)
	     (expr3 nil)
	     (stmt nil))
	 (sck-cpps-getlex tns)
	 (sck-cpps-getlex-expect tns 'lpa)
	 (unless (eq sck*cpps-token 'sem)
	   (setq expr1 (sck-cpps-expr tns fsys1)))
	 (sck-cpps-getlex-expect tns 'sem)
	 (unless (eq sck*cpps-token 'sem)
	   (setq expr2 (sck-cpps-expr tns fsys1)))
	 (sck-cpps-getlex-expect tns 'sem)
	 (unless (eq sck*cpps-token 'rpa)
	   (setq expr3 (sck-cpps-expr tns fsys1)))
	 (sck-cpps-getlex-expect tns 'rpa)
	 (setq stmt (sck-cpps-stmt tns fsys))
	 (setq ast `(for ,$ (,expr1 ,expr2 ,expr3) ,stmt))))
      ((do)
       (let ((expr nil)
	     (stmt nil))
	 (sck-cpps-getlex tns)
	 (setq stmt (sck-cpps-stmt tns (cons 'while fsys)))
	 (sck-cpps-getlex-expect tns 'while)
	 (sck-cpps-getlex-expect tns 'lpa)
	 (setq expr (sck-cpps-expr tns (cons 'rpa fsys)))
	 (sck-cpps-getlex-expect tns 'rpa)
	 (sck-cpps-getlex-expect tns 'sem)
	 (setq ast `(do ,$ ,stmt ,expr))))
      ((break)
       (sck-cpps-getlex tns)
       (sck-cpps-getlex-expect tns 'sem)
       (setq ast `(break ,$)))
      ((continue)
       (sck-cpps-getlex tns)
       (sck-cpps-getlex-expect tns 'sem)
       (setq ast `(continue ,$)))
      ((goto)
       (sck-cpps-getlex tns)
       (let ((label nil))
	 (cond
	  ((memq sck*cpps-token '(ident typename))
	   ;; ł ident Ƃ typename B
	   ;; typedef int L; f(){goto L;}
	   (setq label `(ident ,$ ,(cadr sck*cpps-token-value)))
	   (sck-cpps-getlex tns))
	  (t
	   (sck-cpps-error "missing label")
	   (setq label `(ident ,$ "@error"))))
	 (sck-cpps-getlex-expect tns 'sem)
	 (setq ast `(goto ,$ ,label))))
      ((return)
       (let ((expr nil))
	 (sck-cpps-getlex tns)
	 (unless (eq sck*cpps-token 'sem)
	   (setq expr (sck-cpps-expr tns (cons 'sem fsys))))
	 (sck-cpps-getlex-expect tns 'sem)
	 (setq ast `(return ,$ ,expr))))
      ((lbr)
       (sck-cpps-getlex tns)
       (let ((stmts (make-que))
	     (fsys1 (union sck=cpps-stmtbegsys (cons 'rbr fsys)))
	     (saved-tns tns))
	 (while (memq sck*cpps-token sck=cpps-stmtbegsys)
	   (let ((stmt (sck-cpps-stmt tns fsys1)))
	     (append-que stmts stmt)
	     (when (eq (car stmt) 'declaration)
	       (setq tns (sck-cpps-enter-typename stmt tns)))))
	 (sck-cpps-getlex-expect saved-tns 'rbr)
	 (setq ast `(block ,$ ,@(que-to-list stmts)))))
      ((sem)
       (sck-cpps-getlex tns)
       (setq ast `(empty ,$)))
      (t
       (cond
	((memq sck*cpps-token '(ident typename))
	 (cond
	  ;; x
	  ;; ӁF typedef int L; f(){L:;}
	  ((eq (sck-cpps-getlex-lookahead tns) 'col)
	   (setq ast `(label ,$
			     ,(prog1
				  `(ident ,$ ,(cadr sck*cpps-token-value))
				(sck-cpps-getlex tns)
				(sck-cpps-getlex-expect tns 'col))
			     ,(sck-cpps-stmt tns fsys))))
	  ;; typedef int L; f(){L *p;}
	  ((eq sck*cpps-token 'typename)
	   (setq ast (sck-cpps-decl-fundecl tns fsys)))
	  (t
	   (setq ast (sck-cpps-expr tns (cons 'sem fsys)))
	   (sck-cpps-getlex-expect tns 'sem))))
	((memq sck*cpps-token sck=cpps-specbegsys)
	 (setq ast (sck-cpps-decl-fundecl tns fsys)))
	(t
	 (setq ast (sck-cpps-expr tns (cons 'sem fsys)))
	 (sck-cpps-getlex-expect tns 'sem)))))
    ;;
    (sck-cpps-getlex-skip tns fsys)
    ;;
    ast))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ̉

(defun sck-cpps-expr (tns fsys)
  (let (($ sck*cpps-token-pos)
	(fsys1 (cons 'com fsys))
	(ast nil))
    (setq ast (sck-cpps-expr-assign tns fsys1))
    (while (eq sck*cpps-token 'com)
      (sck-cpps-getlex tns)
      (setq ast `(comma ,$ ,ast ,(sck-cpps-expr-assign tns fsys1))))
    ;;
    ast))

(defconst sck=cpps-setop-hashtab
  (let ((htab (make-hash-table)))
    (dolist (a '((equ set) (astequ setmul) (slaequ setdiv)
		 (perequ setmod) (plsequ setadd)
		 (hypequ setsub) (ampequ setband)
		 (hatequ setbxor) (pipequ setbor)
		 (leslesequ setlsh) (gregreequ setrsh)))
      (setf (gethash (car a) htab) (cadr a)))
    htab))

(defun sck-cpps-expr-assign (tns fsys)
  (let (($ sck*cpps-token-pos)
	(fsys1 (union sck=cpps-setopsys fsys))
	(ast nil))
    (setq ast (sck-cpps-expr-cond tns fsys1))
    (when (memq sck*cpps-token sck=cpps-setopsys)
      (let ((setop  (gethash sck*cpps-token sck=cpps-setop-hashtab)))
	(sck-cpps-getlex tns)
	(setq ast `(,setop
		    ,$
		    ,ast
		    ,(sck-cpps-expr-assign tns fsys)))))
    ;;
    ast))

(defun sck-cpps-expr-cond (tns fsys)
  (let (($ sck*cpps-token-pos)
	(fsys1 (cons 'que fsys))
	(ast nil))
    (setq ast (sck-cpps-expr-binary tns fsys1 0))
    (when (eq sck*cpps-token 'que)
      (let ((th nil)
	    (el nil)
	    (fsys2 (cons 'col fsys1)))
	(sck-cpps-getlex tns)
	(setq th (sck-cpps-expr tns fsys2))
	(sck-cpps-getlex-expect tns 'col)
	(setq el (sck-cpps-expr-cond tns fsys))
	(setq ast `(cond ,$ ,ast ,th ,el))))
    ;;
    ast))

(defconst sck=cpps-binop-sys
  [(pippip) (ampamp) (pip) (hat) (amp) (equequ excequ) (les gre lesequ greequ)
   (gregre lesles) (pls hyp) (ast sla per)])

(defconst sck=cpps-binop-hashtab
  (let ((htab (make-hash-table)))
    (dolist (a '((ast mul) (sla div) (per mod) (pls add) (hyp sub)
		 (gregre rsh) (lesles lsh)
		 (les tstlt) (gre tstgt) (lesequ tstle)
		 (greequ tstge) (equequ tsteq) (excequ tstne)
		 (amp band) (hat bxor)
		 (pip bor) (ampamp land) (pippip lor)))
      (setf (gethash (car a) htab) (cadr a)))
    htab))

(defun sck-cpps-expr-binary (tns fsys prec)
  (let (($ sck*cpps-token-pos)
	(ast nil))
    (cond
     ((<= (length sck=cpps-binop-sys) prec)
      (setq ast (sck-cpps-expr-unary tns fsys)))
     (t
      (let* ((binopsys (aref sck=cpps-binop-sys prec))
	     (fsys1 (union binopsys fsys)))
	(incf prec)
	(setq ast (sck-cpps-expr-binary tns fsys1 prec))
	(while (memq sck*cpps-token binopsys)
	  (let (($ sck*cpps-token-pos)
		(binop (gethash sck*cpps-token sck=cpps-binop-hashtab)))
	    (assert binop)
	    (sck-cpps-getlex tns)
	    (setq ast `(,binop ,$
			       ,ast
			       ,(sck-cpps-expr-binary tns fsys1 prec))))))))
    ;;
    ast))

(defconst sck=cpps-unaop-hashtab
  (let ((htab (make-hash-table)))
    (dolist (a '((exc lnot) (til bnot) (plspls preinc) (hyphyp predec)
		 (hyp neg) (pls uadd) (ast mem) (amp addr)))
      (setf (gethash (car a) htab) (cadr a)))
    htab))

(defun sck-cpps-expr-unary (tns fsys)
  (let (($ sck*cpps-token-pos)
	(ast nil))
    (case sck*cpps-token
      ;; sizeof unary
      ;; sizeof (type)
      ;; sizeof (expr)
      (sizeof
       (sck-cpps-getlex tns)
       (cond
	((eq sck*cpps-token 'lpa)
	 (cond
	  ((memq (sck-cpps-getlex-lookahead tns) sck=cpps-specbegsys)
	   (setq ast `(sizeof ,$ ,(sck-cpps-expr-unary/specdecl tns fsys))))
	  (t
	   (setq ast `(sizeof ,$ ,(sck-cpps-expr-factor tns fsys))))))
	(t
	 (setq ast `(sizeof ,$ ,(sck-cpps-expr-unary tns fsys))))))
      ;; ! ~ ++ -- + - * &  unary
      ((exc til plspls hyphyp pls hyp ast amp)
       (let ((op (gethash sck*cpps-token sck=cpps-unaop-hashtab)))
	 (sck-cpps-getlex tns)
	 (setq ast `(,op ,$ ,(sck-cpps-expr-unary tns fsys)))))
      ;; (...)
      (lpa
       (cond
	;; (type)unary
	((memq (sck-cpps-getlex-lookahead tns) sck=cpps-specbegsys)
	 (setq ast `(cast ,$
			  ,(sck-cpps-expr-unary/specdecl tns fsys)
			  ,(sck-cpps-expr-unary tns fsys))))
	(t
	 (setq ast (sck-cpps-expr-primary tns fsys)))))
      ;; primary
      (t
       (setq ast (sck-cpps-expr-primary tns fsys))))
    ;;
    ast))

(defun sck-cpps-expr-unary/specdecl (tns fsys)
  (assert (eq sck*cpps-token 'lpa))
  (sck-cpps-getlex tns)
  (prog1
      (sck-cpps-decl-simple-declaration tns (cons 'rpa fsys))
    (sck-cpps-getlex-expect tns 'rpa)))

(defun sck-cpps-expr-primary (tns fsys)
  (let* (($ sck*cpps-token-pos)
	 (fsys1 (union sck=cpps-primopsys fsys))
	 (ast nil))
    (setq ast (sck-cpps-expr-factor tns fsys1))
    (while (memq sck*cpps-token sck=cpps-primopsys)
      (ecase sck*cpps-token
	;; factor(...)
	(lpa
	 (sck-cpps-getlex tns)
	 (setq ast `(call
		     ,$
		     ,ast
		     ,@(sck-cpps-delimited-list
			tns fsys sck=cpps-exprbegsys
			'sck-cpps-expr-assign 'com 'sep 'rpa t))))
	;; factor[...]
	(lbk
	 (sck-cpps-getlex tns)
	 (setq ast `(aref
		     ,$
		     ,ast
		     ,(sck-cpps-expr tns (union '(rbk) fsys))))
	 (sck-cpps-getlex-expect tns 'rbk))
	;; factor.ident
	;; factor->ident
	((dot hypgre)
	 (setq ast `(,(ecase sck*cpps-token (dot 'member) (hypgre 'pmember))
		     ,$
		     ,ast
		     ,(progn
			(sck-cpps-getlex tns)
			(if (eq sck*cpps-token 'ident)
			    (sck-cpps-expr-factor tns fsys1)
			  (prog1
			      `(ident ,sck*cpps-token-pos "@error")
			    (sck-cpps-error "missing identifier")
			    (sck-cpps-getlex tns)))))))
	;; factor++
	;; factor--
	((plspls hyphyp)
	 (setq ast `(,(ecase sck*cpps-token (plspls 'posinc) (hyphyp 'posdec))
		     ,$
		     ,(progn
			(sck-cpps-getlex tns)
			ast))))))
    ;;
    (sck-cpps-getlex-skip tns fsys)
    ;;
    ast))

(defun sck-cpps-expr-factor (tns fsys)
  (sck-cpps-getlex-skip tns (union sck=cpps-exprbegsys fsys))
  (let (($ sck*cpps-token-pos)
	(ast nil))
    (case sck*cpps-token
      ((ident iconst fconst cconst)
       (setq ast (list* sck*cpps-token
			$
			(cdr sck*cpps-token-value)))
       (sck-cpps-getlex tns))
      (typename
       (sck-cpps-error "illegal usage of typename `%s"
		       (cadr sck*cpps-token-value))
       (setq ast `(ident ,$ "@error"))
       (sck-cpps-getlex tns))
      (sconst
       (let ((chars (cadr sck*cpps-token-value))
	     (Lsuff (caddr sck*cpps-token-value)))
	 (sck-cpps-getlex tns)
	 (while (eq sck*cpps-token 'sconst)
	   (setq chars (append (butlast chars) (cadr sck*cpps-token-value))
		 Lsuff (or Lsuff (caddr sck*cpps-token-value)))
	   (sck-cpps-getlex tns))
	 (setq ast (list* 'sconst
			  $
			  chars
			  (if Lsuff '(L) nil)))))
      ((lpa)
       (sck-cpps-getlex tns)
       (setq ast (sck-cpps-expr tns (union '(rpa) fsys)))
       (sck-cpps-getlex-expect tns 'rpa))
      (t
       (sck-cpps-error "missing expression")
       (setq ast `(ident ,sck*cpps-token-pos "@error"))))
    ;;
    (sck-cpps-getlex-skip tns fsys)
    ;;
    ast))

;;;
(provide 'sck-cpps)
