;;;; 
;;;; LastChange: Sat Apr 14 14:24:06 2007

;;;   plain AST
;;; 
;;; <Expr> ::=   <Iconst> | <Fconst>                /* A */
;;;            | <Cconst> | <Sconst>                /* A */
;;;            | <Ident>                            /* ʎq */
;;;            | (<Uop> <Pos> <Expr>)               /* PZ */
;;;            | (<Bop> <Pos> <Expr> <Expr>)        /* 񍀉Z */
;;;            | (cond <Pos> <Expr> <Expr> <Expr>)  /* OZ */
;;;            | (cast <Pos> <Declaration> <Expr>)  /* LXg */
;;;            | (sizeof <Pos> <Declaration>)       /* sizeof ^ */
;;;            | (sizeof <Pos> <Expr>)              /* sizeof  */
;;;            | (aref <Pos> <Expr> <Expr>)         /* v[i] */
;;;            | (call <Pos> <Expr> <Expr>..)       /* f(...) */
;;;            | (member <Pos> <Expr> <Ident>)      /* s.i */
;;;            | (pmember <Pos> <Expr> <Ident>)     /* p->i */
;;; 
;;; <Uop> ::=  preinc | predec | posinc | posdec /* ++x, --x, x++, x-- */
;;;          | addr | mem                   /* &x, *x */
;;;          | uadd | neg | bnot | lnot     /* +x, -x, ~x, !x */
;;; 
;;; <Bop> ::= com                           /* ,  */
;;;          | set                          /* = */
;;;          | setmul | setdiv | setmod | setadd | setsub | setlsh  /* op= */
;;;          | setrsh | setband | setbxor | setbor                  /* op= */
;;;          | lor | land                   /* ||, && */
;;;          | bor | bxor | band            /* |, ^, & */
;;;          | tsteq | tstne | tstlt | tstgt | tstle | tstge  /* ==, !=, ... */
;;;          | lsh | rsh                    /* <<, >> */
;;;          | add | sub | mul | div | mod  /* +, -, *, /, % */
;;;
;;; <Iconst> ::= (iconst <Pos> <Ipart> <ISuffixes>)
;;; <Fconst> ::= (fconst <Pos> (<Ipart> <Fpart> <Epart>) <FSuffixes>)
;;; <Cconst> ::= (cconst <Pos> (<Char>) <CSuffixes>)
;;; <Sconst> ::= (sconst <Pos> (<Char>..) <SSuffixes>)
;;; <Ipart>, <Fpart>, <Epart> ::= <String:<Integer>>



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; PAST  <Initializer> ͂B

(defun sck-cps-initializer-expr (env type past staticp)
  ;; PAST: <Initializer> ͂B
  ;; <Initializer> ::=  <Expr> | (list <Pos> <Initializer>..)
  ;; Return: <sck-ast-expr> | (inits <<sck-ast-expr>|(zeros <Integer>)>..)
  (let* (($ (cadr past))
	 (exprl (sck-cps-initializer-list-expr env past))
	 (inits (sck-cps-bind-initializer type exprl)))
    ;;
    (when staticp
      (dolist (expr inits)
	(when (sck-ast-expr-p expr)
	  (let (($ (sck-ast-expr-pos expr)))
	    (unless (sck-ast-asmconst-expr-p expr)
	      (sck-ast-error
	       "too complicated initializer for static variable"))))))
    ;;
    (if (and (= (length inits) 1)
	     (not (eq (car-safe (car inits)) 'zeros)))
	(car inits)
      (cons 'inits inits))))


(defun sck-cps-initializer-list-expr (env past)
  ;; PAST: <Initializer>
  ;; <Initializer> ::=  <Expr> | (list <Pos> <Initializer>..)
  ;; <Return> ::= <sck-ast-expr> | (<Return>..)
  (if (eq (car past) 'list)
      (mapcar #'(lambda (e)
		  (sck-cps-initializer-list-expr env e))
	      (cddr past))
    (sck-ast-expr env past)))


(defvar sck*cps-bind-initializer-result nil)

(defun sck-cps-bind-initializer (type exprl)
  (let ((sck*cps-bind-initializer-result nil))
    (sck-cps-bind-initializer-rec type exprl (consp exprl))
    (nreverse sck*cps-bind-initializer-result)))

(defun sck-cps-bind-initializer-push (x)
  (assert (or (sck-ast-expr-p x)
	      (eq (car-safe x) 'zeros)))
  ;;
  ;; ̎ fold ȂȉLɂB
  ;; (when (sck-ast-expr-p x) (setq x (sck-ast-fold-const x)))
  ;;
  (push x sck*cps-bind-initializer-result))

(defun sck-cps-bind-initializer-push-zeros (n)
  ;; (zeros n) vbVB
  (cond
   ;; zeros }[WgC
   ((eq (car-safe (car sck*cps-bind-initializer-result)) 'zeros)
    (setf (cadr (car sck*cps-bind-initializer-result))
	  (bn-try-to-elnum
	   (bn-add n (cadr (car sck*cps-bind-initializer-result))))))
   ;;
   (t
    (sck-cps-bind-initializer-push `(zeros ,(bn-try-to-elnum n))))))

(defun sck-cps-bind-initializer-rec (type exprl brace)
  ;; BRACE  Non-nil ȂAnꂽ EXPRL  {1,2,3} ̂悤
  ;; BRACE w肳ĂƂB
  ;; TYPE ƑΉʂL^A EXPRL ԂB
  (setq type (sck-ast-skip-typedef-and-qual type))
  (case (car-safe type)
    (array (sck-cps-bind-initializer-array type exprl brace))
    (struct (sck-cps-bind-initializer-struct type exprl brace))
    (union  (sck-cps-bind-initializer-union type exprl brace))
    (function (assert nil))
    (t
     (sck-cps-bind-initializer-scalar type exprl brace))))

(defun sck-cps-bind-initializer-scalar (type exprl brace)
  (cond
   ((sck-ast-expr-p exprl)
    (sck-cps-bind-initializer-push (sck-ast-make-conv-expr type exprl))
    nil)
   ((null exprl)
    (sck-ast-error "empty initializer for scalar")
    nil)
   (brace
    (sck-ast-warning "initializer for scalar within brace")
    (when (< 1 (length exprl))
      (sck-ast-warning "too many initializer for scalar"))
    (sck-cps-bind-initializer-scalar type (car exprl) brace)
    nil)
   (t
    (sck-cps-bind-initializer-scalar type (pop exprl) t)
    exprl)))

(defun sck-cps-bind-initializer-array (type exprl brace)
  (let ((nelm (cadr type))
	(etype (caddr type)))
    (cond
     ((null (sck-ast-sizeof etype))
      (sck-ast-error "initializer for array of incomplete element type")
      nil)
     ;; char s[] = "string";
     ((and (sck-ast-expr-p exprl)
	   (sck-ast-sconst-expr-p exprl)
	   (sck-ast-type-equal
	    etype
	    (caddr (sck-ast-expr-type exprl))))
      (sck-cps-bind-initializer-array/string type exprl))
     ((sck-ast-expr-p exprl)
      (sck-ast-error "illegal initializer for array")
      nil)
     (t
      (let ((esize (sck-ast-sizeof etype))
	    (i 0))
	(while (and exprl (if nelm (bn-< i nelm) exprl))
	  (setq exprl (sck-cps-bind-initializer-element etype exprl)
		i (bn-add i 1)))
	(when (null nelm)
	  ;; initializer ̗vf^⊮
	  (setq nelm (bn-try-to-elnum i))
	  (setf (cadr type) nelm))
	(when (bn-< i nelm)
	  (sck-cps-bind-initializer-push-zeros
	   (bn-mul esize (bn-sub nelm i))))
	(when (and exprl brace)
	  (sck-ast-error "too many initializer for array"))
	exprl)))))

(defun sck-cps-bind-initializer-array/string (type exprl)
  (let* ((ctype (caddr (sck-ast-expr-type exprl)))
	 (chars (sck-ast-expr-value exprl))
	 (exprls (mapcar #'(lambda (c) (sck-ast-make-iconst-expr ctype c))
			 chars)))
    (sck-cps-bind-initializer-array type exprls t)))

(defun sck-cps-bind-initializer-element (type exprl)
  (cond
   ((listp (car exprl))
    (sck-cps-bind-initializer-rec type (car exprl) t)
    (cdr exprl))
   (t
    (sck-cps-bind-initializer-rec type exprl nil))))
  
(defun sck-cps-bind-initializer-struct (type exprl brace)
  (let ((size (sck-ast-sizeof type)))
    (cond
     ((null size)
      (sck-ast-error "initializer for incomplete struct")
      nil)
     ((sck-ast-expr-p exprl)
      (sck-ast-error "illegal initializer for struct")
      nil)
     (t
      (assert (sck-ast-env-p (cadr type)))
      (let ((end 0))
	(dolist (mident (sck-ast-env-idents (cadr type)))
	  (when (null exprl) (return))
	  (let ((off (sck-ast-ident-value mident))
		(mtype (sck-ast-ident-type mident)))
	    (assert (bn-integer-p off))
	    (when (bn-< end off)
	      (sck-cps-bind-initializer-push-zeros (bn-sub off end)))
	    (setq exprl (sck-cps-bind-initializer-element mtype exprl)
		  end (bn-add off (sck-ast-sizeof mtype)))))
	(when (bn-< end size)
	  (sck-cps-bind-initializer-push-zeros (bn-sub size end)))
	(when (and exprl brace)
	  (sck-ast-error "too many initializer for array"))
	exprl)))))

(defun sck-cps-bind-initializer-union (type exprl brace)
  (let ((size (sck-ast-sizeof type)))
    (cond
     ((null size)
      (sck-ast-error "initializer for incomplete union")
      nil)
     (t
      (sck-cps-bind-initializer-rec
       (sck-ast-ident-type (car (sck-ast-env-idents (cadr type))))
       exprl
       brace)))))

     


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  PAST  <sck-ast-expr> ɕϊB

(defun sck-ast-expr (env past)
  ;; ̉͂̓B
  ;; Return: <sck-ast-expr>
  (let (($ (cadr past)))
    (case (car past)
      (iconst (sck-cps-iconst-expr past))
      (fconst (sck-cps-fconst-expr past))
      (cconst (sck-cps-cconst-expr past))
      (sconst (sck-cps-sconst-expr past))
      (ident  (sck-cps-vref-expr env past))
      (sizeof (sck-ast-sizeof-expr env past))
      (cast   (sck-cps-cast-expr env past))
      ((member pmember) (sck-cps-member-expr env past))
      (call   (sck-cps-call-expr env past))
      (t ; (<OP> <Pos> <Expr>..) Ƃ`
       (sck-cps-uniform-expr env past)))))

(defun sck-cps-iconst-expr (past)
  ;; PAST:
  ;; <Iconst> ::= (iconst <Pos> <String:<Integer>> <ISuffixes>)
  (let* ((numstr (caddr past))
	 (num (str-to-bn numstr))
	 (suffix (cadddr past))
	 (radix (if (string-match "^#[ox]" numstr) 'octhex 'decimal))
	 (type (sck-ast-typeof-iconst num radix suffix)))
    (sck-ast-make-iconst-expr type
			      num
			      :overflowmsg
			      "integer constant too big")))

(defun sck-cps-fconst-expr (past)
  ;; PAST:
  ;; <Fconst> ::= (fconst <Pos> (<Ipart> <Fpart> <Epart>) <FSuffixes>)
  (let* ((ifenum (caddr past))
	 (suffix (cadddr past))
	 (type (sck-ast-typeof-fconst suffix)))
    (sck-ast-make-fconst-expr type (read (apply 'format "%s.%se%s" ifenum)))))
			      
(defun sck-cps-cconst-expr (past)
  ;; PAST:
  ;; <Cconst> ::= (cconst <Pos> (<Char>) <CSuffixes>)
  (let* ((num (caddr past))
	 (suffix (cadddr past))
	 (type (sck-ast-typeof-cconst suffix)))
    (sck-ast-make-iconst-expr type
			      num
			      ;;  char c = '\xabe'; ̌
			      :overflowmsg "character constant too big"
			      )))

(defun sck-cps-sconst-expr (past)
  ;; PAST:
  ;; <Sconst> ::= (sconst <Pos> (<Char>..) <SSuffixes>)
  (let* ((nums (caddr past))
	 (suffix (cadddr past))
	 (arytype (sck-ast-typeof-sconst (length nums) suffix)))
    (sck-ast-make-expr 'const
		       :type arytype
		       :value (mapcar
			       #'(lambda (n)
				   (sck-ast-make-iconst-expr-value
				    (caddr arytype)
				    n
				    :overflowmsg
				    "a character in string is out of range"))
			       nums))))

(defun sck-cps-vref-expr (env past)
  ;; ident (kind=var) ̕ϐƂĂ̎Q
  (assert (eq (car past) 'ident))
  (let* ((name (caddr past))
	 (ident (sck-ast-search-ident env name)))
    (cond
     ;; o^Ă
     (ident
      (ecase (sck-ast-ident-kind ident)
	(var (sck-ast-make-vref-expr ident))
	(const
	 (assert (sck-ast-enum-type-p (sck-ast-ident-type ident)))
	 (sck-ast-make-iconst-expr (sck-ast-ident-type ident)
				   (sck-ast-ident-value ident)))))
     (t
      ;; `Ȃ error ^Ƃēo^B
      ;;  錾ĂȂ֐QƂ̌B
      (setq ident (sck-ast-enter-ident env
				       (sck-ast-make-ident past
							   :kind 'var
							   :type 'error)))
      ;;  cpps ŃG[wEς݂ @error Ȃ烁bZ[WoȂĂ悢B
      (sck-ast-error "undefined variable `%s'" ident)
      (sck-ast-make-vref-expr ident)))))

(defun sck-ast-sizeof-expr (env past)
  (assert (eq (car past) 'sizeof))
  (let ((arg (caddr past)))
    (cond
     ((eq (car arg) 'declaration)
      (multiple-value-bind (type ident paramidents)
	  (sck-cps-decl-declaration-simple env arg)
	(when ident
	  (sck-ast-warning "unexpected identifier `%s'" ident))
	(sck-ast-make-sizeof-expr type)))
     (t
      (sck-ast-make-sizeof-expr
       (sck-ast-expr-type
	(sck-ast-expr env arg)))))))


(defun sck-cps-cast-expr (env past)
  ;; (cast <Pos> <Declaration> <Expr>)
  (let ((declaration (caddr past))
	(expr (cadddr past)))
    (multiple-value-bind (type ident paramidents)
	(sck-cps-decl-declaration-simple env declaration)
      (when ident
	(sck-ast-warning "unexpected identifier `%s'" ident))
      (sck-ast-complete-expr
       (sck-ast-make-expr 'cast
			  :type type
			  :args (list (sck-ast-expr env expr)))))))

(defun sck-cps-member-expr (env past)
  ;; (member <Pos> <Expr> <Ident>)
  ;; (pmember <Pos> <Expr> <Ident>)
  (sck-ast-complete-expr
   (sck-ast-make-expr
    (car past)
    :args (list 
	   (sck-ast-expr env (caddr past))
	   (sck-ast-make-vref-expr
	    (sck-ast-make-ident (cadddr past)
				:kind 'var
				:type 'int ; _~[
				))))))

(defun sck-cps-call-expr (env past)
  ;; KvȂ錾ĂȂ֐
  ;; foo(){bar();}
  ;; ̓o^_ sck-cps-uniform-expr ƓB
  (when (eq (car (caddr past)) 'ident)
    (let* ((pident (caddr past))
	   (name (caddr pident)))
      (unless (sck-ast-search-ident env name)
	;; O[oɓo^
	(sck-ast-enter-ident sck*cps-global-env
			     (sck-ast-make-ident
			      pident
			      :kind 'var
			      :type sck=ast-default-function-type)))))
  ;;
  (sck-cps-uniform-expr env past))
  

(defun sck-cps-uniform-expr (env past)
  ;; (<OP> <Pos> <Expr> <Expr>..)
  (sck-ast-complete-expr
   (sck-ast-make-expr (car past)
		      :args (mapcar #'(lambda (e) (sck-ast-expr env e))
				    (cddr past)))))


;;;
(provide 'sck-cps-expr)
