;;;; 
;;;; LastChange: Wed Apr 11 01:08:44 2007


;;;   plain AST
;;; 
;;; <Stmt> ::=  <Expr>
;;;          | <Declaration>
;;;          | <Block>
;;;          | (label <Pos> <Ident> <Stmt>)  typedef int L; f() {L:;} ̌...
;;;          | (case <Pos> <Expr> <Stmt>)
;;;          | (default <Pos> <Stmt>)
;;;          | (if <Pos> <Expr> <Stmt> <<Stmt>|nil>)
;;;          | (switch <Pos> <Expr> <Stmt>)
;;;          | (while <Pos> <Expr> <Stmt>)
;;;          | (do <Pos> <Stmt> <Expr>)
;;;          | (for <Pos> (<<Expr>|nil> <<Expr>|nil> <<Expr>|nil>) <Stmt>)
;;;          | (goto <Pos> <Ident>)
;;;          | (continue <Pos>)
;;;          | (break <Pos>)
;;;          | (return <Pos> <<Expr>|nil>)
;;;          | (empty <Pos>)
;;; 
;;; <Block> ::= (block <Pos> <Stmt>..)




(defvar sck*cps-current-function-ident nil)
(defvar sck*cps-current-function-type nil)
(defvar sck*cps-current-function-label-alist nil)

(defvar sck*cps-stmt-stack
  ;; ((<while|do|for|switch> <StmtName> ...)..)
  nil)

(defun sck-cps-function-body (paramenv fident past)
  (assert (eq (car past) 'block))
  (let ((sck*cps-stmt-stack nil)
	(sck*cps-current-function-ident fident)
	(sck*cps-current-function-type (sck-ast-ident-type fident))
	(sck*cps-current-function-label-alist nil))
    (assert (sck-ast-function-type-p sck*cps-current-function-type))
    (prog1
	`(lambda ,sck*cps-current-function-type
	   ,paramenv
	   ,(sck-cps-block-stmt paramenv past))
      ;; `x̃`FbN
      ;; G[bZ[W̍sԍɖ肠B
      (dolist (a sck*cps-current-function-label-alist)
	(unless (cadr a)
	  (sck-ast-error "undefined label `%s'" (car a))))
      )))
    
(defun sck-cps-search-stmt-stack (syms)
  (assoc* syms sck*cps-stmt-stack :test #'(lambda (syms s) (memq s syms))))

(defun sck-cps-search-label (name)
  ;; sck*cps-current-function-label-alist
  ;;  ((<Name> <Defined>)..) ƂXgB
  ;; <Defined> ::= nil | t
  ;; GgȂΖ`ŒǉAGgԂB
  (let ((a (assoc name sck*cps-current-function-label-alist)))
    (unless a
      (setq a (list name nil))
      (push a sck*cps-current-function-label-alist))
    a))

(defun sck-cps-defined-label-p (name)
  ;; `ς݂eXgB
  (cadr (sck-cps-search-label name)))

(defun sck-cps-define-label (name)
  ;; `B
  (let ((a (sck-cps-search-label name)))
    (assert (not (cadr a)))
    (setf (cadr a) t)))


(defun sck-cps-stmt (env past)
  ;; Return: <Stmt> | nil
  (let (($ (cadr past)))
    (case (car past)
      ;; 錾͕ƍ݂ėǂB
      (declaration (sck-cps-decl-declaration-local env past) nil)
      ;; ݂ cpps  struct foo {int f(){};}; ĂB
      (function (sck-ast-error "cannot define local function") nil)
      (block (sck-cps-block-stmt env past))
      (if (sck-cps-if-stmt env past))
      (switch (sck-cps-switch-stmt env past))
      (case (sck-cps-case-stmt env past))
      (default (sck-cps-default-stmt env past))
      (while (sck-cps-while-stmt env past))
      (do (sck-cps-do-stmt env past))
      (for (sck-cps-for-stmt env past))
      (break (sck-cps-break-stmt env past)) 
      (continue (sck-cps-continue-stmt env past))
      (label (sck-cps-label-stmt env past))
      (goto (sck-cps-goto-stmt env past))
      (return (sck-cps-return-stmt env past))
      (empty nil)
      (t
       (sck-ast-expr env past)
       ))))


(defun sck-cps-stmt-condition (env past)
  (sck-ast-make-conv-bool-expr (sck-ast-expr env past)))
    
(defun sck-cps-block-stmt (env past)
  ;; PAST: (block <Pos> <Stmt>..)
  (let ((newenv (sck-ast-make-env env))
	(stmts nil))
    (dolist (pstmt (cddr past))
      (let ((stmt (sck-cps-stmt newenv pstmt)))
	(when stmt
	  (push stmt stmts))))
    (prog1
	`(block ,newenv ,@(nreverse stmts))
      ;;  local static ̃l[
      )))

(defun sck-cps-if-stmt (env past)
  `(if ,(sck-cps-stmt-condition env (elt past 2))
       ,(sck-cps-stmt env (elt past 3))
     ,(and (elt past 4)
	   (sck-cps-stmt env (elt past 4)))))

(defun sck-cps-switch-stmt (env past)
  ;; PAST: (switch <Pos> <Expr> <Stmt>)
  (let* ((sck*cps-stmt-stack sck*cps-stmt-stack)
	 (name (sck-ast-genname "switch"))
	 (stmt (copy-list `(switch ,name nil nil)))
	 (expr nil)
	 (body nil))
    (push (list 'switch name) sck*cps-stmt-stack)
    (setq expr (sck-ast-make-decay-expr (sck-ast-expr env (caddr past)))
	  body (sck-cps-stmt env (cadddr past)))
    (unless (sck-ast-integral-type-p (sck-ast-expr-type expr))
      (sck-ast-error "illegal switch selector type"))
    (setf (elt stmt 2) (sck-ast-make-conv-expr sck=ast-switch-type expr)
	  (elt stmt 3) body)
    ;;
    stmt))

(defun sck-cps-case-stmt (env past)
  (let ((sstmt (sck-cps-search-stmt-stack '(switch)))
	(cexpr (sck-ast-fold-const
		(sck-ast-make-decay-expr
		 (sck-ast-expr env (caddr past)))))
	(stmt (sck-cps-stmt env (cadddr past))))
    (cond
     ((null sstmt)
      (sck-ast-error "unexpected case label")
      stmt)
     ((not (sck-ast-iconst-expr-p cexpr))
      (sck-ast-error "illegal case label")
      stmt)
     (t
      (let ((cval (sck-ast-expr-value cexpr)))
	(cond
	 ((member cval (cddr sstmt))
	  (sck-ast-error "duplicated case label %s" cval)
	  stmt)
	 (t
	  (nconc sstmt (list cval))
	  `(case ,(cadr sstmt) ,cval ,stmt))))))))
	  
(defun sck-cps-default-stmt (env past)
  (let ((sstmt (sck-cps-search-stmt-stack '(switch))))
    (cond
     ((null sstmt)
      (sck-ast-error "unexpected default label")
      nil)
     ((member 'default (cddr sstmt))
      (sck-ast-error "duplicated default label")
      nil)
     (t
      (nconc sstmt (list 'default))
      `(default ,(cadr sstmt) ,(sck-cps-stmt env (caddr past)))))))

(defun sck-cps-while-stmt (env past)
  (let ((sck*cps-stmt-stack sck*cps-stmt-stack)
	(stmt (copy-list `(while ,(sck-ast-genname "while")
			    ,(sck-cps-stmt-condition env (elt past 2))
			    nil))))
    (push stmt sck*cps-stmt-stack)
    (setf (cadddr stmt) (sck-cps-stmt env (elt past 3)))
    stmt))

(defun sck-cps-for-stmt (env past)
  ;; (for <Pos> (<<Expr>|nil> <<Expr>|nil> <<Expr>|nil>) <Stmt>)
  (let* ((sck*cps-stmt-stack sck*cps-stmt-stack)
	 (e3 (caddr past))
	 (body (cadddr past))
	 (stmt
	  ;;  copy-list ͉ (setf (elt stmt 3)...) 
	  ;; R[ĥ󂳂Ȃ߁BoCgRpCƂ̊댯B
	  ;; ̕ł copy-list ̈Ӑ}lłB
	  (copy-list 
	   `(for ,(sck-ast-genname "for")
		 (,(and (elt e3 0) (sck-ast-expr env (elt e3 0)))
		  ,(and (elt e3 1) (sck-cps-stmt-condition env (elt e3 1)))
		  ,(and (elt e3 2) (sck-ast-expr env (elt e3 2))))
		 nil))))
    (push stmt sck*cps-stmt-stack)
    (setf (elt stmt 3) (sck-cps-stmt env body))
    stmt))

(defun sck-cps-do-stmt (env past)
  (let ((sck*cps-stmt-stack sck*cps-stmt-stack)
	(stmt (copy-list
	       `(do ,(sck-ast-genname "do")
		    ,(sck-cps-stmt-condition env (elt past 3))
		  nil))))
    (push stmt sck*cps-stmt-stack)
    (setf (cadddr stmt) (sck-cps-stmt env (elt past 2)))
    stmt))

(defun sck-cps-break-stmt (env past)
  (let ((lpstmt (sck-cps-search-stmt-stack '(for while do switch)))
	(label nil))
    (cond
     (lpstmt
      `(break ,(cadr lpstmt)))
     (t
      (sck-ast-error "unexpected break")
      nil))))

(defun sck-cps-continue-stmt (env past)
  (let ((lpstmt (sck-cps-search-stmt-stack '(for while do))))
    (cond
     (lpstmt
      `(continue ,(cadr lpstmt)))
     (t
      (sck-ast-error "unexpected continue")
      nil))))

(defun sck-cps-label-stmt (env past)
  (let* ((lident (caddr past))
	 (lname (caddr lident))
	 (stmt (cadddr past)))
    (cond
     ((sck-cps-defined-label-p lname)
      (sck-ast-error "duplicated label `%s'" lname))
     (t
      (sck-cps-define-label lname)))
    ;;
    `(label ,lname ,(sck-cps-stmt env stmt))))

(defun sck-cps-goto-stmt (env past)
  (let* ((lident (caddr past))
	 (lname (caddr lident)))
    (sck-cps-search-label lname)
    `(goto ,lname)))
	

(defun sck-cps-return-stmt (env past)
  (let ((e (caddr past))
	(rtype (caddr sck*cps-current-function-type)))
    (when e
      (setq e (sck-ast-make-conv-expr rtype (sck-ast-expr env e))))
    (if (sck-ast-void-type-p rtype)
	'(return nil)
      `(return ,e))))

;;;
(provide 'sck-cps-stmt)
