;;;; C Parser
;;;; LastChange: Fri Apr 13 12:45:40 2007

;;; Export:
;;; sck-cps : <cpps's output> -> <CpsOutput>
;;;
;;; Type:
;;; <CpsInput> ::= ((name <String>) 
;;;                 (source <String>))
;;;
;;; <CpsOutput> ::= ((ast <AST>)
;;;                  (messages (<String>..)))

;;(require 'sck-clex)
(require 'sck-cpps)
(require 'sck-ast)

(require 'sck-cps-decl)
(require 'sck-cps-expr)
(require 'sck-cps-stmt)


(defconst sck=cps-input-example
  "/* cps ̃fpvO */
struct data {
  int id;
  char *name;
  int flags[10];
};

struct data datav[] = {{0,\"foo\",{1,2,3}},  {1,\"bar\",{4}},  {2}, 0};

struct node {
  struct node *next;
  int key;
  struct data data;
};

typedef struct node Node;


int flagok(int flags[10])
{
  int flag;
  int i;

  for (i = 0, flag = 0; i < 10; i++)
    flag |= flags[i];

  return flag == 0xffffffff;
}

Node *search(Node *root, int key, int (*flagok)(int [10]))
{
  Node *p;

  for (p = root; p; p = p->next)
    if (key == p->key && flagok(p->data.flags))
      return p;

  return (Node*)0;
}
")


(defun sck-cps-demo ()
  (interactive)
  (sck-cps-str sck=cps-input-example))

(defun sck-cps-buffer ()
  "obt@̃vvς C-\[X ͂ AST \B"
  (interactive)
  (sck-cps-region (point-min) (point-max)))

(defun sck-cps-region (beg end)
  "[W̃vvς C-\[X ͂ AST \B"
  (interactive "r")
  (sck-cps-str (buffer-substring-no-properties beg end)))

(defun sck-cps-str (str)
  "\̃vvς C-\[X ͂ AST \B"
  (with-display-buffer "*sck-cps*"
    (emacs-lisp-mode)
    (sck-ast-font-lock)
    (let ((output (sck-cps
		   (sck-cpps
		    (sck-clex
		     `((name "stdin")
		       (source ,str)))))))
      (insert (sck-ast-pp-output output))
      output)))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; gbvx

(defun sck-cps (cpps-output)
  "INPUT: <cpps's output>
Return: <CpsOutput>"
  (sck-letalist cpps-output
      (plain-ast file-names messages)
    (sck-ast-message-init file-names)
    (sck-ast-genname-init)
    (let ((iast-module (sck-cps-module plain-ast))
	  (messages (sck-ast-sort-message-strings
		     (append messages
			     (sck-ast-message-strings)))))
      ;;
      `((ast ,(sck-ast-module-to-sexp iast-module))
	(messages ,messages)))))

				       
;;;
(provide 'sck-cps)
