;;;
;;; $Id: sck-dom.el,v 1.4 2007/01/31 07:55:29 heavy Exp $
;;;
;;; <sck-dom-input> ::= ((bblocks ((<Label> (<Succ>..))..)))
;;;
;;; <Label> ::= <Symbol>
;;; <Succ> ::= <Label>
;;;
;;; <sck-dom-output> ::= ((dominators ((<Label> (<Dom>..))..)))
;;;
;;; <Dom> ::= <Label>
;;; 
(defconst sck=dom-sample-input
  '((bblocks (("B1" ("B2"))
	      ("B2" ("B3" "B4"))
	      ("B3" ("B1"))
	      ("B4" ("B1"))))))

(defstruct sck-dom-bblock
  label
  succ
  pred
  dom)

;;;
;;;
;;;
(defun sck-dom-demo ()
  "sck-dom $B$r%G%b$9$k(B"
  (interactive)
  (let ((input sck=dom-sample-input)
	(buf (get-buffer-create "*sck-dom*")))
    (save-excursion
      (set-buffer buf)
      (erase-buffer)
      (insertf "Input:\n%s\nOutput:\n%s"
	       (pp-to-string input)
	       (pp-to-string (sck-dom input)))
      (display-buffer buf))))

;;;
;;;
;;;
(defun sck-dom (input)
  (sck-letalist input
      ((bblocks input-bblocks))
    (let ((bblocks (mapcar #'(lambda (b)
			       (make-sck-dom-bblock
				:label (car b)))
			   input-bblocks)))
      ;; init succ,pred
      (dolist (b bblocks)
	(setf (sck-dom-bblock-succ b)
	      (mapcar #'(lambda (succlabel)
			  (let ((bs (find succlabel bblocks
					  :key 'sck-dom-bblock-label
					  :test 'equal)))
			    (push b (sck-dom-bblock-pred bs))
			    bs))
		      (cadr (assoc (sck-dom-bblock-label b) input-bblocks)))))
      (sck-dom-make-dom bblocks)
      ;; make outputdata
      (list (list 'dominators
		  (mapcar #'(lambda (b)
			      (list (sck-dom-bblock-label b)
				    (sck-dom-bblock-dom b)))
			  bblocks))))))


(defun sck-dom-make-dom (bblocks)
  (let ((changed t)
	(init-dom (mapcar 'sck-dom-bblock-label bblocks)))
    ;; init dom
    (setf (sck-dom-bblock-dom (car bblocks))
	  (cons (sck-dom-bblock-label (car bblocks)) nil))
    (dolist (b (cdr bblocks))
      (setf (sck-dom-bblock-dom b) init-dom))
    ;; make dom
    (while changed
      (setq changed nil)
      (dolist (bb (cdr bblocks))
	;; DOM(B) = union ({B} (intersection DOM(PREDS(B))))
	(let* ((pred-dom (mapcar 'sck-dom-bblock-dom (sck-dom-bblock-pred bb)))
	       (dom (car pred-dom)))
	  (dolist (is (cdr pred-dom))
	    (setq dom (intersection dom is)))
	  ;; $B<+J,<+?H$rF~$l$k(B
	  (unless (member (sck-dom-bblock-label bb) dom)
	    (push (sck-dom-bblock-label bb) dom))
	  ;; $BJQ$C$F$$$?$i(B dom $B$r%;%C%H$7(B changed
	  (unless (equal dom (sck-dom-bblock-dom bb))
	    (setf (sck-dom-bblock-dom bb) dom)
	    (setq changed t)))))))


(provide 'sck-dom)
