;;; ex4-ans.el

(load "ex3-ans.el")


(defun my-foldconst (exp)
  ;; <exp> = <symbol> | <number> | (<op> <exp> <exp>)
  ;; <op>  = + | - | * | /
  ;; 
  ;; (my-foldconst '(+ (+ a (* (/ (- b 0) (- 2 1)) (* c (/ 3 (+ 1 2))))) 0))
  ;; => (+ a (* b c))
  (if (or (symbolp exp) (numberp exp))
      exp
    (let ((op (car exp))
	  (lexp (my-foldconst (cadr exp)))
	  (rexp (my-foldconst (caddr exp))))
      (cond
       ((and (numberp lexp) (numberp rexp))
	(funcall op lexp rexp))
       ((or (and (memq op '(+ -)) (equal rexp 0))
	    (and (memq op '(* /)) (equal rexp 1)))
	lexp)
       (t (list op lexp rexp))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *my-eval-toplevel-env*
  ;; $B%7%s%\%k$N%H%C%W%l%Y%k$K$*$1$kCM$rJ];}$9$k%H%C%W%l%Y%k4D6-!J(Ba-list$B!K!#(B
  '(;; $BDj?t!JDj5A:Q$_JQ?t!K(B
    (nil nil)
    (t t)
    ;; $B%W%j%_%F%#%V4X?t(B
    (car my-car)
    (cdr my-cdr)
    (cons my-cons)
    (atom my-atom)
    (equal my-equal)))

(defun my-eval-symbol-value (sym env)
  ;; $B%7%s%\%k(B sym $B$NCM$r%m!<%+%k4D6-(B env $B$+$iC5$7$FJV$9!#(B
  ;; env $B$KEPO?$5$l$F$$$J$1$l$P!"%H%C%W%l%Y%k4D6-(B *my-eval-toplevel-env*
  ;; $B$+$iC5$9!#(B
  ;; (my-eval-symbol-value 'foo '((foo 1) (bar 2))) => 1
  ;; (my-eval-symbol-value 'car nil) => my-car
  ;; $B%H%C%W%l%Y%k4D6-$K$bL5$1$l$P0J2<$N%(%i!<$r=P$9!#(B
  ;; (error "Unbound variable: %s" sym)
  (let ((a (or (assq sym env)
	       (assq sym *my-eval-toplevel-env*))))
    (unless a
      (error "Unbound variable: %s" sym))
    (cadr a)))

(defun my-eval-setq (sym val)
  ;; $B%7%s%\%k(B sym $B$N%H%C%W%l%Y%k$K$*$1$kCM$r(B val $B$K%;%C%H$7!"(Bval $B$rJV$9!#(B
  ;; (my-eval-setq 'poi '(p o i)) => poi
  ;; (my-eval-symbol-value 'poi nil) => (p o i)
  ;; sym $B$,%7%s%\%k$G$J$$>l9g!"0J2<$N%(%i!<$r=P$9!#(B
  ;; (error "Illegal symbol: %s" sym)
  (unless (symbolp sym)
    (error "Illegal symbol: %s" sym))
  (setq *my-eval-toplevel-env*
	(cons (list sym val) *my-eval-toplevel-env*))
  val)

(defun my-eval-toplevel (sexp)
  ;; $B%U%)!<%`(B sexp $B$r6u$N%m!<%+%k4D6-$N2<$GI>2A$9$k!#(B
  ;; (my-eval-toplevel '(car '(a b))) => a
  (my-eval sexp nil))

(defun my-eval (sexp env)
  ;; $B%U%)!<%`(B sexp $B$r%m!<%+%k4D6-(B env $B$N2<$GI>2A$9$k!#(B
  ;; $B%9%Z%7%c%k%U%)!<%`$O(B if, setq, defun $B$r%5%]!<%H!#(B
  ;; $B%W%j%_%F%#%V4X?t$O(B car, cdr, cons, atom, equal $B$r%5%]!<%H!#(B
  ;; (my-eval '(cons x y) '((x a) (y (b c)))) => (a b c)
  ;; $BK\4X?t$G$O!"(Bsexp $B$,%7%s%\%k5Z$S$=$NB>$N%"%H%`$N>l9g$N=hM}$r9T$$!"(B
  ;; sexp $B$,(B nil $B$G$J$$%j%9%H$N>l9g$O(B my-eval-call $B$KG$$;$k!'(B
  ;; ... (my-eval-call (car sexp) (cdr sexp) env)
  (cond
   ((symbolp sexp) (my-eval-symbol-value sexp env))
   ((atom sexp) sexp)
   (t (my-eval-call (car sexp) (cdr sexp) env))))

(defun my-eval-call (fun args env)
  ;; $B%9%Z%7%c%k%U%)!<%`$^$?$O4X?t(B fun $B$N=hM}$r9T$&!#(B
  ;; (my-eval-call 'if '(t a b) '((a 1))) => 1
  ;; (my-eval-call 'if '(nil a b) '((a 1))) => Error: Unbound variable: b
  ;; $BK\4X?tFb$G$O!"%9%Z%7%c%k%U%)!<%`$N>l9g$N=hM}$r9T$$!"(B
  ;; $B4X?t8F$S=P$7!J%9%Z%7%c%k%U%)!<%`0J30!K$N>l9g$O(B my-eval-apply $B$KG$$;$k!#(B
  ;; my-eval-apply $B$K$O!"(Bfun $B$rI>2A$7$F<h$j=P$7$?4X?tK\BN!"(B
  ;; $B5Z$S(B args $B$N3FMWAG$rI>2A$7$?%j%9%H$rEO$9!#(B
  (case fun
   ;; Special forms
   ((quote) (car args))
   ((if)
    (if (my-eval (car args) env)
	(my-eval (cadr args) env)
      (my-eval (caddr args) env)))
   ((setq)
    (my-eval-setq (car args) (my-eval (cadr args) env)))
   ((defun)
    (my-eval-setq (car args) (list 'lambda (cadr args) (caddr args)))
    (car args))
   ;; Functoin call
   (t
    (my-eval-apply (my-eval fun env)
		   (mapcar '(lambda (s) (my-eval s env)) args)))))
  
(defun my-eval-apply (funobj args)
  ;; $B4X?tK\BN(B funobj $B$r!JI>2A:Q$_!K0z?t(B args $B$KE,MQ$9$k!#(B
  ;; $BK\4X?t$G$O(B funobj $B$,%7%s%\%k$N>l9g$O(B my-eval-apply-primitive $B$r(B
  ;; $B$=$&$G$J$1$l$P%i%`%@<0$G$"$k$H2>Dj$7$F(B my-eval-apply-lambda $B$r8F$S=P$9!#(B
  (if (symbolp funobj)
      (my-eval-apply-primitive funobj args)
    (my-eval-apply-lambda funobj args)))

(defun my-eval-apply-primitive (prim args)
  ;; $B%W%j%_%F%#%V(B prim $B$r!JI>2A:Q$_!K0z?t(B args $B$KE,MQ$9$k!#(B
  ;; $B%W%j%_%F%#%V$O(B my-FOO $B$H$$$&%7%s%\%k$G$"$j!"(BEmacsLisp $B$N(B FOO $B$HF1$80UL#!#(B
  ;; (my-eval-apply-primitive 'my-cons '(a (b c))) => (a b c)
  ;; prim $B$,(B my-{car,cdr,cons,atom,equal} $B0J30$N>l9g$O0J2<$N%(%i!<$r=P$9!#(B
  ;; (error "Unknwon primitive function: %s" prim)
  (case prim
   (my-car (car (car args)))
   (my-cdr (cdr (car args)))
   (my-cons (cons (car args) (cadr args)))
   (my-atom (atom (car args)))
   (my-equal (equal (car args) (cadr args)))
   (t
    (error "Unknwon primitive function: %s" prim))))

(defun my-eval-apply-lambda (lam args)
  ;; $B%i%`%@<0(B lam $B$r!JI>2A:Q$_!K0z?t$KE,MQ$9$k!#(B
  ;; (my-eval-apply-lambda '(lambda (x y) (cons x y)) '(a (b c))) => (a b c)
  ;; lam $B$,(B lambda $B$G;O$^$k%j%9%H$G$J$1$l$P0J2<$N%(%i!<$r=P$9!#(B
  ;; (error "Illegal function: %s" lam)$B!!(B
  ;; $B@5Ev$J%i%`%@<0$+$I$&$+$K$D$$$F$N!"$h$j>\:Y$J%A%'%C%/$OITMW!#(B
  (unless (eq (car-safe lam) 'lambda)
    (error "Illegal function: %s" lam))
  (my-eval (caddr lam) (my-zip (cadr lam) args)))


(defun my-eval-test ()
  ;; elim $B%i%$%V%i%j$K$h$j%F%9%H$r9T$&!#(B
  ;; ESC x my-eval-test $B$G5/F0!#(B
  (interactive)
  (elim "*my-eval*") 
  (setq elim*eval 'my-eval-toplevel))
