;;; ex5-ans.el

(load "ex4-ans.el")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; while, dotimes, dolist

(defun my-member-by-while (x y)
  (while (and (consp y)
	      (not (equal x (car y))))
    (setq y (cdr y)))
  y)

(defun my-elt-by-dotimes (x n)
  (dotimes (i n)
    (setq x (cdr x)))
  (car x))

(defun my-reverse-by-dolist (x)
  (let ((rev nil))
    (dolist (e x)
      (setq  rev (cons e rev)))
    rev))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; oCgR[h̖

(defun my-bytecomp (exp)
  ;;  <exp> X^bN}ṼR[h <bytecode> ɕϊB
  ;; <exp> = <symbol> | <number> | (<op> <exp> <exp>)
  ;; <op>  = + | - | * | /
  ;;
  ;; <bytecode>  = (<stacksize> <codeseq>)
  ;; <codeseq>   = <code> | <codeseq> <code>
  ;; <code>      = (vref <symbol>) | (const <number>) | (call <op>)
  ;; <stacksize> = <number>
  ;;
  ;; (my-bytecomp '(+ (- a (* b 3)) 4))
  ;; => (3  (vref a) (vref b) (const 3) (call *) (call -) (const 4) (call +))
  (cons (my-bytecomp-stacksize exp)
	(my-bytecomp-codeseq exp)))

(defun my-bytecomp-stacksize (exp)
  ;; <exp> ]邽߂ɍŏKv <stacksize> ԂBQlFmaxB
  (if (atom exp)
      1
    (max (my-bytecomp-stacksize (cadr exp))
	 (1+ (my-bytecomp-stacksize (caddr exp))))))

(defun my-bytecomp-codeseq (exp)
  ;; <exp>  (<codeseq>) ɕϊBQlFsymbolp, numberpB
  (cond
   ((symbolp exp) `((vref ,exp)))
   ((numberp exp) `((const ,exp)))
   (t `(,@(my-bytecomp-codeseq (cadr exp)) ,@(my-bytecomp-codeseq (caddr exp))
	  (call ,(car exp))))))

(defun my-exec-bytecode (bytecode env)
  ;; <bytecode>  env (a-list) ̉ŕ]BsX^bN
  ;; `ϐQƂꍇ (error "Unbound variable: %s" ...) oB
  ;;
  ;; (my-exec-bytecode (my-bytecomp '(- (* a 2) b)) '((a 2) (b 3)))
  ;; => 1
  (let ((stack (make-vector (car bytecode) nil))
	(sp -1))
    (dolist (code (cdr bytecode))
      (case (car code)
	(const
	 (aset stack (incf sp) (cadr code)))
	(vref
	 (let ((a (assq (cadr code) env)))
	   (unless a
	     (error "Unbound variable: %s" (cadr code)))
	   (aset stack (incf sp) (cadr a))))
	(call
	 (aset stack (1- sp) (funcall (cadr code)
				      (aref stack (1- sp))
				      (aref stack sp)))
	 (decf sp))
	(t
	 (error "Illegal code: %s" code))))
    (aref stack sp)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; KS̃C^v^̏C

(defun my-eval-setq (sym val)
  (unless (symbolp sym)
    (error "Illegal symbol: %s" sym))
  (let ((a (assq sym *my-eval-toplevel-env*)))
    (if a
	(setcar (cdr a) val)
      (setq *my-eval-toplevel-env*
	    (cons (list sym val) *my-eval-toplevel-env*))))
  val)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; IXR[v Lisp

(defconst *my-eval2-startup-alist*
  ;; C^v^ÑV{̒lZbgAbv邽߂ɁA
  ;; my-eval2-init QƂ a-listB
  '(;; 萔i`ςݕϐj
    (nil nil)
    (t t)
    ;; v~eBu֐
    (car my-car)
    (cdr my-cdr)
    (cons my-cons)
    (atom my-atom)
    (equal my-equal)))

;;@oCfBOX^bNB̔zŎBmy-eval2-init ŏB
(defconst *my-eval2-bind-stack-size* 100) ; ȉ̓̔z̃TCY
(defvar *my-eval2-bind-stack-sym*)        ; V{X^bNizj
(defvar *my-eval2-bind-stack-val*)        ; o[X^bNizj
(defvar *my-eval2-bind-stack-sp*)         ; X^bN|C^

;; `\IuWFNg
(defconst *my-eval2-unbound-value* (list 'unbound))


(defun my-eval2-init ()
  ;; {C^v^̏sB
  ;; oCfBOX^bNi̔zj쐬A
  ;; X^bN|C^ -1 ɃZbgB
  ;; *my-eval2-startup-alist* = ((sym val)...) ɏ]Amy-eval2-setq ɂ
  ;; C^v^ÑV{ sym ̒l  val ɐݒB
  (setq *my-eval2-bind-stack-sym* (make-vector *my-eval2-bind-stack-size* nil)
	*my-eval2-bind-stack-val* (make-vector *my-eval2-bind-stack-size* nil)
	*my-eval2-bind-stack-sp* -1)
  (dolist (a *my-eval2-startup-alist*)
    (my-eval2-setq (car a) (cadr a))))

(defun my-eval2-symbol-value-raw (sym)
  ;; V{ sym ̒lԂB
  ;; {V[oCfBO̎ł́AV{ sym ̒lƂāA
  ;; V{ 'my-eval2-value 𗘗pB
  ;; ̑l΂̒lA܂̑ȂΖ`ϐ̒lł
  ;; `li*my-eval2-unbound-value* ̒ljԂBG[ɂ͂ȂB
  ;; ⑫F (get sym 'my-eval2-value) ł͑ꍇ nil A邽߁A
  ;; `Ȃ̂ nil  put Ă̂ʏoȂ̂ōHvKvB
  ;; QlF symbol-plist, plist-memberB
  (let ((kv (plist-member (symbol-plist sym) 'my-eval2-value)))
    (if kv
	(cadr kv)
      *my-eval2-unbound-value*)))

(defun my-eval2-save-bind (sym)
  ;; V{̍̒loCfBOX^bNɃZ[uF
  ;;   1. X^bN|C^CNg
  ;;   2. V{ sym V{X^bNgbvɃZbg(aset)
  ;;   3. V{̒lo[X^bNgbvɃZbg(aset)
  (incf *my-eval2-bind-stack-sp*)
  (aset *my-eval2-bind-stack-sym* *my-eval2-bind-stack-sp* sym)
  (aset *my-eval2-bind-stack-val* *my-eval2-bind-stack-sp*
	(my-eval2-symbol-value-raw sym)))

(defun my-eval2-restore-binds (savedsp)
  ;; oCfBOX^bN *my-eval2-bind-stack-sp* = savedsp ܂Ŗ߂A
  ;; Z[uĂeV{̒lXgAĂB
  (while (< savedsp *my-eval2-bind-stack-sp*)
    (my-eval2-setq (aref *my-eval2-bind-stack-sym* *my-eval2-bind-stack-sp*)
		   (aref *my-eval2-bind-stack-val* *my-eval2-bind-stack-sp*))
    (decf *my-eval2-bind-stack-sp*)))
		      
(defun my-eval2-symbol-value (sym)
  ;; my-eval2-symbol-value-raw ƓlAV{ sym `ϐ
  ;; ꍇAȂ킿̒l *my-eval2-unbound-value* ̒lłꍇɂ
  ;; ȉ̃G[oB
  ;; (error "Unbound variable: %s" sym)
  (let ((val (my-eval2-symbol-value-raw sym)))
    (when (eq val *my-eval2-unbound-value*)
      (error "Unbound variable: %s" sym))
    val))

(defun my-eval2-setq (sym val)
  ;; ex4 ɂ my-eval-setq Ɠl̂ƂsA
  ;; V{ sym  'my-eval2-value  val iputjAval ԂB
  ;; sym V{łȂꍇAȉ̃G[oB
  ;; (error "Illegal symbol: %s" sym)
  (unless (symbolp sym)
    (error "Illegal symbol: %s" sym))
  (put sym 'my-eval2-value val))


(defun my-eval2 (sexp)
  ;; _Ci~bNoCfBÔ߁A_āA
  ;; ex4  my-eval ƓlB
  (cond
   ((symbolp sexp) (my-eval2-symbol-value sexp))
   ((atom sexp) sexp)
   (t (my-eval2-call (car sexp) (cdr sexp)))))

(defun my-eval2-call (fun args)
  ;; _Ci~bNoCfBÔ߁A_āA
  ;; ex4  my-eval-call ƓlB
  (case fun
   ;; Special forms
   ((quote) (car args))
   ((if)
    (if (my-eval2 (car args))
	(my-eval2 (cadr args))
      (my-eval2 (caddr args))))
   ((setq)
    (my-eval2-setq (car args) (my-eval2 (cadr args))))
   ((defun)
    ;; (my-eval2-setq (car args) `(lambda ,(cadr args) ,@(cddr args)))
    (my-eval2-setq (car args) (list* 'lambda (cadr args) (cddr args)))
    (car args))
   ((progn)
    (my-eval2-progn args))
   ((let let*)
    (my-eval2-letX fun (car args) (cdr args)))
   ;; Functoin call
   (t
    (my-eval2-apply (my-eval2 fun) (mapcar 'my-eval2 args)))))

(defun my-eval2-apply (funobj args)
  ;; ex4  my-eval-apply ƓlB
  (if (symbolp funobj)
      (my-eval2-apply-primitive funobj args)
    (my-eval2-apply-lambda funobj args)))

(defun my-eval2-apply-primitive (prim args)
  ;; ex4  my-eval-apply-primitive ƓlB
  (my-eval-apply-primitive prim args))

(defun my-eval2-apply-lambda (lam args)
  ;; ex4 ɂ my-eval-apply-lambda Ɠl̂ƂsA
  ;; V[oCfBO@ɂB
  ;; lam  lambda Ŏn܂郊XgłȂΈȉ̃G[oB
  ;; (error "Illegal function: %s" lam)@
  (unless (eq (car-safe lam) 'lambda)
    (error "Illegal function: %s" lam))
  (let ((savedsp *my-eval2-bind-stack-sp*))
    (dolist (sym (cadr lam))
      (my-eval2-save-bind sym)
      (my-eval2-setq sym (car args))
      (setq args (cdr args)))
    (prog1
	(my-eval2-progn (cddr lam))
      (my-eval2-restore-binds savedsp))))

(defun my-eval2-progn (forms)
  ;; progn ̏
  (let ((val nil))
    (dolist (form forms)
      (setq val (my-eval2 form)))
    val))

(defun my-eval2-letX (letX binds forms)
  ;; let  let* ̏
  ;; <letX>  = let | let*
  ;; <binds> = ((<symbol> <exp>)...)
  ;; <forms> = (<form>...)
  (let ((vars (mapcar 'car binds))
	(exps (mapcar 'cadr binds)))
    (ecase letX
      (let  (my-eval2-let  vars exps forms))
      (let* (my-eval2-let* vars exps forms)))))

(defun my-eval2-let (vars exps forms)
  ;; let ̏
  (let ((savedsp *my-eval2-bind-stack-sp*))
    (prog1
	(let ((vals (mapcar 'my-eval2 exps))) ; VȑOɑS]
	  (dolist (sym vars)
	    (my-eval2-save-bind sym)
	    (my-eval2-setq sym (car vals))
	    (setq vals (cdr vals)))
	  (my-eval2-progn forms))
      (my-eval2-restore-binds savedsp))))

(defun my-eval2-let* (vars exps forms)
  ;; let* ̏ilet Ƃ̔ȈႢɒӁj
  (let ((savedsp *my-eval2-bind-stack-sp*))
    (prog1
	(progn
	  (dolist (sym vars)
	    (my-eval2-save-bind sym)
	    (my-eval2-setq sym (my-eval2 (car exps))) ; ]
	    (setq exps (cdr exps)))
	  (my-eval2-progn forms))
      (my-eval2-restore-binds savedsp))))


(defun my-eval2-test ()
  ;; elim CuɂeXgsB
  (interactive)
  (elim "*my-eval2*")
  (my-eval2-init)
  ;; (my-eval2-load "my-lisp.lsp") ; eXgp֐`̃[h
  (setq elim*eval 'my-eval2))

(defun my-eval2-load (file)
  ;; my-eval2 ŕ]t@C[_B
  (interactive "fMy load file: ")
  (with-temp-buffer
    (insert-file-contents file)
    (while (condition-case nil
	       (progn (my-eval2 (read (current-buffer))) t)
	     (end-of-file nil)))))
