;;; ex8-ans.el

(load "ex7-ans.el")


(defun my-numbering-buffer ()
  (interactive)
  (my-numbering-region (point-min) (point-max)))

(defun my-wordfreq-region (beg end)
  (interactive "r")
  (save-restriction
    (narrow-to-region beg end)
    (my-wordfreq-buffer)))

(defun my-wordfreq-file (file)
  (interactive "fFile: ")
  (with-temp-buffer
    (insert-file-contents file)
    (my-wordfreq-buffer file)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; my-wordfreq

;; my-wordfreq.el  begin...

(require 'cl)

(defun my-wordfreq-batch ()
  ;; % EMACS -batch -l my-wordfreq.el -f my-wordfreq-batch  srcfile [outfile]
  (unless command-line-args-left
    (message "Usage: my-wordfreq srcfile [outfile]")
    (kill-emacs 1))
  (let ((srcfile (car command-line-args-left))
	(outfile (cadr command-line-args-left)))
    (with-temp-buffer
      (insert-file-contents srcfile)
      (my-wordfreq-buffer srcfile)
      (set-buffer "*my-wordfreq*")
      (if outfile
	  (write-region (point-min) (point-max) outfile)
	(princ (buffer-string)))))
  (kill-emacs 0))

(defun my-wordfreq-buffer (&optional srcname)
  (interactive)
  (my-wordfreq-buffer-display (my-wordfreq-buffer-count-words) srcname))

(defun my-wordfreq-buffer-display (alist srcname)
  ;; $B=P8=IQEY>pJs(B alist $B$r(B *my-wordfreq* $B%P%C%U%!$XI=<($9$k!#(B
  (let ((srcbuf (current-buffer))
	(dspbuf (get-buffer-create "*my-wordfreq*")))
    (save-excursion
      (set-buffer dspbuf)
      (erase-buffer)
      (insert (format "The frequency list of the words in %S\n\n"
		      (or srcname (buffer-name srcbuf))))
      (dolist (a alist)
	(insert (format "%-16s : %d\n" (car a) (cadr a))))
      (goto-char (point-min))
      (display-buffer dspbuf))))

(defun my-wordfreq-buffer-count-words ()
  ;; $B%+%l%s%H%P%C%U%!$N1QC18l$N=P8=IQEY$r0J2<$N$h$&$J(B alist $B$H$7$FJV$9!#(B
  ;; $B=g=x$O=P8=IQEY(B <freq> $B$N9b$$=g$K%=!<%H$5$l$F$$$k!#(B
  ;;  <alist> = ({(<words> <freq>)})
  ;;  <words> = <string>
  ;;  <freq>  = <number>
  (save-excursion
    (goto-char (point-min))
    (let ((alist nil)
	  (htab (make-hash-table :test 'equal)))
      (while (re-search-forward "[a-zA-Z][-a-zA-Z]*" nil t)
	(let ((word (match-string 0)))
	  (if (gethash word htab)
	      (incf (gethash word htab))
	    (setf (gethash word htab) 1))))
      (maphash '(lambda (key val)
		  (push (list key val) alist))
	       htab)
      (sort alist '(lambda (x y) (> (cadr x) (cadr y)))))))

;; my-wordfreq.el  end.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Web $B%V%i%&%6(B

(defun my-http-get-cont (url)
  ;; url $B$G;XDj$5$l$?(B http $B%5!<%P$K@\B3$7!"(Burl $B$G;XDj$5$l$F$$$k%U%!%$%k(B
  ;; $B$r$H$j=P$9(B GET $B%3%^%s%I$rAw$j!"%5!<%P$,JV$7$?%G!<%?$r$=$N$^$^2C9)$;$:(B
  ;; $BJ8;zNs$H$7$FJV$9!#(B
  (with-temp-buffer
    ;; url = http://host/file $B$+$i(B host $B$H(B file $B$r<h$j=P$9$?$a$N%^%C%A%s%0(B
    (unless (string-match "^http://\\([^/]+\\)/\\(.*\\)$" url)
      (error "my-http-get-cont: Illiegal url: %s" url))
    (let* ((host (match-string 1 url))
	   (file (match-string 2 url))
	   ;; with-temp-buffer $B$,:n$C$?$3$N%+%l%s%H%P%C%U%!(B
	   (buf (current-buffer))
	   ;; host $B$N(B http $B%5!<%P$X@\B3$9$k2>A[E*%W%m%;%9%*%V%8%'%/%H$r@8@.$7!"(B
	   ;; proc $B$K%P%$%s%I$9$k!#%W%m%;%9$N=PNO$O%P%C%U%!(B buf $B$KA^F~$5$l$k!#(B
	   ;; $B%P%C%U%!(B buf $B$r:o=|$9$k$3$H$G%W%m%;%9$b:o=|$5$l@\B3$,@Z$i$l$k!#(B
	   (proc (open-network-stream "*proc*" buf host "http")))
      ;; proc $B$X(B GET $B%3%^%s%I$rAw$k!#(B
      (process-send-string
       proc
       (format (concat "GET /%s HTTP/1.0\n"
		       "Host: %s\n"
		       "User-Agent: ex8\n"
		       "\n")
	       file host))
      ;; $BFI$_9~$_=hM}(B accept-process-output $B$rA4ItFI$_=*$o$k$^$G7+$jJV$9!#(B
      ;; $B!zK\Ev$O%k!<%W$N>e8B$r7h$a!"D6$($?$i%?%$%`%"%&%H%(%i!<$r=P$9$Y$-!#(B
      (while (/= (buffer-size)
		 (progn (accept-process-output proc 10)
			(buffer-size))))
      ;; $B%X%C%@!<$rD4$Y!"@.8y!J%3!<%I(B 200$B!K$G$J$+$C$?$i%(%i!<$r=P$9!#(B
      (goto-char (point-min))
      (unless (looking-at "HTTP/.* 200 ")
	(error "my-http-get-cont: GET command failed:\n%s"
	       (buffer-string)))
      ;; GET $B$7$?(B url $B$NFbMF$rJ8;zNs$H$7$FJV$9!#(B
      (buffer-string))))

(defun my-http-format-cont (cont)
  ;; http $B%5!<%P$,JV$7$?FbMFJ8;zNs(B cont $B$r0J2<$N$h$&$K$7$F%F%-%9%H$K@07A$7!"(B
  ;; $B@07A$7$?J8;zNs$H$7$FJV$9!#>/$J$/$H$b0J2<$NJQ49$r9T$&!'(B
  ;; 1. $B%X%C%@!<!J@hF,$+$i:G=i$N6u9T!J@55,I=8=!'(B"^$"$B!K$^$G!K$r:o=|$9$k!#(B
  ;; 2. HTML $B%?%0(B <...> $B$N=hM}!#(B<br> $B$O2~9T$K$7!"$=$NB>$OC1$K:o=|$9$k!#(B
  ;; 3. $B0J2<$NCV49$r9T$&!#(B
  ;;   &lt;   -> <
  ;;   &gt;   -> >
  ;;   &quot; -> "
  ;;   &amp;  -> &
  (with-temp-buffer
    (insert cont)
    ;; 1. $B%X%C%@!<$r:o=|(B
    (goto-char (point-min))
    (re-search-forward "^$")
    (delete-region (point-min) (+ (point) 1))
    ;; 2. $B$H(B 3. $B$NCV49(B
    (dolist (a '(("<br>" "\n")
		 ("<[^>]+>" "")
		 ("&lt;" "<")
		 ("&gt;" ">")
		 ("&quot;" "\"")
		 ("&amp;" "&")))
      (goto-char (point-min))
      (while (re-search-forward (car a) nil t)
	(replace-match (cadr a))))
    ;; $B%U%)!<%^%C%H$7$?FbMF$rJ8;zNs$H$7$FJV$9!#(B
    (buffer-string)))

(defun my-http-view (url &optional noformat)
  ;; (my-http-view "http://www.lisp.org/table/history.htm" t)
  ;; (my-http-view "http://www.lisp.org/table/history.htm")
  ;; (my-http-view "http://www.outpost9.com/reference/jargon/jargon_32.html")
  (interactive "surl: \nP")
  (let ((cont (funcall (if noformat 'identity 'my-http-format-cont)
		       (my-http-get-cont url)))
	(buf (get-buffer-create "*my-http-view*")))
    (save-excursion
      (set-buffer buf)
      (erase-buffer)
      (insert cont)
      (goto-char (point-min))
      (display-buffer buf))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun my-wordfreq-url (url)
  ;; (my-wordfreq-url "http://www.cs.cmu.edu/~rgs/alice-I.html")
  (interactive "surl: ")
  (with-temp-buffer
    (insert (my-http-format-cont (my-http-get-cont url)))
    (my-wordfreq-buffer url)))
