;;;
;;; $Id: graphviz.el,v 1.7 2007/01/27 05:44:53 heavy Exp $
;;;
;;; Emacs $B>e$G(B Graphviz $B$r;H$&$?$a$N%U%m%s%H%(%s%I(B
;;;
(defun graphviz-graph-to-dot (graph)
  ;; ([graph|digraph] GRAPH-NAME
  ;;  (graph ((size "...") (label "...") ...))
  ;;  (node ((shape "...") (color "...") ...))
  ;;  (edge ((...) ...))
  ;;  (node-list (NODE-NAME attr-alist...) ...) 
  ;;  (edge-list ((NODE-NAME NODE-NAME) attr-alist ...) ...)
  ;;  (add-attr-list-node (attr-list NODE-NAME NODE-NAME ...))
  ;;  (add-attr-list-edge (attr-list (NODE-NAME NODE-NAME) ...))
  (let ((graphviz*graph-attr (car graph))
	(print-escape-newlines t))
    (with-temp-buffer
      (if (memq graphviz*graph-attr '(graph digraph))
	  (insert (format "%s" (car graph)))
	(error "graph or digraph?"))
      (when (cadr graph)
	(insert (format " %s" (graphviz-to-dot-string (cadr graph)))))
      (graphviz-output-smtp-list (cddr graph))
      (buffer-string))))

(defun graphviz-output-smtp-list (smtp-list)
  (insert " {\n")
  (dolist (smtp smtp-list)
    (graphviz-output-smtp smtp smtp-list))
  (insert "}\n"))

(defun graphviz-output-smtp (smtp smtp-list)
  (cond
   ;; $B$3$l$i0J30$OL5;k$9$k(B
   ;; attr
   ((memq (car smtp) '(graph node edge))
    (when (cdr smtp)
      (insert (format "%s" (car smtp)))
      (graphviz-output-attr-list (cdr smtp))
      (insert ";\n")))
   ;; subgraph
   ((eq (car smtp) 'subgraph)
    (insert (format "subgraph %s" (cadr smtp)))
    (graphviz-output-smtp-list (cddr smtp))
    (insert ";\n"))
   ;; node-list
   ((eq (car smtp) 'node-list)
    (graphviz-output-node-list (cdr smtp)
			       (graphviz-add-attr-list-node smtp-list)))
   ;; edge-list
   ((eq (car smtp) 'edge-list)
    (graphviz-output-edge-list (cdr smtp)
			       (graphviz-add-attr-list-node smtp-list)
			       (graphviz-add-attr-list-edge smtp-list)))))

(defun graphviz-output-attr-list (attr-list)
  (insert " [")
  (dolist (attr attr-list)
    (insert (format "%s=%s," 
		    (graphviz-to-dot-string (car attr))
		    (graphviz-to-dot-string (cadr attr)))))
  (delete-char -1)
  (insert "]"))

(defun graphviz-output-node-list (node-list add-attr-node-list)
  (dolist (node node-list)
    ;; $B$3$3$G(Bport$BBP1~$,I,MW(B
    (let ((attr (cdr node)))
      (dolist (x add-attr-node-list)
	(when (member (car node) (cdr x))
	  (setq attr (append attr (car x)))))
      (when attr
	(insert (format "%s" (graphviz-to-dot-string (car node))))
	(graphviz-output-attr-list attr)))
    (insert ";\n")))

(defun graphviz-output-edge-list (edge-list 
				  add-attr-node-list
				  add-attr-edge-list)
  (let ((edge-op
	 (cond
	  ((eq graphviz*graph-attr 'graph) "--")
	  ((eq graphviz*graph-attr 'digraph) "->"))))
    (dolist (edge edge-list)
      (insert (format "%s%s%s" 
		      (graphviz-to-dot-string (caar edge))
		      edge-op
		      (graphviz-to-dot-string (cadar edge))))
      (let ((attr (cdr edge)))
	(dolist (x add-attr-node-list)
	  (when (or
		 (member (caar edge) (cdr x))
		 (member (cadar edge) (cdr x)))
	    (setq attr (append attr (car x)))))
	(dolist (x add-attr-edge-list)
	  (when (member (car edge) (cdr x))
	    (setq attr (append attr (car x)))))
	(when attr
	  (graphviz-output-attr-list attr)))
      (insert ";\n"))))

;;;
;;; Graphviz $B$NJ8;zNs$KJQ49(B
;;;
;;; \a -> $B:8B7$((B(\l)
;;; \n -> $B%;%s%?%j%s%0(B(\n)
;;; \r -> $B1&B7$((B(\r)
;;;
(defun graphviz-to-dot-string (str)
  (cond 
   ((symbolp str) (symbol-name str))
   (t
    (assert (stringp str))
    (format 
     "\"%s\""
     (replace-regexp-in-string
      "\r" "\\r"
      (replace-regexp-in-string 
       "\a" "\\l" 
       (replace-regexp-in-string
	"\n" "\\n"
	(replace-regexp-in-string "\"" "\\\"" str t t) t t) t t) t t)))))

;;
(defun graphviz-add-attr-list-edge (smtp-list)
  (let (al)
    (dolist (x smtp-list)
      (when (eq (car x) 'add-attr-edge-list)
	(push (cadr x) al)))
    al))

(defun graphviz-add-attr-list-node (smtp-list)
  (let (al)
    (dolist (x smtp-list)
      (when (eq (car x) 'add-attr-node-list)
	(push (cadr x) al)))
    al))

;;
;; $B%G!<%?:n@.4X?t7O(B
;;
(defun graphviz-make-graph (type id  &optional graph-attr node-attr edge-attr)
  "$B%0%i%U$r:n$k(B"
  (list type id
	(cons 'graph graph-attr)
	(cons 'node node-attr)
	(cons 'edge edge-attr)
	(cons 'node-list nil)
	(cons 'edge-list nil)))

(defun graphviz-set-alist (id val alist)
  ;; alist (alist-name alist-body...)
  (let ((m (assoc id (cdr alist))))
    (if m 
	(setf (cadr m) val)
      (setf (cdr alist) (cons (list id val) (cdr alist))))))

(defun graphviz-graph-label (graph label &optional loc just)
  "$B%0%i%U%i%Y%k@_Dj(B label:$B%i%Y%kL>(B loc:labelloc just:labeljust"
  (let ((attr (nth 2 graph)))
    (if (not (eq (car attr) 'graph))
	(error ""))
    (graphviz-set-alist 'label label attr)
    (when loc 
      (graphviz-set-alist 'labelloc loc attr)
    (when just
      (graphviz-set-alist 'labeljust just attr)))))

(defun graphviz-graph-set-edge (graph edge &optional attr setonly nocheck)
  "$B%0%i%U$K%(%C%8$r:n@.(B"
  (let (m edge-list node-list)
    (block add-edge
      (dolist (x (cddr graph))
	(when (eq (car x) 'edge-list)
	  (setq edge-list x)
	  (setq m (assoc edge (cdr edge-list)))
	  (when m
	    (return-from add-edge)))))
    (unless edge-list (error ""))
    ;; $BF1$8%N!<%I4V$G$b0c$&%(%C%8$r$D$1$k>l9g$,$"$k(B
    (if (and (not nocheck) m)
	(when attr
	  (dolist (x attr)
	    (graphviz-set-alist (car x) (cadr x) m)))
      (unless setonly
	;; add-attr-node-list $B$J$I$KBP1~$9$k$?$a(B
	;; node-list $B$KEPO?$r9T$&(B
	(setq node-list (assoc 'node-list (cddr graph)))
	(unless node-list (error ""))
	(block out-edge-1
	  (dolist (x (cdr node-list))
	    (when (equal (car x) (car edge))
	      (return-from out-edge-1)))
	  (setf (cdr node-list)
		(cons (cons (car edge) nil) (cdr node-list))))
	(block out-edge-2
	  (dolist (x (cdr node-list))
	    (when (equal (car x) (cadr edge))
	      (return-from out-edge-2)))
	  (setf (cdr node-list)
		(cons (cons (cadr edge) nil) (cdr node-list))))
	(setf (cdr edge-list) 
	      (cons (cons edge attr) (cdr edge-list)))))))

(defun graphviz-graph-delete-edge (graph edge)
  "$B%(%C%8$r>C$9(B"
  (dolist (x (cddr graph))
    (when (eq (car x) 'edge-list)
      (let ((entry (assoc edge (cdr x))))
	(delete entry x)))))

(defun graphviz-graph-set-node-attr (graph node attr)
  "$B%N!<%I$K%"%H%j%S%e!<%H$rIU$1$k(B"
  (let (m node-list)
    (block set-node-attr
      (dolist (x (cddr graph))
	(when (eq (car x) 'node-list)
	  (setq node-list x)
	  (setq m (assoc node (cdr node-list)))
	  (when m
	    (return-from set-node-attr)))))
    (unless node-list (error ""))
    (if m
	(dolist (x attr)
	  (graphviz-set-alist (car x) (cadr x) m))
      (setf (cdr node-list)
	    (cons (cons node attr) (cdr node-list))))))

(defun graphviz-graph-rename-node (graph node newname)
  "$B%N!<%I$N(BID$B$rJQ99(B"
  (dolist (x (cddr graph))
    (cond
     ((eq (car x) 'node-list)
      (let ((m (assoc node (cdr x))))
	(when m
	  (setf (car m) newname))))
     ((eq (car x) 'edge-list)
      (dolist (y (cdr x))
	(when (equal (caar y) node)
	  (setf (caar y) newname))
	(when (equal (cadar y) node)
	  (setf (cadar y) newname)))))))

;;
;; (union a b :test 'equal)
;;
(defun graphviz-union (l1 l2)
  (cond ((null l1) l2) ((null l2) l1)
	((equal l1 l2) l1)
	(t
	 (or (>= (length l1) (length l2))
	     (setq l1 (prog1 l2 (setq l2 l1))))
	 (while l2
	   (or (member (car l2) l1)
	       (push (car l2) l1))
	   (pop l2))
	 l1)))

(defun graphviz-graph-set-add-attr-key-list (graph key attr nodes)
  (block set-attr
    (dolist (x (cddr graph))
      (when (and (eq (car x) key)
		 (equal (caadr x) attr))
	(setf (cdadr x) (graphviz-union nodes (cdadr x)))
	(return-from set-attr)))
    (while (cdr graph) (setq graph (cdr graph)))
    (setf (cdr graph)
	  (cons (list key (cons attr nodes)) nil))))

(defun graphviz-graph-set-add-attr-node-list (graph attr nodes)
  ""
  (graphviz-graph-set-add-attr-key-list
   graph
   'add-attr-node-list
   attr
   nodes))

(defun graphviz-graph-set-add-attr-edge-list (graph attr edges)
  ""
  (graphviz-graph-set-add-attr-key-list
   graph
   'add-attr-edge-list
   attr
   edges))

(defun graphviz-graph-clear-add-attr-key-list (graph key attr nodes)
  ;; nodes $B$,(B nil $B$J$i$9$Y$F%/%j%"(B
  (dolist (x (cddr graph))
    (when (and (eq (car x) key)
	       (equal (caadr x) attr))
      (if nodes
	  (dolist (y nodes)
	    (setf (cdadr x) (delete y (cdadr x))))
	(setf (cdadr x) nil)))))

(defun graphviz-graph-clear-add-attr-node-list (graph attr nodes)
  ""
  (graphviz-graph-clear-add-attr-key-list
   graph
   'add-attr-node-list
   attr
   nodes))

(defun graphviz-graph-clear-add-attr-edge-list (graph attr edges)
  ""
  (graphviz-graph-clear-add-attr-key-list
   graph
   'add-attr-edge-list
   attr
   edges))

;;
;; DEMO
;;
(defun graphviz-dot-display-buffer ()
  "DOT$B7A<0%U%!%$%k$N%P%C%U%!$r(B Graphviz $B$rDL$7$FI=<((B"
  (interactive)
  (graphviz-dot-display-region (point-min) (point-max)))

(defun graphviz-dot-display-region (beg end)
  "DOT$B7A<0$G=q$+$l$?%j!<%8%g%s$r(B Graphviz $B$rDL$7$FI=<((B"
  (interactive "r")
  (let ((tempfile 
	 (expand-file-name "graphviz.gif" 
			   temporary-file-directory)))
    (call-process-region beg
			 end
			 "dot"
			 nil
			 nil
			 nil
			 "-Tgif" 
			 (format "-o%s" tempfile))
    (w32-shell-execute "open" tempfile)))

(defun graphviz-graph-display (graph)
  "graph $B2hA|$rI=<((B"
  (interactive "XGraph S-Exp: ")
  (with-temp-buffer 
    (insert (graphviz-graph-to-dot graph))
    (graphviz-dot-display-buffer)))

(defun graphviz-demo ()
  "graphviz $B%G%b(B
$B3F%"%H%j%S%e!<%H$O(B alist($B%"%H%j%S%e!<%H(B $BJ8;zNs(B) $B$G@_Dj(B
$B%"%H%j%S%e!<%H$N>\$7$$FbMF$O(B Graphviz $BK\BN$N%^%K%e%"%k;2>H(B"
  (interactive)
  (let ((buf (get-buffer-create "*graphviz demo*"))
	graph)
    ;; $B%0%i%U$N=i4|2=(B
    (setq graph 
	  (graphviz-make-graph 
	   ;; $BM-8z%0%i%U(B(digraph)$B!?L58z%0%i%U(B(graph)$B$N@_Dj(B
	   'graph
	   ;; ID($B$I$s$J$b$N$G$b9=$o$J$$(B)
	   "test"
	   ;; $B0J2<$OL5$/$F$b9=$o$J$$(B
	   ;; $B%0%i%UA4BN$N%G%U%)%k%H%"%H%j%S%e!<%H(B
	   '((size "3,3"))
	   ;; $B%N!<%I$N%G%U%)%k%H%"%H%j%S%e!<%H(B
	   '((shape "cube") (color "black"))
	   ;; $B%(%C%8$N%G%U%)%k%H%"%H%j%S%e!<%H(B
	   '((color "black"))))
    ;; $B%i%Y%k@_Dj(B
    (graphviz-graph-label 
     graph 
     ;; $B%i%Y%kL>(B
     "Test Graph" 
     ;; $B%i%Y%k$NI=<(0LCV;XDj(B($BL5$/$F$b9=$o$J$$(B)
     t nil)
    ;; $B%(%C%8@_Dj(B
    (graphviz-graph-set-edge 
     graph 
     '("a" "b"))
    (graphviz-graph-set-edge 
     graph 
     '("b" "d"))
    (graphviz-graph-set-edge 
     graph 
     '("c" "d"))
    ;; $B%(%C%8$K%"%H%j%S%e!<%H$rF1;~$K@_Dj(B
    (graphviz-graph-set-edge 
     graph 
     '("a" "c")
     '((style "dotted") (color "red")))
    ;; $B%N!<%I$K%"%H%j%S%e!<%H$r@_Dj(B
    (graphviz-graph-set-node-attr 
     graph
     "c" 
     '((label "--C--") (color "blue") (shape "circle") (fontcolor "blue")))
    ;;
    ;; $B%G%b%P%C%U%!I=<((B
    ;;
    (save-excursion
      (set-buffer buf)
      (erase-buffer)
      (insert 
       "//\n// Graphviz dot $B7A<0(B\n// $B@_DjFbMF$O(B graphviz-demo $B$N%=!<%9;2>H(B\n//\n")
      (insert (graphviz-graph-to-dot graph))
      (display-buffer buf))
    ;;
    ;; $B:G=*E*$J%0%i%U$r2hA|%U%!%$%k$K$7$FI=<((B
    ;;
    (graphviz-graph-display graph)))

;;
;; sgraph
;;
;; sgraph-to-graphviz-graph : <Sgraph> -> <Graphviz-Graph>
;;
;; <Input> ::=
;; ((node-list ((<NodeID> <Content> (<Successor>..))..)))
;;
;; <Output> :: <Graphviz-Graph>
;;
;; <NodeID> ::= <Symbol>
;; <Content> ::= <String> | nil
;;
;; <Successor> ::= <NodeID> | (<NodeID> <EdgeLabel>)
;;
;; <EdgeLabel> ::= <String>
;;
(defun sgraph-to-graphviz-graph (sgraph)
  (sck-letalist sgraph
      ((node-list nlist))
    (let ((graph (graphviz-make-graph 
		  'digraph 
		  ""
		  nil
		  '((shape "cube") (fontname "Courier") (labeljust "l"))
		  )))
      (dolist (x nlist)
	;; Content
	(when (cadr x)
	  (graphviz-graph-set-node-attr
	   graph
	   (symbol-name (car x))
	   (list (list 'label (format "%s:\n%s" 
				      (symbol-name (car x))
				      (cadr x))))))
	;; Successor
	(dolist (y (caddr x))
	  (graphviz-graph-set-edge graph
				   (list (symbol-name (car x))
					 (symbol-name 
					  (if (listp y) (car y) y)))
				   (when (listp y)
				     (list (list 'label (cadr y))
					   (list 'fontcolor "blue")
					   )))))
      graph)))

;;
(provide 'graphviz)
