;;;; {ZpbP[W
;;;; LastChange: Fri Apr 13 01:47:38 2007

;;; Public:
;;; bn-add bn-sub bn-neg bn-mul bn-div bn-mod bn-gcd bn-lcm bn-divmod
;;; bn-cmp bn-= bn-/= bn-< bn-<= bn-> bn->= bn-zerop
;;; bn-floor bn-ceiling bn-truncate bn-round
;;; str-to-bn bn-to-str bn-try-to-elnum
;;; bn-signed bn-unsigned
;;; bn-expt bn-expt2 bn-fact bn-perm bn-combi
;;; bn-ash bn-logand bn-logior bn-logxor bn-lognot
;;; bn-z-to-bitvec bn-bitvec-to-z bn-bitvec-to-n


(require 'bitvec)
(require 'cache)


;;; <Number> ::= <Integer> | <Rational> | <Floating>
;;;            | <String>  | <Fixnum>   | <Flonum>
;;;
;;; <Integer> ::= <IntegerExternalForm> | <String> | <Fixnum>
;;; <IntegerExternalForm> ::= (bn-z . <InternalFormZ>)
;;; <String> ::= <Integer> ̕\
;;; <Fixnum> ::= Emacs Lisp  integer
;;;
;;; <Rational> ::= <RationalExternalForm> | <String>
;;; <RationalExternalForm> ::= (bn-q . <InternalFormZ>)
;;; <String> ::= <Rational> ̕\
;;;
;;; <Floating> ::= <FloatingExternalForm> | <String> | <Flonum>
;;; <FloatingExternalForm> ::= (bn-f . <InternalFormZ>)
;;; <String> ::= <Floating> ̕\
;;; <Flonum> ::= Emacs Lisp  float
;;;
;;; <Elnum> ::= <Fixnum> | <Flonum>
;;; 
;;; <ExternalForm> ::= <IntegerExternalForm> 
;;;                  | <RationalExternalForm>
;;;                  | <FloatingExternalForm>
;;;
;;; <ExternalForm>  bn pbP[Wɂ鐔̕\łA
;;; bn-add ̃I[o[[h֐iʊ֐jŎg\łB
;;; AInternalFormX ͌^ X ̓\łAX̌^ɏ]ĉZs
;;; bn-ifZadd ̉ʊ֐ŎgB^`FbN^킹͏ʊ֐sA
;;; ^̃^OOAʊ֐ĂяoB

;;;  ̕ύX\ internal form -> raw form,  external form -> internal form
;;; O^ defstruct gB


(defun bn-number-p (x)
  "X  <Number> ǂ肷B"
  (or (numberp x)
      (memq (car-safe x) '(bn-z bn-q bn-f))
      (and (stringp x)
	   (condition-case nil
	       (progn (str-to-bn x) t)
	     (error nil)))))

(defun bn-integer-p (x)
  "X  <Integer> ǂ肷B"
  (or (integerp x)
      (eq (car-safe x) 'bn-z)
      (and (stringp x)
	   (eq (car-safe
		(condition-case nil
		    (str-to-bn x)
		  (error nil)))
	       'bn-z))))

(defun bn-rational-p (x)
  "X  <Rational> ǂ肷B"
  (or (eq (car-safe x) 'bn-q)
      (and (stringp x)
	   (eq (car-safe
		(condition-case nil
		    (str-to-bn x)
		  (error nil)))
	       'bn-q))))

(defun bn-floating-p (x)
  "X  <Floating> ǂ肷B"
  (or (floatp x)
      (eq (car-safe x) 'bn-f)
      (and (stringp x)
	   (eq (car-safe
		(condition-case nil
		    (str-to-bn x)
		  (error nil)))
	       'bn-f))))



(defun bn-add (&rest args)
  "ARGS:(<Number>..) ̘aԂB"
  (bn-apply 'bn-add args))

(defun bn-mul (&rest args)
  "ARGS:(<Number>..) ̐ςԂB"
  (bn-apply 'bn-mul args))

(defun bn-sub (x y)
  "X,Y:<Number> ̍ԂB"
  (bn-apply 'bn-sub (list x y)))

(defun bn-neg (x)
  "X:<Number> ̉Z̋tԂB"
  (bn-apply 'bn-neg (list x)))

(defun bn-div (x y)
  "X,Y:<Number> X  Y ŊԂB
Example:
  ;;  divmod  quorem ̈ႢB"
  (bn-apply 'bn-div (list x y)))

(defun bn-mod (x y)
  "X,Y:<Integer> X  Y Ŋ]ԂB"
  (bn-apply 'bn-mod (list x y)))

(defun bn-divmod (x y)
  "X,Y:<Integer> X  Y ŊƏ]̃XgԂB
Return: (<Integer> <Integer>)"
  (bn-apply 'bn-divmod (list x y)))

(defun bn-quo (x y)
  "X,Y:<Number> X  Y ŊԂB"
  (bn-apply 'bn-quo (list x y)))

(defun bn-rem (x y) 
  "X,Y:<Integer> X  Y Ŋ]ԂB"
  (bn-apply 'bn-rem (list x y)))

(defun bn-quorem (x y)
  "X,Y:<Integer> X  Y ŊƏ]̃XgԂB
Return: (<Integer> <Integer>)"
  (bn-apply 'bn-quorem (list x y)))

(defun bn-gcd (&rest args)
  "ARGS:(<Integer>..) ̍ő񐔂ԂB"
  (bn-apply 'bn-gcd args))

(defun bn-lcm (&rest args)
  "ARGS:(<Integer>..) ̍ŏ{ԂB"
  (bn-apply 'bn-lcm args))

(defun bn-cmp (x y)
  "X,Y:<Number> rAX<Y Ȃ -1AX=Y Ȃ 0A X>Y Ȃ 1 ԂB"
  (bn-apply 'bn-cmp (list x y)))

(defun bn-=   (x y)
  "X,Y:<Number> rAX=Y Ȃ tAȂ nil ԂB"
  (bn-apply 'bn-=   (list x y)))

(defun bn-/=  (x y)
  "X,Y:<Number> rAX!=Y Ȃ tAȂ nil ԂB"
  (bn-apply 'bn-/=  (list x y)))

(defun bn-<   (x y)
  "X,Y:<Number> rAX<Y Ȃ tAȂ nil ԂB"
  (bn-apply 'bn-<   (list x y)))

(defun bn-<=  (x y)
  "X,Y:<Number> rAX<=Y Ȃ tAȂ nil ԂB"
  (bn-apply 'bn-<=  (list x y)))

(defun bn->   (x y)
  "X,Y:<Number> rAX>Y Ȃ tAȂ nil ԂB"
  (bn-apply 'bn->   (list x y)))

(defun bn->=  (x y)
  "X,Y:<Number> rAX>=Y Ȃ tAȂ nil ԂB"
  (bn-apply 'bn->=  (list x y)))

(defun bn-floor (x) 
  "X:<Number> ȉ X ɍł߂ԂB"
  (bn-apply 'bn-floor  (list x)))

(defun bn-ceiling (x)
  "X:<Number> ȏ X ɍł߂ԂB"
  (bn-apply 'bn-ceiling  (list x)))

(defun bn-truncate (x)
  "X:<Number>  i0<XAX<0jȂ X iȉAȏj X ɍł߂ԂB"
  (bn-apply 'bn-truncate  (list x)))

(defun bn-round (x)
  "X:<Number> ɍł߂ԂB"
  (bn-apply 'bn-round  (list x)))


(defconst bn=get-apply-table
  (let ((hashtab (make-hash-table :test 'equal)))
    (dolist (a '((bn-add      (Z Q F)  bn-apply-add-type)
		 (bn-mul      (Z Q F)  bn-apply-mul-type)
		 (bn-neg      (Z Q F)  bn-apply-neg-type)
		 (bn-sub      (Z Q F)  bn-apply-sub-type)
		 (bn-div      (Z Q F)  bn-apply-sub-type)
		 (bn-mod      (Z)      bn-apply-sub-type)
		 (bn-divmod   (Z)      bn-apply-divmod-type)
		 (bn-quo      (Z)      bn-apply-sub-type)
		 (bn-rem      (Z)      bn-apply-sub-type)
		 (bn-quorem   (Z)      bn-apply-divmod-type)
		 (bn-gcd      (Z)      bn-apply-add-type)
		 (bn-lcm      (Z)      bn-apply-mul-type)
		 (bn-cmp      (Z Q F)  bn-apply-cmp-type)
		 (bn-=        (Z Q F)  bn-apply-cmp-type)
		 (bn-/=       (Z Q F)  bn-apply-cmp-type)
		 (bn-<        (Z Q F)  bn-apply-cmp-type)
		 (bn-<=       (Z Q F)  bn-apply-cmp-type)
		 (bn->        (Z Q F)  bn-apply-cmp-type)
		 (bn->=       (Z Q F)  bn-apply-cmp-type)
		 (bn-floor    (Z Q F)  bn-apply-neg-type)
		 (bn-ceiling  (Z Q F)  bn-apply-neg-type)
		 (bn-truncate (Z Q F)  bn-apply-neg-type)
		 (bn-round    (Z Q F)  bn-apply-neg-type)))
      (let ((op (car a)))
	(dolist (typename (cadr a))
	  (let ((type (ecase typename (Z 'bn-z) (Q 'bn-q) (F 'bn-f)))
		(iffun (intern (format "bn-if%S%s"
				       typename
				       (substring (symbol-name op) 3))))
		(applyfun (caddr a)))
	    (puthash (list op type) (list iffun applyfun) hashtab)))))
    hashtab))

(defun bn-get-apply (op type)
  ;; (bn-get-apply 'bn-add 'bn-q) => (bn-ifQadd bn-apply-add-type)
  (gethash (list op type) bn=get-apply-table))

(defun bn-apply (op args)
  (let* ((xforms (bn-adjust-types args))
	 (type (if xforms (caar xforms) 'bn-z))
	 (iforms (mapcar 'cdr xforms))
	 (a (bn-get-apply  op type)))
    (unless a
      (signal 'wrong-type-argument (cons op args)))
    (funcall (cadr a) op (car a) type args iforms)))

(defun bn-apply-add-type (op iffun type args iforms)
  (let ((ret (cdr (bn-coerce type (str-to-bn "0")))))
    (dolist (i iforms)
      (setq ret (funcall iffun ret i)))
    (cons type ret)))

(defun bn-apply-mul-type (op iffun type args iforms)
  (let ((ret (cdr (bn-coerce type (str-to-bn "1")))))
    (dolist (i iforms)
      (setq ret (funcall iffun ret i)))
    (cons type ret)))

(defun bn-apply-neg-type (op iffun type args iforms)
  (bn-apply-check-nargs op args 1)
  (cons type (funcall iffun (car iforms))))

(defun bn-apply-sub-type (op iffun type args iforms)
  (bn-apply-check-nargs op args 2)
  (cons type (apply iffun iforms)))

(defun bn-apply-divmod-type (op iffun type args iforms)
  (bn-apply-check-nargs op args 2)
  (mapcar #'(lambda (x) (cons type x))
	     (apply iffun iforms)))

(defun bn-apply-cmp-type (op iffun type args iforms)
  (bn-apply-check-nargs op args 2)
  (apply iffun iforms))

(defun bn-apply-check-nargs (op args n)
  (unless (= (length args) n)
    (error "Wrong number of arguments: %s: %s" op args)))


(defun bn-adjust-types (args)
  ;; O^AA邢 Fixnum ̃Xg ARGS ̗vf
  ;; O`ɂA^킹XgԂB
  ;; (bn-adjust-types '(1 "2" "3/4"))
  ;; => ((bn-q 1 (1) (1)) (bn-q 1 (2) (1)) (bn-q 1 (3) (4)))
  (let* ((extforms (mapcar 'any-to-bn args))
	 (bn-type (foldl #'(lambda (x y) (bn-type-max x (car y)))
			 'bn-z
			 extforms)))
    (mapcar #'(lambda (extform) (bn-coerce bn-type extform))
	    extforms)))

(defun bn-type-max (t1 t2)
  ;; (bn-type-max 'bn-z 'bn-f)
  (let ((ta '((bn-z 0) (bn-q 1) (bn-f 2))))
    (aref [bn-z bn-q bn-f] (max (cadr (assq t1 ta))
				(cadr (assq t2 ta))))))

(defun bn-coerce (totype x)
  "^ϊiւ̂݁j
Example:
 (bn-coerce 'bn-q (str-to-bn \"123\"))"
  (ecase totype
    (bn-z
     (ecase (car x)
       (bn-z x)))
    (bn-q
     (ecase (car x)
       (bn-z (cons 'bn-q (bn-ifZ-to-ifQ (cdr x))))
       (bn-q x)))
    (bn-f
     (ecase (car x)
       (bn-z (cons 'bn-f (bn-ifZ-to-ifF (cdr x))))
       (bn-q (cons 'bn-f (bn-ifQ-to-ifF (cdr x))))
       (bn-f x)))))


(defun str-to-bn (str)
  "O`ɕϊB
Z  Q ɂĂ [#0][bBxXoO] Ƃ radix prefix T|[gB
ӁFoctal  #o... ܂ 0o... łA0...  decimal ƈB
Example:
 (str-to-bn \"007\")
 (str-to-bn \"-#xABE\")
 (str-to-bn \"1234567890\")
 (str-to-bn \"-123/456\")
 (str-to-bn \"#b1101/11100\")
 (str-to-bn \"-123.456e-78\")"
  (with-cache "(str-to-bn str)" str
    (save-match-data
      (unless (string-match
	       "\\`\\([-+]\\)?\\([#0][bBxXoO]\\)?\\(.+\\)\\'" str)
	(error "Illegal number syntax: %s" str))
      (let* ((sign (if (= (aref str 0) ?-) "-" ""))
	     (radix (if (match-beginning 2)
			(ecase (aref (match-string 2 str) 1)
			  ((?b ?B) 2)
			  ((?o ?O) 8)
			  ((?x ?X) 16))
		      10))
	     (numstr (concat sign (match-string 3 str)))
	     (xform (dolist (a '((bn-str-to-ifZ bn-z)
				 (bn-str-to-ifQ bn-q)
				 (bn-str-to-ifF bn-f))
			       nil)
		      (let ((ifx (funcall (car a) numstr radix)))
			(when ifx
			  (return
			   ;; ^Oǉ
			   (cons (cadr a) ifx)))))))
	(or xform
	    (error "Illegal number syntax: %s" str))))))

(defun bn-to-str (x &optional radix uppercase)
  "X:<Number> 𕶎iftHgV^bNXjɕϊB
Example:
 (bn-to-str (bn-add 2750) 16 t)"
  (setq x (any-to-bn x))
  (concat (case radix (2 "#b") (8 "#o") (10 "") (16 "#x")
		;;  #36r...  Ƃ
		(t ""))
	  (funcall (ecase (pop x) ; ^O폜
		     (bn-z 'bn-ifZ-to-str)
		     (bn-q 'bn-ifQ-to-str)
		     (bn-f 'bn-ifF-to-str))
		   x
		   radix
		   uppercase)))


(defconst bn=from-any-special-alist nil)
    
(defun any-to-bn (num)
  "NUM:<Number> AΉ <ExteranlForm> ɕϊB"
  (or bn=from-any-special-alist
      (setq bn=from-any-special-alist
	    `((0  ,(str-to-bn "0"))
	      (1  ,(str-to-bn "1"))
	      (2  ,(str-to-bn "2"))
	      (-1 ,(str-to-bn "-1"))
	      (,bn=fixnum-min ,(str-to-bn (format "%d" bn=fixnum-min)))
	      (,bn=fixnum-max ,(str-to-bn (format "%d" bn=fixnum-max))))))
  (cond
   ((cadr (assq num bn=from-any-special-alist)))
   ((or (integerp num)
	(floatp num))
    (str-to-bn (number-to-string num)))
   ((stringp num)
    (str-to-bn num))
   ((memq (car-safe num) '(bn-z bn-q bn-f))
    num)
   (t
    (error "any-to-bn: cannot convert to bn-Number: %S" num))))

(defun bn-try-to-elnum (x)
  "X:<Number> \Ȃ Emacs Lisp ̐ɁAȂ <String> ɕϊB"
  (cond
   ((numberp x) x)
   ((eq (car-safe x) 'bn-z)
    (if (and (bn-<= bn=fixnum-min x)
	     (bn-<= x bn=fixnum-max))
	;; 
	;; (string-to-int "-134217728") => -134217728.0
	;; (read "-134217728") => -134217728
	(read (bn-to-str x))
      (bn-to-str x)))
   ((eq (car-safe x) 'bn-f)
    (bn-ifF-double (cdr x)))
   ((eq (car-safe x) 'bn-q)
    (bn-to-str x))
   ((stringp x)
    (bn-try-to-elnum (str-to-bn x)))
   (t
    (assert nil))))

    
;;; elim pv^̒`

(when (featurep 'elim)
  (elim-add-printer
   #'(lambda (x) (memq (car-safe x) '(bn-z bn-q bn-f)))
   #'(lambda (x) (format "%S" (bn-try-to-elnum x))))
  ;; Xgp
  (elim-add-printer
   #'(lambda (x) (memq (car-safe (car-safe x)) '(bn-z bn-q bn-f)))
   #'(lambda (x) (format "%S" (mapcar 'bn-try-to-elnum x)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 牺̉ʊ֐ł͑S InternalForm ɏ]ďsBɂA
;;; ʂ ExternalForm ƓƗɁA\ύX邱Ƃ\łB
;;; ł bn-z, bn-q, bn-f Ƃ^̃^O͂ȂAExternalForm  cdr
;;; łAInternalForm ɏ]āAvZsBI[o[[h֐ȂB
;;; 񂩂̕ϊ֐͐V^bNXłȂȂ nil ԂdlłA
;;; V^bNX`FbJ[˂ĂB
;;; radix prefix (0x... #x... ) ͂ł͏ȂB

(defconst bn=fixnum-max (lsh -1 -1))
(defconst bn=fixnum-min (1+ bn=fixnum-max))
(defconst bn=radixwidth (floor (/ (log bn=fixnum-max 10) 2)))
(defconst bn=radix (expt 10 bn=radixwidth))
;; (assert (< (+ (* bn=radix bn=radix) bn=radix) bn=fixnum-max))

(defun bn-small-to-ifN (small)
  (unless (and (natnump small)
	       (< small bn=radix))
    (error "Illegal small number: %S" small))
  (list small))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; RZ
;;; <InternalFormN> ::=(<SmallNum>..)
;;; <SmallNum> ::= {n | 0 <= n < bn=radix}
;;; R̕\͒PȃXgłAdefstruct ͂ĂȂB

(defconst bn=ifN-0 (bn-small-to-ifN 0))
(defconst bn=ifN-1 (bn-small-to-ifN 1))
(defconst bn=ifN-2 (bn-small-to-ifN 2))


(defun bn-str-to-ifN (str &optional radix)
  ;; radix ([2..36]) iL@ str  <InternalFormN> ɕϊB
  ;; V^bNXG[Ȃ nil ԂB
  ;; (bn-str-to-ifN "ABE" 16)
  ;; (bn-str-to-ifN "222" 2)
  (setq radix (or radix 10))
  (case radix
    (10
     (bn-str-to-ifN/10 str))
    (t
     (let ((ret bn=ifN-0)
	   (radix-ifN (bn-small-to-ifN radix)))
       (unless (and (<= 2 radix) (<= radix 36))
	 (error "Illegal radix: %d" radix))
       (block exit
	 (dolist (c (string-to-list str))
	   (setq ret (bn-ifNadd (bn-ifNmul ret radix-ifN)
				(let ((n (bn-str-to-ifN/dig-to-num c)))
				  (if (and n (< n radix))
				      (bn-small-to-ifN n)
				    ;; Illegal digit
				    (return-from exit nil))))))
	 ret)))))

(defun bn-str-to-ifN/dig-to-num (c)
  ;; (bn-str-to-ifN/dig-to-num ?1)
  ;; (bn-str-to-ifN/dig-to-num ?a)
  ;; (bn-str-to-ifN/dig-to-num ?Z)
  ;; (bn-str-to-ifN/dig-to-num ?@)
  (cond
   ((and (<= ?0 c) (<= c ?9))
    (- c ?0))
   ((and (<= ?a c) (<= c ?z))
    (- c ?a -10))
   ((and (<= ?A c) (<= c ?Z))
    (- c ?A -10))
   (t
    nil)))

(defun bn-str-to-ifN/10 (str)
  ;; \iL@̕ <InternalFormN> ɕϊB
  ;; V^bNXG[Ȃ nil ԂB
  ;; (bn-str-to-ifN "0")
  ;; (bn-str-to-ifN "007")
  ;; (bn-str-to-ifN "1234567890")
  ;; (bn-str-to-ifN "error")
  (save-match-data
    (and (string-match "\\`[0-9]+\\'" str)
	 (let* ((str1 (concat
		       (make-string (mod (- (length str)) bn=radixwidth) ?0)
		       str))
		(r nil))
	   (dotimes (i (/ (length str1) bn=radixwidth))
	     (push (string-to-int (substring str1
					     (* bn=radixwidth i)
					     (* bn=radixwidth (1+ i))))
		   r))
	   (bn-ifN0sup (nreverse r))))))
			  

(defun bn-ifN-to-str (n &optional radix uppercase)
  ;; <InternalFormN>  radix i@ɕϊB
  ;; uppercase  Non-nil Ȃ abc... łȂ ABC...
  ;; (bn-ifN-to-str (bn-str-to-ifN "2750") 16 t)
  (setq radix (or radix 10))
  (case radix
    (10
     (bn-ifN-to-str/10 n))
    (t
     (let ((ret nil)
	   (radix-ifN (list radix))
	   (r 0))
       (unless (and (<= 2 radix) (<= radix 36))
	 (error "Illegal radix: %d" radix))
       (while (not (bn-ifNzerop n))
	 (multiple-value-setq (n r) (bn-ifNdivmod n radix-ifN))
	 (setq r (car r))
	 (push (if (<= r 9)
		   (+ r ?0)
		 (+ r -10 (if uppercase ?A ?a)))
	       ret))
       (list-to-string ret)))))

(defun bn-ifN-to-str/10 (n)
  ;; <InternalFormN> \iL@̕ɕϊB
  ;; (bn-ifN-to-str (bn-str-to-ifN "01234567890"))
  (let ((fmt (format "%%0%dd" bn=radixwidth)))
    (apply 'concat
	   (cons (format "%d" (car n))
		 (mapcar #'(lambda (x) (format fmt x)) (cdr n))))))
    

(defun bn-n= (x y) (= (bn-ifNcmp x y) 0))
(defun bn-n< (x y) (< (bn-ifNcmp x y) 0))

(defun bn-ifNcmp (x y)
  (bn-ifNcmp-1 x (length x) y (length y)))

(defun bn-ifNcmp-1 (x xl y yl)
  ;; Returns a fixnum whose sign is the same to that of a-b.
  (while (and (<= 2 xl) (zerop (car x)))
    (setq x (cdr x)
	  xl (1- xl)))
  (while (and (<= 2 yl) (zerop (car y)))
    (setq y (cdr y)
	  yl (1- yl)))
  (cond
   ((< xl yl) -1)
   ((> xl yl) 1)
   (t
    (let ((i 0))
      (while (and (< i xl) (= (car x) (car y)))
	(setq x (cdr x)
	      y (cdr y)
	      i (1+ i)))
      (if (< i xl)
	  (- (car x) (car y))
	0)))))

(defun bn-ifN0sup (x)
  ;; Skip leading zeros.
  ;; (bn-ifN0sup '(0 0 0))
  (while (and (cdr x) (zerop (car x)))
    (setq x (cdr x)))
  x)

(defun bn-ifNzerop (x)
  ;; (bn-ifNzerop '(0 0))
  ;; (bn-ifNzerop '(0 1))
  (= 0 (car (bn-ifN0sup x))))

(defun bn-ifNlength (x)
  (* bn=radixwidth (length x)))

; (defun bn-ifNcarry-old (x len)
;   ;; Destructively propagete carries from end.
;   ;; ċAŋtɒH鏈c
;   (if (zerop len)
;       0
;     (let* ((c (bn-ifNcarry (cdr x) (1- len)))
;            (n (+ (car x) c))
;            (m (mod n bn=radix)))
;       (setcar x m)
;       (/ (- n m) bn=radix))))

(defun bn-ifNcarry (x)
  ;; Destructively propagete carries from end.
  ;; ǂ nreverse
  (let* ((xr (nreverse x))
	 (c 0)
	 (n nil)
	 (r xr))
    (while r
      (setq n (+ (car r) c)
	    c (floor n bn=radix))
      (setcar r (mod n bn=radix))
      (setq r (cdr r)))
    (nreverse xr)))

(defun bn-ifNadd (x y) (bn-ifNaddsub x y '+))
(defun bn-ifNsub (x y) (bn-ifNaddsub x y '-))

(defun bn-ifNaddsub (x y op)
  (let ((lx (length x))
	(ly (length y))
	r a r1)
    (unless (<= ly lx)
      (setq ly (prog1 lx (setq lx ly)))
      (setq y (prog1 x (setq x y))))
    (setq r (cons 0 (copy-sequence x))
	  a y
	  r1 (nthcdr (1+ (- lx ly)) r))
    (while r1
      (setcar r1 (funcall op (car r1) (car a)))
      (setq r1 (cdr r1)
	    a (cdr a)))
    (setq r (bn-ifNcarry r))
    (bn-ifN0sup r)))

(defun bn-ifNmul (x y)
  (let* ((lx (length x))
	 (ly (length y))
	 (r (make-list (+ lx ly) 0))
	 (r1 (cdr r))
	 (cl (1+ ly))
	 r2)
    (dolist (xd x)
      (setq r2 r1)
      (dolist (yd y)
	;;  bn=radix ̐
	(setcar r2 (+ (car r2) (* xd yd)))
	(setq r2 (cdr r2)))
      (setq r (bn-ifNcarry r))
      (setq cl (1+ cl))
      (setq r1 (cdr r1)))
    (bn-ifN0sup r)))

(defun bn-ifNdiv (x y) (car (bn-ifNdivmod x y)))
(defun bn-ifNmod (x y) (cadr (bn-ifNdivmod x y)))

(defun bn-ifNdivmod (x y)
  ;; Returns list of quotient and remainder.
  ;; This code is based on the Knuth's book.
  (cond
   ((bn-n= y bn=ifN-0) (error "Divide by zero"))
   ((bn-n< x y) (list (list 0) x))
   (t
    (let* ((d (/ bn=radix (1+ (car y))))
	   (x (bn-ifNdivmod/1 x d))
	   (y (bn-ifNdivmod/1 y d))
	   (xl (length x))
	   (yl (length y))
	   (r (list 0))
	   (q nil))
      (dotimes (i (1+ (- xl yl)))
	(let ((nasty (not (zerop (car r)))))
	  (push (bn-ifNdivmod/2 (if nasty r x) nasty y yl)
		q))
	(setq r x
	      x (cdr x)))
      (list (bn-ifN0sup (nreverse q))
	    (bn-ifN0sup (bn-ifNdivmod/4 r d)))))))

(defun bn-ifNdivmod/1 (x a)
  ;; Returns ax
  (let ((x1 (cons 0 (copy-sequence x))))
    (dorest (d (cdr x1))
      (setcar d (* (car d) a)))
    (setq x1 (bn-ifNcarry x1))
    (bn-ifN0sup x1)))

(defun bn-ifNdivmod/2 (x nasty y yl)
  ;; x -= qy
  ;; Returns q
  (cond
   ;; xl = yl+1
   (nasty
    (let ((qguess (- (if (<= (car y) (car x))
			 (1- bn=radix)
		       ;;  bn=radix ̐
		       (/ (+ (* (car x) bn=radix) (cadr x)) (car y)))
		     2))
	  (xl (1+ yl)))
      (bn-ifNdivmod/3 x t y yl qguess)
      ;; This loop will be executed at most twice.
      (while (not (< (bn-ifNcmp-1 x xl y yl) 0))
	(setq qguess (1+ qguess))
	(bn-ifNdivmod/3 x t y yl 1))
      qguess))
   ;; xl = yl
   ((< (bn-ifNcmp-1 x yl y yl) 0) 0)
   (t (bn-ifNdivmod/3 x nil y yl 1)
      1)))

(defun bn-ifNdivmod/3 (x nasty y yl q)
  ;; x -= q*y
  (let ((x1 (if nasty (cdr x) x)))
    (dotimes (i yl)
      ;;  bn=radix ̐
      (setcar x1 (- (car x1) (* q (car y))))
      (setq x1 (cdr x1)
	    y (cdr y)))
    (setq x (bn-ifNcarry x))))

(defun bn-ifNdivmod/4 (x a)
  ;; Returns x/a
  (if (= a 1)
      x
    (let ((x1 (copy-sequence x))
	  (r 0)
	  (n nil))
      (dorest (d x1)
	;;  bn=radix ̐
	(setq n (+ (car d) (* r bn=radix)))
	(setcar d (/ n a))
	(setq r (mod n a)))
      (assert (zerop r))
      (setq x1 (bn-ifNcarry x1))
      x1)))

(defun bn-ifNgcd (x y)
  (while (not (bn-ifNzerop y))
    (setq x (prog1 y (setq y (bn-ifNmod x y)))))
  x)

(defun bn-ifNlcm (x y)
  (if (or (bn-ifNzerop x) (bn-ifNzerop y))
      bn=ifN-0
    (bn-ifNmul x (bn-ifNdiv y (bn-ifNgcd x y)))))
	  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Z
;;; <InternalFormZ> ::= (< -1 | 1 > <InternalFormN>)

(defstruct (bn-ifZ (:type list)) (sign 1) num)
(defconst bn=ifZ-0 (make-bn-ifZ :num bn=ifN-0))
(defconst bn=ifZ-1 (make-bn-ifZ :num bn=ifN-1))

(defun bn-str-to-ifZ (str &optional radix)
  ;; (bn-str-to-ifZ "-123")
  ;; (bn-str-to-ifZ "-ABE" 16)
  ;; (bn-str-to-ifZ "123" 2)
  ;; (bn-str-to-ifZ "syntax error") => nil
  (save-match-data
    (and (string-match "\\`\\([-+]\\)?\\([0-9a-zA-Z]+\\)\\'" str)
	 (let ((num (bn-str-to-ifN (match-string 2 str) radix)))
	   (and num
		(make-bn-ifZ :sign (if (= (aref str 0) ?-) -1 1)
			     :num num))))))

(defun bn-ifZ-to-str (x &optional radix uppercase)
  ;; (bn-ifZ-to-str (bn-str-to-ifZ "-ABE" 16) 16 t)
  (format "%s%s"
	  (if (= (bn-ifZ-sign x) -1) "-" "")
	  (bn-ifN-to-str (bn-ifZ-num x) radix uppercase)))


(defun bn-ifZadd (x y)
  (let ((xs (bn-ifZ-sign x))
	(xa (bn-ifZ-num x))
	(ys (bn-ifZ-sign y))
	(ya (bn-ifZ-num y)))
    (bn-ifZ-correct-sign-of-zero
     (cond
      ((= xs ys)
       (make-bn-ifZ :sign xs :num (bn-ifNadd xa ya)))
      ((bn-n< xa ya)
       (make-bn-ifZ :sign ys :num (bn-ifNsub ya xa)))
      (t
       (make-bn-ifZ :sign xs :num (bn-ifNsub xa ya)))))))


(defun bn-ifZ-correct-sign-of-zero (x)
  ;; Žʂ 0 ̏ꍇA -1 ɂȂĂ܂ꍇ̂ŁA
  ;;  1 ɏCB
  (if (and (bn-ifNzerop (bn-ifZ-num x))
	   (= -1 (bn-ifZ-sign x)))
      (make-bn-ifZ :num (bn-ifZ-num x))
    x))

(defun bn-ifZneg (x)
  (bn-ifZ-correct-sign-of-zero
   (make-bn-ifZ :sign (- (bn-ifZ-sign x))
		:num (bn-ifZ-num x))))

(defun bn-ifZsub (x y)
  (bn-ifZadd x (bn-ifZneg y)))

(defun bn-ifZmul (x y)
  (bn-ifZ-correct-sign-of-zero
   (make-bn-ifZ :sign (* (bn-ifZ-sign x) (bn-ifZ-sign y))
		:num (bn-ifNmul (bn-ifZ-num x) (bn-ifZ-num y)))))

(defun bn-ifZdiv (x y) (car (bn-ifZdivmod x y)))
(defun bn-ifZmod (x y) (cadr (bn-ifZdivmod x y)))

(defun bn-ifZdivmod (x y)
  ;; Return: list of truncate(x/y),  r = x - q*y
  (let* ((xs (bn-ifZ-sign x))
	 (ys (bn-ifZ-sign y))
	 (xa (bn-ifZ-num x))
	 (ya (bn-ifZ-num y))
	 (qr (bn-ifNdivmod xa ya))
	 (q (car qr))
	 (r (cadr qr)))
    (mapcar 'bn-ifZ-correct-sign-of-zero
	    (list (make-bn-ifZ :sign (* xs ys) :num q)
		  (make-bn-ifZ :sign xs :num r)))))

(defun bn-ifZquo (x y) (car (bn-ifZquorem x y)))
(defun bn-ifZrem (x y) (cadr (bn-ifZquorem x y)))

(defun bn-ifZquorem (x y)
  ;; Return: list of floor(x/y),  r = x - q*y
  (multiple-value-bind (q r) (bn-ifZdivmod x y)
    (when (if (bn-ifZ< y bn=ifZ-0) (bn-ifZ> r bn=ifZ-0) (bn-ifZ< r bn=ifZ-0))
      (setq r (bn-ifZadd r y)
	    q (bn-ifZsub q bn=ifZ-1)))
    (list (bn-ifZ-correct-sign-of-zero q)
	  (bn-ifZ-correct-sign-of-zero r))))
	  
(defun bn-ifZ=  (x y) (=  (bn-ifZcmp x y) 0))
(defun bn-ifZ/= (x y) (/= (bn-ifZcmp x y) 0))
(defun bn-ifZ<  (x y) (<  (bn-ifZcmp x y) 0))
(defun bn-ifZ<= (x y) (<= (bn-ifZcmp x y) 0))
(defun bn-ifZ>  (x y) (>  (bn-ifZcmp x y) 0))
(defun bn-ifZ>= (x y) (>= (bn-ifZcmp x y) 0))

(defun bn-ifZcmp (x y)
  (let ((xs (bn-ifZ-sign x))
	(xa (bn-ifZ-num x))
	(ys (bn-ifZ-sign y))
	(ya (bn-ifZ-num y)))
    (cond
     ((= xs ys)
      (* xs (bn-ifNcmp xa ya)))
     ((and (bn-ifNzerop xa) (bn-ifNzerop ya))
      0)
     (t xs))))

(defun bn-ifZgcd (x y)
  (bn-ifZ-correct-sign-of-zero
   (make-bn-ifZ :num (bn-ifNgcd (bn-ifZ-num x) (bn-ifZ-num y)))))

(defun bn-ifZlcm (x y)
  (bn-ifZ-correct-sign-of-zero
   (make-bn-ifZ :num (bn-ifNlcm (bn-ifZ-num x) (bn-ifZ-num y)))))

(defun bn-ifZfloor (x) x)
(defun bn-ifZceiling (x) x)
(defun bn-ifZtruncate (x) x)
(defun bn-ifZround (x) x)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; LZ
;;; <InternalFormQ> ::= (< -1 | 1 > <Numerator> <Denominator>)
;;; <Numerator> ::= <InternalFormN>
;;; <Denominator> ::= <InternalFormN>

(defstruct (bn-ifQ (:type list)) (sign 1) num (den (list 1)))
(defun bn-ifQ-znum (x) x)
(defconst bn=ifQ-0 (make-bn-ifQ :num bn=ifN-0 :den bn=ifN-1))
(defconst bn=ifQ-1 (make-bn-ifQ :num bn=ifN-1 :den bn=ifN-1))
(defconst bn=ifQ-1/2 (make-bn-ifQ :num bn=ifN-1 :den bn=ifN-2))

(defun bn-str-to-ifQ (str &optional radix)
  ;; (bn-str-to-ifQ "-1/2")
  ;; (bn-str-to-ifQ "-1/2" 2)
  ;; (bn-str-to-ifQ "-1/@")
  ;; (bn-str-to-ifQ "error")
  (save-match-data
    (and (string-match
	  "\\`\\([-+]\\)?\\([0-9a-zA-Z]+\\)/\\([0-9a-zA-Z]+\\)\\'" str)
	 (let* ((sign (if (= (aref str 0) ?-) -1 1))
		(num (bn-str-to-ifN (match-string 2 str) radix))
		(den (bn-str-to-ifN (match-string 3 str) radix)))
	   (and num
		den
		(bn-ifQreduce
		 (make-bn-ifQ :sign sign
			      :num num
			      :den den)))))))

(defun bn-ifQ-to-str (x &optional radix uppercase)
  (format "%s%s/%s"
	  (if (= (bn-ifQ-sign x) -1) "-" "")
	  (bn-ifN-to-str (bn-ifQ-num x) radix uppercase)
	  (bn-ifN-to-str (bn-ifQ-den x) radix uppercase)))

(defun bn-ifQadd (x y)
  (let* ((xznum (bn-ifQ-znum x))
	 (xnden (bn-ifQ-den x))
	 (yznum (bn-ifQ-znum y))
	 (ynden (bn-ifQ-den y))
	 (rznum (bn-ifZadd
		 (bn-ifZmul xznum (make-bn-ifZ :num ynden))
		 (bn-ifZmul yznum (make-bn-ifZ :num xnden))))
	 (rnden (bn-ifNmul xnden ynden)))
    (bn-ifQreduce (make-bn-ifQ :sign (bn-ifZ-sign rznum)
			       :num (bn-ifZ-num rznum)
			       :den rnden))))

(defun bn-ifQneg (x)
  (make-bn-ifQ :sign (- (bn-ifQ-sign x))
	       :num (bn-ifQ-num x)
	       :den (bn-ifQ-den x)))

(defun bn-ifQsub (x y)
  (bn-ifQadd x (bn-ifQneg y)))

(defun bn-ifQmul (x y)
  (let* ((xznum (bn-ifQ-znum x))
	 (xnden (bn-ifQ-den x))
	 (yznum (bn-ifQ-znum y))
	 (ynden (bn-ifQ-den y))
	 (rznum (bn-ifZmul xznum yznum))
	 (rnden (bn-ifNmul xnden ynden)))
    (bn-ifQreduce (make-bn-ifQ :sign (bn-ifZ-sign rznum)
			       :num (bn-ifZ-num rznum)
			       :den rnden))))

(defun bn-ifQdiv (x y)
  (bn-ifQmul x (make-bn-ifQ :sign (bn-ifQ-sign y)
			    :num (bn-ifQ-den y)
			    :den (bn-ifQ-num y))))

(defun bn-ifQ=  (x y) (=  (bn-ifQcmp x y) 0))
(defun bn-ifQ/= (x y) (/= (bn-ifQcmp x y) 0))
(defun bn-ifQ<  (x y) (<  (bn-ifQcmp x y) 0))
(defun bn-ifQ<= (x y) (<= (bn-ifQcmp x y) 0))
(defun bn-ifQ>  (x y) (>  (bn-ifQcmp x y) 0))
(defun bn-ifQ>= (x y) (>= (bn-ifQcmp x y) 0))

(defun bn-ifQcmp (x y)
  (let ((diff (bn-ifQsub x y)))
    (cond
     ((= (bn-ifQ-sign diff) -1) -1)
     ((bn-ifNzerop (bn-ifQ-num diff)) 0)
     (t 1))))

(defun bn-ifQreduce (x)
  (let ((gcd (bn-ifNgcd (bn-ifQ-num x) (bn-ifQ-den x))))
    (make-bn-ifQ :sign (bn-ifQ-sign x)
		 :num (bn-ifNdiv (bn-ifQ-num x) gcd)
		 :den (bn-ifNdiv (bn-ifQ-den x) gcd))))

(defun bn-ifQfloor (x)
  (bn-ifZdiv (bn-ifQ-znum x)
	   (make-bn-ifZ :num (bn-ifQ-den x))))

(defun bn-ifQceiling (x)
  (multiple-value-bind (q r) (bn-ifZdivmod x (make-bn-ifZ :num (bn-ifQ-den x)))
    (if (bn-ifZ= r bn=ifZ-0)
	q
      (bn-ifZadd q bn=ifZ-1))))

(defun bn-ifQtruncate (x)
  (multiple-value-bind (q r) (bn-ifZdivmod x (list 1 (caddr x)))
    (if (or (= 1 (bn-ifQ-sign x))
	    (bn-ifZ= r bn=ifZ-0))
	q
      (bn-ifZadd q bn=ifZ-1))))

(defun bn-ifQround (x)
  (bn-ifQfloor (bn-ifQadd x bn=ifQ-1/2)))

(defun bn-ifZ-to-ifQ (z)
  (make-bn-ifQ
   :sign (bn-ifZ-sign z)
   :num (bn-ifZ-num z)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; _Z
;;; <InternalFormF> ::= (< -1 | 1 > <Mantissa> <Precision> <Exp>   <Double>)
;;; <Mantissa> ::= <InternalFormF>
;;; <Precision> ::= <Fixnum>
;;; <Exp> ::= <Fixnum>
;;; <Double> ::= Emacs Lisp's double floating Ƃ肠

(defstruct (bn-ifF (:type list)) sign man pre exp  double)
(defconst bn=ifF-0 (make-bn-ifF :double 0.0))
(defconst bn=ifF-1 (make-bn-ifF :double 1.0))

(defun bn-str-to-ifF (str &optional radix)
  ;; (bn-str-to-ifF "+1.2")
  ;; (bn-str-to-ifF "-.2")
  ;; (bn-str-to-ifF "1.")
  ;; (bn-str-to-ifF ".0")
  ;; (bn-str-to-ifF "1.2e-10")
  ;; (bn-str-to-ifF "1E-10")
  ;; (bn-str-to-ifF "-1") => nil
  ;; (bn-str-to-ifF "error") => nil
  (and (memq radix '(nil 10))
       (save-match-data
	 (and
	  (string-match
   "\\`\\([-+]\\)?\\([0-9]+\\)?\\(\\.[0-9]*\\)?\\([eE]\\([-+]?[0-9]+\\)\\)?\\'"
   str)
	  (or (match-beginning 3) ; \(\.[0-9]*\)?
	      (match-beginning 4)); \([eE]\\([-+]?[0-9]+\)\)?
	  (let ((sign (if (and (match-beginning 1)
			       (= (aref (match-string 1 str) 0) ?-))
			  -1
			1))
		(manl (if (match-beginning 2)
			  (match-string 2 str)
			"0"))
		(manr (if (match-beginning 3)
			  (let ((s (substring (match-string 3 str) 1)))
			    (if (zerop (length s))
				"0"
			      s))
			"0"))
		(exp  (if (match-beginning 4)
			  (string-to-int (match-string 5 str))
			0)))
	    (make-bn-ifF
	     :sign sign
	     ;;  man  pre ̒l͉R
	     :man (bn-str-to-ifN manl) 
	     :pre (bn-str-to-ifN manr)
	     :exp exp
	     :double (string-to-number (match-string 0 str)))
	    )))))


(defun bn-ifF-to-str (f &optional radix uppercase)
  (unless (memq radix '(nil 10))
    (error "bn-ifF-to-str: Illegal radix: %S" radix))
  (format "%s" (bn-ifF-double f)))


(defun bn-ifFadd (x y)
  (make-bn-ifF :double (+ (bn-ifF-double x) (bn-ifF-double y))))

(defun bn-ifFsub (x y)
  (make-bn-ifF :double (- (bn-ifF-double x) (bn-ifF-double y))))

(defun bn-ifFneg (x)
  (make-bn-ifF :double (- (bn-ifF-double x))))

(defun bn-ifFmul (x y)
  (make-bn-ifF :double (* (bn-ifF-double x) (bn-ifF-double y))))

(defun bn-ifFdiv (x y)
  (make-bn-ifF :double (/ (bn-ifF-double x) (bn-ifF-double y))))

(defun bn-ifF=  (x y) (=  (bn-ifFcmp x y) 0))
(defun bn-ifF/= (x y) (/= (bn-ifFcmp x y) 0))
(defun bn-ifF<  (x y) (<  (bn-ifFcmp x y) 0))
(defun bn-ifF<= (x y) (<= (bn-ifFcmp x y) 0))
(defun bn-ifF>  (x y) (>  (bn-ifFcmp x y) 0))
(defun bn-ifF>= (x y) (>= (bn-ifFcmp x y) 0))

(defun bn-ifFcmp (x y)
  (let ((xd (bn-ifF-double x))
	(yd (bn-ifF-double y)))
    (cond
     ((= xd yd) 0)
     ((> xd yd) 1)
     (t -1))))

(defun bn-ifFfloor (x) (make-bn-ifF :double (floor (bn-ifF-double x))))
(defun bn-ifFceiling (x) (make-bn-ifF :double  (ceiling (bn-ifF-double x))))
(defun bn-ifFtruncate (x) (make-bn-ifF :double (truncate (bn-ifF-double x))))
(defun bn-ifFround (x) (make-bn-ifF :double  (round (bn-ifF-double x))))


(defun bn-ifZ-to-ifF (z)
  ;; (bn-ifZ-to-ifF (cdr (str-to-bn "-1234567890")))
  (make-bn-ifF
   :double (let* ((n (bn-ifZ-num z))
		  ;;  ifN Ɉˑ
		  (head (list (car n)))
		  (tail (or (cdr n) '(0)))
		  (str (format "%s%s.%sE%d"
			       (if (= -1 (bn-ifZ-sign z)) "-" "")
			       (bn-ifN-to-str head)
			       (bn-ifN-to-str tail)
			       (bn-ifNlength (cdr n)))))
	     (string-to-number str))))

(defun bn-ifQ-to-ifF (x)
  ;; (bn-ifQ-to-ifF (cdr (str-to-bn "1234/12345678")))
  (let* ((doubleprec 20)
	 (num (bn-ifQ-num x))
	 (den (bn-ifQ-den x))
	 (digdiff (- (bn-ifNlength num) (bn-ifNlength den)))
	 ;;  ifN Ɉˑ
	 (geta (cons 1 (make-list (/ (max bn=radixwidth (- doubleprec digdiff))
				     bn=radixwidth)
				  0))))
    (make-bn-ifF
     :double (* (bn-ifF-double
		 (bn-ifZ-to-ifF
		  (make-bn-ifZ :sign (bn-ifQ-sign x)
			       :num (bn-ifNdiv (bn-ifNmul num geta) den))))
		(expt 10.0 (- (* bn=radixwidth (1- (length geta)))))))))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; ܂܂ȉZ

(defun bn-zerop (x)
  "X:<Number>  0 Ȃ tAȂ nil ԂB"
  (bn-= 0 x))

(defun bn-expt (x n)
  "X:<Number  NiRjԂB
Example:
 (bn-expt \"2/3\" 20) => \"1048576/3486784401\""
  (setq n (bn-try-to-elnum n))
  (unless (natnump n)
    (signal 'wrong-type-argument (list 'natnump n)))
  (let ((r 1))
    (while (< 0 n)
      (when (= 1 (logand n 1))
	(setq r (bn-mul r x)))
      (setq x (bn-mul x x)
	    n (lsh n -1)))
    r))

(defun bn-expt2 (n)
  "2  NiRjԂB"
  (with-cache "(bn-expt2 n)" n
    (bn-expt 2 n)))

(defun bn-perm (n r)
  "N  R 鏇̐ N*(N-1)*...*(N-N+1) ԂB
Example:
 (bn-perm 100 10) => \"5653408585997652480000\""
  (let ((p (any-to-bn 1)))
    (while (bn-<= 1 r)
      (setq p (bn-mul p n)
	    n (bn-sub n 1)
	    r (bn-sub r 1)))
    p))

(defun bn-fact (n)
  "N ̊KԂB"
  (let ((ret 1))
    (dotimes (i (bn-try-to-elnum n))
      (setq ret (bn-mul ret (1+ i))))
    ret))

(defun bn-combi (n r)
  "N  R gݍ킹̐ nPr/r! ԂB
Example:
 (bn-comb 50 10) => \"10272278170\"
 (bn-add (bn-combi 49 10) (bn-combi 49 9)) => \"10272278170\""
  (bn-div (bn-perm n r) (bn-fact r)))


(defun bn-signed (x n)
  "X:<Integer>  N:<Fixnum> rbg̕t֕ϊB
Example:
 (bn-signed #xff 8) => -1"
  (unless (natnump n)
    (signal 'wrong-type-argument (list 'natnump n)))
  (let* ((m (bn-expt2 n))
	 (r (bn-rem x m)))
    (if (bn-< r (bn-expt2 (1- n)))
	r
      (bn-sub r m))))

(defun bn-unsigned (x n)
  "X:<Integer>  N:<Fixnum> rbg̕Ȃ֕ϊB
Example:
 (bn-unsigned #xff 8) = \"255\""
  (unless (natnump n)
    (signal 'wrong-type-argument (list 'natnump n)))
  (bn-rem x (bn-expt2 n)))

(defun bn-ash (x n)
  "X:<Integer> * 2^(N:<Integer>) ԂB"
  (if (bn-<= 0 n)
      (bn-mul x (bn-expt2 n))
    (bn-quo x (bn-expt2 (bn-neg n)))))

(defun bn-bitwise-apply (bitop args)
  ;; args:(<Integer>..)
  ;; 24 rbgƂɋ؂ bitop KpĂB
  ;;  28 MM bn-try-to-elnum  Fixnum ɕϊoȂB
  (assert (= bn=fixnum-max 134217727)) ; 28 rbg Fixnum B
  (let ((bits nil))
    (loop
     (let* ((qrs (mapcar #'(lambda (x) (bn-quorem x #x1000000)) args))
	    (qs (mapcar 'car qrs))
	    (rs (mapcar 'cadr qrs)))
       (setq args qs)
       (push (apply bitop (mapcar 'bn-try-to-elnum rs)) bits)
       (when (every #'(lambda (x) (or (bn-= x 0) (bn-= x -1))) qs)
	 (push (apply bitop (mapcar 'bn-try-to-elnum qs)) bits)
	 (return))))
    (let ((ret (pop bits)))
      (dolist (b bits)
	(setq ret (bn-add (bn-mul ret #x1000000) (logand #xffffff b))))
      ret)))

(defun bn-logand (&rest args)
  "ARGS:(<Integer>..) ̃rbgƂ̘_ςԂB"
  (bn-bitwise-apply #'logand args))

(defun bn-logior (&rest args)
  "ARGS:(<Integer>..) ̃rbgƂ̘_aԂB"
  (bn-bitwise-apply #'logior args))

(defun bn-logxor (&rest args)
  "ARGS:(<Integer>..) ̃rbgƂ̔rI_aԂB"
  (bn-bitwise-apply #'logxor args))

(defun bn-lognot (x)
  "X:<Integer> ̃rbgƂ̘_ےԂB"
  (bn-bitwise-apply #'lognot (list x)))


(defun bn-z-to-bitvec (x nbits)
  " X  2 ̕␔\ɊÂ NBITS ̃rbgxN^ɕϊB
Example:
 (bn-z-to-bitvec -3 100)"
  ;;  1 rbgƂɏ̂͌B
  (setq x (any-to-bn x))
  (unless (eq (car-safe x) 'bn-z)
    (error "bn-z-to-bitvec: Not an integer: %S" x))
  (let ((bs nil)
	(r 0))
    (dotimes (i nbits)
      (multiple-value-setq (x r) (bn-quorem x 2))
      (push (if (bn-= r 0) 0 1) bs))
    (list-to-bitvec bs)))

(defun bn-bitvec-to-n (bv)
  "bn-z-to-bitvec ̋tBR֕ϊB
Example:
  (bn-bitvec-to-n (str-to-bitvec \"#*111\"))"
  (str-to-bn 
   (concat "#b" (substring (bitvec-to-str bv) 2))))

(defun bn-bitvec-to-z (bv)
  "bn-z-to-bitvec ̋tB֕ϊB
Example:
  (bn-bitvec-to-z (str-to-bitvec \"#*111\"))"
  (if (= 1 (bitvec-ref bv (1- (bitvec-length bv))))
      (bn-neg (bn-bitvec-to-n (bitvec-neg bv)))
    (bn-bitvec-to-n bv)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; eXg

(defun bn-test ()
  (interactive)
  ;; (a+b)^n = nC0*a^n*b^0 + ... + nCn*a^0*b^n
  (let ((a "355/113")
	(b "-19992001/98765")
	(n 10))
    (assert (bn-= (bn-expt (bn-add a b) n)
		  (let ((sum 0))
		    (dotimes (i (1+ n))
		      (setq sum (bn-add sum (bn-mul (bn-combi n i)
						    (bn-expt a (- n i))
						    (bn-expt b i)))))
		    sum))))
  ;;
  (message "Done"))

;;;
(provide 'bn)
