;;;; M̑SoHT@iSCK vWFNg̃vO@Ƃāj
;;;; LastChange: Fri Apr 13 01:30:19 2007

;;; Export:
;;; hitohude : <HitohudeInput> -> <HitohudeOutput>
;;; SoH̃XgԂB
;;;
;;; Type:
;;; <HitohudeInput> ::= ((sart <Node>)
;;;                      (end <Node>)
;;;                      (figure (((<Node> <Node>) (<Edge>..))..)))
;;; <Node> ::= <Fixnum>
;;; <Edge> ::= <Symbol>
;;;
;;; <HitohudeOutput> ::= ((routes ((<Edge>..)..)))
;;;
;;; Demo:
;;; ESC x hitohude-demo RET


(defconst hitohude=input-example
  ;; ÓTIȖ
  ;;
  ;;     0
  ;;    / \
  ;;   a   b
  ;;  /     \
  ;; 1---c---2
  ;; |\     /|
  ;; | e   f |
  ;; d  \ /  g
  ;; |  / \  |
  ;; | /   \ |
  ;; 3---h---4
  ;;
  ;; SoH 44
  '((start 3)
    (end 4)
    (figure (((0 1) (a))
	     ((0 2) (b))
	     ((1 2) (c))
	     ((1 3) (d))
	     ((1 4) (e))
	     ((2 3) (f))
	     ((2 4) (g))
	     ((3 4) (h))))))
	      
(defconst hitohude=input-example-takeshi
  ;; ̃R}l`wwȂ
  ;;
  ;;           +-b-+   +-e-+   +-h-+   +-k-+ 
  ;;          /     \ /     \ /     \ /     \
  ;; 0---a---1---c---2---f---3---i---4---l---5---n---6
  ;;          \     / \     / \     / \     /
  ;;           +-d-+   +-g-+   +-j-+   +-m-+ 
  ;;
  ;; SoH 10368 = 6*12^3
  '((start 0)
    (end 6)
    (figure (((0 1) (a))
	     ((1 2) (b c d))
	     ((2 3) (e f g))
	     ((3 4) (h i j))
	     ((4 5) (k l m))
	     ((5 6) (n))))))

(defconst hitohude=input-example-takeshi-small
  ;; ̃R}l`wwȂ
  ;;
  ;;           +-b-+   +-e-+   +-h-+
  ;;          /     \ /     \ /     \
  ;; 0---a---1---c---2---f---3---i---4---k---5
  ;;          \     / \     / \     /
  ;;           +-d-+   +-g-+   +-j-+
  ;;
  ;; SoH 864 = 6*12^2
  '((start 0)
    (end 5)
    (figure (((0 1) (a))
	     ((1 2) (b c d))
	     ((2 3) (e f g))
	     ((3 4) (h i j))
	     ((4 5) (k))))))


(defun hitohude (input)
  ;; C
  `((routes ,(hitohude-find-routes input))))

(defun* hitohude-demo (&optional (inputvarsym 'hitohude=input-example))
  (interactive (hitohude-interactive-readvar "hitohude=input-example"))
  (with-display-buffer "*hitohude-demo*"
    (let ((input (symbol-value inputvarsym)))
    (sck-letalist (hitohude (symbol-value inputvarsym))
	(routes)
      (insert ";;; HitohudeInput\n\n")
      (insertf "%s" (pp-to-string input))
      (newline)
      (insert ";;; HitohudeOutput\n\n")
      (insert "((route (\n")
      (let ((n 0))
	(dolist (route routes)
	  (insertf "%S ; %d\n"  route (incf n))))
      (insert ")))\n")))))

(defun hitohude-interactive-readvar (default)
  ;; (interactive (hitohude-interactive-readvar "hitohude=input-example"))
  (list
   (intern
    (completing-read (format "Variable (default %s): " default)
		     obarray
		     #'(lambda (sym)
			 (and (boundp sym)
			      (string-match default (symbol-name sym))))
		     nil
		     ""
		     nil
		     default))))


(defstruct hitohude-node
  (no -1) ; <Node>
  succ   ; ((<Edge> <hitohude-node>)..)
  )

(defun hitohude-node-format (node)
  ;; node : <hitohude-node> | (<hitohude-node>..)
  ;; 𕶎ɕϊB
  (cond
   ((listp node)
    (with-temp-buffer
      (dolist (n node)
	(insertf "%s\n" (hitohude-node-format n)))
      (buffer-string)))
   (t
    (format "%d %s"
	    (hitohude-node-no node)
	    (mapcar #'(lambda (x) (list (car x) (hitohude-node-no (cadr x))))
		    (hitohude-node-succ node))))))

(when (featurep 'elim)
  (elim-add-printer
   #'(lambda (x) (or (hitohude-node-p x) (hitohude-node-p (car-safe x))))
   #'(lambda (x) (hitohude-node-format x))))


(defun hitohude-make-nodes (input)
  "input : <HitohudeInput>
A(<hitohude-node>..) B
Example:
  (hitohude-make-nodes hitohude=input-example)"
  (sck-letalist input
      (start end figure)
    (let ((nodes (mapcar #'(lambda (no) (make-hitohude-node :no no))
			 (sort (union (mapcar 'caar figure)
				      (mapcar 'cadar figure))
			       '<))))
      (dolist (f figure)
	;; f : ((<Node> <Node>) (<Edge>..))
	(destructuring-bind ((from to) edges) f
	  (let ((fromn (hitohude-no-to-node nodes from))
		(ton   (hitohude-no-to-node nodes to)))
	    (dolist (edge edges)
	      (push (list edge ton) (hitohude-node-succ fromn))
	      (push (list edge fromn) (hitohude-node-succ ton))))))
      ;;
      nodes)))

(defun hitohude-no-to-node (nodes no)
  (or (find no nodes :key 'hitohude-node-no)
      (error "hitohude-no-to-node: no=%d not found" no)))


(defun hitohude-find-routes (input)
  "input:<HitohudeInput> SĂ̌oH̃Xg ((<Edge>..)..) B"
  (sck-letalist input
      (start end figure)
    (let ((nodes (hitohude-make-nodes input))
	  (edges (foldr 'union nil (mapcar 'cadr figure)))
	  (routesq (make-que)))
      (hitohude-find-routes/rec (hitohude-no-to-node nodes start)
				(hitohude-no-to-node nodes end)
				(length edges)
				nil
				routesq)
      (que-to-list routesq))))

(defun hitohude-find-routes/rec (node endnode remain path routesq)
  "ċAIɌsB
node, endnode : <hitohude-node>
remain : <Fixnum>
path : (<Edge>..)
routeseq : <que>"
  (cond
   ;; SĂ̓ǂāA
   ((= remain 0)
    ;; endnode ֒BĂA
    (when (eq node endnode)
      ;; o̓L[ path L^B
      (append-que routesq (reverse path))))
   (t
    ;; ̓ ipath ɖjɂāA
    (dolist (s (hitohude-node-succ node))
      ;; s : (<Edge> <hitohude-node>)
      (unless (memq (car s) path)
	;; ċAIɒB
	(hitohude-find-routes/rec (cadr s)
				  endnode
				  (1- remain)
				  (cons (car s) path)
				  routesq))))))
