;;;; LIR $BA`:n$N$?$a$N4pK\E*4X?t(B
;;;; LastChange: Tue Apr 03 15:46:00 2007

(require 'sck-lir-foldconst)
(require 'sck-lir-jopt)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; $BDj?tDj5A(B

(defconst sck=lir-integral-types
  '(i8 i16 i32 i64 i128))

(defconst sck=lir-floating-types
  '(f32 f64 f128))

(defconst sck=lir-types
  (append sck=lir-integral-types sck=lir-floating-types))

(defconst sck=lir-const-ops
  '(iconst fconst))

(defconst sck=lir-addr-ops
  '(static frame label))

(defconst sck=lir-reg-ops
  '(reg subreg))

(defconst sck=lir-arith-ops
  '(neg add sub mul divs divu mods modu))

(defconst sck=lir-conv-ops
  '(convsx convzx convit convfx convft convfi convsf convuf))

(defconst sck=lir-bit-ops
  '(band bor bxor bnot))

(defconst sck=lir-shift-ops
  '(lshs lshu rshs rshu))

(defconst sck=lir-tst-ops
  '(tsteq tstne tstlts tstles tstgts tstges
	        tstltu tstleu tstgtu tstgeu))

(defconst sck=lir-tst-not-alist
  ;; $B??56$r5U$K$9$kHf3S!'(B
  ;; x c1 y != x c2 y
  '((tsteq     tstne)
    (tstne     tsteq)
    (tstlts    tstges)
    (tstltu    tstgeu)
    (tstles    tstgts)
    (tstleu    tstgtu)
    (tstgts    tstles)
    (tstgtu    tstleu)
    (tstges    tstlts)
    (tstgeu    tstltu)))

(defconst sck=lir-tst-xargs-alist
  ;; $B0z?t$N:81&8r49!'(B
  ;; x c1 y == y c2 x
  '((tsteq     tsteq)
    (tstne     tstne)
    (tstlts    tstgts)
    (tstltu    tstgtu)
    (tstles    tstges)
    (tstleu    tstgeu)
    (tstgts    tstlts)
    (tstgtu    tstltu)
    (tstges    tstles)
    (tstgeu    tstleu)))

(defconst sck=lir-mem-ops
  '(mem))

(defconst sck=lir-set-ops
  '(set))

(defconst sck=lir-jump-ops
  '(jump jumpc jumpn))

(defconst sck=lir-untyped-ops
  (append sck=lir-jump-ops
	  '(deflabel call prologue epilogue parallel use clobber)))

(defconst sck=lir-pure-ops
  (append sck=lir-arith-ops
	  sck=lir-conv-ops
	  sck=lir-bit-ops
	  sck=lir-shift-ops
	  sck=lir-tst-ops))

(defconst sck=lir-atomic-ops
  (append sck=lir-const-ops
	  sck=lir-addr-ops
	  sck=lir-reg-ops))

(defconst sck=lir-nonatomic-ops
  (append sck=lir-pure-ops
	  sck=lir-mem-ops
	  sck=lir-set-ops))

(defconst sck=lir-typed-ops
  (append sck=lir-atomic-ops
	  sck=lir-nonatomic-ops))

(defconst sck=lir-unary-ops
  (append '(neg bnot mem)
	  sck=lir-conv-ops))

(defconst sck=lir-binary-ops
  (set-difference sck=lir-nonatomic-ops sck=lir-unary-ops))
   
(defconst sck=lir-commutative-ops
  '(add mul band bxor bor tsteq tstne))

(defconst sck=lir-associative-ops
  '(add mul band bxor bor))

(defconst sck=lir-exp-ops
  (append sck=lir-typed-ops sck=lir-untyped-ops))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; L type

(defun sck-lir-type-kind (type)
  "TYPE:<Ltype> $B$N<oN`$rJV$9!#(B
Example:
 (sck-lir-type-kind 'i32) => floating
 (sck-lir-type-kind 'f32) => floating
 (sck-lir-type-kind 'a100) => aggregate"
  (ecase (aref (symbol-name type) 0)
    (?i 'integral)
    (?f 'floating)
    (?a 'aggregate)))

(defun sck-lir-type-p (x)
  (and (symbolp x)
       (memq (aref (symbol-name x) 0) '(?i ?f ?a))
       (string-match "\\`[0-9]+\\'"
		     (substring (symbol-name x) 1))))

(defun sck-lir-integral-type-p (type)
  (eq (sck-lir-type-kind type) 'integral))

(defun sck-lir-floating-type-p (type)
  (eq (sck-lir-type-kind type) 'floating))

(defun sck-lir-type-bits (type)
  "Example:
 (sck-lir-type-bits 'i32) => 32
 (sck-lir-type-bits 'a100) => 100"
  (string-to-int (substring (symbol-name type) 1)))

(defun sck-lir-type-bytes (type)
  "Example:
 (sck-lir-type-bytes 'i32) => 4
 (sck-lir-type-bytes 'a100) => 12"
  (/ (string-to-int (substring (symbol-name type) 1)) 8))


(defun sck-lir-type (exp)
  (assert (sck-lir-typed-exp-p exp))
  (cadr exp))

(defun sck-lir-reg-name (exp)
  "EXP:<RegExp> $B$N%l%8%9%?L>$rJV$9!#(B
Example:
 (sck-lir-reg-name '(reg I32 \"foo\")) => \"foo\"
 (sck-lir-reg-name '(subreg I32 (reg I16 \"bar\") 0)) => \"bar\""
  (assert (sck-lir-reg-exp-p exp))
  (ecase (car exp)
    (reg (caddr exp))
    (subreg (sck-lir-reg-name (caddr exp)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; predicates

(defun sck-lir-exp-p (exp)
  (memq (car-safe exp) sck=lir-exp-ops))

(defun sck-lir-const-exp-p (exp)
  (memq (car-safe exp) sck=lir-const-ops))

(defun sck-lir-typed-exp-p (exp)
  (memq (car-safe exp) sck=lir-typed-ops))

(defun sck-lir-pure-exp-p (exp)
  (memq (car-safe exp) sck=lir-pure-ops))

(defun sck-lir-untyped-exp-p (exp)
  (memq (car-safe exp) sck=lir-untyped-ops))

(defun sck-lir-atomic-exp-p (exp)
  (memq (car-safe exp) sck=lir-atomic-ops))

(defun sck-lir-set-exp-p (exp)
  (eq (car-safe exp) 'set))

(defun sck-lir-set-mem-exp-p (exp)
  "(set (mem ...) ...) $B$+!)(B"
  (and (eq (car-safe exp) 'set)
       (eq (car (caddr exp)) 'mem)))

(defun sck-lir-set-reg-exp-p (exp)
  "REG $B$^$?$O(B SUBREG $B$X$NBeF~$+!)(B"
  (and (eq (car-safe exp) 'set)
       (sck-lir-reg-exp-p (caddr exp))))

(defun sck-lir-reg-exp-p (exp)
  "REG $B$^$?$O(B SUBREG $B$+!)(B"
  (memq (car-safe exp) sck=lir-reg-ops))

(defun sck-lir-real-reg-p (exp)
  "$B<B%l%8%9%?$+!)(B
Example:
 (sck-lir-real-reg-p '(reg I32 \"%r0\")) => t
 (sck-lir-real-reg-p '(reg I32 \"r0\")) => nil"
  (and (sck-lir-reg-exp-p exp)
       (= ?% (aref (sck-lir-reg-name exp) 0))))

(defun sck-lir-commutative-exp-p (exp)
  (memq (car-safe exp) sck=lir-commutative-ops))

(defun sck-lir-associative-exp-p (exp)
  (memq (car-safe exp) sck=lir-associative-ops))

(defun sck-lir-tst-exp-p (exp)
  (memq (car-safe exp) sck=lir-tst-ops))

(defun sck-lir-unary-exp-p (exp)
  (memq (car-safe exp) sck=lir-unary-ops))

(defun sck-lir-binary-exp-p (exp)
  (memq (car-safe exp) sck=lir-binary-ops))


(defun sck-lir-const-zero-p (exp)
  (and (sck-lir-const-exp-p exp)
       (bn-zerop (caddr exp))))

(defun sck-lir-const-one-p (exp)
  (and (sck-lir-const-exp-p exp)
       (bn-= 1 (caddr exp))))

(defun sck-lir-const-minusone-p (exp)
  (and (sck-lir-const-exp-p exp)
       ;; $B!z(B iconst $B$NCM$O(B signed $B$H$$$&5,Ls$r2>Dj$7$F$$$k!#(B
       (bn-= -1 (caddr exp))))


(defun sck-lir-exp-include (exp pred)
  "EXP:<Lexp> $B$NItJ,<0!J(BEXP $B<+?H$r4^$`!K$G(B PRED:<Lexp> -> <Boolean> 
$B$rK~$?$9$b$N$,$"$l$P(B Non-nil $B$J$1$l$P(B nil $B$rJV$9!#(B
Example:
 (sck-lir-exp-include '(add i32 (mem i32 (frame i32 \"a\")) (iconst i32 1))
                       '(lambda (e) (eq (car-safe e) 'frame)))
 => t"
  (or (funcall pred exp)
      (case (car exp)
	;;
	;; untyped exp
	;;
	((prologue epilogue parallel use)
	 (sck-lir-exps-include (cdr exp) pred))
	((clobber)
	 ;; $B!zItJ,<0(B defs $B$N9MN8(B
	 (funcall pred exp))
	((deflabel)
	 (funcall pred exp))
	((call) ; (call <TypedExp> (<TypedExp>..) (<InterfaceVar>..))
	 (or (funcall pred (cadr exp))
	     (sck-lir-exps-include (caddr exp) pred)
	     (sck-lir-exps-include (cadddr exp) pred)))
	((jump)
	 (funcall pred (cadr exp)))
	((jumpc)
	 (sck-lir-exps-include (cdr exp) pred))
	((jumpn) (assert nil))
	;;
	;; typed exp
	;;
	(t
	 (if (sck-lir-atomic-exp-p exp)
	     nil
	   (sck-lir-exps-include (cddr exp) pred))))))

(defun sck-lir-exps-include (exps pred)
  "sck-lir-exp-include $B$HF1MM$@$,!"(Bexps:(<Lexp>..) $B$K$D$$$F%F%9%H$r9T$&!#(B"
  (dolist (e exps nil)
    (when (sck-lir-exp-include e pred)
      (return t))))

(defun sck-lir-function-include (lfunc pred)
  "LFUNC:<Lfunc> $B$N(B <Lexp>.. $B$NItJ,<0$G(B PRED:<Lexp> -> <Boolean> $B$r(B
$BK~$?$9$b$N$,$"$l$P(B Non-nil $B$J$1$l$P(B nil $B$rJV$9!#(B"
  (sck-lir-exps-include (sck-lir-function-exps lfunc)
			pred))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; $B%"%/%;%94X?t(B

(defmacro* sck-lir-doargs ((ep exp) &rest body)
  "EXP: (... <TypedExp>.. ...)
EXP $B$N:G=i$K8=$l$kO"B3$7$?(B <TypedExp> $B$r(B
$B$=$l$>$l(B EP $B$KB+G{$7$F(B BODY $B$r<B9T$9$k!#(B"
  `(let ((--tmp-- (sck-lir-exp-beg ,exp))
	 (,ep nil))
     (while (sck-lir-typed-exp-p (car --tmp--))
       (setq ,ep (car --tmp--))
       ,@body
       (setq --tmp-- (cdr --tmp--)))))

(put 'sck-lir-doargs 'lisp-indent-function 1)

(defmacro* sck-lir-doargs* ((ep exp) &rest body)
  "EXP: (... <TypedExp>.. ...)
EXP $B$N:G=i$K8=$l$kO"B3$7$?(B <TypedExp> $B$N%W%l!<%9!J$=$l$r(B car $B$H$9$k%3%s%9!K$r(B
$B$=$l$>$l(B EP $B$KB+G{$7$F(B BODY $B$r<B9T$9$k!#(B"
  `(let ((,ep (sck-lir-exp-beg ,exp)))
     (while (sck-lir-typed-exp-p (car ,ep))
       ,@body
       (setq ,ep (cdr ,ep)))))

(put 'sck-lir-doargs* 'lisp-indent-function 1)

(font-lock-add-keywords 'emacs-lisp-mode
  `((,(concat "(" (regexp-opt
		   '("sck-lir-doargs"
		     "sck-lir-doargs*"
		     ) t)
	      "\\>")
     1
     font-lock-keyword-face)))

(defun sck-lir-exp-beg (exp)
  "$B:G=i$N(B <TypedExp> $B$N(B place $B$rJV$9!#(B
Example:
 (sck-lir-exp-beg '(add i32 (sub) (mul))) => ((sub) (mul))"
  (while (and exp
	      (not (sck-lir-typed-exp-p (car exp))))
    (setq exp (cdr exp)))
  ;;
  exp)

(defun sck-lir-exp-end (exp)
  "$B:G8e$N(B <TypedExp> $B$h$j8e$m$N%j%9%H$rJV$9!#(B
Example:
 (sck-lir-exp-end '(add i32 (sub) (mul) &z 1)) => (&z 1)"
  (setq exp (sck-lir-exp-beg exp))
  (while (and exp (sck-lir-typed-exp-p (car exp)))
    (setq exp (cdr exp)))
  ;;
  exp)

(defun sck-lir-args (exp)
  "EXP: (... <TypedExp>.. ...)
EXP $B$N:G=i$K8=$l$kO"B3$7$?(B <TypedExp> $B$N%j%9%H$rJV$9!#(B
$BCm0U!'$3$N%j%9%H$O$b$H$N(B EXP $B$H6&M-9=B$$O$J$$$N$G(B EXP $B$rGK2uE*$K(B
$B=$@5$9$k$?$a$N%j%9%H$H$7$F$O;H$($J$$$,!"(Bsetf $B%a%=%C%I$rDj5A$7$F$"$k$N$G!"(B
\(setf (sck-lir-args exp) newargs) $B$K$h$j=$@5$9$k$3$H$,=PMh$k!#(B
$B$3$N>l9g!"(Bnewargs $B$OGK2u$7$J$$!#(B"
  (let ((-args- nil))
    (sck-lir-doargs (e exp)
      (push e -args-))
    (nreverse -args-)))

(defun sck-lir-args-set (exp newargs)
  (let ((beg (sck-lir-exp-beg exp))
	(end (sck-lir-exp-end exp)))
    (assert beg)
    (assert newargs)
    (setq newargs (append newargs end))
    (setf (car beg) (car newargs)
	  (cdr beg) (cdr newargs))
    exp))

(defsetf sck-lir-args sck-lir-args-set)

(defun sck-lir-xchange-args (exp)
  "EXP:(... <TypedExp> <TypedExp> ...) $B$N$U$?$D$N0z?t$r8r49$7!"(BEXP $B$rJV$9!#(B"
  (let ((args (sck-lir-args exp)))
    (assert (= 2 (length args)))
    (setf (sck-lir-args exp) (reverse args))
    exp))

(defun sck-lir-putprop (lobj key val)
  "LOBJ: <Lexp> | <GlobalEnt> | <LocalEnt>
$B$KB0@-$rCV$/!#B0@-(B KEY $B$O(B & $B$G;O$^$k%7%s%\%k$G$"$k$3$H!#(B"
  (assert (consp lobj))
  (assert (= (aref (symbol-name key) 0) ?&))
  (let ((a (memq key lobj)))
    (cond
     (a
      (assert (consp (cdr a)))
      (setf (cadr a) val))
     (t
      (nconc lobj (list key val))))
    ;;
    val))

(defun sck-lir-getprop (lobj key)
  "LOBJ: <Lexp> | <GlobalEnt> | <LocalEnt>
$B$NB0@-(B KEY $B$NCM$r!"B0@-$,$J$1$l$P(B nil $B$rJV$9!#(B"
  (assert (consp lobj))
  (assert (= (aref (symbol-name key) 0) ?&))
  (let ((a (memq key lobj)))
    (cond
     (a
      (assert (consp (cdr a)))
      (cadr a))
     (t
      nil))))

(defun sck-lir-remprop (lobj key)
  "LOBJ: <Lexp> | <GlobalEnt> | <LocalEnt>
$B$NB0@-(B KEY $B$r:o=|$9$k!#(B"
  (assert (consp lobj))
  (assert (= (aref (symbol-name key) 0) ?&))
  (let ((lobj1 lobj))
    (block break
      (while lobj1
	(when (eq (cadr lobj1) key)
	  (setf (cdr lobj1) (cdddr lobj1))
	  (return-from break))
	(setq lobj1 (cdr lobj1))))
    ;;
    lobj))


(defun sck-lir-without-plist (lobj)
  "$BItJ,<0$b4^$a!"A4$F$N(B &-$BB0@-$r:o=|$7$?(B lobj $B$rJV$9!#(B
$B85$N<0$OGK2u$7$J$$!#(B
Example:
 (sck-lir-without-plist '(a (b &key val) (c &kkk 11) &mmm lll &kkk 1))
 => (a (b) (c))"
  (cond
   ((atom lobj) lobj)
   ((and (symbolp (car lobj))
	 (= ?& (aref (symbol-name (car lobj)) 0)))
    nil)
   (t
    (cons (sck-lir-without-plist (car lobj))
	  (sck-lir-without-plist (cdr lobj))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; module, function $B%"%/%;%94X?t(B

(defun sck-lir-module-name (module)
  (assert (eq (car-safe module) 'module))
  (cadr module))

(defun sck-lir-module-defs (module)
  (assert (eq (car-safe module) 'module))
  (assq 'defs module))

(defun sck-lir-module-symtab (module)
  (assert (eq (car-safe module) 'module))
  (assq 'symtab module))

(defun sck-lir-module-body (module)
  (assert (eq (pop module) 'module))
  (assert (stringp (pop module)))
  (while (memq (car-safe (car module)) '(defs symtab))
    (pop module))
  module)


(defun sck-lir-function-name (lfunc)
  "LFUNC:<Lfunc> $B$N(B name $B$rJV$9!#(B"
  (assert (eq (car-safe lfunc) 'function))
  (cadr lfunc))

(defun sck-lir-function-symtab (lfunc)
  "LFUNC:<Lfunc> $B$N(B symtab $B$rJV$9!#(B"
  (assert (eq (car-safe lfunc) 'function))
  (assert (eq (car (caddr lfunc)) 'symtab))
  (caddr lfunc))

(defun sck-lir-function-exps (lfunc)
  "LFUNC:<Lfunc> $B$N(B <Lexp>.. $B$rJV$9!#(B"
  (assert (eq (car-safe lfunc) 'function))
  (assert (eq (car (caddr lfunc)) 'symtab))
  (cdddr lfunc))

(defsetf sck-lir-function-exps (lfunc) (exps)
  `(setf (cdddr ,lfunc) ,exps))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; $BFCDj$N7A$N<0$r:n$k(B

(defun sck-lir-make-set-exp (x y)
  (copy-list `(set ,(cadr x) ,x ,y)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; $BItJ,<0$NCV49(B

(defun sck-lir-subst-exp (exp func)
  "EXP:<Lexp> $B$NItJ,<0$r2<$+$i>e$KGK2uE*$K(B FUNC $B$G=$@5$9$k!#(B"
  ;; $B!zItJ,<0$r:F5"E*$K$a$0$k=hM}!"(BUntypedExp$B!"(Bdefs $B$N7o!#(B
  (case (car exp)
    ;;
    ;; untyped exp
    ;;
    ((prologue epilogue parallel use)
     ;; $B!zB0@-$N=hM}!D(B
     (funcall func (cons (car exp)
			 (mapcar #'(lambda (e) (sck-lir-subst-exp e func))
				 (cdr exp)))))
    ((clobber) (funcall func exp)) ; $B!z(B defs ...
    ((deflabel) (funcall func exp))
    ((call) ; (call <TypedExp> (<TypedExp>..) (<InterfaceVar>..))
     ;; $B!zB0@-$N=hM}!D(B
     (funcall func
	      `(call ,(sck-lir-subst-exp (cadr exp) func)
		     ,(mapcar #'(lambda (e) (sck-lir-subst-exp e func))
			      (caddr exp))
		     ,(mapcar #'(lambda (e) (sck-lir-subst-exp e func))
			      (cadddr exp)))))
    ((jump jumpc)
     ;; $B!zB0@-$N=hM}!D(B
     (funcall func
	      (cons (car exp)
		    (mapcar #'(lambda (e) (sck-lir-subst-exp e func))
			    (cdr exp)))))
    ((jumpn) (assert nil))
    ;;
    ;; typed exp
    ;;
    (t
     (sck-lir-doargs* (ep exp)
       (setf (car ep) (sck-lir-subst-exp (car ep) func)))
     (funcall func exp))))


(defun sck-lir-subst-exp-function (lfunc func)
  "LFUNC:<Lfunc> $B$NItJ,(BL-$B<0$rGK2uE*$K(B func $B$GCV$-49$($k!#(B"
  (dorest (ep (sck-lir-function-exps lfunc))
    (setf (car ep) (sck-lir-subst-exp (car ep) func)))
  lfunc)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; symtab $B4X78(B

(defstruct (sck-lir-gsent (:type list))
  name
  kind ; reg | static
  type ; <Ltype> | nil
  align
  ;; $B0J2<$O(B kind=static $B$N$_(B
  segment
  linkage)

(defstruct (sck-lir-lsent (:type list))
  name
  kind ; reg | frame
  type
  align
  offset)


(defun sck-lir-lsymtab-frame-size (lsymtab)
  "frame $B$N%5%$%:$rJV$9!#(B
offset=nil $B$OL5;k!#(B
$B!z(B bn-
Example:
 (sck-lir-lsymtab-frame-size '(symtab (\"e\" frame i32 4 -100))) => 100"
  (let ((maxoffset 0))
    (dolist (lsent (cdr lsymtab))
      (when (and (eq (sck-lir-lsent-kind lsent) 'frame)
		 (sck-lir-lsent-offset lsent))
	(setq maxoffset (min maxoffset (sck-lir-lsent-offset lsent)))))
    (- maxoffset)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; $B%F%s%]%i%jJQ?t$N:n@.(B

(defun sck-lir-new-vreg (type lsymtab &optional regset)
  "$B?7$?$J2>A[%l%8%9%?$r:n$j!"(Blsymtab $B$KEPO?$7!":n$C$?%l%8%9%?$rJV$9!#(B
regset $B$,(B Non-nil $B$J$i%(%s%H%j$N(B &regset $BB0@-$b%;%C%H$9$k!#(B"
  (let* ((newname (make-new-name
		   #'(lambda (n) (assoc n lsymtab))
		   "_"))
	 (ent (copy-list
	       ;; <RegEnt> ::= REG <Ltype> <Align> <Offset>
	       `(,newname
		 reg
		 ,type
		 ;; $B!z(B
		 ,(case type
		    (i8 1)
		    (i16 2)
		    ((i32 f32) 4)
		    ((i64 f64) 8))
		 nil))))
    (when regset
      (sck-lir-putprop ent '&regset regset))
    (nconc lsymtab (list ent))
    (copy-list `(reg ,type ,newname))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; $BJ,4t%i%Y%k4X78(B

(defun sck-lir-label-exp-labelname (labelexp)
  "$B%i%Y%k<0(B LABELEXP $B$N%i%Y%kL>J8;zNs$rJV$9!#(B
Example:
  (sck-lir-label-exp-labelname '(label i32 \"L1\")) => \"L1\""
  (caddr labelexp))

(defun sck-lir-make-label-exp (labelname)
  "$B%i%Y%kJ8;zNs(B LABELNAME $B$+$i%i%Y%k;2>H<0$r:n$k!#(B
Example:
 (sck-lir-make-label-exp \"L1\") => (label i32 \"L1\")"
  `(label
    i32 ; $B!z%^%7%s5-=R$+$i;}$C$F$/$k$Y$-(B
    ,labelname))

(defun sck-lir-deflabel-exp-labelname (deflabel)
  "deflabel $B<0(B DEFLABEL $B$N%i%Y%kJ8;zNs$rJV$9!#(B
Example:
 (sck-lir-deflabel-exp-labelname '(deflabel \"L1\")) => \"L1\""
  (cadr deflabel))

(defun sck-lir-make-deflabel-exp (labelname)
  "$B%i%Y%kJ8;zNs(B LABELNAME $B$+$i%i%Y%kDj5A<0$r:n$k!#(B
Example:
 (sck-lir-make-deflabel-exp \"L1\") => (deflabel \"L1\")"
  `(deflabel ,labelname))
  
(defun sck-lir-make-labeltab (module)
  "$B%b%8%e!<%k(B MODULE $B$NA4$F$N%i%Y%k$,EPO?$5$l$?%i%Y%k%F!<%V%k$rJV$9!#(B
$B%i%Y%k%F!<%V%k$O(B (nil <String:labelname>..) $B$H$$$&%j%9%H!#(B"
  (assert (eq (car-safe module) 'module))
  (let ((labelnames nil))
    (dolist (fun module)
      (when (eq (car-safe fun) 'function)
	(dolist (e (cdddr fun))
	  (when (eq (car-safe e) 'deflabel)
	    (let ((labelname (sck-lir-deflabel-exp-labelname e)))
	    (assert (null (member labelname labelnames)))
	    (push labelname labelnames))))))
    ;;
    (cons nil
	  (nreverse labelnames))))

(defun sck-lir-new-labelname (labeltab*)
  "sck-lir-make-labeltab $B$G:n$i$l$?(B LABELTAB $B$K$J$$%i%Y%kJ8;zNs$r:n$j!"(B
$B$=$l$r(B LABELTAB $B$KEPO?$7$F$+$iJV$9!#(B
Example:
 (let ((ltab (list nil \"L1\")))
   (list (sck-lir-new-labelname ltab) ltab))
 => 
 (\"L2\"
  (nil \"L1\" \"L2\"))"
  (assert labeltab*)
  (let ((newname (make-new-name #'(lambda (n) (member n labeltab*)) "L")))
    (nconc labeltab* (list newname))
    newname))
  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; $BDj?t<0(B <TypedExp> $B$NI>2A(B

(defun sck-lir-eval-pure-exp (exp)
  "EXP: <PureExp> $B$rI>2A$9$k!#(B
EXP $B$ODj?t<0$KI>2A=PMh$k(B <PureExp> $B$G$J$1$l$P$J$i$J$$!#(B
Return: <ConstExp>

<TypedExp> ::= <ConstExp> | <AddrExp> | <RegExp>
             | <MemExp> | <SetExp> | <PureExp>

<PureExp>           ::= (<PureOps> <Ltype> <TypedExp> {<TypedExp>})
<PureOps>           ::= <ArithOps> | <ConvOps> | <BitOps> | <ShiftOps>
                      | <TstOps> 
<ArithOps>          ::= neg | add | sub | mul | divs | divu | mods | modu
<ConvOps>           ::= convsx | convzx | convit | convfx
                      | convft | convfi | convsf | convuf
<BitOps>            ::= band | bor | bxor | bnot
<ShiftOps>          ::= lshs | lshu | rshs | rshu
<TstOps>            ::= tsteq | tstne | tstlts | tstles | tstgts | tstges
                                      | tstltu | tstleu | tstgtu | tstgeu
Example:
 (sck-lir-eval-pure-exp '(add i32 (iconst i32 1) (iconst i32 1)))
 => (iconst i32 2)"
  (case (car exp)
    ;; <ConstExp> ::= <IntConstExp> | <FloatConstExp>
    ((iconst fconst)
     exp)
    ;; <AddrExp> ::= <StaticExp> | <FrameExp> | <LabelExp>
    ((static frame label)
     (assert nil))
    ;; <RegExp> ::= <SimpleRegExp> | <SubRegExp>
    ((reg subreg)
     (assert nil))
    ;; <MemExp> ::= (MEM <Ltype> <TypedExp>)
    (mem
     (assert nil))
    ;; <SetExp> ::= (SET <Ltype> < <MemExp> | <RegExp> > <TypedExp>)
    (set
     (assert nil))
    ;; <PureExp> ::= (<PureOps> <Ltype> <TypedExp> {<TypedExp>})
    (t
     (assert (sck-lir-pure-exp-p exp))
     (let* ((type (sck-lir-type exp))
	    (bits (sck-lir-type-bits type))
	    (args (sck-lir-args exp))
	    (a (car args))
	    (atype (cadr a))
	    (abits (sck-lir-type-bits atype))
	    (vals (let ((args1 (mapcar 'sck-lir-eval-pure-exp args)))
		    (assert (every 'sck-lir-const-exp-p args1))
		    (mapcar 'caddr args1)))
	    (val (ecase (car exp)
		   (neg (apply 'bn-neg vals))
		   (add (apply 'bn-add vals))
		   (sub (apply 'bn-sub vals))
		   (mul (apply 'bn-mul vals))
		   (divs (apply 'bn-div vals))
		   (mods (apply 'bn-mod vals))
		   (divu (apply 'bn-div
				(mapcar #'(lambda (n) (bn-unsigned n bits))
					vals)))
		   (modu (apply 'bn-mod
				(mapcar #'(lambda (n) (bn-unsigned n bits))
					vals)))
		   (bnot (apply 'bn-lognot vals))
		   (bor (apply 'bn-logior vals))
		   (bxor (apply 'bn-logxor vals))
		   (band (apply 'bn-logand vals))
		   (lshs (bn-ash (car vals)
				 (cadr vals)))
		   (rshs (bn-ash (car vals)
				 (bn-neg (cadr vals))))
		   (lshu (bn-ash (bn-unsigned (car vals) abits)
				 (cadr vals)))
		   (rshu (bn-ash (bn-unsigned (car vals) abits)
				 (bn-neg (cadr vals))))
		   (convsx (car vals))
		   (convzx (bn-unsigned (car vals) abits))
		   (convit (car vals))
		   (convfx (car vals))
		   (convft (car vals))
		   (convfi (bn-truncate (car vals)))
		   (convsf (bn-add 0.0 (car vals)))
		   (convuf (bn-add 0.0 (bn-unsigned (car vals) abits)))
		   (tsteq (if (apply 'bn-=  vals) 1 0))
		   (tstne (if (apply 'bn-/= vals) 1 0))
		   (tstlts (if (apply 'bn-<  vals) 1 0))
		   (tstles (if (apply 'bn-<= vals) 1 0))
		   (tstgts (if (apply 'bn->  vals) 1 0))
		   (tstges (if (apply 'bn->= vals) 1 0))
		   (tstltu (if (apply 'bn-< (mapcar #'(lambda (n)
							(bn-unsigned n abits))
						    vals))
			       1 0))
		   (tstleu (if (apply 'bn-<= (mapcar #'(lambda (n)
							 (bn-unsigned n abits))
						     vals))
			       1 0))

		   (tstgtu (if (apply 'bn-> (mapcar #'(lambda (n)
							(bn-unsigned n abits))
						    vals))
			       1 0))
		   (tstgeu (if (apply 'bn->= (mapcar #'(lambda (n)
							 (bn-unsigned n abits))
						     vals))
			       1 0))
		   )))
       ;;
       (ecase (sck-lir-type-kind type)
	 (integral
	  `(iconst ,type
		   ;; $BCM<+BN$O$+$J$i$:(B signed $B$GI=8=$9$k!#(B
		   ;; $B$3$l$K$h$j(B signed $B1i;;$O0z?tJQ49$,ITMW!#(B
		   ;; $B!z(B $B$3$N5,Ls$r30$7$F!"(Bsigned/unsigned $B$K0MB8$9$k1i;;$O(B
		   ;;    $B$D$M$K0z?t$rJQ49$9$k$h$&$K$7$F$bBg$7$FCY$/$J$i$J$$(B
		   ;;    $B$J$i$P!"5,Ls$O30$7$?$[$&$,$h$$$+$b$7$l$J$$!#(B
		   ,(bn-try-to-elnum (bn-signed val bits))))
	 (floating
	  `(fconst ,type ,(bn-try-to-elnum val))))))))
  
  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Pretty printer

(defun sck-lir-pp (lir &optional col)
  "LIR $B$N%W%j%F%#!<%W%j%s%H$r9T$$!"7k2L$rJ8;zNs$H$7$FJV$9!#(B"
  (with-temp-buffer
    (sck-lir-pp/aux lir (or col 0))
    ;; $B@hF,$NL5BL$J2~9T$r:o$k!#(B
    (goto-char (point-min))
    (skip-chars-forward "\n")
    (delete-region (point-min) (point))
    (buffer-string)))

(defun sck-lir-pp/aux (lir col)
  "$B%U%)!<%`$K1~$8$F(B sck-lir-pp/styles $B$r;H$$=hM}$r9T$&!#(B
$B>\:Y$O(B sck-lir-pp/styles $B$N%X%k%W$r;2>H$N$3$H!#(B"
  (cond
   ((atom lir)
    (sck-lir-pp/styles lir col nil))
   ;; ((...) ...)
   ((consp (car lir))
    (sck-lir-pp/styles lir col '(nil 1)))
   ;; deflabel $B$NA0$K6u9T$rF~$l$k!#(B
   ((eq (car lir) 'deflabel)
    (sck-lir-pp/styles lir col '(1)))
   ;; defs
   ((eq (car lir) 'defs)
    (sck-lir-pp/styles lir col '(nil 6)))
   ;; symtab
   ((eq (car lir) 'symtab)
    (sck-lir-pp/styles lir col '(nil 8)))
   ;; jumpc
   ((eq (car lir) 'jumpc)
    (sck-lir-pp/styles lir col '(nil t 7 t)))
   ;; function
   ((eq (car lir) 'function)
    (sck-lir-pp/styles lir col '(1 t 3)))
   ;; data
   ((eq (car lir) 'data)
    (sck-lir-pp/styles lir col '(1 t 3)))
   ;; $B0J2<!"0lHLE*$J>l9g$N=hM}!#(B
   ;; $B$^$:!">.$5$$$b$N$O@^$i$J$$!#(B
   ((<= (length (format "%S" lir)) 45)
    (sck-lir-pp/styles lir col nil))
   ;; $BD9$$<0$N%G%U%)%k%H$N@^$jJ}!#(B
   ((<= 2 (length lir))
    (sck-lir-pp/styles lir col '(nil t 2)))
   ;; $B$=$NB>$N>l9g$O@^$i$J$$!#(B
   (t
    (sck-lir-pp/styles lir col nil))))

(defun sck-lir-pp/styles (lir col styles)
  "STYLES: (<Newlines> <Style>..) | nil
<Newlines> ::= nil | <Fixnum>
<Style> ::= <Fixnum> | t

<Newlines> $B$O%U%)!<%`$N:G=i$KA^F~$9$k2~9T$N?t!"(Bnil $B$J$i(B 0$B!#(B
<Style>.. $B$O%U%)!<%`$N(B cdr $B$N%$%s%G%s%H$r;XDj$9$k!'(B
  <Fixnum> $B$J$i!"I,MW$K1~$8$F2~9T$7!"(Bcol+<Style> $B%+%i%`$XCV$/!#(B
  t $B$J$i2~9T$;$:1&$KCV$/!#(B
$B%U%)!<%`$NMWAG$,(B <Style>.. $B$h$jD9$$>l9g$O!":G8e$N(B <Style> $B$,E,MQ$5$l$k!#(B"
  (when (and (consp styles)
	     (integerp (car styles)))
    ;; styles $B$N:G=i$NMWAG$O2~9T;X<((B
    (unless (bolp)
      (newline))
    (newline (car styles)))
  (setq styles (cdr styles))
  ;; col $B$^$G%$%s%G%s%H$7!"(B
  (when (< col (current-column))
    (newline))
  (indent-to col)
  ;; styles $B$K=>$C$FI=<($9$k!#(B
  (cond
   ((atom lir)
    (insertf "%S" lir))
   ((null styles)
    (insertf "%S" lir))
   (t
    (insert "(")
    (sck-lir-pp/aux (car lir) (current-column))
    (let ((style (pop styles)))
      (dolist (x (cdr lir))
	(cond
	 ;;; nil $B$J$i:#$N>l=j$KCV$/!#(B
	 ;((null style)
	 ; (sck-lir-pp/aux x (current-column)))
	 ;; natnum $B$J$i(B (+ col style) $B$KCV$/!#(B
	 ((natnump style)
	  (sck-lir-pp/aux x (+ col style)))
	 ;; t $B$J$i2~9T$;$:1&$KCV$/!#(B
	 ((eq style t)
	  (sck-lir-pp/aux x (1+ (current-column))))
	 (t
	  (assert nil)))
	(when styles
	  (setq style (pop styles)))))
    (insert ")"))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; LIR mode $B!J2>!K(B

(defun sck-lir-mode ()
  "LIR mode $B!J2>!K(B"
  (interactive)
  (emacs-lisp-mode) 
  (sck-lir-font-lock))

(defun sck-lir-font-lock ()
  (setq font-lock-defaults
	'((lisp-font-lock-keywords)
	  nil nil (("+-*/.<>=!?$%_&~^:" . "w")) beginning-of-defun
	  (font-lock-mark-block-function . mark-defun)))
  (font-lock-mode t))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ELIM $B%W%j%s%?$N@_Dj(B

(defun sck-lir-po (x)
  ;; X $B$,(B LIR $B$C$]$$$+$I$&$+H=Dj$9$k!#(B
  ;; (sck-lir-po '(module "foo" (symtab ))) => t
  ;; (sck-lir-po '(mul i32 (reg i32 "a") (iconst i32 2))) => t
  ;; (sck-lir-po '(+)) => nil
  (and (consp x)
       (symbolp (car x))
       (case (car x)
	 ((module function)
	  (and (stringp (cadr x))
	       (assq 'symtab (cddr x))))
	 (symtab (every #'(lambda (e) (stringp (car e)))
			(cdr x)))
	 (t
	  (or (sck-lir-untyped-exp-p x)
	      (and (sck-lir-typed-exp-p x)
		   (sck-lir-type-p (cadr x))))))))

(when (featurep 'elim)
  (elim-add-printer
   #'sck-lir-po
   #'sck-lir-pp))

;;;
(provide 'sck-lir)
