;;;; Core-AST -> LIR
;;;; LastChange: Fri Apr 13 12:45:15 2007

;;(require 'sck-clex)
;;(require 'sck-cpps)
;;(require 'sck-ast)
(require 'sck-cps)
(require 'sck-lir)


(defun sck-lirgen-demo ()
  (interactive)
  (sck-lirgen-str sck=cps-input-example))

(defun sck-lirgen-buffer ()
  (interactive)
  (sck-lirgen-region (point-min) (point-max)))

(defun sck-lirgen-region (beg end)
  (interactive "r")
  (sck-lirgen-str (buffer-substring-no-properties beg end)))

(defun sck-lirgen-str (str)
  " STR p[XA*sck-lirgen* obt@֏o͂B"
  (with-display-buffer "*sck-lirgen*"
    (emacs-lisp-mode)
    (sck-ast-font-lock)
    (let ((output (sck-lirgen
		   (sck-ast-reduce
		    (sck-cps
		     (sck-cpps
		      (sck-clex
		       `((name "stdin")
			 (source ,str)))))))))
      (insert (sck-lirgen-pp-output output))
      output)))

(defun sck-lirgen-pp-output (output)
  (with-temp-buffer
    (sck-letalist output
	(lir messages)
      (insertf "((lir\n")
      (insertf "%s)\n\n" (sck-lir-pp lir 2))
      (insertf "(messages (\n")
      (dolist (msg messages)
	(insertf "%S\n" msg))
      (insert ")))\n"))
    (buffer-string)))
      
(defun sck-lirgen (reduce-output)
  "reduce ꂽ AST ł REDUCE-OUTPUT  LIR ɕϊB
R}hƂċNꂽꍇ̓|Cgȉ S- REDUCE-OUTPUT ɂȂA
ʂobt@ɕ\B"
  (interactive (list (save-excursion (read (current-buffer)))))
  (sck-letalist reduce-output
      (ast messages)
    (assert (eq (car-safe ast) 'module))
    (let ((lir nil)
	  (output nil))
      (cond
       ;; G[NĂA(module <Name> (symtab)) ɂĂ܂B
       ((sck-ast-error-in-messages-p messages)
	(setq lir (copy-tree `(module ,(cadr ast) (symtab)))))
       ;; m[}P[X
       (t
	(setq lir
	      (sck-lirgen-module
	       (sck-ast-sexp-to-module ast)))))
      ;;
      (setq output `((lir ,lir) (messages ,messages)))
      ;;
      (when (interactive-p)
	(with-display-buffer "*sck-lirgen*"
	  (insert (sck-lirgen-pp-output output))))
      ;;
      output)))


(defun sck-lirgen-module (module)
  ;; ^ AST  LIR 𐶐B
  (assert (eq (car module) 'module))
  (let* ((modname (cadr module))
	 (globalenv (caddr module)))
    `(module ,modname
	     (symtab ,@(sck-lirgen-gsymtab globalenv))
	     ,@(sck-lirgen-gdefs globalenv))))

(defun sck-lirgen-gsymtab (globalenv)
  ;; LIR  global symtab B
  ;; <GlobalSymtab> ::= (symtab {<GlobalEnt>} )
  ;; <GlobalEnt>   ::= (<String> < <RegEnt> | <StaticEnt> >)
  ;; <RegEnt>      ::= reg    <Ltype> <Align> <Offset> 
  ;; <StaticEnt>   ::= static < <Ltype> | nil > <Align> <Segment> <Linkage>
  ;; <Linkage>     ::= ldef | xdef | xref
  (assert (sck-ast-env-p globalenv))
  (let ((gsymtab (sck-ast-make-env nil)))
    (dolist (ident (sck-ast-env-idents globalenv))
      (when (eq (sck-ast-ident-kind ident) 'var)
	(setf ident (copy-sck-ast-ident ident)
	      (sck-ast-ident-env ident) nil)
	(let ((pident (sck-ast-search-ident gsymtab ident)))
	  (cond
	   (pident
	    ;;  sc }[W̌B
	    ;; ͍ŏ sc D悵ĂB
	    )
	   (t
	    (sck-ast-enter-ident gsymtab ident))))))
    ;;
    (mapcar 'sck-lirgen-ident-to-symtabent
	    (sck-ast-env-idents gsymtab))))

(defun sck-lirgen-ident-to-symtabent (ident)
  (let* ((name (sck-ast-ident-name ident))
	 (type (sck-ast-ident-type ident))
	 (ltype (sck-ast-ltypeof type))
	 (align (sck-ast-alignof type))
	 (sc (sck-ast-ident-sc ident)))
    (cond
     ;; local
     ;; <LocalEnt> ::= (<String> <<RegEnt> | <FrameEnt>>)
     ;; <RegEnt>   ::= REG   <Ltype> <Align> <Offset>
     ;; <FrameEnt> ::= FRAME <Ltype> <Align> <Offset>
     ((memq sc '(auto register))
      (copy-list `(,name
		   ,(ecase sc (register 'reg) (auto 'frame))
		   ,ltype
		   ,align
		   nil)))
     ;; <GlobalEnt> ::= (<String> <<RegEnt>|<StaticEnt>>)
     ;; <StaticEnt> ::= STATIC <<Ltype>|nil> <Align> <Segment> <Linkage>
     (t
      (copy-list
       `(,name
	 static
	 ,ltype
	 ,align
	 ,(if (sck-ast-ident-value ident) "data" "bss") ; 
	 ,(ecase sc
	    ((nil) 'xdef)
	    (static 'ldef)
	    (extern 'xref))))))))

(defun sck-lirgen-gdefs (globalenv)
  (let ((gdefsq (make-que)))
    (dolist (ident (sck-ast-env-idents globalenv))
      (let ((value (sck-ast-ident-value ident)))
	(when (eq (sck-ast-ident-kind ident) 'var)
	  (cond
	   ((eq (car-safe value) 'lambda)
	    (append-que gdefsq (sck-lirgen-function ident)))
	   (value
	    (append-que gdefsq (sck-lirgen-data ident)))
	   ((sck-ast-function-type-p (sck-ast-ident-type ident))
	    ;; ֐QƁB
	    )
	   ((not (eq (sck-ast-ident-sc ident) 'extern))
	    (append-que gdefsq (sck-lirgen-bss ident)))))))
    ;;
    (que-to-list gdefsq)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; data, bss

(defun sck-lirgen-data (ident)
  (let ((valexps (sck-ast-ident-value ident)))
    (setq valexps
	  (cond
	   ((sck-ast-expr-p valexps)
	    (list valexps))
	   ((eq (car-safe valexps) 'inits)
	    (cdr valexps))
	   (t
	    (assert nil))))
    (dorest (ep valexps)
      (when (sck-ast-expr-p (car ep))
	(setf (car ep)
	      (sck-lir-foldconst
	       (sck-lirgen-expr (car ep))))))
    ;; ȉ̌
    ;; (data "S1" (iconst i8 115) (iconst i8 116) (iconst i8 0))
    ;; ->
    ;; (data "S1" (i8 115 116 i8 0))
    `(data ,(sck-ast-ident-name ident)
	   ,@valexps)))

(defun sck-lirgen-bss (ident)
  `(data ,(sck-ast-ident-name ident)
	 (space ,(sck-ast-sizeof (sck-ast-ident-type ident)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; function

(defun sck-lirgen-function (fident)
  (let* ((fname (sck-ast-ident-name fident))
	 (lambda (sck-ast-ident-value fident))
	 (paramenv (caddr lambda))
	 (block (cadddr lambda))
	 (localenv (cadr block))
	 (stmts (cddr block))
	 (lsymtab nil)
	 (exprs nil)
	 ;;
	 (prologue (cons 'prologue
			 (mapcar #'(lambda (ident)
				     (sck-lirgen-expr
				      (sck-ast-make-vref-expr ident)))
				 (sck-ast-env-idents paramenv))))
	 (lsymtab (cons 'symtab
			(mapcar 'sck-lirgen-ident-to-symtabent
				 (sck-ast-env-idents localenv))))
	 (exps (mapcar 'sck-lirgen-stmt stmts)))
    ;;
    ;; prologue ŏ̍ŏ̊{ubN̐擪֑}B
    ;;
    (assert (eq (caar exps) 'deflabel))
    (setq exps `(,(car exps)
		 ,prologue
		 ,@(cdr exps)))
    ;;
    `(function ,fname
	       (symtab ,@(mapcar 'sck-lirgen-ident-to-symtabent
				 (filter
				  #'(lambda (ident)
				      (eq (sck-ast-ident-kind ident) 'var))
				  (append (sck-ast-env-idents paramenv)
					  (sck-ast-env-idents localenv)))))
	       ,@exps)))


(defun sck-lirgen-stmt (stmt)
  (if (sck-ast-expr-p stmt)
      (sck-lirgen-expr stmt)
    (ecase (car stmt)
      (label
       `(deflabel ,(cadr stmt)))
      (goto
       `(jump ,(sck-lirgen-label (cadr stmt))))
      ((if)
       `(jumpc ,(sck-lirgen-expr (cadr stmt))
	       ,(sck-lirgen-label (cadr (caddr stmt)))
	       ,(sck-lirgen-label (cadr (cadddr stmt)))))
      (return
       (cond
	((cadr stmt)
	 `(epilogue ,(sck-lirgen-expr (cadr stmt))))
	(t
	 '(epilogue))))
      ((set call)
       (sck-lirgen-expr stmt)))))

(defun sck-lirgen-label (name)
  ;; (sck-lirgen-label "L1") => (label i32 "L1")
  `(label ,(sck-ast-ltypeof sck=ast-pointer-implement-type) ,name))

(defconst sck=cps-lirgen-pointer-ltype
  (sck-ast-ltypeof sck=ast-pointer-implement-type))
	  

(defun sck-lirgen-expr (expr)
  (let* ((op (sck-ast-expr-op expr))
	 (args (sck-ast-expr-args expr))
	 (type (sck-ast-expr-type expr))
	 (ltype (sck-ast-ltypeof type)))
    (ecase op
      ;; 萔
      (const
       (sck-lirgen-const expr))
      ;; ϐ
      (vref
       (sck-lirgen-vref expr))
      ;; signed/unsigned Ɋ֌WȂ
      ((neg add sub mul
	    tsteq tstne
	    bnot bor bxor band)
       (cond
	((and (memq op '(add sub))
	      (some 'sck-ast-pointer-type-p
		    (mapcar 'sck-ast-expr-type args)))
	 ;; pointer +/- integer, pointer - pointer
	 (sck-lirgen-pointer-addsub expr))
	(t
	 `(,op ,ltype ,@(mapcar 'sck-lirgen-expr args)))))
      ;; signed/unsigned ̋ʂKvȎ
      ((div mod lsh rsh
	    tstlt tstgt tstle tstge)
       `(,(intern (format "%s%s" op (if (sck-ast-unsigned-scalar-type-p
					 (sck-ast-expr-type (car args)))
					"u"
				      "s")))
	 ,ltype
	 ,@(mapcar 'sck-lirgen-expr args)))
      ;; ϊ
      (conv
       (sck-lirgen-conv expr))
      ;; 
      (set
       (cond
	;; (set ret (call ...))
	((eq (sck-ast-expr-op (cadr args)) 'call)
	 (let* ((largs (mapcar 'sck-lirgen-expr args))
		(lret (car largs))
		(lcall (cadr largs)))
	   (setf (cadddr lcall) (list lret))
	   lcall))
	(t
	 `(set ,ltype ,@(mapcar 'sck-lirgen-expr args)))))
      ;; &x, decay
      ((addr decay)
       (let ((alexp (sck-lirgen-expr (car args))))
	 (assert (eq (car alexp) 'mem))
	 (caddr alexp)))
      ;; *x
      (mem
       `(mem ,ltype ,(sck-lirgen-expr (car args))))
      ;; x.y
      (member
       (let ((alexp (sck-lirgen-expr (car args)))
	     (ident (sck-ast-expr-value (cadr args))))
	 (assert (eq (car alexp) 'mem))
	 (assert (sck-ast-ident-p ident))
	 `(mem ,ltype (add ,sck=cps-lirgen-pointer-ltype
			   ,(caddr alexp)
			   ,(sck-lirgen-make-const
			     sck=cps-lirgen-pointer-ltype
			     (sck-ast-ident-value ident))))))
      ;; ֐Ăяo
      (call
       (let ((largs (mapcar 'sck-lirgen-expr args)))
	 (copy-list
	  `(call ,(car largs) ,(cdr largs) nil)))))))

(defun sck-lirgen-const (expr)
  (let* ((type (sck-ast-expr-type expr))
	 (ltype (sck-ast-ltypeof type))
	 (value (sck-ast-expr-value expr)))
    (cond
     ((sck-ast-scalar-type-p type)
      (sck-lirgen-make-const ltype value))
     (t
      ;; string 萔͕\Ȃ͂B
      (assert nil)))))

(defun sck-lirgen-make-const (ltype value)
  (ecase (sck-ast-kindof-ltype ltype)
    (integral `(iconst ,ltype
		       ;; iconst  signed ŕ\B
		       ,(bn-try-to-elnum
			 (bn-signed value
				    (* sck=ast-byte-in-bits
				       (sck-ast-sizeof-ltype ltype))))))
    (floating `(fconst ,ltype ,(bn-try-to-elnum value)))))
     

(defun sck-lirgen-vref (expr)
  (let* ((type (sck-ast-expr-type expr))
	 (ltype (sck-ast-ltypeof type))
	 (ident (sck-ast-expr-value expr))
	 (name (sck-ast-ident-name ident))
	 (sc (sck-ast-ident-sc ident)))
    (ecase sc
      (register
       `(reg ,ltype ,name))
      (auto
       `(mem ,ltype (frame ,sck=cps-lirgen-pointer-ltype ,name)))
      ((static extern nil)
       `(mem ,ltype (static ,sck=cps-lirgen-pointer-ltype ,name))) 
      (t
       (assert nil)))))

(defun sck-lirgen-conv (expr)
  ;; convsx, convzx, convit : integral->integral
  ;; convfx, convft         : floating->floating
  ;; convsf, convuf         : integral->floating
  ;; convfi                 : floating->integral
  (let* ((args (sck-ast-expr-args expr))
	 (largs (mapcar 'sck-lirgen-expr args))
	 (type (sck-ast-expr-type expr))
	 (ltype (sck-ast-ltypeof type))
	 (atype (sck-ast-expr-type (car args)))
	 (altype (sck-ast-ltypeof atype)))
    (if (eq ltype altype)
	;; ltype Ȃ̂܂
	(car largs)
      ;; {ɕϊKvȏꍇ
      `(,(cond
	  ;; integral->integral
	  ((and (sck-ast-integral-scalar-type-p atype)
		(sck-ast-integral-scalar-type-p type))
	   (if (< (sck-ast-sizeof-ltype altype)
		  (sck-ast-sizeof-ltype ltype))
	       (if (sck-ast-signed-scalar-type-p atype)
		   'convsx
		 'convzx)
	     'convit))
	  ;; floating->floating
	  ((and (sck-ast-floating-type-p atype)
		(sck-ast-floating-type-p type))
	   (if (< (sck-ast-sizeof-ltype altype)
		  (sck-ast-sizeof-ltype ltype))
	       'convfx
	     'convft))
	  ;; integral->floating
	  ((and (sck-ast-integral-type-p atype)
		(sck-ast-floating-type-p type))
	   (if (sck-ast-signed-scalar-type-p atype)
	       'convsf
	     'convuf))
	  ;; floating->integral
	  ((and (sck-ast-floating-type-p atype)
		(sck-ast-integral-type-p type))
	   'convfi)
	  ;;
	  (t
	   (assert nil)))
	;;
	,ltype
	;;
	,@largs))))


(defun sck-lirgen-pointer-addsub (expr)
  (let* ((op (sck-ast-expr-op expr))
	 (type (sck-ast-expr-type expr))
	 (args (sck-ast-expr-args expr))
	 (atypes (mapcar 'sck-ast-expr-type args)))
    (cond
     ;; pointer +/- integer
     ((and (sck-ast-pointer-type-p (car atypes))
	   (sck-ast-integral-type-p (cadr atypes)))
      (let ((ltype (sck-ast-ltypeof type))
	    (largs (mapcar 'sck-lirgen-expr args))
	    (otype (cadr (sck-ast-skip-typedef-and-qual (car atypes)))))
	;; pointer +/- integer*sizeof(object)
	`(,op ,ltype
	      ,(car largs)
	      (mul ,ltype
		   ,(cadr largs)
		   ,(sck-lirgen-make-const ltype
					       (sck-ast-sizeof otype))))))
     ;; integer + pointer
     ((and (sck-ast-integral-type-p (car atypes))
	   (sck-ast-pointer-type-p (cadr atypes)))
      ;; pointer + integer ɏCďB
      (setf (sck-ast-expr-args expr) (reverse args))
      (sck-lirgen-pointer-addsub expr))
     ;; pointer - pointer
     ((and (eq op 'sub)
	   (and (sck-ast-pointer-type-p (car atypes))
		(sck-ast-pointer-type-p (cadr atypes))))
      (let ((ltype (sck-ast-ltypeof type))
	    (largs (mapcar 'sck-lirgen-expr args))
	    (otype (cadr (sck-ast-skip-typedef-and-qual (car atypes)))))
	`(divs ,ltype
	       (sub ,ltype ,(car largs) ,(cadr largs))
	       ,(sck-lirgen-make-const ltype (sck-ast-sizeof otype)))))
     ;;
     (t
      (assert nil)))))

;;;
(provide 'sck-lirgen)
