;;;
;;; $Id: sck-loop.el,v 1.9 2007/03/15 09:37:49 heavy Exp $
;;;
;;; <sck-loop-input> ::= ((bblocks ((<Label> (<Succ>..))..)))
;;;
;;; <Label> ::= <Symbol>
;;; <Succ> ::= <Label>
;;;
;;; <sck-loop-output> ::= ((loops (<Loop>..))
;;;			   (bblocks ((<Label> (Succ)..))))
;;;
;;; <Loop> ::= (<PriHeader> (<Label>|<Loop>).. )
;;; <PriHeader> ::= <Label>
;;;
(require 'sck-dom)

(defconst sck=loop-sample-input-old
  '((bblocks (("B1" ("B2"))
	      ("B2" ("B3" "B4"))
	      ("B3" ("B1"))
	      ("B4" ("B1"))))))

(defconst sck=loop-sample-input-o
  '((bblocks (("B1" ("B2"))
	      ("B2" ("B3"))
	      ("B3" ("B1" "C1"))
	      ("C1" ("C2"))
	      ("C2" ("C3"))
	      ("C3" ("C1"))))))

(defconst sck=loop-sample-input
  '((bblocks (("D1" ("A1"))
	      ("A1" ("B1"))
	      ("B1" ("B2"))
	      ("B2" ("B3" "B4"))
	      ("B3" ("B1"))
	      ("B4" ("B1" "A2"))
	      ("A2" ("A1" "C1"))
	      ("C1" ("C2"))
	      ("C2" ("C3"))
	      ("C3" ("C1" "D2"))
	      ("D2" ("D3"))
	      ("D3")
	      ))))

;;;
(defstruct sck-loop-bblock
  label
  succ
  pred
  dom)

;;;
;;;
;;;
(defun sck-loop-demo ()
  "sck-loop $B$r%G%b$9$k(B"
  (interactive)
  (let ((input sck=loop-sample-input)
	(buf (get-buffer-create "*sck-loop*")))
    (save-excursion
      (set-buffer buf)
      (erase-buffer)
      (insertf "Input:\n%s\nOutput:\n%s"
	       (pp-to-string input)
	       (pp-to-string (sck-loop input)))
      (display-buffer buf))))

;;;
;;;
;;;
(defun sck-loop (input)
  '(setq LLL input)
  (sck-letalist input
      ((bblocks input-bblocks))
    (let ((bblocks (mapcar #'(lambda (b)
			       (make-sck-loop-bblock
				:label (car b)))
			   input-bblocks))
	  (init-dom (sck-dom input))
	  loops
	  (sck*loop-label-count 0)
	  sck*loop-labels)
      (sck-letalist init-dom
	  ((dominators input-dom))
	;; init succ,pred,dom,labels
	(dolist (b bblocks)
	  (push (sck-loop-bblock-label b) sck*loop-labels)
	  '(message "sck*loop-labels %s" sck*loop-labels)

	  (setf (sck-loop-bblock-succ b)
		(mapcar #'(lambda (succlabel)
			    (let ((bs (find succlabel bblocks
					    :key 'sck-loop-bblock-label
					    :test 'equal)))
			      (push b (sck-loop-bblock-pred bs))
			      bs))
			(cadr (assoc (sck-loop-bblock-label b) input-bblocks))))
	  (setf (sck-loop-bblock-dom b)
		(mapcar #'(lambda (domlabel)
			    (find domlabel bblocks
				  :key 'sck-loop-bblock-label))
			(cadr (assoc (sck-loop-bblock-label b) input-dom)))))
	;; $B%k!<%W$r:n$k(B
	(dolist (b bblocks)
	  '(message "b (%s (succ %s) (dom %s)"
		   (sck-loop-bblock-label b)
		   (mapcar 'sck-loop-bblock-label (sck-loop-bblock-succ b))
		   (mapcar 'sck-loop-bblock-label (sck-loop-bblock-dom b)))

	  (dolist (succ (sck-loop-bblock-succ b))
	    ;; succ $B$,(B dom $B$K4^$^$l$F$$$l$P%k!<%W(B
	    '(message "succ %s (dom %s)"
		      (sck-loop-bblock-label succ)
		      (mapcar 'sck-loop-bblock-label (sck-loop-bblock-dom b)))

	    (when (member succ (sck-loop-bblock-dom b))
	      '(message "loop in (D %s) (B %s)"
		       (sck-loop-bblock-label succ)
		       (sck-loop-bblock-label b))

	      ;; $B%k!<%W$NMWAG$rA4$F=8$a$k(B
	      (push (cons succ 
			  (cons b (sck-loop-make-loop-bblocks succ b)))
		    loops))))


	;;(message "make loop OK\n")
	;;(dolist (x loops)
	;;(message "%S\n" (mapcar 'sck-loop-bblock-label x)))

	;; $B%"%&%H%W%C%H%G!<%?$r:n$k(B
	(list 
	 (list 'loops
	       (sck-loop-make-output-loops
		(sck-loop-make-loop-tree 
		 (sck-loop-make-priheader 
		  (sck-loop-cleanup-non-natural-loop loops bblocks) bblocks))))
	 (list 'bblocks
	       (sck-loop-make-output-bblocks bblocks)))))))

;;
(defun sck-loop-make-output-loops (l)
  (let (output)
    (dolist (x l)
      (push (if (sck-loop-bblock-p x)
		(sck-loop-bblock-label x)
	      (sck-loop-make-output-loops x)) output))
    (nreverse output)))

(defun sck-loop-make-output-bblocks (bblocks)
  (let (output)
    (dolist (b bblocks)
      (push 
       (list (sck-loop-bblock-label b)
	     (mapcar #'sck-loop-bblock-label (sck-loop-bblock-succ b)))
       output))
    (nreverse output)))

;;
(defun sck-loop-make-loop-tree (loops)
  ;; $B%k!<%W%D%j!<$r:n$k(B
  '(message "sck-loop-make-loop-tree loops %S" 
	    (sck-loop-make-output-loops loops))

  (let (newloops)
    ;; $BC;$$%j%9%H$+$i?F;R$r8+$D$1$k(B
    (setq loops (sort loops #'(lambda (x y) (< (length x) (length y)))))
    '(message "sort loops \n%S\n" 
	      (sck-loop-make-output-loops loops))

    (while loops
      (let ((c (car loops))
	    (l (cdr loops)))
	(setq loops l)
	'(message "c %S\nl %S\n" 
		  (sck-loop-make-output-loops c)
		  (sck-loop-make-output-loops l))
	
	(block 'loop
	  (while l
	    ;; Head $B$,4^$^$l$k$J$i(B ($B?F(Bloop $B;R(Bloop) $B$N7A$G?F;R4X78$r:n$k(B
	    (when (member
		   ;; $BBP>]$,?F$N>l9g$,$"$k$N$G!D(B
		   (cadr (sck-loop-make-loop-tree/top c))
		   (sck-loop-make-loop-tree/top (car l)))
	      (setf (car l) (list (car l) c))
	      (return-from 'loop))
	    (setq l (cdr l)))
	  ;; $B$I$3$H$b?F;R$N4X78$,L5$1$l$P(B
	  (push c newloops))
	;; block loop end
	))

    '(message "newloops %S\n" (sck-loop-make-output-loops newloops))

    ;; $B?F;R4X78$+$i=EJ#$7$?%i%Y%k$r=|$/(B
    (let (output)
      (dolist (x newloops)
	(push (sck-loop-make-loop-tree/clean x) output))
      output)))

(defun sck-loop-make-loop-tree/top (loop)
  ;; $B?F%k!<%W$rJV$9(B
  (if (listp (car loop))
      (sck-loop-make-loop-tree/top (car loop))
    loop))

(defun sck-loop-make-loop-tree/clean (loop)
  ;; $B?F;R4X78$+$i=EJ#$7$?%i%Y%k$r=|$/(B
  '(message "sck-loop-make-loop-tree/clean loops %S\n"
	   (sck-loop-make-output-loops loops))

  (if (listp (car loop))
      (let* ((mainloop (if (listp (caar loop))
			   (sck-loop-make-loop-tree/clean (car loop))
			 (car loop)))
	     (subloops (cadr loop))
	     (count 0)
	     point
	     output)

	'(message "mainloop %S\nsubloops %S\n"
		  (sck-loop-make-output-loops mainloop)
		  (sck-loop-make-output-loops subloops))
	

	(dolist (x mainloop)
	  (if (member x (if (listp (car subloops)) (car subloops) subloops))
	      (unless point (setq point count))
	    (push x output)
	    (incf count)))
	
	'(message "output %S point %s\n"
		  (sck-loop-make-output-loops output)
		  point)

	(setq output (nreverse output))
	(setf (nthcdr point output)
	      (cons (sck-loop-make-loop-tree/clean subloops) 
		    (nthcdr point output)))
	output)
    loop))

;;
(defun sck-loop-make-priheader (loops bblocks)
  ;; $B%W%j%X%C%@$r:n$k(B
  (let (newloops)
    (dolist (x loops)
      (let ((priheader (make-sck-loop-bblock
			:label (sck-loop-make-label))))
	;; bblocks $B$KF~$l$k(B
	(setf (cdr bblocks) (cons priheader (cdr bblocks)))
	
	(setf (sck-loop-bblock-succ priheader) (list (car x)))
	(setf (sck-loop-bblock-pred priheader) (sck-loop-bblock-pred (car x)))
	;; $B%X%C%I$N(B pred $B$+$i$O(B tail $B$OH4$/(B
	(setf (sck-loop-bblock-pred priheader)
	      (delq (cadr x) (sck-loop-bblock-pred priheader)))
	(dolist (y (sck-loop-bblock-pred priheader))
	  (push priheader (sck-loop-bblock-succ y))
	  (setf (sck-loop-bblock-succ y)
		(delq (car x) (sck-loop-bblock-succ y))))
	(setf (sck-loop-bblock-pred (car x)) (list priheader))
	;; $B%X%C%@IU$N%k!<%W$r:n$k(B
	(push (cons priheader x) newloops)))
    newloops))

;;
(defun sck-loop-make-loop-bblocks (d b)
  ;; (C | C $B$+$i(B d $B$rDL$i$:$K(B b $B$KC#$9$kO)$,$"$k(B)$B$r=8$a$k(B
  (sck-loop-make-loop-bblocks-sub d b b nil))

(defun sck-loop-make-loop-bblocks-sub (d b c l)

  '(message "sck-loop-make-loop-bblocks-sub\nd %S\nb %S\nc %S\nl %S\n"
	    (sck-loop-bblock-label d)
	    (sck-loop-bblock-label b)
	    (sck-loop-bblock-label c)
	    (mapcar 'sck-loop-bblock-label l))

  (dolist (cc (sck-loop-bblock-pred c))
    (unless (or (eq cc d) 
		(eq cc b) 
		(memq cc l))
      
      '(message "sck-loop-make-loop-bblocks-sub \ncc %S \nl %S"
		(sck-loop-bblock-label cc)
		(mapcar 'sck-loop-bblock-label l))

      (push cc l)
      (setq l (sck-loop-make-loop-bblocks-sub d b cc l))))
  l)

;;
(defun sck-loop-cleanup-non-natural-loop (loops bblocks)
  ;; Non-Natural-Loop $B$rL5$/$9(B
  ;; $B%^!<%8$5$l$?%k!<%W$K$O?7$?$J%V%m%C%/$,8e$m$K:n$i$l$k(B
  (let (newloops)
    (while loops
      '(message "loops len %d" (length loops))

      (let (hs newhs)
	;; $B$^$:F1$8%X%C%I$r;}$D$b$N$r=8$a$k(B
	(dolist (b (mapcar #'(lambda (x) 
			       (when (equal (caar loops) (car x)) x))
			   loops))
	  (when b 
	    '(message "(car b) %s" (sck-loop-bblock-label (car b)))

	    (push b hs)
	    ;; $B=8$a$?$b$N$r85$+$i8:$i$9(B
	    (setq loops (delq b loops))))
	;; $B=8$a$?$b$N$,(B natural-loop $B$+D4$Y$F$=$&$G$J$1$l$P$R$H$D$K$^$H$a$k(B
	;; Feb. 13, 2007 $B$R$H$D$N>l9g$O$=$N$^$^EPO?(B
	(if (cdr hs)
	    (dolist (b hs)
	      (if (sck-loop-cleanup-non-natural-loop/subetps-p b hs)
		  (push b newloops)
		(push b newhs)))
	  (push (car hs) newloops))
	
	'(message "newhs %S" 
		  (mapcar #'(lambda (x)
			      (mapcar 'sck-loop-bblock-label x))
			  newhs))

	;; Non-Natural-Loop $B$r%^!<%8$9$k(B
	(when newhs
	  ;; $B?7$?$J%V%m%C%/$r:n$C$F(B succ,pred $B$r@_Dj(B
	  (let ((tail (make-sck-loop-bblock
		       :label (sck-loop-make-label))))
	    ;; bblocks $B$KF~$l$k(B
	    (setf (cdr bblocks) (cons tail (cdr bblocks)))

	    (setf (sck-loop-bblock-succ tail) (list (caar newhs)))
	    (push tail (sck-loop-bblock-pred (caar newhs)))
	    (dolist (x newhs)
	      (push x (sck-loop-bblock-pred tail))
	      (setf (sck-loop-bblock-pred (car x))
		    (delq (cadr x) (sck-loop-bblock-pred (car x))))
	      (push tail (sck-loop-bblock-succ (cadr x)))
	      (setf (sck-loop-bblock-succ (cadr x))
		    (delq (car x) (sck-loop-bblock-succ (cadr x)))))
	    ;; $B%^!<%8$7$?%k!<%W$r:n$k(B
	    (push 
	     (cons (caar newhs) 
		   (cons tail (foldr 'union nil (mapcar 'cdr newhs))))
	     newloops)))))
    newloops))

(defun sck-loop-cleanup-non-natural-loop/subetps-p (b hs)
  ;; b $B$,(B hs $B$NA4$F$KBP$7$F$I$A$i$+$,(B subset $B$N4X78$K$"$k$+!)(B
  (block 'test
    (dolist (x hs)
      '(message "\nx %S\nb %S\n"
	       (mapcar 'sck-loop-bblock-label x)
	       (mapcar 'sck-loop-bblock-label b))
      (when
	  (and
	   (if (> (length x) (length b))
	       (subsetp b x) 
	     (subsetp x b))
	   ;; Feb. 13, 2007 $BCf?H$,F1$8>l9g$O%^!<%8$5$l$k$Y$-(B
	   (not (eq (length x) (length b))))
	(return-from 'test t)))
    nil))

;;
(defun sck-loop-make-label ()
  ;; $B=E$J$i$J$$%i%Y%k$r:n$k(B
  (let ((label (format "LLoop%d" (incf sck*loop-label-count))))
    (while (member label sck*loop-labels)
      (setq label (format "LLoop%d" (incf sck*loop-label-count))))
    (push label sck*loop-labels)
    '(message "make-label label (%S) labels %S"
	      label
	      sck*loop-labels)
    label))

(provide 'sck-loop)
