;;;
;;; $Id: sck-regalloc.el,v 1.22 2007/03/29 09:04:27 heavy Exp $
;;;
;;; OtgWX^蓖
;;;
;;; <Input-to-regalloc> ::=
;;; ((interference-graph         ((<VirtualReg> <VirtualReg>)..))
;;;  (coalesceable-edges         ((<VirtualReg> <VirtualReg>)..))
;;;  (real-register-sets         ((<RealRegSet> (<RealReg>..))..))
;;;  (real-register-constraints  ((<RealReg> <RealReg>)..))
;;;  (possible-assignments       ((<VirtualReg> <RealRegSet>)..))
;;;  (spill-priority		((<VirtualReg> <Number>)..))
;;;  )
;;;
;;; <VirtualReg> ::= <List>
;;; <RealReg> ::= <String>
;;; <RealRegSet> ::= <Symbol>
;;;
;;; <Output-from-regalloc> ::=
;;; ((register-assignments ((<VirtualReg> <RealReg>)..))
;;;  (spilled-registers    (<VirtualReg>..)) )
;;;
(require 'graphviz)

;;
;; graphviz init attribute
;;
(defconst sck=regalloc-graphviz-simplify-add-attr 
  '((color "white") (fontcolor "white")))
(defconst sck=regalloc-graphviz-coalesce-add-attr
  '((color "blue") (fontcolor "blue") (label "alias")))
(defconst sck=regalloc-graphviz-filename "IG")

;;
;; f[^\`ƃANZX֐
;;
(defstruct sck-regalloc-interference-node
  (name)
  (reg)
  (edge)
  (coalesce)
  (alias)
  (id)
  (regset)
  (maxregs)
  (priority)
  ;; debug flags
  (spill)
  (freeze))

(defvar sck*regalloc-node-id-table)

;;
;; bitvector Ή
;;
(defun sck-regalloc-make-node-id-table (nl)
  "m[h̃Xg nl e[uAm[h id ݒ"
  (setq sck*regalloc-node-id-table (vconcat nl))
  (let ((n 0))
    (dolist (x nl)
      (setf (sck-regalloc-interference-node-id x) n)
      (incf n))))

(defun sck-regalloc-edge-list (x)
  "x ̃GbW̃XgԂ"
  (sck-regalloc-interference-node-edge x))

(defun sck-regalloc-check-edge-p (x y)
  "x ̃GbW y ܂܂邩H"
  (memq y (sck-regalloc-edge-list x)))

(defun sck-regalloc-set-edge (x y)
  "x ̃GbW y ǉ"
  (unless (sck-regalloc-check-edge-p x y)
    (setf
     (sck-regalloc-interference-node-edge x)
     (cons y (sck-regalloc-interference-node-edge x)))))

(defun sck-regalloc-del-edge (x y)
  "x ̃GbW y 𔲂"
  (setf (sck-regalloc-interference-node-edge x) 
	(delq y (sck-regalloc-interference-node-edge x))))

(defun sck-regalloc-edge-count (x)
  "GbW̌Ԃ"
  (length (sck-regalloc-edge-list x)))

;;
;;
;;
(defun sck-regalloc-make-interference-edge (x y)
  (assert (and x y))
  ;; xy ԂɃGbW
  (sck-regalloc-set-edge x y)
  (sck-regalloc-set-edge y x))

(defun sck-regalloc-del-interference-node (x)
  ;; x Ot甲
  ;; x ̂̃GbW͂̂܂
  (dolist (y (sck-regalloc-edge-list x))
    (sck-regalloc-del-edge y x)))

(defun sck-regalloc-make-alias-interference-node (x y regset maxregs)
  ;; x  y ̃GCAXΏۂɂ
  ;; x  y ̃GbWƃGCAX,coalesce}[W
  ;; x  coalesce  x ̃GCAX폜
  ;; y Ot폜
  ;; x ̃WX^Zbg֌Wݒ肵Ȃ
  (sck-regalloc-set-interference-alias x y)
  ;; edge
  (dolist (z (sck-regalloc-edge-list y))
    (sck-regalloc-make-interference-edge x z))
  ;; alias
  (dolist (z (sck-regalloc-interference-node-alias y))
    (sck-regalloc-set-interference-alias x z))
  (dolist (z (sck-regalloc-interference-node-coalesce y))
    (unless (eq x z)
      (sck-regalloc-set-interference-coalesce x z)))
  ;;
  (setf
   (sck-regalloc-interference-node-coalesce x)
   (let ((coalesce (sck-regalloc-interference-node-coalesce x)))
     (dolist (z (sck-regalloc-interference-node-alias x))
       (setq coalesce (delq z coalesce)))
     coalesce))

  (sck-regalloc-del-interference-node y)
  (setf (sck-regalloc-interference-node-regset x) regset)
  (setf (sck-regalloc-interference-node-maxregs x) maxregs))

(defun sck-regalloc-set-interference-coalesce (x y)
  ;; xy Ԃcoalecse
  (unless (memq y (sck-regalloc-interference-node-coalesce x))
    (setf
     (sck-regalloc-interference-node-coalesce x)
     (cons y (sck-regalloc-interference-node-coalesce x))))
  (unless (memq x (sck-regalloc-interference-node-coalesce y))
    (setf
     (sck-regalloc-interference-node-coalesce y)
     (cons x (sck-regalloc-interference-node-coalesce y)))))

(defun sck-regalloc-set-interference-alias (x y)
  ;; x  alias  y 
  (unless (memq y (sck-regalloc-interference-node-alias x))
    (setf
     (sck-regalloc-interference-node-alias x)
     (cons y (sck-regalloc-interference-node-alias x)))))

(defun sck-regalloc-get-list-all-edges (list)
  '(message "sck-regalloc-get-list-all-edges\nlist %s\n" list)
  (cond 
   ((null (cdr list))
    nil)
   (t
    (let (edges)
      (dolist (x (cdr list))
	(push 
	 (if (string< (format "%S"
			      (sck-regalloc-interference-node-name (car list)))
		      (format "%S"
			      (sck-regalloc-interference-node-name x)))
	     (list (sck-regalloc-interference-node-name (car list))
		   (sck-regalloc-interference-node-name x))
	   (list (sck-regalloc-interference-node-name x)
		 (sck-regalloc-interference-node-name (car list))))
	 edges))
      '(message "edges %S" edges)
      (append edges
	      (sck-regalloc-get-list-all-edges (cdr list)))))))

;;
;; 
;;
(defun sck-regalloc-print-interference-node (x &optional debug buf)
  ;; m[h\֐
  (let ((s 
	 (format 
	  (if debug
	      "%s: (R %s) (E %s) (C %s) (A %s) (RS %s) (MR %s) d:%s\n"
	    "%s: (R %s) (E %s) (C %s) (A %s) (RS %s) (MR %s)\n"
	    )
	  (sck-regalloc-interference-node-name x)
	  (sck-regalloc-interference-node-reg x)
	  (let (z)
	    (dolist (y (sck-regalloc-edge-list x))
	      (push (sck-regalloc-interference-node-name y) z)) z)
	  (let (z)
	    (dolist (y (sck-regalloc-interference-node-coalesce x))
	      (push (sck-regalloc-interference-node-name y) z)) z)
	  (let (z)
	    (dolist (y (sck-regalloc-interference-node-alias x))
	      (push (sck-regalloc-interference-node-name y) z)) z)
	  (sck-regalloc-interference-node-regset x)
	  (sck-regalloc-interference-node-maxregs x)
	  (when debug
	    (format "(s %s) (f %s)"
		    (sck-regalloc-interference-node-spill x)
		    (sck-regalloc-interference-node-freeze x)))
	  )))
    (if buf
	(progn
	  (set-buffer buf)
	  (insert s))
      (princ s))))

(defun sck-regalloc-print-interference-graph (glist &optional debug buf)
  (dolist (x glist)
    (if (bufferp buf)
	(sck-regalloc-print-interference-node x debug buf)
      )))

;;
;; (defun sck-regalloc-briggs-coalesce-p (x y z) nil)
;;
(defun sck-regalloc-briggs-coalesce-p (x y z)
  "z ̃WX^ɑ΂ x y  coalesce Ă悢H"
  (let ((n 0)
	(l (sck-regalloc-edge-list x)))
    (dolist (u (sck-regalloc-edge-list y))
      (unless (memq u l)
	(push u l)))
    (dolist (u l)
      (when (>= (sck-regalloc-edge-count u) z)
	(incf n)))
    (< n z)))

;;
;; (defun sck-regalloc-george-coalesce-p (x y z) nil)
;;
(defun sck-regalloc-george-coalesce-p (x y z)
  "z ̃WX^ɑ΂ x y  coalesce Ă悢H"
  (block geroge
    (dolist (u (sck-regalloc-edge-list x))
      (unless
	  (or (sck-regalloc-check-edge-p u y)
	      (< (sck-regalloc-edge-count u) z))
	(return-from geroge)))
    t))

;;
;;
;;
(defun sck-regalloc-graphviz-coalesce-edges (graph coalesce)
  (let (ret)
    (dolist (edge 
	     (mapcar #'(lambda (x)
			 (list 
			  (format "%S" (car x))
			  (format "%S" (cadr x))))
		     (sck-regalloc-get-list-all-edges coalesce)))
      (dolist (edge-list (cddr graph))
	(when (eq (car edge-list) 'edge-list)
	  '(message "\nedge %S edge-list %S\n" edge edge-list)
	  (when (assoc edge (cdr edge-list))
	    (push edge ret)))))
    ret))

;;
;; FƃWX^蓖
;;
(defun sck-regalloc-colorling-by-symplefication (nodelist
						 &optional buf graphviz)
  ;;
  (let (selectstack
	simplify-worklist 
	coalesce-worklist
	spill-worklist
	(graphviz-number (when graphviz 1)))
    ;; make worklist
    (dolist (x nodelist)
      (cond 
       ((>= (sck-regalloc-edge-count x) 
	    (sck-regalloc-interference-node-maxregs x))
	(push x spill-worklist))
       ((sck-regalloc-interference-node-coalesce x)
	(push x coalesce-worklist))
       (t
	(push x simplify-worklist))))
    ;;
    (when buf
      (insert ";; sck-regalloc-colorling-by-symplefication\n")
      (insert ";;\n;; make worklist\n;;\n")
      (insert ";; spill-worklist\n")
      (sck-regalloc-print-interference-graph spill-worklist t buf)
      (insert ";; coalesce-worklist\n")
      (sck-regalloc-print-interference-graph coalesce-worklist t buf)
      (insert ";; simplify-worklist\n")
      (sck-regalloc-print-interference-graph simplify-worklist t buf))
    ;;
    (while (or simplify-worklist
	       coalesce-worklist
	       spill-worklist)
      (cond
       ;; simplify
       (simplify-worklist
	(dolist (x simplify-worklist)
	  (sck-regalloc-del-interference-node x)
	  (push x selectstack)
	  (setq simplify-worklist (delq x simplify-worklist))
	  ;;
	  ;; graphviz
	  ;;
	  (when graphviz
	    (graphviz-graph-set-add-attr-node-list
	     (car graphviz)
	     sck=regalloc-graphviz-simplify-add-attr
	     (cons (format "%S"
			   (sck-regalloc-interference-node-name x)) nil))
	    ;; alias Ăꍇɂ͑Ώې
	    (when (sck-regalloc-interference-node-alias x)
	      (dolist (edge (sck-regalloc-graphviz-coalesce-edges
			     (car graphviz)
			     (cons
			      x (sck-regalloc-interference-node-alias x))))
		(graphviz-graph-clear-add-attr-edge-list
		 (car graphviz)
		 sck=regalloc-graphviz-coalesce-add-attr
		 (cons edge nil)))
	      (dolist (y (sck-regalloc-interference-node-alias x))
		(graphviz-graph-set-add-attr-node-list
		 (car graphviz)
		 sck=regalloc-graphviz-simplify-add-attr
		 (cons (format "%S"
			       (sck-regalloc-interference-node-name y)) nil))))
	    (sck-regalloc-write-graphviz-file (car graphviz) 
					      (cadr graphviz)
					      graphviz-number)
	    (incf graphviz-number)))
	;; spill-worklist ړł͈̂ړ
	(dolist (x spill-worklist)
	  (when (< (sck-regalloc-edge-count x)
		   (sck-regalloc-interference-node-maxregs x))
	    (setq spill-worklist (delq x spill-worklist))
	    (if (sck-regalloc-interference-node-coalesce x)
		(push x coalesce-worklist)
	      (push x simplify-worklist)))))
       ;; coalesce
       (coalesce-worklist
	(block coalesce
	  (dolist (x coalesce-worklist)
	    ;; regset Ɖ肵Ă͂Ȃ
	    (let* ((y (car (sck-regalloc-interference-node-coalesce x)))
		   ;; VȃWX^Zbg
		   (regset
		    (if (eq (sck-regalloc-interference-node-regset x)
			    (sck-regalloc-interference-node-regset y))
			(sck-regalloc-interference-node-regset x)
		      (intersection 
		       (sck-regalloc-interference-node-regset x)
		       (sck-regalloc-interference-node-regset y)
		       :test 'equal)))
		   ;; VȃWX^Zbǧ
		   (nonewregs (length regset))
		   ;; 肷̂͑傫
		   (maxregs 
		    (max (sck-regalloc-interference-node-maxregs x)
			 (sck-regalloc-interference-node-maxregs y))))
	      (when 
		  (and regset
		       ;; Ώۂt[YĂȂ𒲂ׂ
		       (not (sck-regalloc-interference-node-freeze y))
		       ;; GbWɊ܂܂ĂȂ
		       (not (sck-regalloc-check-edge-p y x))
		       (or (sck-regalloc-briggs-coalesce-p x y maxregs)
			   ;; W[W̏ꍇΏې悪n[hWX^
			   ;; coalesce-worklist ɂ邩H
			   (and
			    (or (= (sck-regalloc-interference-node-maxregs y) 1)
				(memq y coalesce-worklist))
			    (sck-regalloc-george-coalesce-p x y maxregs))))
		(sck-regalloc-make-alias-interference-node x y regset nonewregs)
		(setq coalesce-worklist (delq y coalesce-worklist))
		(setq spill-worklist (delq y spill-worklist))
		;; coalesce-worklist ړ͈̂ړ
		(if (sck-regalloc-interference-node-coalesce x)
		    (unless (< (sck-regalloc-edge-count x)
			       (sck-regalloc-interference-node-maxregs x))
		      (setq coalesce-worklist (delq x coalesce-worklist))
		      (push x spill-worklist))
		  (setq coalesce-worklist (delq x coalesce-worklist))
		  (if (< (sck-regalloc-edge-count x)
			 (sck-regalloc-interference-node-maxregs x))
		      (push x simplify-worklist)
		    (push x spill-worklist)))
		;;
		;; graphviz
		;;
		(when graphviz
		  (dolist (edge (sck-regalloc-graphviz-coalesce-edges
				 (car graphviz)
				 (cons 
				  x (sck-regalloc-interference-node-alias x))))
		    (graphviz-graph-set-add-attr-edge-list
		     (car graphviz)
		     sck=regalloc-graphviz-coalesce-add-attr
		     (cons edge nil)))
		  (sck-regalloc-write-graphviz-file (car graphviz) 
						    (cadr graphviz)
						    graphviz-number)
		  (incf graphviz-number))
		;; spill-worklist ړ͈̂ړ
		(dolist (x spill-worklist)
		  (when (< (sck-regalloc-edge-count x)
			   (sck-regalloc-interference-node-maxregs x))
		    (setq spill-worklist (delq x spill-worklist))
		    (if (sck-regalloc-interference-node-coalesce x)
			(push x coalesce-worklist)
		      (push x simplify-worklist))))
		(return-from coalesce))))
	  ;; freeze
	  (push (car coalesce-worklist) simplify-worklist)
	  (setf (sck-regalloc-interference-node-freeze 
		 (car coalesce-worklist)) t)
	  (setq coalesce-worklist (cdr coalesce-worklist))))
       ;; potential spill
       (spill-worklist
	'(message "potential spill in")

	(let (spill-node (priority 10000000000.0))
	  (block spill
	    ;;
	    ;; Jan. 4, 2007 spill-priority Č߂
	    ;;
	    (dolist (x spill-worklist)
	      (unless (= (sck-regalloc-interference-node-maxregs x) 1)
		'(message "spill x (%s) priority (%s) node priority (%s)"
			  (sck-regalloc-interference-node-name x)
			  priority
			  (sck-regalloc-interference-node-priority x))


		(when (> priority (sck-regalloc-interference-node-priority x))
		  (setq spill-node x)
		  (setq priority (sck-regalloc-interference-node-priority x)))))
	    (when spill-node
	      (when (sck-regalloc-interference-node-coalesce spill-node)
		(setf (sck-regalloc-interference-node-freeze spill-node) t))
	      (return-from spill))
	    ;; ŏIIɃn[hWX^Xs
	    (setq spill-node (car spill-worklist))
	    )				; out spill block
	  '(message "spill-node (%s) priority (%s)" 
		    (sck-regalloc-interference-node-name spill-node)
		    priority)

	  (setf (sck-regalloc-interference-node-spill spill-node) t)
	  (push spill-node simplify-worklist)
	  (setq spill-worklist (delq spill-node spill-worklist))
	  ))
       )

      (when buf
	(insert ";;\n;; worklist loop\n;;\n")
	(insert ";; spill-worklist\n")
	(sck-regalloc-print-interference-graph spill-worklist t buf)
	(insert ";; coalesce-worklist\n")
	(sck-regalloc-print-interference-graph coalesce-worklist t buf)
	(insert ";; simplify-worklist\n")
	(sck-regalloc-print-interference-graph simplify-worklist t buf)))
    ;; return
    (when buf
      (insert ";;\n;; after select\n;;\n;; selectstack\n")
      (sck-regalloc-print-interference-graph selectstack t buf))
    selectstack
    ))

;;
(defun sck-regalloc-write-graphviz-file (graph dir number)
  (let ((filename
	 (expand-file-name (format "%s%04d.gif" 
				   sck=regalloc-graphviz-filename
				   number)
			   dir)))
    (with-temp-buffer 
      (insert (graphviz-graph-to-dot graph))
      (call-process-region (point-min)
			   (point-max)
			   "dot"
			   nil
			   nil
			   nil
			   "-Tgif" 
			   (format "-o%s" filename)))))

;;
;;
;;
(defun sck-regalloc-get-regconst (reg regset regconst)
  ;; reg ɃRXgCgĂ regset ̃WX^XgɂĕԂ

  '(message "sck-regalloc-get-regconst\nreg %S\nregset\n%s\nregconst\n%s"
	   reg (pp-to-string regset) (pp-to-string regconst))

  (let (reglist)
    (dolist (x regconst)
      (cond 
       ((equal reg (car x))
	(when (member (cadr x) regset)
	  (push (cadr x) reglist)))
       ((equal reg (cadr x))
	(when (member (car x) regset)
	  (push (car x) reglist)))))
    reglist))

(defun sck-regalloc-assign-color-reg (selectstack regconst)
  ;; m[hɃWX^蓖Ă
  ;; Xsꂽm[hŊ蓖ĂȂꍇ nil 
  (dolist (node selectstack)

    '(message "Node %s\n" 
	      (with-temp-buffer
		(sck-regalloc-print-interference-node node t (current-buffer))
		(buffer-string)))
    
    (let ((vec (make-bool-vector 
		(sck-regalloc-interference-node-maxregs node) nil))
	  (n 0))
      ;; ꂽm[hŎgpĂ郌WX^𒲂ׂ
      (dolist (x (sck-regalloc-edge-list node))
	;; RXgCgĂꍇɂȂꍇ̂list
	(let ((reglist (cons (sck-regalloc-interference-node-reg x) nil)))
	  (when (and
		 (car reglist)
		 (or
		  ;; WX^Zbg
		  (eq (sck-regalloc-interference-node-regset x)
		      (sck-regalloc-interference-node-regset node))
		  ;; WX^Zbg̃WX^gĂ
		  (member (car reglist) 
			  (sck-regalloc-interference-node-regset node))
		  ;; WX^ZbgɃRXgCgꂽWX^
		  (setq reglist 
			(sck-regalloc-get-regconst 
			 (car reglist) 
			 (sck-regalloc-interference-node-regset node)
			 regconst))))
	    ;; gpĂ郌WX^̃tO𗧂Ă
	    '(message "--- reglist ( %s) maxregs ( %d) regset ( %s) ----"
		      reglist 
		      (sck-regalloc-interference-node-maxregs node)
		      (sck-regalloc-interference-node-regset node))
	    
	    (dolist (y reglist)
	      (aset vec 
		    (- (sck-regalloc-interference-node-maxregs node)
		       (length
			(member 
			 y (sck-regalloc-interference-node-regset node))))
		    t)))))

      '(message "vec %s" vec)

      ;; tO𒲂ׂċ󂢂Ă郌WX^ݒ
      (block check-flag
	(while (< n (sck-regalloc-interference-node-maxregs node))

	  '(message "vec(%d) %s" n (aref vec n))

	  (unless (aref vec n) 
	    (setf (sck-regalloc-interference-node-reg node)
		  (nth n (sck-regalloc-interference-node-regset node)))
	    (return-from check-flag))
	  (incf n))
	(setf (sck-regalloc-interference-node-reg node) nil))
      ;; GCAXݒ
      (when (sck-regalloc-interference-node-alias node)
	(dolist (x (sck-regalloc-interference-node-alias node))
	  (unless (sck-regalloc-interference-node-reg x)
	    (setf (sck-regalloc-interference-node-reg x) 
		  (sck-regalloc-interference-node-reg node)))))
      )))

;;
;;
;;
(defun sck-regalloc-init-get-node (name nl)
  (find name nl
	:key 'sck-regalloc-interference-node-name
	:test 'equal))

;;
;;
;;
(defun sck-regalloc-graphviz-init (nodelist outdir)
  (let ((graph (graphviz-make-graph 
		'graph "IG"
		nil			;'((size "3,3"))
		'((shape "circle"))
		nil)))
    (graphviz-graph-label graph "Interference Graph" t nil)
    (dolist (x nodelist)
      (graphviz-graph-set-node-attr
       graph 
       (format "%S" (sck-regalloc-interference-node-name x))
       '((color "black")))
      (dolist (y (sck-regalloc-edge-list x))
	(graphviz-graph-set-edge 
	 graph 
	 (sort (list (format "%S"
			     (sck-regalloc-interference-node-name x))
		     (format "%S"
			     (sck-regalloc-interference-node-name y)))
	       'string<)))
      (dolist (y (sck-regalloc-interference-node-coalesce x))
	(graphviz-graph-set-edge
	 graph
	 (sort (list (format "%S"
			     (sck-regalloc-interference-node-name x))
		     (format "%S"
			     (sck-regalloc-interference-node-name y)))
	       'string<)
	 '( ;;(style "dotted")
	   (color "red")
	   (label "Coalecse")
	   (weight "2")
	   )
	 )))
    (sck-regalloc-write-graphviz-file graph outdir 0)
    (list graph outdir)))

(defun sck-regalloc-graphviz-output (nodelist graphviz)
  (let ((graph (car graphviz))
	(outdir (cadr graphviz)))
    (dolist (x nodelist)
      (graphviz-graph-rename-node 
       graph
       (format "%S" (sck-regalloc-interference-node-name x))
       (format "%S=%S" 
	       (sck-regalloc-interference-node-name x) 
	       (sck-regalloc-interference-node-reg x))))
    (graphviz-graph-clear-add-attr-node-list
     graph
     sck=regalloc-graphviz-simplify-add-attr
     nil)
    (graphviz-graph-clear-add-attr-edge-list
     graph
     sck=regalloc-graphviz-coalesce-add-attr
     nil)
    (sck-regalloc-write-graphviz-file graph outdir 1000)
    (w32-shell-execute "open" 
		       (expand-file-name
			(format "%s%04d.gif" 
				sck=regalloc-graphviz-filename 1000)
			(cadr graphviz)))))

;;
;; 
;;
(defun sck-regalloc-init-make-node (name regsets maxregs)
  (let ((node (make-sck-regalloc-interference-node)))
    (setf (sck-regalloc-interference-node-name node) name)
    (setf (sck-regalloc-interference-node-regset node) regsets)
    (setf (sck-regalloc-interference-node-maxregs node) maxregs)
    node))

(defun* sck-regalloc (input &key debug-buf graphviz-outdir)
  "
input: alist
debug-buf: fobOAEgvbgobt@
graphviz-outdir: Graphvizt@Co͂fBNg
"
  '(message "sck-regalloc\n%s" (pp-to-string input))
  
  (sck-letalist input
      ((interference-graph edges)
       (coalesceable-edges coalesces)
       (real-register-sets regsets)
       (real-register-constraints regconst)
       (possible-assignments assign)
       (spill-priority priority))
    ;;(print (format "(edges %S) (coalesces %S)\n" edges coalesces))
    (let (nodelist hardregs graphviz)
      ;; make node structure & settings
      (dolist (x assign)
	(let ((regsets (cadr (assoc (cadr x) regsets))))
	  (push (sck-regalloc-init-make-node (car x) 
					     regsets 
					     (length regsets))
		nodelist)))
      ;; n[hWX^Node
      ;; @@@ mɊ܂܂̂ priority
      (dolist (x priority)
	(unless (sck-regalloc-init-get-node (car x) nodelist)
	  (let ((node (sck-regalloc-init-make-node (car x) (list (car x)) 1)))
	    (push node hardregs)
	    (push node nodelist))))
      
      '(progn
	(message "--nodelist--")
	(dolist (x nodelist)
	  (message "n:%S" (sck-regalloc-interference-node-name x)))
	(message "--hardregs--")
	(dolist (x hardregs)
	  (message "n:%S" (sck-regalloc-interference-node-name x))))

      (sck-regalloc-make-node-id-table nodelist)
      ;; make edge
      (dolist (x edges)
	(when (cdr x)
	  (sck-regalloc-make-interference-edge
	   (sck-regalloc-init-get-node (car x) nodelist)
	   (sck-regalloc-init-get-node (cadr x) nodelist))))
      ;; n[hWX^̈ˑ֌W
      (when hardregs
	(dolist (x (sck-regalloc-get-list-all-edges hardregs))
	  (let ((n1 (sck-regalloc-init-get-node (car x) nodelist))
		(n2 (sck-regalloc-init-get-node (cadr x) nodelist)))
	    ;; WX^ZbggĂ镨assert
	    '(message "n1 regset %s\nn2 regset %s"
		      (sck-regalloc-interference-node-regset n1)
		      (sck-regalloc-interference-node-regset n2))
	    (if (equal (sck-regalloc-interference-node-regset n1)
		       (sck-regalloc-interference-node-regset n2))
		(assert nil nil "regalloc error reg %S reg %S" 
			(sck-regalloc-interference-node-name n1)
			(sck-regalloc-interference-node-name n2))
	      (sck-regalloc-make-interference-edge n1 n2)))))
      ;; priority vZ
      (dolist (x priority)
	(let ((node (sck-regalloc-init-get-node (car x) nodelist)))
	  (assert node)
	  (setf (sck-regalloc-interference-node-priority node)
		(/ (float (cadr x)) (sck-regalloc-edge-count node)))

	  '(message "node %S priority (%s)" 
		    (sck-regalloc-interference-node-name node)
		    (sck-regalloc-interference-node-priority node))
	))
      ;; make coalesce
      (dolist (x coalesces)
	(sck-regalloc-set-interference-coalesce
	 (sck-regalloc-init-get-node (car x) nodelist)
	 (sck-regalloc-init-get-node (cadr x) nodelist)))
      ;;
      ;; colorling by symplefication
      ;;
      (when graphviz-outdir
	(unless (file-directory-p graphviz-outdir)
	  (error "No such directory: %s" graphviz-outdir))
	(dolist
	    (x (file-expand-wildcards 
		(format "%s/%s*.gif" 
			graphviz-outdir
			sck=regalloc-graphviz-filename)))
	  (delete-file x))
	(setq graphviz (sck-regalloc-graphviz-init nodelist graphviz-outdir)))
      (sck-regalloc-assign-color-reg
       (sck-regalloc-colorling-by-symplefication nodelist debug-buf graphviz)
       regconst)
      (when debug-buf
	(insert ";; nodelist\n")
	(sck-regalloc-print-interference-graph nodelist t debug-buf))
      (when graphviz
	(sck-regalloc-graphviz-output nodelist graphviz))
      ;;
      ;; fomput output
      ;;
      (list 
       (list 'register-assignments
	     (let (l)
	       (dolist (x nodelist)
		 (when (sck-regalloc-interference-node-reg x)
		   (push
		    (list (copy-tree (sck-regalloc-interference-node-name x))
			  (copy-tree (sck-regalloc-interference-node-reg x)))
		    l)))
	       l))
       (list 'spilled-registers
	     (let (l)
	       (dolist (x nodelist)
		 (unless (sck-regalloc-interference-node-reg x)
		   (push (copy-tree (sck-regalloc-interference-node-name x))
			 l)))
	       l))))))

;;
;; for debug 
;; 
(defun sck-regalloc-debug (input &optional graphviz-outdir)
  (let ((buf (get-buffer-create "*colorling-by-symplefication*")))
    (save-excursion
      (set-buffer buf)
      (erase-buffer)
      (prog1
	  (sck-regalloc input :debug-buf buf :graphviz-outdir graphviz-outdir)
	(display-buffer buf)
	))))

;; test
(setq sck-regalloc-test-input
      '((interference-graph 
	 ((("a" 1) ("b" 2)) (("a" 1) ("d" 4)) (("a" 1) ("g" 7)) (("a" 1) ("f" 6))
	  (("b" 2) ("c" 3)) (("b" 2) ("d" 4)) (("b" 2) ("f" 6))
	  (("b" 2) ("g" 7)) (("c" 3) ("g" 7))
	  (("d" 4) ("g" 7))
	  (("f" 6) ("g" 7))))
	(coalesceable-edges
	 ())
	(real-register-sets 
	 ((regset1 ("R0" "R1" "R2" "R3"))))
	(real-register-constraints
	 ())
	(possible-assignments
	 ((("a" 1) regset1)
	  (("b" 2) regset1)
	  (("c" 3) regset1)
	  (("d" 4) regset1)
	  (("f" 6) regset1)
	  (("g" 7) regset1)))
	 (spill-priority
	  ((("a" 1) 1)
	   (("b" 2) 1)
	   (("c" 3) 1)
	   (("d" 4) 1)
	   (("f" 6) 1)
	   (("g" 7) 1)))
	))

(setq sck-regalloc-test-input2
      '((interference-graph
	 (("a" "b") ("a" "d") ("a" "e") ("a" "f")
	 ("b" "c") ("b" "f") ("b" "g")
	 ("c" "d") ("c" "f") ("c" "g")
	 ("d" "e") ("d" "g")
	 ("e" "f") ("e" "g")
	 ))
	(coalesceable-edges
	 (("f" "g")))
	(real-register-sets 
	 ((regset1 ("R0" "R1" "R2" "R3"))
	  (regset2 ("R1" "R2"))
	  (regset3 ("R0A" "R1A" "R2A"))
	  ))
	(real-register-constraints
	 (("R0A" "R0")
	  ("R1A" "R1")
	  ("R2A" "R2")
	  ("R3A" "R3")))
	(possible-assignments
	 (("a" regset1)
	  ("b" regset3)
	  ("c" regset1)
	  ("d" regset3)
	  ("e" regset1)
	  ("f" regset2)
	  ("g" regset1)))
	(spill-priority
	 (("a" 1)
	  ("b" 1)
	  ("c" 1)
	  ("d" 1)
	  ("e" 1)
	  ("f" 1)
	  ("g" 1)))
	 ))

;;
;;
;;
(provide 'sck-regalloc)
