;;;
;;; $Id: sck-regassign.el,v 1.2 2007/04/09 09:50:30 heavy Exp $
;;;
;;; <input> ::= ((module-defs <Defs>)
;;;		 (lfunc <Lfunc>))
;;;
;;; <output> ::= ((lfunc <Lfunc>)
;;;	          (spilled-registers (<Reg>...)))
;;; 
(require 'sck-regalloc)
(require 'sck-live)
(require 'sck-loop)
(require 'sck-lir)
(require 'sck-tmd)
;;
(require 'sck-regassign-demo)

(defvar sck*regassign-after-restruct-hook nil)
(defvar sck*regassign-after-instsel-hook nil)
(defvar sck*regassign-after-spill-hook nil)

(defvar sck*regassign-module-defs)

;;;
;;;
;;;
(defstruct sck-regassign-bblock
  label 
  insts 
  succ  
  pred  
  livein
  liveout)

(defun sck-regassign-format-bblock (bblock-s)
  ;; bblock $B$^$?$O$=$N%j%9%H$rJ8;zNs$KJQ49$9$k!#(B
  (with-temp-buffer
    (cond
     ((sck-regassign-bblock-p bblock-s)
      (insertf "%S ; %S->%S ; livein=%s liveout=%s\n"
	       (sck-regassign-bblock-label bblock-s)
	       (mapcar 'sck-regassign-bblock-label
		       (sck-regassign-bblock-pred bblock-s))
	       (mapcar 'sck-regassign-bblock-label
		       (sck-regassign-bblock-succ bblock-s))
	       (sck-regassign-bblock-livein bblock-s)
	       (sck-regassign-bblock-liveout bblock-s))
      (dolist (inst (sck-regassign-bblock-insts bblock-s))
	(if (and (eq (car inst) 'set)
		 (equal (caddr inst) (cadddr inst)))
	    (insertf "  ;; %S\n" inst)
	  (insertf "  %S\n" inst))
	))
     (t
      (dolist (b bblock-s)
	(insertf "%s" (sck-regassign-format-bblock b)))))
    (buffer-string)))

;;;
;;;
;;;
(defun sck-regassign (input &optional graphviz-outdir)
  ;; <input> ::= ((module-defs <Defs>)
  ;;		 (lfunc <Lfunc>))
  ;; <output> ::= ((lfunc <Lfunc>)
  ;;	          (spilled-registers (<Reg>...)))
  (sck-letalist input
      ((module-defs defs)
       (lfunc func))
    ;; $BGK2u$9$k$?$a(B
    (setq func (copy-tree func))
    (let* ((sck*regassign-module-defs defs)
	   (bblocks (sck-regassign-make-bblocks func))
	   (symtab (cdr (assq 'symtab func))))
      '(message "%s" (sck-lir-pp defs))
      '(message "symtab %s" (sck-lir-pp symtab))
      (sck-letalist 
	  (sck-regalloc
	   (sck-regassign-make-regalloc-input
	    bblocks 
	    symtab 
	    (sck-regassign-make-real-register-sets defs)
	    (sck-regassign-make-real-register-constraints defs))
	   :graphviz-outdir graphviz-outdir)
	  ;;(sck-regalloc-debug 
	  ;;(sck-regassign-make-regalloc-input bblocks symtab regsets regconst)
	  ;;graphviz-outdir)
	  ((register-assignments assigns)
	   (spilled-registers spills))
	'(message "assigns\n%s" (sck-lir-pp assigns))
	(if spills
	    (setq func (sck-regassign-spilled-remake-function func spills))
	  (setq symtab (sck-regassign-assign-regsters-symtab symtab assigns))
	  (sck-regassign-assign-regsters bblocks assigns)
	  '(message "symtab\n%s" (sck-lir-pp symtab))
	  (setq func 
		(sck-regassign-make-output-function 
		 (cadr func)
		 symtab
		 (sck-regassign-sort-bblock bblocks))))
	'(message "lfunc %s" (sck-lir-pp func))
	;; make output
	(list
	 (list 'lfunc func)
	 (list 'spilled-registers spills))))))

;;;
;;;
;;;
(defun* sck-cgen (input &key graphviz-outdir)
  ;; @@@ OBSOLETE
  ;; ((lir <LIR>))
  ;; -> ((lir <LIR>))
  (sck-letalist input
      ((lir lir))
    ;; lir $B<+BN$,(B module $B$G$$$$(B?
    (sck-cgen-module lir :graphviz-outdir graphviz-outdir)))

(defun* sck-cgen-module (module &key graphviz-outdir)
  ;; @@@ OBSOLETE
  (setq module (sck-tmd-restruct-add-reginfo module))
  (let* (output
	 ;; regsets $B$H(B regconst $B$r:n@.(B
	 (sck*regassign-module-defs (assq 'defs module))
	 (regsets (sck-regassign-make-real-register-sets
		   sck*regassign-module-defs))
	 (regconst (sck-regassign-make-real-register-constraints
		    sck*regassign-module-defs)))
    '(message "defs %s" (sck-lir-pp sck*regassign-module-defs))
    (assert (and sck*regassign-module-defs regsets))
    ;; function $B$rJQ49(B
    (dolist (func module)
      (cond
       ((eq (car-safe func) 'function)
	'(message "\nfunction %S\n" (cadr func))
	(push (sck-cgen-function func regsets regconst
				 :graphviz-outdir graphviz-outdir)
	      output)
	(when graphviz-outdir
	  (y-or-n-p
	   "Graphviz $B$rBT$A$^$9(B(Y/N)$B$I$A$i$+$rF~NO$7$?$i?J$_$^$9(B"))
	)
       (t
	(push func output))))
    (nreverse output)))

(defun* sck-cgen-function (func regsets regconst &key graphviz-outdir)
  ;; @@@ OBSOLETE
  (assert (eq (car-safe func) 'function))
  '(message "sck-regassign-function %s" (sck-lir-pp func))
  ;; $BCf?H$r=q$-49$($k$N$G(B
  (setq func (copy-tree func))
  (let ((loopf t)
	bblocks
	symtab)
    ;; spill $B$,L5$/$J$k$^$G%k!<%W(B
    ;; $B:G=i$@$1(B restruct
    (setq func (sck-tmd-restruct func))
    ;; print hook
    (when sck*regassign-after-restruct-hook
      (funcall sck*regassign-after-restruct-hook func))
    (while loopf
      (let ((old-symtab (copy-tree symtab)))
	(setq func (sck-tmd-instsel func))
	'(message "func\n%S" (sck-lir-pp func))
	(setq bblocks (sck-regassign-make-bblocks func)
	      symtab (cdr (assq 'symtab func)))
	'(message "%s\n" (sck-regassign-format-bblock bblocks))
	'(message "symtab %S\n" symtab)
	'(message "old-symtab %S\n" old-symtab)
	;; $B%9%T%k$5$l$?8e$K:n$i$l$?2>A[%l%8%9%?$N%W%i%$%*%j%F%#$r@_Dj(B
	;; @@@ tmd $BB&$G$d$k(B
	'(when old-symtab
	   (dolist (sym symtab)
	     (unless (assoc (car sym) old-symtab)
	       (sck-lir-putprop sym '&spillcost 10000))))
	)
      ;; print hook
      (when sck*regassign-after-instsel-hook
	(funcall sck*regassign-after-instsel-hook func))
      (sck-letalist 
	  (sck-regalloc
	   (sck-regassign-make-regalloc-input bblocks symtab regsets regconst)
	   :graphviz-outdir graphviz-outdir)
	  ;;(sck-regalloc-debug 
	  ;; (sck-regassign-make-regalloc-input bblocks symtab regsets regconst)
	  ;;graphviz-outdir)
	  ((register-assignments assigns)
	   (spilled-registers spills))
	
	'(message "assigns\n%s\nspilled-registers\n%s" 
		  (sck-lir-pp assigns)
		  (sck-lir-pp spills))
	
	(if spills
	    (progn
	      (setq func (sck-regassign-spilled-remake-function func spills))
	      ;; print hook
	      (when sck*regassign-after-spill-hook
		(funcall sck*regassign-after-spill-hook func))
	      (when graphviz-outdir
		(y-or-n-p
		 "Graphviz $B$rBT$A$^$9(B(Y/N)$B$I$A$i$+$rF~NO$7$?$i?J$_$^$9(B")))

	  (setq symtab (sck-regassign-assign-regsters-symtab symtab assigns))
	  (sck-regassign-assign-regsters bblocks assigns)
	  (setq loopf nil))))
    (sck-regassign-make-output-function (cadr func)
					symtab
					(sck-regassign-sort-bblock bblocks))))

;;;
;;;
;;;
(defun sck-regassign-make-output-function (name symtab bblocks)
  ;; bblocks -> <Lfunc>
  (append (list 'function name
		(cons 'symtab symtab))
	  (let (insts)
	    (dolist (b bblocks)
	      (push (list 'deflabel (sck-regassign-bblock-label b)) insts)
	      (dolist (inst (sck-regassign-bblock-insts b))
		(push inst insts)))
	    (nreverse insts))))

;;;
;;;
;;;
(defun sck-regassign-assign-regsters-symtab (symtab assigns)
  ;; assign $B$5$l$?(B symtab $B$r:n$jD>$9(B
  (let (newsymtab)
    (dolist (s symtab)
      (unless (eq (cadr s) 'reg) 
	(push s newsymtab)))
    (nreverse newsymtab)))

(defun sck-regassign-assign-regsters (bblocks assigns)
  ;; inst$B$r(Bassign$B$5$l$?%l%8%9%?$G=q$-49$($k(B
  ;; $BI,MW$NL5$$(Binst$B$O:o=|(B
  (dolist (b bblocks)
    (let ((insts (sck-regassign-bblock-insts b)))
      (while insts
	(sck-regassign-assign-regsters/inst (car insts) assigns)
	(when (and (eq (caar insts) 'set)
		   (equal (caddr (car insts)) (cadddr (car insts))))
	  (setf (car insts) nil))
	(setq insts (cdr insts))))
    (setf (sck-regassign-bblock-insts b) 
	  (delq nil (sck-regassign-bblock-insts b)))))

(defun sck-regassign-assign-regsters/inst (inst assigns)
  (while inst
    (cond
     ((atom (car inst)))
     ((sck-regassign-vreg-p (car inst))
      (let ((reg (assoc (car inst) assigns)))
	
	'(message "inst %S reg %S" (car inst) reg)

	(assert reg)
	(setf (car inst) (cadr reg))))
     (t
      (sck-regassign-assign-regsters/inst (car inst) assigns)))
    (setq inst (cdr inst))))

;;
;; 
;;
(defun sck-regassign-spilled-remake-function (func spill)
  '(message "\nsck-regassign-spilled-remake-function in\nspill %S\n" spill)
  (dolist (inst (cddr func))
    (case (car inst)
      (symtab
       (dolist (reg spill)
	 '(message "reg %S" reg)
	 (let ((conv (assoc (caddr reg) (cdr inst))))
	   (when conv 
	     (setf (cadr conv) 'frame)))))
      ;;(prologue)
      ;;(epilogue)
      (deflabel)
      (jump)
      (t
       (dolist (x (cdr inst))
	 (sck-regassign-spilled-remake-function/inst x spill)))))

  '(message "sck-regassign-spilled-remake-function out\n%s\n"
	    (sck-lir-pp func))

  func)

(defun sck-regassign-spilled-remake-function/inst (inst spill)
  '(message 
    "sck-regassign-spilled-remake-function/inst inst %S spill %S" inst spill)
  (if (atom inst)
      nil
    (cond
     ((or (eq (car inst) 'reg)
	  (eq (car inst) 'subreg))

      '(message "inst %S spill %S" inst spill)

      (when (member inst spill)
	(setf (car inst) 'mem)
	(setf (caddr inst) (list 'frame (cadr inst) (caddr inst)))))
     (t
      (dolist (x (cdr inst))
	(sck-regassign-spilled-remake-function/inst x spill))))))

;;
;; @@@ $B$3$l$O8e$GJL%U%!%$%k$K$9$k(B
;;
(defun sck-regassign-sort-bblock (bblocks)
  ;; @@@ Mar. 13, 2007
  ;; (setq bblocks (sck-regassign-remove-useless-bblock bblocks))
  ;; $B%k!<%W$rH4$1$k(Bbblock$B$N0\F0(B
  (sck-letalist (sck-loop (list
			   (list 
			    'bblocks
			    (mapcar 
			     #'(lambda (b)
				 (list 
				  (sck-regassign-bblock-label b)
				  (mapcar 'sck-regassign-bblock-label
					  (sck-regassign-bblock-succ b))))
			     bblocks))))
      ((loops loops)
       (bblocks out-bblocks))
    '(message "loops\n%s" (sck-lir-pp loops))
    (dolist (l loops)
      (setq bblocks (sck-regassign-sort-bblock/loop bblocks l))))
  ;; @@@ Feb. 18, 2007 $B$3$l$OF~$l$J$$(B @@@ Mar. 13, 2007
  ;; jumpc$B$N8e$m$K(Bbblock$B$r0\F0(B
  ;;(sck-regassign-sort-bblock/jumpc bblocks)
  bblocks
  )

(defun sck-regassign-remove-useless-bblock (bblocks)
  ;; $BL5BL$J(Bbblock$B$N:o=|(B
  (let (newbblocks)
    (dolist (b bblocks)
      (let ((inst (car (sck-regassign-bblock-insts b))))
	;; $B:G=i$,(Bjump$B$J$i(Binst$B$,$R$H$D$7$+L5$$$O$:(B
	(if (eq (car inst) 'jump)
	    (let* ((srclabel (sck-regassign-bblock-label b))
		   (dstlabel (sck-regassign-bblock-label 
			      (car (sck-regassign-bblock-succ b))))
		   (dst (find dstlabel
			      bblocks
			      :key 'sck-regassign-bblock-label
			      :test 'string=)))

	      '(message "srclabel %S dstlabel %S" srclabel dstlabel)
	      (dolist (pred (sck-regassign-bblock-pred b))
		(let ((lastinst (sck-regassign-bblock-last-inst pred)))

		  '(message "\npred %s\nlastinst %S\n\n" 
			    (sck-regassign-format-bblock pred)
			    lastinst)
		  (case (car lastinst)
		    (jump
		     (assert (string= (caddr (cadr lastinst)) srclabel))
		     (setf (caddr (cadr lastinst)) dstlabel)
		     (setf (car (sck-regassign-bblock-succ pred)) dst)
		     (setf (sck-regassign-bblock-pred dst)
			   (delq b (sck-regassign-bblock-pred dst)))
		     (pushnew pred (sck-regassign-bblock-pred dst)))
		    (jumpc
		     (dolist (x (cddr lastinst))
		       (when (string= (caddr x) srclabel)
			 (setf (caddr x) dstlabel)))
		     (setf (sck-regassign-bblock-succ pred)
			   (delq b (sck-regassign-bblock-succ pred)))
		     (pushnew dst (sck-regassign-bblock-succ pred))
		     (setf (sck-regassign-bblock-pred dst)
			   (delq b (sck-regassign-bblock-pred dst)))
		     (pushnew pred (sck-regassign-bblock-pred dst)))
		    (t
		     (assert nil))))))
	  (push b newbblocks))))

    '(message "sck-regassign-remove-useless-bblock\n%s"
	      (sck-regassign-format-bblock (reverse newbblocks)))

    (nreverse newbblocks)))

(defun sck-regassign-jumpc-out-label (bblock)
  ;; jumpc$B$GH4$1$k@h$N%i%Y%k$rF@$k(B
  (let ((inst (sck-regassign-bblock-last-inst bblock)))
    (when (and inst
	       (eq (car inst) 'jumpc))
      (caddr (cadddr inst)))))

(defun sck-regassign-sort-bblock/jumpc (bblocks)
  ;; jumpc$B$GH4$1$k(Bbblock$B$N8e$m$KHt$S@h$N(Bbblock$B$r0\F0(B
  (let ((changed t)
	l-bblocks
	outbblock
	movebblock)
    ;; $BA4$F$N(Bjumpc$B$,JRIU$/$^$G%k!<%W(B
    (while changed
      (setq changed nil)
      (setq l-bblocks bblocks)
      (setq movebblock
	    (block 'find-movebblock
	      (while l-bblocks
		(let ((succ (sck-regassign-jumpc-out-label (car l-bblocks))))
		  (when succ
		    (unless (string= 
			     (sck-regassign-bblock-label (cadr l-bblocks))
			     succ)
		      (setq outbblock (car l-bblocks))
		      (return-from 'find-movebblock
			(find succ bblocks
			      :key 'sck-regassign-bblock-label
			      :test 'string=)))))
		(setq l-bblocks (cdr l-bblocks)))))
      (when movebblock
	(setq changed t)
	(let (newbblocks)
	  (while bblocks
	    (let ((label (sck-regassign-bblock-label (car bblocks))))
	      (cond 
	       ((string= label (sck-regassign-bblock-label movebblock)) nil)
	       ((string= label (sck-regassign-bblock-label outbblock))
		;; $B8e$m$KA^F~(B
		(push outbblock newbblocks)
		(push movebblock newbblocks))
	       (t
		(push (car bblocks) newbblocks)))
	      (setq bblocks (cdr bblocks))))
	  (setq bblocks (nreverse newbblocks)))))
    bblocks))

(defun sck-regassign-sort-bblock/loop (bblocks loop)
  ;; $B%k!<%W30$KHt$V(Bbblock$B$r8+$D$1$F$=$l$rHt$S@h$NA0$K0\F0(B
  '(message "sck-regassign-sort-bblock/loop in %S" loop)
  (let ((flat-loop (flat loop))
	outbblock
	movebblock
	newbblocks)
    (dolist (l loop)
      (if (listp l)
	  (setq bblocks (sck-regassign-sort-bblock/loop bblocks l))
	;; $B%k!<%W30$KHt$S=P$9(Bbblock$B$r8+$D$1$k(B
	(unless movebblock
	  (let ((b (find l bblocks
			 :key 'sck-regassign-bblock-label
			 :test 'string=)))
	    (when b
	      (dolist (succ (sck-regassign-bblock-succ b))
		(unless (member (sck-regassign-bblock-label succ) flat-loop)
		  (setq outbblock succ)
		  (setq movebblock b))))))))

    '(message "\nmovebblock %S\noutbblock %S" movebblock outbblock)
    (assert (and movebblock outbblock))
    ;; jumpc $B$NJQ99$,I,MW$J$iJQ99$9$k(B
    (unless (string= (sck-regassign-jumpc-out-label movebblock)
		     (sck-regassign-bblock-label outbblock))
      (let ((inst (sck-regassign-bblock-last-inst movebblock)))
	(assert (and (car inst) 'jumpc
		     (string= (caddr (caddr inst)) 
			      (sck-regassign-bblock-label outbblock))))
	(setf (caadr inst) (cadr (assq (caadr inst) sck=lir-tst-not-alist)))
	(setf (cddr inst) (list (cadddr inst) (caddr inst)))))
    ;; bblocks $B$rJB$SBX$($k(B
    (while bblocks
      (let ((label (sck-regassign-bblock-label (car bblocks))))
	(cond 
	 ((string= label (sck-regassign-bblock-label movebblock)) nil)
	 ((string= label (sck-regassign-bblock-label outbblock))
	  ;; $BA0$KA^F~(B
	  (push movebblock newbblocks)
	  (push outbblock newbblocks))
	 (t
	  (push (car bblocks) newbblocks)))
	(setq bblocks (cdr bblocks))))
    (nreverse newbblocks)))

;;;
;;; make bblocks
;;;
(defun sck-regassign-jump-labels (inst)
  ;; inst $B$,(B jump/jumpc $B$J$i$=$N%i%Y%k$N%j%9%H$r!"(B
  ;; $B$=$l0J30$NL?Na$J$i(B nil $B$rJV$9!#(B
  (case (car inst)
    (jump (mapcar #'(lambda (x) (caddr x)) (cdr inst)))
    (jumpc (mapcar #'(lambda (x) (caddr x)) (cddr inst)))
    ;;(jumpn (cddr inst))
    (t nil)))

(defun sck-regassign-bblock-last-inst (bblock)
  ;; bblock $B$N:G8e$NL?Na$rJV$9!#(B
  ;; $B$J$1$l$P(B nil$B!#(B
  (car (last (sck-regassign-bblock-insts bblock))))

(defun sck-regassign-make-bblocks (function)
  (let (bblockalist
	binsts
	label)
    ;; $B$^$:(B bblockalist = ((Label inst...) ...) $B$r:n$k!#(B
    (dolist (inst (cddr function))
      (cond
       ((eq (car inst) 'deflabel)
	(when label
	  (push (cons label (nreverse binsts)) bblockalist))
	(setq binsts nil)
	(setq label (cadr inst)))
       (t
	(push inst binsts))))
    ;; $B:G8e$N#1$D(B
    (push (cons label (nreverse binsts)) bblockalist)
    (setq bblockalist (nreverse bblockalist))

    '(message "bblockalist\n%s" (sck-lir-pp bblockalist))

    ;; $B$=$l$+$i(B bblock $B$r:n$j!"$=$N%j%9%H$rJV$9(B
    (let ((bblocks (mapcar '(lambda (x)
			      ;; x = (Label inst...)
			      (make-sck-regassign-bblock
			       :label (car x)
			       :insts (cdr x)))
			   bblockalist)))
      ;; succ/pred $B$r%;%C%H(B
      (dolist (b bblocks)
	(setf (sck-regassign-bblock-succ b)
	      (mapcar #'(lambda (label)
			  (or (find label bblocks
				    :key 'sck-regassign-bblock-label
				    :test 'string=)
			      (error "Label %s not found" label)))
		      (sck-regassign-jump-labels
		       (sck-regassign-bblock-last-inst b))))
	(dolist (bs (sck-regassign-bblock-succ b))
	  (push b (sck-regassign-bblock-pred bs))))
      '(message "%s\n" (sck-regassign-format-bblock bblocks))
      ;; live{in,out} $B$r7W;;$7$FJV$9(B
      (sck-regassign-life-analysis bblocks))))

(defun sck-regassign-vreg-p (inst)
  (or
   (eq (car inst) 'reg)
   (eq (car inst) 'subreg)))

(defun sck-regassign-inst-defrefs (inst)
  "inst $B$N(B out $B$H(B in $B$N2>A[%l%8%9%?$N%j%9%H$rJV$9(B"
  (let (defs
	 sck*regassign-inst-defrefs-refs)
    ;; def 
    (case (car inst)
      (set
       (when (sck-regassign-vreg-p (caddr inst))
	 (push (caddr inst) defs)))
      (prologue
       (dolist (x (cdr inst))
	 (let ((defrefs (sck-regassign-inst-defrefs x)))
	   (dolist (def (car defrefs))
	     (push def defs)))))
      (parallel
       (dolist (x (cdr inst))
	 (let ((defrefs (sck-regassign-inst-defrefs x)))
	   (dolist (def (car defrefs))
	     (push def defs)))))
      (clobber
       (if (symbolp (cadr inst))
	   ;; symbol$B$N>l9g(B
	   (dolist (x (cadr (assq (cadr inst) sck*regassign-module-defs)))
	     (push x defs))
	 (push (cadr inst) defs))))
    ;; ref
    (sck-regassign-inst-defrefs/ref inst)
    '(message "defref inst %S\nout %S" 
	      inst
	      (list defs sck*regassign-inst-defrefs-refs))
    (list defs sck*regassign-inst-defrefs-refs)))

(defun sck-regassign-inst-defrefs/ref (inst)
  '(message "/ref %S\n" inst)
  (case (car inst)
    (jump nil)
    (iconst nil)
    (fconst nil)
    (static nil)
    (frame nil)
    (label nil)
    (deflabel nil)
    (reg
     (pushnew inst sck*regassign-inst-defrefs-refs))
    (subreg
     (pushnew inst sck*regassign-inst-defrefs-refs))
    (set
     (when (not (sck-regassign-vreg-p (caddr inst)))
       (sck-regassign-inst-defrefs/ref (caddr inst)))
     (sck-regassign-inst-defrefs/ref (cadddr inst)))
    (epilogue 
     (sck-regassign-inst-defrefs/ref (cadr inst)))
    (jumpc
     (sck-regassign-inst-defrefs/ref (cadr inst)))
    (use
     (sck-regassign-inst-defrefs/ref (cadr inst)))
    (parallel
     (dolist (x (cdr inst))
       (sck-regassign-inst-defrefs/ref x)))
    (prologue
     (dolist (x (cdr inst))
       (sck-regassign-inst-defrefs/ref x)))
    (call
     (dolist (x (cadddr inst))
       (sck-regassign-inst-defrefs/ref x)))
    (t
     (dolist (x (cddr inst))
       (sck-regassign-inst-defrefs/ref x)))))

(defun sck-regassign-life-analysis (bblocks)
  "bblock $B$N%j%9%H(B BBLOCKS $B$r<u$1<h$j!"=q$/(B bblock $B$N(B live{in,out} $B$r7W;;!#(B
BBLOCKS $B$rJV$9!#%U%m!<2r@O%Q%C%1!<%8(B sck-live $B$r;H$&!#(B"
  (let* ((input (list
		 (list 'basic-blocks
		       (mapcar
			#'(lambda (b)
			    (list (sck-regassign-bblock-label b)
				  (mapcar 'sck-regassign-inst-defrefs
					  (sck-regassign-bblock-insts b))
				  (mapcar 'sck-regassign-bblock-label
					  (sck-regassign-bblock-succ b))))
			bblocks))))
	 ;; <basic-blocks> ::= ((<Label> (<Inst>..) (<Succ>..))..)
	 (output (sck-live input)))
    
    '(message "\nsck-live input %s output %s\n"
	      (sck-lir-pp input)
	      (sck-lir-pp output))

    (sck-letalist output
	;; <liveness-info> ::= ((<Label> (<LiveIn>..) (<LiveOut>..))..)
	(liveness-info)
      (dolist (i liveness-info)
	(let ((b (find (car i) bblocks
		       :key 'sck-regassign-bblock-label
		       :test 'string=)))
	  (setf (sck-regassign-bblock-livein b) (cadr i)
		(sck-regassign-bblock-liveout b) (caddr i)))))
    ;;
    bblocks))

;;;
;;; regalloc input
;;;
(defun sck-regassign-make-real-register-sets (defs)
  (cdr defs))

(defun sck-regassign-make-real-register-constraints (defs)
  (let (regconst)
    (dolist (regset (cdr defs))
      '(message "regset\n%s" (sck-lir-pp regset))
      (when (eq (caaadr regset) 'subreg)
	(dolist (reg (cadr regset))
	  (push (list reg (caddr reg)) regconst))))
    '(message "regconst\n%s" (sck-lir-pp regconst))
    regconst))

(defun sck-regassign-make-regalloc-input (bblocks symtab regsets regconst)
  (let (regs)
    ;; $B;HMQ$7$F$$$k%l%8%9%?$r=8$a$k(B
    (dolist (b bblocks)
      (dolist (inst (sck-regassign-bblock-insts b))
	(let* ((defrefs (sck-regassign-inst-defrefs inst))
	       (defs (car defrefs))
	       (refs (cadr defrefs)))
	  (dolist (reg defs) (pushnew reg regs :test 'equal))
	  (dolist (reg refs) (pushnew reg regs :test 'equal)))))
    (dolist (x symtab)
      (when (eq (cadr x) 'reg)
	(let ((reg (list 'reg (caddr x) (car x))))
	  (unless (member reg regs)
	    (pushnew reg regs :test 'equal)))))

    '(message "sck-regassign-make-regalloc-input\nregs %s" (sck-lir-pp regs))
    ;;
    (list 
     (list 'interference-graph (sck-regassign-make-interference-graph bblocks))
     (list 'coalesceable-edges (sck-regassign-make-coalesceable-edges bblocks))
     (list 'real-register-sets regsets)
     (list 'real-register-constraints regconst)
     (list 'possible-assignments 
	   (sck-regassign-make-possible-assignments regs symtab regsets))
     (list 'spill-priority
	   (sck-regassign-make-spill-priority bblocks regs symtab)))))

(defun sck-regassign-make-spill-priority (bblocks regs symtab)
  (let (ret)
    (sck-letalist (sck-loop (list
			     (list 
			      'bblocks
			      (mapcar 
			       #'(lambda (b)
				   (list 
				    (sck-regassign-bblock-label b)
				    (mapcar 'sck-regassign-bblock-label
					    (sck-regassign-bblock-succ b))))
			       bblocks))))
	((loops loops)
	 (bblocks out-bblocks))
      '(message "loops\n%s" (sck-lir-pp loops))
      (dolist (reg regs)
	(let* ((sp (memq '&spillcost (assoc (caddr reg) symtab)))
	       (bp (if sp (cadr sp) 1)))
	  (push 
	   (list reg
		 ;; loop$B$N%W%i%$%*%j%F%#$O$3$3$G7h$a$k(B
		 (* bp 
		    (+ (sck-regassign-make-spill-priority/count 
			reg bblocks)
		       (sck-regassign-make-spill-priority/inloop
			reg bblocks loops)
		       ))) ret))))
    
    '(message "spill-priority\n%s" (sck-lir-pp ret))
    
    ret))

(defun sck-regassign-make-spill-priority/inloop (reg bblocks loops)
  ;; loop$B$N?<$5(B*10$B$rJV$9(B
  (* (sck-regassign-make-spill-priority/inloop-in reg bblocks loops 0) 10))

(defun sck-regassign-make-spill-priority/inloop-in (reg bblocks loops depth)
  '(message "sck-regassign-make-spill-priority reg %S" reg)
  (let ((count 0))
    (dolist (x loops)
      (let ((newcount 0))
	(if (listp x)
	    (setq newcount 
		  (sck-regassign-make-spill-priority/inloop-in
		   reg bblocks x (1+ depth)))
	  (let ((b (find x bblocks
			 :key 'sck-regassign-bblock-label
			 :test 'string=)))
	    (when (and b
		       (or 
			(member reg (sck-regassign-bblock-livein b))
			(member reg (sck-regassign-bblock-liveout b))))
	      (setq newcount depth))))
	(when (> newcount count)
	  (setq count newcount))))
    '(message "sck-regassign-make-spill-priority count %d" count)
    count))

(defun sck-regassign-make-spill-priority/count (reg bblocks)
  ;; $B%l%8%9%?$N;H$o$l$F$$$k?t$rJV$9(B
  '(message "\nsck-regassign-make-spill-priority/count reg %S" reg)
  (let ((count 0))
    (dolist (b bblocks)
      (dolist (inst (sck-regassign-bblock-insts b))
	(let ((defrefs (sck-regassign-inst-defrefs inst)))
	  (when (member reg (car defrefs))
	    (incf count))
	  (when (member reg (cadr defrefs)) (incf count)))))
    '(message "sck-regassign-make-spill-priority/count count %d\n" count)
    count))

;;;
(defun sck-regassign-make-possible-assignments (regs symtab regsets)
  '(message "sck-regassign-make-possible-assignments in\nregs %s"
	    (sck-lir-pp regs))
  
  (let (ret)
    (dolist (reg regs)
      ;; $B%O!<%I%l%8%9%?$N(Bregset$B$O8!:w$9$k(B
      (let ((regset (memq '&regset (assoc (caddr reg) symtab))))
	(when regset 
	  (push (list reg (cadr regset)) ret))))
    ;; $BL5$$$b$N$O%O!<%I%l%8%9%?$J$O$:(B
    ;; $B%O!<%I%l%8%9%?$O@_Dj$7$J$/$J$C$?(B
    ret))

(defun sck-regassign-make-coalesceable-edges (bblocks)
  "coalesceable-edge $B$r:n$k(B"
  (let ((edges nil))
    (dolist (b bblocks)
      (let ((insts (sck-regassign-bblock-insts b)))
	(setq edges
	      (union edges
		     (mapcar #'(lambda (setxy)
				 (list (caddr setxy) (cadddr setxy)))
			     (filter
			      #'(lambda (inst)
				  (and (eq (car inst) 'set)
				       (sck-regassign-vreg-p (caddr inst))
				       (sck-regassign-vreg-p (cadddr inst))))
			      insts))
		     :test 'equal))))
    edges))

;;;
(defun sck-regassign-make-interference-graph (bblocks)
  (let (conflicts)
    (dolist (b bblocks)
      (let ((liveout (sck-regassign-bblock-liveout b)))
	(dolist (inst (reverse (sck-regassign-bblock-insts b)))
	  (let* ((defrefs (sck-regassign-inst-defrefs inst))
		 (defs (car defrefs))
		 (refs (cadr defrefs)))
	    (dolist (ref (union liveout defs :test 'equal))
	      (dolist (def defs)
		(unless (equal def ref)
		  (pushnew (list def ref) conflicts :test 'equal))))
	    ;; defs $B$rH4$-!"(Brefs $B$rDI2C!#(B
	    (setq liveout (set-difference liveout defs :test 'equal)
		  liveout (union liveout refs :test 'equal))))
	;; liveout $B$O(B bblock-livein $B$KEy$7$$%O%:!#(B
	(unless (set-equal liveout (sck-regassign-bblock-livein b))
	  (error 
	   "sck-regassign-make-interference-graph: Assertion failed: liveout"))
	))
    conflicts))

(provide 'sck-regassign)
