;;;; Setting-up SCK
;;;; LastChange: Mon Apr 16 15:46:15 2007

;;; Load this file to setting-up SCK.
;;; See $SCK/README.txt for details.

(defconst sck=version "0.2"
  "SCK ̃o[W")

(require 'cl)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; sck=root-directory@($SCK) ̒`

(defconst sck=root-directory
  (substring (file-name-directory load-file-name) 0 -1)
  "SCK ̃[gfBNg
wvhLgɂāÃfBNg $SCK ŎQƂB")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SCK ̒񋟂́A$SCK ̃t@CƃfBNgɊւȉ̏
;;; ((<File/Directory> <Description> <Owner>)..)
;;;  sck=contents-alist ɃZbgB

(defconst sck=contents-alist
  '(("README.txt" "Getting started on SCK" "abe")
    ;;
    ;; Basic environment
    ;;
    ("sck.el" "Setting up SCK" "abe")
    ("lib" "Common libraries" "abe")
    ("graphviz" "Emacs interface to the Graphviz (ATT)" "yamazaki")
    ("yayacc" "LALR(1) based parser generator" "yamane")
    ;;
    ;; Parser related stuffs
    ;;
    ("clex" "C lexical analyzer (C-Source -> Clex-Output)" "abe")
    ("cpps" "C Plain ParSer (Clex-Output -> C-Plain-AST)" "abe")
    ("cpps-yayacc" "C Plain ParSer (Clex-Output -> C-Plain-AST)" "yamazaki")
    ("ast" "AST basic library" "abe")
    ("cps" "C ParSer (C-Plain-AST -> Core-AST)" "abe")
    ("lirgen" "LIR Generator (Core-AST -> LIR)" "abe")
    ;;
    ;; LIR related stuffs
    ;; 
    ("lir" "LIR basic library" "abe")
    ("liri" "LIR interpreter" "abe")
    ;;
    ;; Code generator related stuffs
    ;;
    ("tmd" "Target Machine Description (restruct, instsel, asmout)" "abe")
    ("asl" "Allocation and Scheduling Language (obsolete)" "abe")
    ("live" "Liveness analysis" "abe")
    ("dom" "Dominator analysis" "yamazaki")
    ("loop" "Loop detection" "yamazaki")
    ("regalloc" "Register allocator" "yamazaki")
    ("regassign" "Register assigner" "yamazaki")
    ;;
    ;; Compiler drivers
    ;;
    ("cc" "C compiler driver" "abe")
    ;;
    ;; Miscellaneous
    ;;
    ("doc" "Document" "abe")
    ("bin" "Executables (shcc, sck-cc, etc.)" "abe")
    ("elex" "Emacs Lisp exercise collection" "abe")
    ("csample" "C sample programs" "yamazaki"))
  "$SCK ̃t@CƃfBNgɊւȉ̏
 ((<File/Directory> <Description> <Owner>)..)")

(defun sck-describe-contents ()
  "$SCK/* ̈ꗗƂ̃I[i[obt@ *sck-describe-contents* ɕ\B
ob`[hĂ΂ꂽꍇ͕Wo͂֕\B"
  (interactive)
  (let ((str (with-temp-buffer
	       (insertf "%50s\n\n" "Brief description of $SCK/*")
	       (insertf "%-14s %30s %30s\n\n"
			"File/Directory"   "Description"    "Owner")
	       (dolist (c sck=contents-alist)
		 (multiple-value-bind (name description owner) c
		   (insertf "%-14s %-55s %s\n"
			    (if (file-directory-p
				 (concat sck=root-directory "/" name))
				(concat name "/")
			      name)
			    description
			    owner)))
	       (buffer-string))))
    (if noninteractive
	(princ str)
      (with-display-buffer "*sck-describe-contents*"
	(insert str)))))

(defun sck-settingup-error (fmt &rest args)
  "SCK ̃ZbgAbvi$SCK/sck.el ̃[hjɔG[
*SCK-settingup-error* obt@ցiob`[hŋNꂽꍇ
Wo͂ցj\B"
  (let ((msgstr (apply 'format fmt args)))
    (if noninteractive
	(princ (concat "\n" msgstr))
      (let ((buf (get-buffer-create "*SCK-settingup-error*")))
	(with-current-buffer buf
	  (when (zerop (buffer-size))
	    (insert "SCK setting-up (loading $SCK/sck.el) failed:\n"))
	  (goto-char (point-max))
	  (insert "\n" msgstr)
	  (display-buffer buf)
	  (set-window-point (get-buffer-window buf) 1)
	  (set-buffer-modified-p nil))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; feature ֌W

(defvar sck=features-before-sck-setup
  (copy-list features)
  "SCK ̃ZbgAbvȑO provide Ă feature ̃XgB")

(defconst sck=features-alist
  ;; (<FeatureSymbol> <FileName>)
  (let ((alist nil))
    (dolist (c sck=contents-alist)
      (multiple-value-bind (name description owner) c
	(let ((dir (concat sck=root-directory "/" name)))
	  (when (file-directory-p dir)
	    (dolist (el (file-expand-wildcards (concat dir "/*.el")))
	      (with-temp-buffer
		(insert-file-contents el)
		(when (re-search-forward "^(provide " nil t)
		  (beginning-of-line)
		  (push (list (cadr (cadr (read (current-buffer))))
			      el)
			alist))))))))
    (nreverse alist))
  "$SCK ɂ provide Ă feature Ƃ̃\[XȂ
 ((<Name> <File>)..) Ƃ A-List B")

(defun sck-require-all ()
  "SCK ̑SW[ require ɂ胍[hB"
  (interactive)
  (dolist (a sck=features-alist)
    (when (and (interactive-p)
	       (not (featurep (car a))))
      (message "sck-require-all: (require '%s)" (car a)))
    (require (car a)))
  (when (interactive-p)
    (message "sck-require-all: Done")))

;;; Check conflicts of features.

(let ((conflicts nil))
  (dolist (a sck=features-alist)
    (when (memq (car a) sck=features-before-sck-setup)
      (push a conflicts)))
  (when conflicts
    (sck-settingup-error
     (with-temp-buffer
       (dolist (a conflicts)
	 (insert (format "Error: Conflicting feature `%s' provided in %s\n"
			 (car a) (cadr a))))
       (buffer-string)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Emacs Lisp R[h̒uĂ $SCK ̃fBNg 
;;; load-path ɒǉ

(dolist (c sck=contents-alist)
  (let ((dir (concat sck=root-directory "/" (car c))))
    (when (and (file-directory-p dir)
		 (file-expand-wildcards (concat dir "/*.el")))
      (unless (member dir load-path)
	(nconc load-path (list dir))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ʃCu $SCK/lib/sck-lib.el ̃[h

(condition-case err
    ;; SCK Cu $SCK/lib/sck-lib.el [h
    (require 'sck-lib)
  ;; [hɃG[ꍇ *sck-message* ɃG[bZ[W\
  (error
   (sck-settingup-error
    (with-temp-buffer
      (insert-buffer-substring "*Messages*")
      (goto-char (point-max))
      (when (re-search-backward "^Loading .*/sck.el (source)\\.\\.\\." nil t)
	(delete-region (point-min) (point))
	(goto-char (point-max)))
      (insert (format "Error: %s\n" (error-message-string err)))
      (buffer-string)))))
      

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; gnuserv ̋N

(unless noninteractive
  (condition-case err
      (unless (get-process "server")
	(require 'gnuserv)
	(when (and (featurep 'gnuserv)
		   (not (get-process "server")))
	  (gnuserv-start)))
    (error
     (sck-settingup-error "Warning: Attempt to run gnuserv failed:
%s
SCK uses the provided feature and thus some functions,
including $SCK/bin/shcc, cannot be used.\n"
			  (error-message-string err)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; eval ̐̒

(setq max-lisp-eval-depth (max max-lisp-eval-depth 1000)
      max-specpdl-size (max max-specpdl-size ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SẴR}hi(interactive ...) 𔺂֐`j autoload `B

(defconst sck=sck-cmds-without-doc nil
  "hLg̏ĂȂR}h̃XgB")

(unless noninteractive
  (condition-case err
      (dolist (c sck=contents-alist)
	(multiple-value-bind (name description owner) c
	  (let ((dir (concat sck=root-directory "/" name)))
	    (when (and (file-directory-p dir)
		       (not (member name '("elex"))))
	      (dolist (el (file-expand-wildcards (concat dir "/*.el")))
		(with-temp-buffer
		  (insert-file-contents el)
		  (goto-char (point-max))
		  (while (re-search-backward "^[^;\n]+(interactive[^-]" nil t)
		    (beginning-of-defun)
		    (let ((def (save-excursion (read (current-buffer)))))
		      (when (memq (car-safe def) '(defun defun*))
			(let ((fsym (cadr def))
			      (doc (if (stringp (cadddr def))
				       (cadddr def)
				     nil))
			      (int (assq 'interactive def)))
			  (when (and int
				     (not (fboundp fsym)))
			    ;; ӁF
			    ;; $SCK/{lib,graphviz} ȊOł $SCK/foo/ 
			    ;; R}h` sck-foo-bar.el Œ`Ă
			    ;; ꍇłAȒP̂߂ $SCK/foo/*.el Ŝ
			    ;; [hB̂߁Aautoload  FILE
			    ;; ƂĂ sck-foo-bar ł͂Ȃ sck-foo 
			    ;; ݒ肷B
			    (autoload
			      fsym
			      (if (member name '("lib" "graphviz"))
				  el
				(concat dir (format "/sck-%s" name)))
			      doc
			      int)
			    ;; hLg̏ĂȂR}h̃`FbN
			    (unless doc
			      (push fsym sck=sck-cmds-without-doc)))))))))))))
    (error
     (sck-settingup-error "Defining autoloads failed:\n%s\n"
			  (error-message-string err)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; oCgRpC̊ƂĖ{t@C[hꍇɕKvƂȂR[h

(when noninteractive
  ;; R[h܂ރ\[XRpCꍇɕKv
  (when (fboundp 'set-language-environment)
    (set-language-environment "Japanese"))
  ;; JgfBNgpXɒǉ
  (pushnew nil load-path))


;;;
(provide 'sck)
