;;;; rule structure
;;;; LastChange: Thu Apr 12 23:33:46 2007


(defstruct sck-tmd-rule
  name ; <Symbol>
  nt ; non-terminal symbol
  pat
  ;;
  ;; ȉ body
  ;;
  check  ; <LispForm> | nil
  cost ; <LispForm>
  asm ; (asmarg ...) | (asminst ...) | nil
  regset ; ((<PatVarId> <RegSetSymbol>)..)
  regeqs ; ((<PatVarId> <PatVarId>)..) ; l
  clobber ; (<RegExp>..)
  ;;
  ;;
  ;; ȉ̓[Uw肷̈ȊOŁA
  ;; ̓s゠炩ߌvZĂB
  ;;
  patvars ; p^[̃p^[ϐ̃Xg
  ;;
  ;; type tB[h́A chain rule ݂̂B
  ;; defcode ͂h chain rule ł pat ̌^񂪎A
  ;; chain rule ̃^COł܂}b`N蓾̂ŁA
  ;; ̎ꂽ^L^B
  ;; e.g. sparc  (defcode load-con32-I32 (SET I32 reg con)...)
  ;; h load-con32-I32.regB
  ;;
  type)


(defun sck-tmd-patvar-p (x)
  ;; (sck-tmd-patvar-p '($ 1 reg)) => t
  (and (eq (car-safe x) '$)
       (integerp (cadr x))))

(defun sck-tmd-patvar-nt (patvar)
  ;; (sck-tmd-patvar-nt '($ 1 reg)) => reg
  (caddr patvar))

(defun sck-tmd-chain-rule-p (rule)
  (sck-tmd-patvar-p (sck-tmd-rule-pat rule)))

(defun sck-tmd-chain-rule-nt (rule)
  ;;  rule-chain-rhs-nt
  (sck-tmd-patvar-nt (sck-tmd-rule-pat rule)))


(defun sck-tmd-rule-to-sexp (rule)
  "RULE:<sck-tmd-rule>  S- (defrule ...) tH[ɕϊB"
  `(defrule
     ,(sck-tmd-rule-name rule)
     ,(sck-tmd-rule-nt rule)
     ,(sck-tmd-rule-pat rule)
     ,@(remq nil
	     `(,(and (sck-tmd-rule-check rule)
		     `(check ,(sck-tmd-rule-check rule)))
	       ,(and (sck-tmd-rule-cost rule)
		     `(cost ,(sck-tmd-rule-cost rule)))
	       ,(sck-tmd-rule-asm rule)
	       ,(and (sck-tmd-rule-regset rule)
		     `(regset ,@(sck-tmd-rule-regset rule)))
	       ,@(mapcar #'(lambda (pvs) `(regeq ,(car pvs) ,(cadr pvs)))
			 (sck-tmd-rule-regeqs rule))
	       ,(and (sck-tmd-rule-clobber rule)
		     `(clobber ,@(sck-tmd-rule-clobber rule)))))))


(defun sck-tmd-defrule (defrule)
  "DEFRULE: (defrule <Name> <Nt> <Pat> . <Body>)  <sck-tmd-rule> B

<Name>, <Nt> ::= <Symbol>
<Body> :: (<Item>..)
<Item> ::= (check <LispForm>)
         | (asmarg <LispForm>) | (asminst <LispForm>..)
         | (cost <LispForm>)
         | (regset (<PatVar> <RegSetSymbol>)..) |
         | (regeq <PatVar> <PatVar>)
         | (clobber <RegExp>)
<PatVar> ::= ($ <Fixnum> ...)

Example:
 (sck-tmd-rule-to-sexp
   (sck-tmd-defrule 
     '(defrule index-2 index (MUL I32 reg (INTCONST I32 2))
        (asmarg `(index ,$1 2)))))
 => (defrule index-2 index
      (MUL I32 ($ 0 reg) (INTCONST I32 2))
      (cost 0)
      (asmarg `(index ,($ 1) 2)))"
  (assert (eq (car-safe defrule) 'defrule))
  (let* ((name (cadr defrule))
	 (nt (caddr defrule))
	 ;;  (add t reg ...) ̏ԈႢ̃`FbNB
	 (pat (sck-tmd-desugar-pat (cadddr defrule)))
	 (body (sck-tmd-desugar-$n (cddddr defrule)))
	 ;;
	 (rule (make-sck-tmd-rule :name name
				  :nt nt
				  :pat pat)))
    (dolist (item body)
      (case (car-safe item)
	(check
	 (unless (and (null (sck-tmd-rule-check rule))
		      (= 2 (length item)))
	   (error "Illegal defrule (check): %S" defrule))
	 (setf (sck-tmd-rule-check rule) (cadr item)))
	(cost
	 (unless (and (null (sck-tmd-rule-cost rule))
		      (= 2 (length item)))
	   (error "Illegal defrule (cost): %S" defrule))
	 (setf (sck-tmd-rule-cost rule) (cadr item)))
	(asmarg
	 (unless (null (sck-tmd-rule-asm rule))
	   (error "Illegal defrule (asmarg): %S" defrule))
	 (setf (sck-tmd-rule-asm rule) item))
	(asminst
	 (unless (null (sck-tmd-rule-asm rule))
	   (error "Illegal defrule (asminst): %S" defrule))
	 (setf (sck-tmd-rule-asm rule) item))
	(regset
	 (unless (and (null (sck-tmd-rule-regset rule))
		      (every #'(lambda (pvrs)
				 (and (sck-tmd-patvar-p (car pvrs))
				      (symbolp (cadr pvrs))))
			     (cdr item)))
	   (error "Illegal defrule (regset): %S" defrule))
	 (setf (sck-tmd-rule-regset rule)
	       (mapcar #'(lambda (vs)
			   (list (cadr (car vs))
				 (cadr vs)))
		       (cdr item))))
	(regeq
	 ;; l
	 (unless (and (= 2 (length (cdr item)))
		      (every 'sck-tmd-patvar-p (cdr item)))
	   (error "Illegal defrule (regeqs): %S" defrule))
	 (push (mapcar 'cadr (cdr item)) ; 
	       (sck-tmd-rule-regeqs rule)))
	(clobber
	 (unless (and (null (sck-tmd-rule-clobber rule))
		      (every 'sck-lir-reg-exp-p (cdr item)))
	   (error "Illegal defrule (clobber): %S" defrule))
	 (setf (sck-tmd-rule-clobber rule) (cdr item)))))
    ;;
    ;; ftHg cost=0 [B
    (setf (sck-tmd-rule-cost rule) (or (sck-tmd-rule-cost rule) 0))
    ;; chain rule  asm ̕[
    (when (and (sck-tmd-chain-rule-p rule)
	       (null (sck-tmd-rule-asm rule)))
      (setf (sck-tmd-rule-asm rule) `(asmarg ($ 1))))
      
    ;;
    ;; ̂߂ɂ炩ߌvZĂB
    (setf (sck-tmd-rule-patvars rule)
	  (sck-tmd-make-pat-patvars pat))
    ;;
    rule))


(defun sck-tmd-desugar-pat (pat)
  ;; pat  _nt p^[ϐ ($ <Fixnum> nt) ɕϊ
  ;;
  ;; (sck-tmd-desugar-pat '(A _reg (B _) (C _con)))
  ;; => (A ($ 0 reg) (B ($ -1)) (C ($ 1 con)))
  ;;
  ;; (sck-tmd-desugar-pat '(A ($ 0 reg) (B ($ -1)) (C ($ 1 con))))
  ;; => (A ($ 0 reg) (B ($ -1)) (C ($ 1 con)))
  ;;
  (let ((idplace (list 0)))
    (sck-tmd-desugar-pat/aux pat idplace)))

(defun sck-tmd-desugar-pat/aux (pat idplace)
  (cond
   ((and (symbolp pat)
	 (string-match "^_" (symbol-name pat)))
    (cond
     ((eq pat '_)
      '($ -1))
     (t
      (prog1 `($ ,(car idplace) ,(intern (substring (symbol-name pat) 1)))
	(incf (car idplace))))))
   ((atom pat)
    pat)
   ((sck-tmd-patvar-p pat)
    pat)
   (t
    (mapcar #'(lambda (pat)
		(sck-tmd-desugar-pat/aux pat idplace))
	    pat))))

(defun sck-tmd-make-pat-patvars (pat)
  ;; p^[̃p^[ϐ̃Xg(ԍŃ\[g)ԂB
  ;; (sck-tmd-make-pat-patvars '(A ($ 0 reg) (B ($ -1)) (C ($ 1 con))))
  ;; => (($ 0 reg) ($ 1 con))
  (let ((varq (make-que)))
    (sck-tmd-make-pat-patvars/aux pat varq)
    ;; \[g͂ȂnYAÔ߁B
    (sort (que-to-list varq)
	  #'(lambda (x y) (< (cadr x) (cadr y))))))

(defun sck-tmd-make-pat-patvars/aux (pat varq)
  (cond
   ((atom pat))
   ((eq (car pat) '$)
    (unless (= -1 (cadr pat))
      (append-que varq pat)))
   (t
    (dolist (p pat)
      (sck-tmd-make-pat-patvars/aux p varq)))))
      
(defun sck-tmd-desugar-$n (pat)
  ;; (sck-tmd-desugar-$n '(a $1 `(b ,$2 ,@$3)))
  ;; => (a ($ 1) `(b ,($ 2) ,@($ 3)))
  (cond
   ((symbolp pat)
    (let ((str (symbol-name pat)))
      (if (string-match "^\\$\\([0-9]+\\)$"  str)
	  `($ ,(string-to-int (match-string 1 str)))
	pat)))
   ((atom pat)
    pat)
   (t
    (mapcar 'sck-tmd-desugar-$n pat))))


;;;
(provide 'sck-tmd-rule)
