;;;
;;; 
;;;
(require 'sck-yayacc)
(load "findscc")

(defconst sck=yayacc-cc-c-grammar
  '((object-language "elisp")
    (associativity ((left ("+" "-"))
		    (left ("*" "/"))
		    (left (" NEG"))
		    ))
    (attribute-grammar 
     (
      ((start	file " EOF"))

      ;; Declaration
      ((file	external-definition))
      ((file	file external-definition))
 
      ((external-definition	function-definition))
      ((external-definition	declaration))

      ((function-definition
	declaration-specifiers-opt 
	declarator
	declaration-list-opt
	compound-statement))

      ((declaration
	declaration-specifiers init-declarator-list-opt ";"))

      ((declaration-specifiers-opt))
      ((declaration-specifiers-opt	declaration-specifiers))

      ((declaration-specifiers
	storage-class-specifier declaration-specifiers-opt))
      ((declaration-specifiers
	type-specifier declaration-specifiers-opt))

      ((init-declarator-list-opt	))
      ((init-declarator-list-opt	init-declarator-list))

      ((init-declarator-list	init-declarator))
      ((init-declarator-list	init-declarator-list "," init-declarator))

      ((init-declarator	declarator))
      ((init-declarator	declarator "=" initializer))

      ((storage-class-specifier	"typedef"))
      ((storage-class-specifier	"extern"))
      ((storage-class-specifier	"static"))
      ((storage-class-specifier	"auto"))
      ((storage-class-specifier	"register"))

      ((type-specifier	"void"))
      ((type-specifier	"char"))
      ((type-specifier	"short"))
      ((type-specifier	"int"))
      ((type-specifier	"long"))
      ((type-specifier	"float"))
      ((type-specifier	"unsigned"))
      ((type-specifier	"double"))
      ((type-specifier	"signed"))
      ((type-specifier	"const"))
      ((type-specifier	"volatile"))
      ((type-specifier	struct-or-union-specifier))
      ((type-specifier	enum-specifier))
      ;;((type-specifier	typedef-name))
      ;; $B%?!<%_%J%k$K(B
      ((type-specifier	"typedef-name"))

      ;;((typedef-name	"identifier"))

      ((struct-or-union-specifier
	struct-or-union identifier-opt "{" struct-declaration-list "}"))
      ((struct-or-union-specifier
	struct-or-union "identifier"))

      ((struct-or-union	"struct"))
      ((struct-or-union	"union"))

      ((struct-declaration-list	struct-declaration))
      ((struct-declaration-list	struct-declaration-list struct-declaration))
      ((struct-declaration	type-specifier-list struct-declarator-list))

      ((type-specifier-list-opt	))
      ((type-specifier-list-opt	type-specifier-list))

      ((type-specifier-list	type-specifier))
      ((type-specifier-list	type-specifier-list type-specifier))

      ((struct-declarator-list	struct-declarator))
      ((struct-declarator-list	struct-declarator-list "," struct-declarator))

      ((struct-declarator	declarator))
      ((struct-declarator	declarator-opt ":" constant-expression))

      ((enum-specifier	"enum" identifier-opt "{" enumerator-list "}"))
      ((enum-specifier	"enum" "identifier"))

      ((enumerator-list	enumerator))
      ((enumerator-list	enumerator-list "," enumerator))

      ((enumerator	enumeration-constant))
      ((enumerator	enumeration-constant "=" constant-expression))

      ((enumeration-constant	"identifier"))

      ((declarator-opt	))
      ((declarator-opt	declarator))

      ((declarator	pointer-opt direct-declarator))

      ((direct-declarator	"identifier"))
      ((direct-declarator	"(" declarator ")"))
      ((direct-declarator
	direct-declarator "[" constant-expression-opt "]"))
      ((direct-declarator	direct-declarator "(" parameter-type-list ")"))
      ((direct-declarator
	direct-declarator "(" identifier-list-opt ")"))

      ((pointer-opt	))
      ((pointer-opt	pointer))

      ((pointer	"*" type-specifier-list-opt))
      ((pointer	"*" type-specifier-list-opt pointer))

      ((parameter-type-list	parameter-list))
      ((parameter-type-list	parameter-list "," "..."))

      ((parameter-list	parameter-declaration))
      ((parameter-list	parameter-list "," parameter-declaration))

      ((parameter-declaration	declaration-specifiers declarator))
      ((parameter-declaration	declaration-specifiers abstract-declarator-opt))

      ((identifier-list-opt	))
      ((identifier-list-opt	identifier-list))

      ((identifier-list	"identifier"))
      ((identifier-list	identifier-list "," "identifier"))

      ((type-name	type-specifier-list abstract-declarator-opt))

      ((abstract-declarator-opt	))
      ((abstract-declarator-opt	abstract-declarator))

      ((abstract-declarator	pointer))
      ((abstract-declarator	pointer-opt direct-abstract-declarator))

      ((direct-abstract-declarator-opt	))
      ((direct-abstract-declarator-opt	direct-abstract-declarator))

      ((direct-abstract-declarator	"(" abstract-declarator ")"))
      ((direct-abstract-declarator
	direct-abstract-declarator-opt "[" constant-expression "]"))
      ((direct-abstract-declarator
	direct-abstract-declarator-opt "(" parameter-type-list ")"))

      ((initializer	assignment-expression))
      ((initializer	"{" initializer-list "}"))
      ((initializer	"{" initializer-list "," "}"))

      ((initializer-list	initializer))
      ((initializer-list	initializer-list "," initializer))
     
      ;; Statement
      ((statement	labeled-statement))
      ((statement	compound-statement))
      ((statement	expression-statement))
      ((statement	selection-statement))
      ((statement	iteration-statement))
      ((statement	jump-statement))
     
      ((labeled-statement	"identifier" ":" statement))
      ((labeled-statement	"case" constant-expression ":" statement))
      ((labeled-statement	"default" ":" statement))
     
      ((compound-statement
	"{" declaration-list-opt statement-list-opt "}"))

      ((declaration-list-opt))
      ((declaration-list-opt	declaration-list))
     
      ((declaration-list	declaration))
      ((declaration-list	declaration-list declaration))

      ((statement-list-opt	))
      ((statement-list-opt	statement-list))
     
      ((statement-list	statement))
      ((statement-list	statement-list statement))
     
      ((expression-statement	expression-opt ";"))
     
      ((selection-statement
	"if" "(" expression ")" statement))
      ((selection-statement
	"if" "(" expression ")" statement "else" statement))
      ((selection-statement
	"switch" "(" expression ")" statement))

      ((iteration-statement
	"while" "(" expression")" statement))
      ((iteration-statement
	"do" statement "while" "(" expression ")" ";"))       
      ((iteration-statement
	"for" 
	"(" expression-opt ";" expression-opt ";" expression-opt ")"
	statement))

      ((jump-statement	"goto" "identifier" ";"))
      ((jump-statement	"continue" ";"))
      ((jump-statement	"break" ";"))
      ((jump-statement	"return" expression-opt ";"))

      ;; Expression
      ((expression-opt	))
      ((expression-opt	expression))

      ((expression	assignment-expression))
      ((expression	expression "," assignment-expression))
     
      ((assignment-expression
	conditional-expression))
      ((assignment-expression
	unary-expression assignment-operator assignment-expression))

      ((assignment-operator	"="))
      ((assignment-operator	"*="))
      ((assignment-operator	"/="))
      ((assignment-operator	"%="))
      ((assignment-operator	"+="))
      ((assignment-operator	"-="))
      ((assignment-operator	"<<="))
      ((assignment-operator	">>="))
      ((assignment-operator	"&="))
      ((assignment-operator	"^="))
      ((assignment-operator	"|="))

      ((conditional-expression	logical-OR-expression))
      ((conditional-expression
	logical-OR-expression "?" expression ":" conditional-expression))

      ((logical-OR-expression
	logical-AND-expression))
      ((logical-OR-expression
	logical-OR-expression "||" logical-AND-expression))

      ((logical-AND-expression
	inclusive-OR-expression))
      ((logical-AND-expression
	logical-AND-expression "&&" inclusive-OR-expression))

      ((inclusive-OR-expression
	exclusive-OR-expression))
      ((inclusive-OR-expression
	inclusive-OR-expression "|" exclusive-OR-expression))
     
      ((exclusive-OR-expression
	AND-expression))
      ((exclusive-OR-expression
	exclusive-OR-expression "^" AND-expression))

      ((AND-expression
	equality-expression))
      ((AND-expression
	AND-expression "&" equality-expression))

      ((equality-expression	relational-expression))
      ((equality-expression	equality-expression "==" relational-expression))
      ((equality-expression	equality-expression "!=" relational-expression))

      ((relational-expression	shift-expression))
      ((relational-expression	relational-expression "<"  shift-expression))
      ((relational-expression	relational-expression ">"  shift-expression))
      ((relational-expression	relational-expression "<=" shift-expression))
      ((relational-expression	relational-expression ">=" shift-expression))

      ((shift-expression	addive-expression))
      ((shift-expression	shift-expression "<<" addive-expression))
      ((shift-expression	shift-expression ">>" addive-expression))

      ((addive-expression	multiplicative-expression))
      ((addive-expression
	addive-expression "+" multiplicative-expression))
      ((addive-expression
	addive-expression "-" multiplicative-expression))

      ((multiplicative-expression	cast-expression))
      ((multiplicative-expression
	multiplicative-expression "*" cast-expression))
      ((multiplicative-expression
	multiplicative-expression "/" cast-expression))
      ((multiplicative-expression
	multiplicative-expression "%" cast-expression))

      ((cast-expression	unary-expression))
      ((cast-expression	"(" type-name ")" cast-expression))

      ((unary-expression	postfix-expression))
      ((unary-expression	"++" unary-expression))
      ((unary-expression	"--" unary-expression))
      ((unary-expression	unary-operator cast-expression))
      ((unary-expression	"sizeof" unary-expression))
      ((unary-expression	"sizeof" "(" type-name ")"))
     
      ((unary-operator	"&"))
      ((unary-operator	"*"))
      ((unary-operator	"+"))
      ((unary-operator	"-"))
      ((unary-operator	"~"))
      ((unary-operator	"!"))

      ((postfix-expression	primary-expression))
      ((postfix-expression
	postfix-expression "[" expression "]"))
      ((postfix-expression
	postfix-expression "(" argument-expression-list-opt ")"))
      ((postfix-expression
	postfix-expression "."  "identifier"))
      ((postfix-expression
	postfix-expression "->" "identifier"))
      ((postfix-expression
	postfix-expression "++"))
      ((postfix-expression
	postfix-expression "--"))

      ((argument-expression-list-opt	))
      ((argument-expression-list-opt	argument-expression-list))

      ((argument-expression-list	
	assignment-expression))
      ((argument-expression-list	
	argument-expression-list "," assignment-expression))

      ((primary-expression	"identifier"))
      ((primary-expression	"constant"))
      ((primary-expression	"string-literal"))
      ((primary-expression	"(" expression ")"))

      ((constant-expression-opt	))
      ((constant-expression-opt	constant-expression))

      ((constant-expression	conditional-expression))
      
      ((identifier-opt))
      ((identifier-opt	"identifier"))

      ))
    (output-filename "yayacc-cgrammer-output.el")))


(defun sck-yayacc-test-c-grammar ()
  (interactive)
  (sck-letalist sck=yayacc-cc-c-grammar
      ((object-language language)
       (associativity associativity)
       (attribute-grammar grammar)
       (output-filename outfile))
    (sck-yayacc-lr (list (list 'productions (mapcar 'car grammar))))
    (sck-yayacc-dump)
    (sck-yayacc-graphviz-display)))



(defun sck-yayacc-grammar-to-doty (grammar)
  (let ((buf (get-buffer-create "*sck-yayacc-grammar-to-doty*")))
    (save-excursion
      (set-buffer buf)
      (erase-buffer)
      (let ((cg grammar))
	(while cg
	  (let ((left (caar cg)) assoc)
	    (insert (format "%s:\n\t%s\n"
			    (sck-yayacc-c-grammar-to-doty-format left)
			    (if (cdar cg)
				(mapconcat
				 'sck-yayacc-c-grammar-to-doty-format
				 (cdar cg)
				 " ")
			      "/* empty */")))
	    (setq cg (cdr cg))
	    (while (setq assoc (assoc left cg))
	      (insert (format "\t| %s\n"
			      (if (cdr assoc)
				  (mapconcat 
				   'sck-yayacc-c-grammar-to-doty-format
				   (cdr assoc)
				   " ")
				"/* empty */")))
	      (setq cg (delete assoc cg)))
	    (insert ";\n")
	    )))
      (display-buffer buf))))

(defun sck-yayacc-c-grammar-to-doty ()
  (interactive)
  (sck-letalist sck=yayacc-cc-c-grammar
      ((object-language language)
       (associativity associativity)
       (attribute-grammar grammar)
       (output-filename outfile))
    (sck-yayacc-grammar-to-doty (mapcar 'car grammar))))

(defun sck-yayacc-c-grammar-to-doty-format (r)
  (cond
   ((stringp r)
    (format "%S" r))
   (t
    (replace-regexp-in-string "\-" "_" (format "%S" r)))))
