;;;
;;; $Id: sck-yayacc-lib.el,v 1.2 2007/01/07 07:09:42 heavy Exp $
;;;
(defun sck-yayacc-alist-to-doty-file (alist file)
  (let ((buf (get-buffer-create file)))
    (save-excursion
      (set-buffer buf)
      (erase-buffer)
      (insert (sck-yayacc-alist-to-doty alist))
      (write-file file)
      (kill-buffer buf))))

(defun sck-yayacc-alist-to-doty-buffer (alist buf)
  (save-excursion
    (set-buffer buf)
    (erase-buffer)
    (insert (sck-yayacc-alist-to-doty alist))
    (display-buffer buf)))

(defun sck-yayacc-alist-to-doty (alist)
  (sck-letalist alist
      ((object-language language)
       (associativity associativity)
       (attribute-grammar grammar)
       (output-filename outfile))
    (with-temp-buffer
      (insert "/* sck-yayacc-alist-to-doty output */\n")
      (insert (sck-yayacc-associativity-to-doty associativity))
      (insert "%%\n")
      (insert (sck-yayacc-grammar-to-doty (cdr (mapcar 'car grammar))))
      (insert "%%\n")
      (buffer-string))))

(defun sck-yayacc-associativity-to-doty (associativity)
  (with-temp-buffer 
    (dolist (x associativity)
      (insert (format "%%%s %s\n"
		      (car x)
		      (mapconcat
		       'sck-yayacc-alist-to-doty-format
		       (cadr x)
		       " "))))
    (buffer-string)))

(defun sck-yayacc-grammar-to-doty (grammar)
  (with-temp-buffer 
    (while grammar
      (let ((left (caar grammar)) g)
	(insert (format "%-24s: %s\n"
			(sck-yayacc-alist-to-doty-format left)
			(if (cdar grammar)
			    (mapconcat
			     'sck-yayacc-alist-to-doty-format
			     (cdar grammar)
			     " ")
			  "/* empty */")))
	(setq grammar (cdr grammar))
	(while (setq g (assoc left grammar))
	  (insert (format "\t\t\t| %s\n"
			  (if (cdr g)
			      (mapconcat 
			       'sck-yayacc-alist-to-doty-format
			       (cdr g)
			       " ")
			    "/* empty */")))
	  (setq grammar (delete g grammar)))
	(insert "\t\t\t;\n")))
    (tabify (point-min) (point-max))
    (buffer-string)))

(defun sck-yayacc-alist-to-doty-format (r)
  (cond
   ((stringp r)
    (format "%S" r))
   (t
    (replace-regexp-in-string "\-" "_" (format "%S" r)))))

(provide 'sck-yayacc-lib)
