;;; -*-elisp-*-
;;; $Id: sck-yayacc-yacc-executor.el,v 1.1 2007/02/15 07:24:51 heavy Exp $
;;; sp{
;;;
(defun sck-yayacc-cc-length (production-no)
  ;; ProductionNo -> Length
  (length (cdr (nth production-no sck*yayacc-cc-productions))))

(defun sck-yayacc-cc-lefthandside (production-no)
  ;; ProductionNo -> NonTerminal
  (car (nth production-no sck*yayacc-cc-productions)))

(defun sck-yayacc-cc-action (production-no values lenvalues)
  ;; (ProductionNo values #ofvalues)
  (let ((action (nth production-no sck*yayacc-cc-actions))
	vars)
    (dotimes (x lenvalues)
      (push (intern (format "$%d" x)) vars))
    (progv (nreverse vars) values
      (eval action))))

(defun sck-yayacc-cc-prec (production-no)
  ;; ProductionNo -> prec
  (nth production-no sck*yayacc-cc-precs))

(defun sck-yayacc-cc-associativity-member (terminal)
  ;; return : (Depth (<Associativity> (<Terminal>..)))
  (let ((depth 0) assoc)
    (block 'assoc
      (dolist (x sck*yayacc-cc-associativity)
	(when (member terminal (cadr x))
	  (setq assoc x)
	  (return-from 'assoc))
	(incf depth)))
    (if assoc 
	(list depth assoc))))

;;
;; 
;;
(defun sck-yayacc-cc-executor ()
  (let (token
	sck*yayacc-cc-state-stack
	sck*yayacc-cc-attribute-stack
	sck*yayacc-cc-token-stack
	syn-attribute
	(loopflag t))
    ;; init stack
    (push 0 sck*yayacc-cc-state-stack)
    (push 0 sck*yayacc-cc-token-stack)
    (push 0 sck*yayacc-cc-attribute-stack)
    (setq token (sck-yayacc-lex-get-token))
    (while loopflag

      '(message 
	"In Loop: state-stack (%s) attribute-stack (%s) syn-attribute (%s)"
	sck*yayacc-cc-state-stack
	sck*yayacc-cc-attribute-stack
	syn-attribute)


      (let ((statemoves 
	     (sck-yayacc-cc-table (car sck*yayacc-cc-state-stack)
				  (sck-yayacc-lex-token-id token))))

	'(message"state-stack:%S token:%S statemoves:%S"
		 sck*yayacc-cc-state-stack
		 token statemoves)
	
	(cond
	 ((eq (car statemoves) 'shift)
	  (push (cadr statemoves) sck*yayacc-cc-state-stack)
	  (push (sck-yayacc-lex-token-id token) sck*yayacc-cc-token-stack)
	  (push (sck-yayacc-lex-token-attr token) sck*yayacc-cc-attribute-stack)
	  
	  (setq token (sck-yayacc-lex-get-token)))
	 ((eq (car statemoves) 'reduce)
	  (let* ((len (sck-yayacc-cc-length (cadr statemoves)))
		 action-values)
	    (dotimes (x len)
	      (pop sck*yayacc-cc-state-stack)
	      (pop sck*yayacc-cc-token-stack)
	      (push (pop sck*yayacc-cc-attribute-stack) action-values)
	      )
	    ;; action
	    (setq syn-attribute
		  (sck-yayacc-cc-action (cadr statemoves)
					action-values
					len))
	    ;; statemove
	    (let* ((lefthandside 
		    (sck-yayacc-cc-lefthandside (cadr statemoves)))
		   (sm (sck-yayacc-cc-table 
			(car sck*yayacc-cc-state-stack)
			lefthandside)))
	      (when (eq (car sm) 'shift)
		(push (cadr sm) sck*yayacc-cc-state-stack)
		(push lefthandside sck*yayacc-cc-token-stack)
		(push syn-attribute sck*yayacc-cc-attribute-stack)))
	    ))
	 ((eq (car statemoves) 'error)
	  (setq token (sck-yayacc-cc-error-recovery
		       (list (sck-yayacc-lex-token-pos token) token))))
	 ((eq (car statemoves) 'halt)
	  (setq loopflag nil)))))
    ;;
    syn-attribute))

(defun sck-yayacc-cc-error-recovery (errordata)
  ;; return : token

  '(message "sck-yayacc-cc-error-recovery in")

  (let (statemoves token)
    (setq statemoves
	  (sck-yayacc-cc-table (car sck*yayacc-cc-state-stack) " ERROR"))
    ;; ERROR ŃVtgo܂ŃX^bN߂ăVtg
    ;; X^[gXe[g܂ŗ甲
    (while (and (eq (car statemoves) 'error)
		(not (= (car sck*yayacc-cc-state-stack) 0)))
      (let* ((statedata (assoc 
			 (car sck*yayacc-cc-state-stack)
			 sck*yayacc-cc-lr-table))
	     (productions (cadr statedata))
	     (len (if (listp (cadar productions))
		      (sck-yayacc-cc-length (caar productions))
		    (cadar productions))))

	'(message"state-stack:%S len:%d" 
		 sck*yayacc-cc-state-stack
		 len)

	(dotimes (x len)
	  (pop sck*yayacc-cc-state-stack)
	  (pop sck*yayacc-cc-token-stack)
	  (pop sck*yayacc-cc-attribute-stack))
	(setq statemoves
	      (sck-yayacc-cc-table (car sck*yayacc-cc-state-stack) " ERROR"))))

    '(message"state-stack:%S" sck*yayacc-cc-state-stack)

    ;; X^[gXe[głȂ ERROR ŃVtg
    (when (not (= (car sck*yayacc-cc-state-stack) 0))
      (push (cadr statemoves) sck*yayacc-cc-state-stack)
      (push " ERROR" sck*yayacc-cc-token-stack)
      (push errordata sck*yayacc-cc-attribute-stack))

    ;; ̃g[NŐi߂܂Ńg[Nǂݎ̂Ă
    (setq token (sck-yayacc-lex-get-token))
    (setq statemoves
	  (sck-yayacc-cc-table (car sck*yayacc-cc-state-stack)
			       (sck-yayacc-lex-token-id token)))
    
    '(message"state-stack:%S token:%S statemoves:%S" 
	     sck*yayacc-cc-state-stack
	     token
	     statemoves)

    (while (and (not (string= (sck-yayacc-lex-token-id token) " EOF"))
		(eq (car statemoves) 'error))

      '(message "token:%S statemoves:%S" token statemoves)

      (setq token (sck-yayacc-lex-get-token))
      (setq statemoves
	    (sck-yayacc-cc-table (car sck*yayacc-cc-state-stack)
				 (sck-yayacc-lex-token-id token))))
    
    ;; token  " EOF"  statemoves  error łꍇ
    ;; " EOF" ňȍ~o܂ŃX^bN̂Ă
    (when (and
	   (string= (sck-yayacc-lex-token-id token) " EOF")
	   (eq 'error (car statemoves)))
      (while (and (eq (car statemoves) 'error)
		  (not (= (car sck*yayacc-cc-state-stack) 0)))
	(let* ((statedata (assoc 
			   (car sck*yayacc-cc-state-stack)
			   sck*yayacc-cc-lr-table))
	       (productions (cadr statedata))
	       (len (if (listp (cadar productions))
			(sck-yayacc-cc-length (caar productions))
		      (cadar productions))))

	  '(message"state-stack:%S len:%d" 
		   sck*yayacc-cc-state-stack
		   len)

	  (dotimes (x len)
	    (pop sck*yayacc-cc-state-stack)
	    (pop sck*yayacc-cc-token-stack)
	    (pop sck*yayacc-cc-attribute-stack))
	  (setq statemoves
		(sck-yayacc-cc-table (car sck*yayacc-cc-state-stack) " EOF")
		))))
    
    (assert (not (eq 'error (car statemoves))))

    '(message "sck-yayacc-cc-error-recovery out token:%S statemoves:%S"
	     token statemoves)

    token))


(defun sck-yayacc-cc-table (state vocabulary)
  ;; (stete vocabulary) -> (state-moveID stateNo|productionNo)

  '(message "in sck-yayacc-cc-table (%s %S)" state vocabulary)

  (let* ((statedata (assoc state sck*yayacc-cc-lr-table))
	 (shift (assoc vocabulary (caddr statedata)))
	 ;; (Productionno lookaheadset)
	 reduce default-reduce)
    ;; reduce/reduce conflicts ̏ꍇ lookaheadset 
    ;; " ERROR" ł reduce Ȃ
    (unless (equal vocabulary " ERROR")
      (block 'reduce
	(dolist (x (cadr statedata))
	  (when (listp (cadr x))
	    ;;  reduce 
	    (unless default-reduce
	      (setq default-reduce x))
	    (when ;; Null ̏ꍇ͑S
		(or (null (cadr x))
		    (member vocabulary (cadr x)))
	      (setq reduce x)
	      (return-from 'reduce))))))
    ;; shift/reduce conflicts ̏ꍇ assoc 
    (when (and shift reduce)
      ;; shift/reduce conflict
      (let ((assoc (sck-yayacc-cc-associativity-member vocabulary))
	    (stack-term (and (sck-yayacc-cc-prec (car reduce))
			     (sck-yayacc-cc-associativity-member
			      (sck-yayacc-cc-prec (car reduce))))))

	'(message "assoc:%S stack-term:%S"
		  assoc stack-term)
	
	(if (and assoc stack-term)
	    (cond
	     ;; Ȃ猋Č߂
	     ((= (car assoc) (car stack-term))
	      (cond
	       ((eq (caadr assoc) 'left)
		(setq shift nil))
	       ((eq (caadr assoc) 'right)
		(setq reduce nil))
	       (t
		;; ǂł΂Ƃ肠 shift
		(setq reduce nil))))
	     ;; D揇ʂقI
	     ((> (car assoc) (car stack-term))
	      (setq reduce nil))
	     (t
	      (setq shift nil)))
	  ;; assoc ŉłȂꍇ shift
	  (message "state%d: shift(%d)/reduce(%d) conflict -> shift(%d)"
		   state 
		   (cadr shift)
		   (car reduce)
		   (cadr shift))
	  (setq reduce nil))))
    
    '(message "shift:%S reduce:%S default-reduce %S"
	      shift reduce default-reduce)
    
    (cond
     (shift
      (list 'shift (cadr shift)))
     ;; reduce ݒ肳ĂΖ
     (default-reduce
      (list 'reduce (if reduce (car reduce) (car default-reduce))))
     ;; X^[gXe[gŏI[̏ꍇ͏I
     ((and (= state 0)
	   (equal vocabulary " EOF"))
      (list 'halt))
     (t
      (list 'error)))))
